/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cbuilder.cpp
 *		- methods for verilog tree builder
 *
 ******************************************************************************
 */
						

#include <map>
#include <unistd.h>
#include <vector>
#include <fstream>
#include <set>
#include <ctype.h>
#include "main.h"
#include "cbuilder.h"
#include "cdumpverilog.h"
#include "cmodule.h"
#include "cport.h"
#include "cfref.h"
#include "cvar.h"
#include "cnet.h"
#include "cparam.h"
#include "cnode.h"
#include "cinstance.h"
#include "cfunction.h"
#include "cportdir.h"

static set<CModule*> leafModules;
static set<CModule*> requiredModules;
static set<CModule*> allModules;
static set<CModule*> topLevelModules;
static set<CModule*> undefinedModules;
static set<CModule*> treeModules;
static set<CModule*> processedModules;

static const char*   treeExt = "tree";
static Coord_t	     _loc;
static Coord_t*	     loc = &_loc;

extern CDumpVerilog*  dumpVerilog;
extern char*	     outputPath;
extern char*	     outputDirectory;
extern int           errorCount;
extern map<char*,int> includeFiles;

static int HasSuffix( const char* s, const char* suffix );
static int verbose = 0;
static Message* mTESMS;

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
extern "C" {
CBackend* CreateToolInstance()
{
    return new CBuilder();
}
}

/************************************************
 	Constructor
 ************************************************/
CBuilder::CBuilder( ) 
{
     switches.push_back( "+tree_ext=<ext>" );
     switchDescription["+tree_ext=<ext>"] = "File extension for files to be expanded";
     switches.push_back( "+depend=<filename>" );
     switchDescription["+depend=<filename>"] = "Generate dependency info";
     switches.push_back( "+builder-verbose" );
     switchDescription["+builder-verbose"] = "Dump vebose log of actions";

     mTESMS = Message::RegisterWarning( this, Message::eWARNING, 
		"TESMS", "+tree_ext= switch missing, assuming .tree suffix\n" );
}

/************************************************
 	Activate
	- activate selected tool
 ************************************************/

void CBuilder::Activate()
{
     /*
      * force vectors to be inferred
      */
     inferVectors = 1;
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CBuilder::GetToolName( void )
{
	static char	toolName[] = "builder";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CBuilder::GetToolDescription( void )
{
	static char	toolDescription[] = "Auto route hierarchy";

	return toolDescription;
}

/************************************************
	ResolveInstance
	- returns true if instance of module should
	  be resolved
 ************************************************/
int CBuilder::ResolveInstance( CModule* m, CInstance* ) 
{ 
    if( HasSuffix(m->GetCoord()->filename, treeExt) ) {
	return TRUE; 
    }
    return FALSE;
}
       
/*******************************************************
	CollectArgList
	- build arg list for instance
********************************************************/

static void CollectArgList( list<CNode*>& argList, CNode* n )
{
	if( n == NULL ) {
	} else if( n->GetOp() == eLIST ) {
		CollectArgList( argList, n->Arg<CNode*>(0) );
		CollectArgList( argList, n->Arg<CNode*>(1) );
	} else if( n->GetOp() == eCOMMENT ||
		   n->GetOp() == eVRQ ||
		   n->GetOp() == ePRAGMA ) {
	} else {
	    MASSERT( n->GetOp() == eARG );
	    argList.push_back( n );
	}
}

/*******************************************************
	CollectPortList
	- build list of ports for module
********************************************************/

static void CollectPortList( list<CPort*>& portList, CNode* n )
{
	if( n == NULL ) {
	} else if( n->GetOp() == eLIST ) {
		CollectPortList( portList, n->Arg<CNode*>(0) );
		CollectPortList( portList, n->Arg<CNode*>(1) );
	} else if( n->GetOp() == eCOMMENT ||
		   n->GetOp() == eVRQ ||
		   n->GetOp() == ePRAGMA ) {
	} else {
	    MASSERT( n->GetOp() == ePORT_DEF );
	    portList.push_back( n->Arg<CPort*>(0) );
	}
}
/*******************************************************
	InsertDeclaration
	- traverse list and insert decl statement after IO
********************************************************/
static CNode* InsertDeclaration1( CNode* root, CNode* n )
{
    switch( root->GetOp() ) {
    case ePORT_DECL:
	return NULL;
    case eLIST:
	{
	    CNode* l = InsertDeclaration1( root->Arg<CNode*>(0), n );
	    if( l ) {
		root->Arg<CNode*>(0) = l;
	        return root;
	    }
	    CNode* r = InsertDeclaration1( root->Arg<CNode*>(1), n );
	    if( r ) {
		root->Arg<CNode*>(1) = r;
   	        return root; 
	    }
	    return NULL;
	}
    default:
	return cLIST(n,root);
    }
}
static CNode* InsertDeclaration( CNode* root, CNode* n )
{
    switch( root->GetOp() ) {
    case eLIST:
	return InsertDeclaration1( root, n );
    case ePORT_DECL:
	return cLIST(root,n);
    default:
	return cLIST(n,root);
    }
}
/*******************************************************
	_CollectLHS
	- collect lhs of statements 
********************************************************/


struct ReferencedArgs {
    CDecl* decl;
    int	   count;
};
static int _Referenced( CNode* n, void* arg )
{
    ReferencedArgs* r = (ReferencedArgs*)arg;
    switch( n->GetOp() ) {
    case eNET_REF:
	if( n->Arg<CNet*>(0) == r->decl ) {
	    r->count++;	
	}
	break;
    default:;
    }
    return TRUE;
}

/*******************************************************
	Referenced
	- returns true if decl is referenced in code
********************************************************/
static int Referenced( CNode* code, CDecl* decl )
{
    ReferencedArgs r = { decl, 0 };
    if( code != NULL ) {
	code->PreVisit1( _Referenced, &r );
    }
    return (r.count != 0);
}

/************************************************
	PropagatePortsToInstance
	- fill unspecified ports in instance 
	  from module definition
 ************************************************/
class Syminfo {
public:
    int iCount;
    int oCount;
    int ioCount;
    int bitVector;
    Syminfo() : iCount(0), oCount(0), ioCount(0), bitVector(0) {};
    void BitVector() { bitVector=1; }
    void Inc( Decl_t type ) {
	switch( type ) {
	case eINPUT:
	    iCount++;
	    break;
	case eOUTPUT:
	    oCount++;
	    break;
	case eINOUT:
	    ioCount++;
	    break;
	default:
	    MASSERT( FALSE );
	}
    }
};
static void PropagatePortsToInstance( 
	CModule* parent, 
	set<CSymbol*>& createdNets,
	map<CSymbol*,Syminfo>& sym2info,
	CModule* def, 
	CInstance* inst )
{
    /*
     * make a list of ports in instance and module
     */
    CSymtab<CDecl>& symtab     = parent->GetSymtab();
    CSymtab<CDecl>& portSymtab = parent->GetPortSymtab();
    list<CNode*> instArgs;
    CollectArgList( instArgs, inst->GetArguments() );
    list<CPort*> modulePorts;
    CollectPortList( modulePorts, def->GetPorts() );
    /*
     * build a set of all instance port symbols
     */
    map<CSymbol*,CNode*>   arg2conn;
    set<CSymbol*>          argSymbols;
    list<CNode*>::iterator argPtr;
    for( argPtr = instArgs.begin(); 
		    argPtr != instArgs.end(); ++argPtr ) {
	CSymbol* sym = (*argPtr)->Arg<CSymbol*>(0);
	argSymbols.insert( sym );
	arg2conn[sym] = (*argPtr)->Arg<CNode*>(1); 
    }

      
    /*
     * compare against ports in module definition,
     * creating any missing ports as we go
     */
    list<CPort*>::iterator portPtr;
    for( portPtr = modulePorts.begin(); 
		    portPtr != modulePorts.end(); ++portPtr ) {
	CSymbol* sym = CSymbol::Lookup( (*portPtr)->GetName() );
	CNode* conn = (*portPtr)->GetConnection();
	MASSERT( conn );
	int portWidth = conn->GetWidth();
	if( argSymbols.find( sym ) == argSymbols.end() ) {
	    if( verbose ) {
	        logprintf( "adding port: .%s(%s) to instance %s::%s\n", 
			(*portPtr)->GetName(),
			(*portPtr)->GetName(),
			def->GetName(),
			inst->GetName() );
	    }
	    CDecl* decl = symtab.Lookup(sym);

	    if( decl != NULL && decl->GetClass() == eNET ) { 
		 CDecl* dir = portSymtab.Lookup(sym);
		if( createdNets.find(sym) == createdNets.end() ) {
//		    int referenced = Referenced( parent->GetCodeList(), decl );
		    /*
		     * allow this case for forcing outputs to top
		     */
		    if( dir != NULL && dir->GetType() == eOUTPUT ) {
//			createdNets.insert( sym );
		    } else if( dir != NULL && dir->GetType() == eINPUT 
			    && (*portPtr)->Direction() == eINPUT ) {
			/*
			 * allow this case for explicitly declaring inputs to top
			 */
		    } else if( dir == NULL ) {
			/*
			 * allow ports to connect at top level to a wire
			 * declaration
			 */
//			createdNets.insert( sym );
		    } else { 
			error( NULL, "Declaration for %s already exists in"
				 "module %s, can not promote port", 
				    sym->GetName(), parent->GetName() );
			continue;
		    }
		}
		MASSERT( decl->GetClass() == eNET );
		CNet* net = (CNet*)decl;
		if( net->GetUndefined() && 
		    dir == NULL ) { 
		    createdNets.insert( sym );
		}
		if( portWidth != net->GetWidth() ) {
		    if( !net->GetUndefined() ) {
			error( NULL, "%s in module %s has an incompatible width (%d!=%d),"
				     " can not promote port", 
					sym->GetName(), parent->GetName(),
					portWidth,
					net->GetWidth() );
			continue;
		    }
		}
		inst->SetArguments( cLINK(cARG(sym,cNET_REF(net)), inst->GetArguments()) );
	    } else if( decl != NULL && decl->GetClass() == eVAR ) { 
		CVar* reg = (CVar*)decl;
		if( portWidth != reg->GetWidth() ) {
			error( NULL, "%s in module %s has an incompatible width (%d!=%d),"
				     " can not promote port", 
					sym->GetName(), parent->GetName(),
					portWidth,
					reg->GetWidth() );
			continue;
		}
		inst->SetArguments( cLINK(cARG(sym,cVAR_REF(reg)), inst->GetArguments()) );
	    } else if( decl != NULL ) { 
		error( NULL, "Declaration for %s already exists in"
			 "module %s, can not promote port", 
			    sym->GetName(), parent->GetName() );
		continue;
	    } else {
	        CSimpleType* stype;
		if( portWidth != 1 ) {
		    list<CNode*> rangeList;
		    rangeList.push_back(
		    	cRANGE(cSUB(conn->GetWidthExp(),cINT32(1)),cINT32(0)) );
	       	    stype = new(CNode::CurrentHeap()) CSimpleType(eNO_VAR_TYPE, rangeList );
		} else {
	            stype = new(CNode::CurrentHeap()) 
					CSimpleType(eNO_VAR_TYPE);
		}

		CNet* net = new(CNode::CurrentHeap()) CNet( sym, loc, eWIRE, 
				stype, FALSE );
		parent->Add( net );
		net->SetDeclStatementCreated();
		parent->GetSymtab().Add(sym,net);
		CNode* newNode = cNET_DECL(net,NULL,loc);
		parent->SetCodeList(InsertDeclaration(parent->GetCodeList(),newNode));
		createdNets.insert( sym );
		if( net->GetWidth() == 1 && verbose ) {
		    logprintf( "adding net %s to %s\n", 
			net->GetName(), parent->GetName() );
		} else if( verbose ) {
		    logprintf( "adding net %s[%ld:%ld] to %s\n", 
			net->GetName(), 
			net->GetMsb()->EvalINT32(),
			net->GetLsb()->EvalINT32(),
			parent->GetName() );
		}
		inst->SetArguments( cLINK(cARG(sym,cNET_REF(net)), inst->GetArguments()) );
	    }
	    sym2info[sym].Inc((*portPtr)->Direction());
	} else {
	    /* 
	     * handle special case where port has been renamed
	     */
	    CNode* conn = arg2conn[sym];
	    if( conn && conn->GetOp() == eNET_REF ) {
		CNet* net = conn->Arg<CNet*>(0);
		if( net->GetUndefined() ) {
		    if( portWidth != 1 ) {
		    list<CNode*> rangeList;
		    rangeList.push_back(
		      cRANGE(cSUB(conn->GetWidthExp(),cINT32(1)),cINT32(0)) );
		        net->SetDataType( new(CNode::CurrentHeap()) 
				CSimpleType( eNO_VAR_TYPE, rangeList ) );
		    }
		    net->SetDeclStatementCreated();
		    net->SetUndefined( FALSE );
		    createdNets.insert( net->GetSymbol() );
		    sym2info[net->GetSymbol()].Inc((*portPtr)->Direction());
		    CNode* newNode = cNET_DECL(net,NULL,loc);
		    parent->SetCodeList(InsertDeclaration(parent->GetCodeList(),newNode));
		} else if( createdNets.find(net->GetSymbol()) != 
				createdNets.end() ) {
		    if( portWidth != net->GetWidth() ) {
			error( NULL, "%s in module %s has an incompatible width (%d!=%d),"
				     " can not promote port", 
					net->GetName(), parent->GetName(),
					portWidth,
					net->GetWidth() );
			    continue;
		    }
		    sym2info[net->GetSymbol()].Inc((*portPtr)->Direction());
		}
	    } else if( conn && conn->GetOp() == eARRAY &&
		       conn->Arg<CNode*>(0)->GetOp() == eNET_REF ) {
		CNet* net = conn->Arg<CNode*>(0)->Arg<CNet*>(0);
		if( net->GetUndefined() ) {
		    net->SetDeclStatementCreated();
		    net->SetUndefined( FALSE );
		    createdNets.insert( net->GetSymbol() );
		    CNode* newNode = cNET_DECL(net,NULL,loc);
		    parent->SetCodeList(InsertDeclaration(parent->GetCodeList(),newNode));
		}
		sym2info[net->GetSymbol()].Inc((*portPtr)->Direction());
		sym2info[net->GetSymbol()].BitVector();
	    }
	}
    }
}

/************************************************
	Build
	- build tree modules, starting at leaves
 ************************************************/
static void Build( CModule* m )
{
    int isTreeModule = (treeModules.find(m) != treeModules.end());
    /*
     * if module has been processed, return
     */
    if( processedModules.find(m) != processedModules.end() ) {
	return;
    }
    /*
     * process siblings first and create a list of siblings
     */
    map<CSymbol*,Syminfo> sym2info;
    set<CSymbol*>         createdNets;
    set<CModule*>         siblings;
    vector<CInstance*>& inst = *m->GetInstanceList();
    vector<CInstance*>::iterator p;
    for( p = inst.begin(); p != inst.end(); ++p ) {
	if( (*p)->GetNumberOfUnpackedDimensions() ) {
	    error( (*p)->GetCoord(), "instance arrays are not supported" );
	    continue;
	}
	CModule* def = (*p)->GetDefinition();
	Build( def );
	siblings.insert(def);
	if( isTreeModule ) {
	    PropagatePortsToInstance( m, createdNets, sym2info, def, *p );
	}
    }
    if( !isTreeModule ) {
	return;
    }
    if( verbose ) {
	logprintf( "Building module %s\n", (m)->GetName() );
    }
    /*
     * process each created net
     */
    CSymtab<CDecl>& portSymtab = m->GetPortSymtab();
    CSymtab<CDecl>& symtab     = m->GetSymtab();
    set<CSymbol*>::iterator ptr;
    for( ptr = createdNets.begin(); ptr != createdNets.end(); ++ptr ) {
	Syminfo& info = sym2info[*ptr];
	/*
	 * gather leaf and root info
	 */
	int i  = 0;
	int o  = 0;
	int io = 0;
	CDecl* dir = portSymtab.Lookup(*ptr);
	if( dir != NULL ) {
	    switch( dir->GetType() ) {
	    case eOUTPUT:
		o++;
		break;
	    case eINPUT:
		i++;
		break;
	    case eINOUT:
		io++;
		break;
	    default:
		MASSERT( FALSE );
	    }
	}
/*
	logprintf( "process %s: i=%d o=%d io=%d : i=%d o=%d io=%d\n",
	    (*ptr)->GetName(), 
	    info.iCount,
	    info.oCount,
	    info.ioCount,
	    i, o, io);
*/
	/*
	 * check for errors
	 */
	if( info.oCount+i > 1 && !info.bitVector ) {
	    error( NULL, "more than on output on %s in module %s\n",
		(*ptr)->GetName(), m->GetName() );
	    continue;
	}
	if( info.ioCount+info.iCount+info.oCount == 0 && io > 0 ) {
	    error( NULL, "no connection to inout on %s in module %s\n",
		(*ptr)->GetName(), m->GetName() );
	    continue;
	}
	if( info.oCount == 0 && i == 0 && o == 1) {
	    error( NULL, "no connection to output on %s in module %s\n",
		(*ptr)->GetName(), m->GetName() );
	    continue;
	}
	/*
	 * add port to root if needed
	 */
	int addPort = FALSE;
	Decl_t type;
	if( info.ioCount == 1 && info.iCount+info.oCount == 0 && i+o+io == 0 ) {
	    addPort = TRUE;
	    type = eINOUT;
	} else if( info.oCount > 0 && info.iCount+info.ioCount == 0 && o == 0 ) {
	    addPort = TRUE;
	    type = eOUTPUT;
	} else if( info.iCount > 0 && info.oCount+info.ioCount == 0 && i == 0 ) {
	    addPort = TRUE;
	    type = eINPUT;
	}
	if( addPort ) {
	    CSymbol* sym = *ptr;
	    CDecl* decl = symtab.Lookup( *ptr );
	    MASSERT( decl->GetClass() == eNET );
	    CNet* net = (CNet*)decl;
	    if( net->GetWidth() == 1 && verbose ) {
		logprintf( "need to add %s %s to %s\n", 
		    type == eINOUT ? "inout" :
		    type == eOUTPUT ? "output" :
		    "input",
		    (*ptr)->GetName(), m->GetName() );
	    } else if( verbose ) {
		logprintf( "need to add %s %s[%ld:%ld] to %s\n", 
		    type == eINOUT ? "inout" :
		    type == eOUTPUT ? "output" :
		    "input",
		    (*ptr)->GetName(), 
		    net->GetMsb()->EvalINT32(),
		    net->GetLsb()->EvalINT32(),
		    m->GetName() );
	    }
	    CDataType* stype = net->GetDataType()->Clone(CNode::CurrentHeap());
	    CPortDir* portDir = new(CNode::CurrentHeap()) CPortDir( sym, loc, type, stype );
	    portDir->SetDeclStatementCreated();
	    CPort* port = new(CNode::CurrentHeap()) CPort( sym, FALSE, loc );
	    port->SetConnection( cNET_REF( net, loc ) );
	    port->Direction(type);
	    m->GetSymtab().Add(sym,net);
	    m->SetPorts( cLINK(cPORT_DEF(port),m->GetPorts() ) );
	    m->SetCodeList( cLINK(cPORT_DECL(portDir,loc),m->GetCodeList()) );
	}
    }

    processedModules.insert( m );
}

/************************************************
	HasSuffix
	- returns true if string has given suffix
	  with a period seperator
 ************************************************/

static int HasSuffix( const char* s, const char* suffix )
{
    if( s == NULL ) {
	return FALSE;
    }
    const char* period = strrchr( s, '.' );
    if( period == NULL ) {
	return FALSE;
    }
    if( strcmp( period+1, suffix ) ) {
	return FALSE;
    }
    return TRUE;
}

/************************************************
	CreateModuleList
	- mapping of name to module definition
 ************************************************/

static void CreateModuleList( CNode* n )
{
	if( n == NULL ) {
		return;
	}
        switch( n->GetOp() ) {
        case eLIST:
                CreateModuleList( n->Arg<CNode*>(0) );
                CreateModuleList( n->Arg<CNode*>(1) );
                break;
        case eMODULE_DEF:
		{
			CModule* m = n->Arg<CModule*>(0);
			int required = FALSE;
			allModules.insert(m);
			if( HasSuffix(m->GetCoord()->filename, treeExt) ) {
			    treeModules.insert(m);
			    required = TRUE;
			}
			vector<CInstance*>& inst = *m->GetInstanceList();
			vector<CInstance*>::iterator p;
			for( p = inst.begin(); p != inst.end(); ++p ) {
				CModule* def = (*p)->GetDefinition();
				leafModules.insert(def);
				if( required ) {
					requiredModules.insert(def);
				}
			}
		}
                break;
	case ePRAGMA:
        case eCOMMENT:
	case eVRQ:
                break;
        default:
		MASSERT( FALSE );
        }       
}

/************************************************
	DumpModuleDependencies
	- output module dependency info
 ************************************************/
static void DumpModuleDependencies( FILE* f, CModule* m )
{
    vector<CInstance*>& inst = *m->GetInstanceList();
    vector<CInstance*>::iterator p;
    for( p = inst.begin(); p != inst.end(); ++p ) {
	CModule* def = (*p)->GetDefinition();
	if( !def ) {
	    continue;
	}
	const char* fname = def->GetCoord()->filename;
	fprintf( f, "  " );
	if( fname[0] != '/' ) {
	   char buffer[4096];
	   char* cwd = getcwd( buffer, sizeof(buffer) );
	   fprintf( f, "%s/", cwd );
	}
	fprintf( f, "%s \\\n", def->GetCoord()->filename );
	int isTreeModule = (treeModules.find(def) != treeModules.end());
	if( !isTreeModule ) {
	    continue;
	}
	DumpModuleDependencies( f, def );
    }
}

/************************************************
	DumpDependencies
	- output dependency info for the given module
 ************************************************/
static void DumpDependencies( FILE* f, CModule* m )
{
    fprintf( f, "%s/%s.v: \\\n", outputPath, m->GetName() );
    DumpModuleDependencies( f, m );
    map<char*,int>::iterator p;
    for( p = includeFiles.begin(); p != includeFiles.end(); ++p ) {
	fprintf( f, "  " );
	if( p->first[0] != '/' ) {
	   char buffer[4096];
	   char* cwd = getcwd( buffer, sizeof(buffer) );
	   fprintf( f, "%s/", cwd );
	}
	fprintf( f, "%s \\\n", p->first );
    }
    fprintf( f, "\n" );
}

/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CBuilder::Process( list<CElement>& inputList,
			list<CElement>& outputList )
{
	/*
	 * get input parameters
	 */
	 char* ext = GetPlusArg( "tree_ext=" );
         if( ext == NULL ) {
	    message( NULL, mTESMS );
         } else {
	    treeExt = ext;
	}

	/*
	 * determine if verbose output is desired
	 */
	verbose = GetPlusArg( "builder-verbose" ) != NULL;

	/*
	 * determine if generating dependencies
	 */
	char* dep = GetPlusArg( "depend" );
	int   dependMode = FALSE;
	char* depFilename = NULL;
	if( dep ) {
	    depFilename = strchr( dep, '=' );
	    if( depFilename ) {
		    depFilename++;
	    }
	    dependMode = TRUE;
	}
	
        /*
         * merge all input streams together
         */
	CNode* code = NULL;
        list<CElement>::iterator ptr;
        for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
            code = cLINK( code, ptr->Code() );
        }

	/*
	 * build full module list
	 */
	CreateModuleList( code );
	
	set_difference( allModules.begin(), allModules.end(),
			leafModules.begin(), leafModules.end(),
			inserter(topLevelModules, topLevelModules.end()) );

	set_difference( requiredModules.begin(), requiredModules.end(),
			allModules.begin(), allModules.end(),
			inserter(undefinedModules, undefinedModules.end()) );
	
	set<CModule*>::iterator p;
	for( p = undefinedModules.begin(); p != undefinedModules.end(); ++p ) {
		error( NULL, "module '%s' is undefined\n", (*p)->GetName() );
	}	

	/*
	 * process each top level module
	 * proceeding down the hierarchy
	 */
	if( verbose ) {
	    logprintf( "Top level modules:\n" );
	}
	for( p = topLevelModules.begin(); p != topLevelModules.end(); ++p ) {
	   Build( *p );
	}	
	
	if( !dependMode ) {
	    /*
	     * dump modified tree modules to independent streams
	     */
    	     for( p = treeModules.begin(); p != treeModules.end(); ++p ) {
    	         if( !(*p)->ModuleDirectlyDefined() ) {
    	             continue;
    	         }
		 string ofname = (*p)->GetName();
		 ofname += ".v";
                 outputList.push_back( CElement( ofname.c_str(),
							1, cMODULE_DEF(*p) ) );
	    }
	} else {
	    /*
	     * generate dependencys
	     */
	    FILE* f = stdout;
	    if( depFilename ) {
		f = fopen( depFilename, "w" );
		if( f == NULL ) {
		    error( NULL, "file '%s' could not be created\n", depFilename );
		    return;
		} 
	    }
	    for( p = treeModules.begin(); p != treeModules.end(); ++p ) {
	        if( !(*p)->ModuleDirectlyDefined() ) {
		    continue;
		}
	   	DumpDependencies( f, *p );
	    }
	    fclose( f );
	}
}


