/******************************************************************************
 * 	Copyright (c) 2010 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   cfilter.cc
 *		- methods for tool to filter elements of a design
 *
 ******************************************************************************
 */
						

#include <set>
#include <map>
#include "cnode.h"
#include "cfilter.h"
#include "main.h"

/**
 * \page filter_plugin Filter Plugin
 * The filter tool edits the parse trees, selectively deleting nodes based on 
 * a filter specification that is supplied by command line switches.
 * The filter specification consists of a global policy, either keep or 
 * delete, and per construct keep and delete directives. The per construct
 * directives are applied first, any unmatch constructs are handle by the 
 * default policy. Note not all constructs are allowed in the filter 
 * specification, those that are omited are keep by default. 
 * The follow switches are used to specify the filter specification:
 * - <b>+filter-default-policy-keep</b> Keep all unmatched constructs.
 * - <b>+filter-default-policy-delete</b> Delete all unmatched. This is the 
 *   default if no policy switch is specified.
 * - <b>+filter-<construct_name>-keep</b> Keep <construct_name>.
 * - <b>+filter-<construct_name>-delete</b> Delete <construct_name>.
 *
 * See the \ref manpage for the complete list of which constructs may be
 * specified in the filter specification.
 */

/**
 * \cond PLUGIN_INTERNALS
 */

/*
 * filter specification
 * 0 - cut
 * 1 - keep
 */
static map<NodeOp_t,int> filterSpec;
static int	       defaultPolicy; // 0 - cut, 1 - keep

static NodeOp_t keep[] = {
    eERROR,
    eVCONSTANT,
    eRCONSTANT,
    eELIST,
    eWIDTH,
    eNOP,
    eSUB,
    eMUL,
    eDIV,
    ePOW,
    eADD,
    eLSH,
    eRSH,
    eLSHA,
    eRSHA,
    eMOD,
    eOR,
    eAND,
    eANDANDAND,
    eXOR,
    eXNOR,
    eARRAY,
    eNET_REF,
    eVAR_REF,
    ePARAM_REF,
    ePORT_REF,
    eFWD_REF,
    eGENVAR_REF,
    eEVENT,
    eBLOCK_REF,
    ePOSEDGE,
    eNEGEDGE,
    eEDGE,
    eEVOR,
    eDELAY,
    eMTM,
    eCASEITEM,
    eARG,
    eREPEAT_CONTROL,
    eDELAY_CONTROL,
    eEVENT_CONTROL,
    eEXTERNAL_REF,
    eTABLE_ENTRY,
    eTABLE_SYMBOL,
    ePORTLIST_END,
    eMACRO_EXPR,
    eLIST,
    eRANGE,
    eSLICE,
    ePSLICE,
    eMSLICE,
    eCVRI,
    eCVIR,
    eREP,
    eCAT,
    eUCAT,
    eCOM,
    eNEG,
    ePLUS,
    eNOT,
    eGT,
    eGE,
    eLT,
    eLE,
    eLAND,
    eLOR,
    eCEQ,
    eCNE,
    eEQ,
    eNE,
    eRAND,
    eRNAND,
    eROR,
    eRNOR,
    eRXOR,
    eRXNOR,
    eHOOK,
    ePATH_ASSIGN,
    eIFNONE_PATH_ASSIGN
}; 

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
extern "C" {
CBackend* CreateToolInstance()
{
    return new CFilter();
}
}

/************************************************
 	Constructor
 ************************************************/
CFilter::CFilter() 
{
     switches.push_back( "+filter-default-policy-keep" );
     switchDescription["+filter-default-policy-keep"] = "keep all element by default ";
     switches.push_back( "+filter-default-policy-delete" );
     switchDescription["+filter-default-policy-delete"] = "delete all element by default (tool default behavior)";

     set<string> keepSet;
     for( int i = 0; i < sizeof(keep)/sizeof(NodeOp_t); ++i ) {
	 keepSet.insert( nodeOpName[keep[i]] );
     }

     for( int i = 0; nodeOpName[i]; ++i ) {
	if( keepSet.count( nodeOpName[i] ) ) {
	    continue;
	}
	string sw( nodeOpName[i] );
	for( int j = 0; j < sw.size(); ++j ) {
	    sw[j] = tolower(sw[j]);
	    if( sw[j] == '_' ) {
		sw[j] = '-';
	    }
	}

	string swk("+filter-");
	swk += sw;
	swk += "-keep";
        switches.push_back( swk );
        switchDescription[swk] = "keep "; 
	switchDescription[swk].append(nodeOpDescription[i]);

	string swd("+filter-");
	swd += sw;
	swd += "-delete";
        switches.push_back( swd );
        switchDescription[swd] = "delete "; 
	switchDescription[swd].append(nodeOpDescription[i]);

	deleteMap[swd.substr(8)] = static_cast<NodeOp_t>(i);
	keepMap[swk.substr(8)]   = static_cast<NodeOp_t>(i);
     }
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CFilter::GetToolName( void )
{
	static char	toolName[] = "filter";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CFilter::GetToolDescription( void )
{
	static char	toolDescription[] = "Extract elements of hierarchy";

	return toolDescription;
}

/************************************************
	BuildCutList
	- build node cut list based upon filter 
	  specification.
 ************************************************/
static int BuildCutList( CNode* n, void* arg )
{
    set<CNode*>* cutList = static_cast<set<CNode*>*>(arg);

    switch( n->GetOp() ) {
    case eLIST:
	return 1;
    }
 
    map<NodeOp_t,int>::iterator ptr;
    ptr = filterSpec.find(n->GetOp());
    if( ptr != filterSpec.end() ) {
	if( !ptr->second ) {
	    cutList->insert(n);
	    return 0;
	}
	return 1;
    }

    /*
     * apply default policy 
     */
    if( defaultPolicy == 0 ) {
        cutList->insert(n);
        return 0; 
    }
    return 1;
}

/************************************************
	CutNodes
	- delete nodes in cut list
 ************************************************/
static CNode* CutNodes( CNode* n, void* arg )
{
    set<CNode*>* cutList = static_cast<set<CNode*>*>(arg);
    if( !n ) {
	return NULL;
    }
 
    if( cutList->count(n) ) {
	return NULL;
    }

    switch( n->GetOp() ) {
    case eLIST:
	if( !n->Arg<CNode*>(0) ) {
	     return n->Arg<CNode*>(1);
	} else if( !n->Arg<CNode*>(1) ) {
	     return n->Arg<CNode*>(0);
	}
	break;
    }
    return n; 
}

/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CFilter::Process( list<CElement>& inputList,
   		       list<CElement>& outputList ) 
{
     defaultPolicy = 0;
     if( GetPlusArg( "filter-default-policy-keep" ) ) {
	defaultPolicy = 1;
     } else if( GetPlusArg( "filter-default-policy-delete" ) ) {
	defaultPolicy = 0;
     }

    list<string> args;
    GetPlusArgs( "filter-", args );
    list<string>::iterator aptr;
    for( aptr = args.begin(); aptr != args.end(); ++aptr ) {
	if( keepMap.count(*aptr) ) {	
	    filterSpec[keepMap[*aptr]] = 1;
	}
	if( deleteMap.count(*aptr) ) {	
	    filterSpec[deleteMap[*aptr]] = 0;
	}
    }
    for( int i = 0; i < sizeof(keep)/sizeof(NodeOp_t); ++i ) {
	filterSpec[keep[i]] = 1;
    }

    list<CElement>::iterator ptr;
    set<CNode*> cutList;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
	/*
	 * filter each section of code
	 */
	CNode* code = ptr->Code();
	/*
 	 * build cut list
 	 */
	code->PreVisit1( BuildCutList, &cutList );
	/*
 	 * delete cut elements from tree
 	 */
	code = code->PostSubVisit1( CutNodes, &cutList );
	outputList.push_back( CElement( ptr->Filename(), 
					ptr->Filename()!=NULL, code ) );
    }

}

/**
 * \endcond 
 */

