/******************************************************************************
 * 	Copyright (c) 2008 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   csim.cc
 *		- methods for simulation tool
 *
 ******************************************************************************
 */
						

#include <vector>
#include <set>
#include <fstream>
#include <stack>
#include "cobstack.h"
#include "cinstance.h"
#include "cnode.h"
#include "cmodule.h"
#include "cfunction.h"
#include "cblock.h"
#include "cvar.h"
#include "cnet.h"
#include "cgate.h"
#include "cparam.h"
#include "cport.h"
#include "csim.h"
#include "main.h"
#include "pliext.h"
#include "flags.h"
#include "stats.h"
#include "array.h"

using veriwell::simulator;


/*
 * memory management objects 
 */
static	CObstack 	theBackStack( "backStack" );
static	CObstack*	backStack = &theBackStack;

static stack<File*>     fileStack;

extern  char*		outputPath;

extern int		errorCount;

enum PEMode {
	peRVAL = 0,
	peLVAL_NET,
	peLVAL_REG_NET,
	pePATHIN,
	pePATHOUT,
	peTIMING
};

static tree ParseStatement( CNode* n );
static tree ParseStatementOrNull( CNode* n );
static tree ParseExpList( CNode* n );
static tree ParseExpression( CNode* n, PEMode mode = peRVAL, 
			     int noCheck = 0 );
static tree ParseList( NodeOp_t op, CNode* n, int createBuildTree, 
		PEMode mode = peRVAL, int noCheck = 0 );
static tree currentModule;
static stack<tree> caseStack;
static char* pliLibraryList = NULL;
inline void NI( int condition, const char* feature, Coord_t* coord ) {
    if( !condition ) {
        input_filename = coord->filename;
        lineno         = coord->lineno;
        stmt_lineno    = coord->lineno;
	string buffer;
	buffer = feature;
	buffer += " not supported";
        error( buffer.c_str(), NULL_CHAR, NULL_CHAR );
    }
}



/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
#ifndef BUILTIN_SIM_PLUGIN 
extern "C" {
CBackend* CreateToolInstance()
{
    return new CSim();
}
}
#endif

/************************************************
 	Constructor
 ************************************************/
CSim::CSim() :
	delayType( TYP_DELAY )
{
    switches.push_back( "+sim-interactive" );
    switchDescription["+sim-interactive"] =
                "start simulation in interactive mode only";

    switches.push_back( "+sim-compile-only" );
    switchDescription["+sim-compile-only"] =
                "do not simulate, compile only";

    switches.push_back( "+sim-trace" );
    switchDescription["+sim-trace"] =
                "enable execution tracing";

    switches.push_back( "+sim-mindelays" );
    switchDescription["+sim-mindelays"] =
                "use minimum delays";

    switches.push_back( "+sim-typdelays" );
    switchDescription["+sim-typdelays"] =
                "use typical delays";

    switches.push_back( "+sim-maxdelays" );
    switchDescription["+sim-maxdelays"] =
                "use maximum delays";

    switches.push_back( "+sim-pli=pliLib1+pliLib2+..." );
    switchDescription["+sim-pli=pliLib1+pliLib2+..."] =
                "load pli librarys";

#ifdef BUILD_HISTOGRAMS
    switches.push_back( "+sim-stats" );
    switchDescription["+sim-stats"] =
                "print runtime stats";
#endif

}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CSim::GetToolName( void )
{
	static char	toolName[] = "sim";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CSim::GetToolDescription( void )
{
	static char	toolDescription[] = "Verilog simulator";

	return toolDescription;
}

/************************************************
	ResolveInstance
	- return TRUE for any instance that need
	  to be resolved
 ************************************************/
int CSim::ResolveInstance( CModule* m, CInstance* inst )
{
    return !inst->IsGate();
}

/************************************************
 	GetEscapedIdentifier
	- return tree node with escaped identifier
 ***********************************************/
static tree GetEscapedIdentifier( CSymbol* sym )
{
    return get_identifier( sym->GetName() );
}
/************************************************
 	CVectorToTree
	- convert CVector to tree
 ***********************************************/

static tree CVectorToTree( CVector* v )
{
    tree result = NULL;
    radii base;
    char* buffer;
    int width = v->GetWidth();

    if( !v->Sized() && !v->Based() ) {
        result = build_int_cst(v->GetINT32());
	goto DONE;
    } 

    switch( v->GetPreferredBase() ) {
    case 2:
        base = BIN;
        buffer = v->GetBinary();
	break;
    case 8:
	base = OCT;
	buffer = v->GetOctal();
	break;
    case 10:
        base = DEC;
        buffer = v->GetDecimal();
        break;
    case 16:
        base = HEX;
        buffer = v->GetHex();
	break;
    case 0: {
        buffer = v->GetString();
        result = build_bit_cst(width, STRING_);
	int len = (width+7)/8;
        int ngroups = chars_to_groups(len);
        Group* g = BIT_CST_GROUP(result) + ngroups;

        int i = ((len - 1) % sizeof(Bit)) + 1;
        AVAL(g) = 0;
        for (int ii = i - 1; ii >= 0; ii--) {
#ifdef WORDS_BIGENDIAN 
            *(((char *) (&AVAL(g))) + sizeof(Bit) - 1 - ii) = *buffer++;
#else
            *(((char *) (&AVAL(g))) + ii) = *buffer++;
#endif
	}
        BVAL(g) = 0;
        --g;
        for (i = 1; i <= ngroups; i++, g--) {
	    BVAL(g) = 0;
            for (int ii = sizeof(Bit) - 1; ii >= 0; ii--) {
#ifdef WORDS_BIGENDIAN          //Endianness problem
                *(((char *) (&AVAL(g))) + sizeof(Bit) - 1 - ii) = *buffer++;
#else
                *(((char *) (&AVAL(g))) + ii) = *buffer++;
#endif
	    }

        } 
        } goto DONE;
    } 
    result = build_bit_cst(width, base);
    parse_base_const1g(base, buffer, width, strlen(buffer), 
					BIT_CST_GROUP(result));
DONE:
    return result;
}

/*******************************************************
 *   DeclareParam
 *   - declare parameter
 *********************************************************/
static tree DeclareParam( CParam* r ) 
{
    static tree rangeSpec = make_node(RANGE_HANDLE);

    input_filename = r->GetCoord()->filename;
    lineno = r->GetCoord()->lineno;
    stmt_lineno = r->GetCoord()->lineno;
    
    tree spec;
    CNode* lsb = NULL;
    CNode* msb = NULL;
    if( r->GetNodeType() == eB || r->GetNodeType() == eS ) {
        lsb = r->GetLsb();
        msb = r->GetMsb();
        if( lsb && !r->GetDataType()->IsScalar() ) {
            TREE_OPERAND(rangeSpec, 0) = ParseExpression(msb);
            TREE_OPERAND(rangeSpec, 1) = ParseExpression(lsb);
	}
    }
    spec = make_param_spec( lsb ? rangeSpec : NULL_TREE );
    NI( r->GetDataType()->GetNumberOfPackedDimensions() <= 1, 
		"packed declarations", r->GetCoord() );
    NI( !r->GetDataType()->GetSigned(), "signed parameters", r->GetCoord() );

    tree  decl;
    tree  name = GetEscapedIdentifier( r->GetSymbol() );
    int dims = r->GetNumberOfUnpackedDimensions();
    tree expr = ParseExpression( r->GetExpression() ); 
    if( dims == 0 ) {
        decl = make_param_decl (check_non_reg (name), spec, expr);
    } else {
        NI( FALSE, "parameter arrays", r->GetCoord() );
    }

    if( r->Specify() ) {
	TREE_SET_CODE(decl, SPECPARAM_DECL);
    }
    return decl;
}


/*******************************************************
 *   DeclareReal
 *   - declare  real
 *********************************************************/
static tree DeclareReal( CVar* r ) 
{
    static tree rangeSpec = make_node(RANGE_HANDLE);

    input_filename = r->GetCoord()->filename;
    lineno = r->GetCoord()->lineno;
    stmt_lineno = r->GetCoord()->lineno;
    
    tree spec;
    switch( r->GetDataType()->GetVarDataType() ) {
    case eREAL:
        spec = make_real_spec (NULL_TREE); 
        break;
    case eSHORTREAL:
	NI( FALSE, "shortreal variables", r->GetCoord() );
	break;
    case eREALTIME:
	NI( FALSE, "realtime variables", r->GetCoord() );
	break;
    default:
        ASSERT( FALSE );
    }
    	    
    tree  decl;
    tree  name = GetEscapedIdentifier( r->GetSymbol() );
    int dims = r->GetNumberOfUnpackedDimensions();
    if( dims == 0 ) {
        decl = make_decl(check_reg (name), spec, NULL_TREE, NULL_TREE);
    } else if( dims == 1 ) {
        tree msi = ParseExpression( r->GetUnpackedMsi(0) );
        tree lsi = ParseExpression( r->GetUnpackedLsi(0) );
        decl = make_decl (check_reg(name), spec, msi, lsi);
    } else {
	NI( FALSE, "multi-dimensional real arrays", r->GetCoord() );
    }

    return decl;
}
	
/*******************************************************
 *   DeclareEvent
 *   - declare event
 *********************************************************/
static void DeclareEvent( CVar* r ) 
{
    static tree rangeSpec = make_node(RANGE_HANDLE);

    input_filename = r->GetCoord()->filename;
    lineno = r->GetCoord()->lineno;
    stmt_lineno = r->GetCoord()->lineno;
    
    tree spec = make_event_spec();

    	    
    tree  decl;
    tree  name = GetEscapedIdentifier( r->GetSymbol() );
    int dims = r->GetNumberOfUnpackedDimensions();
    if( dims == 0 ) {
        decl = make_decl (check_non_reg (name), spec, NULL_TREE, NULL_TREE);
    } else {
	NI( FALSE, "multi-dimensional event arrays", r->GetCoord() );
    }

    BLOCK_DECL(current_scope) = chainon (decl, BLOCK_DECL(current_scope));
}

/*******************************************************
 *   DeclarePortDir
 *   - declare port declaration
 *********************************************************/
static tree DeclarePortDir( CPortDir* r, int inFunction ) 
{
    static tree rangeSpec = make_node(RANGE_HANDLE);

    input_filename = r->GetCoord()->filename;
    lineno = r->GetCoord()->lineno;
    stmt_lineno = r->GetCoord()->lineno;
    
    NI( r->GetDataType()->GetNumberOfPackedDimensions() <= 1, 
		"packed declarations", r->GetCoord() );
    
    tree spec;
    CNode* lsb = NULL;
    CNode* msb = NULL;
    if( r->GetDataType()->IsPacked() ) {
	if( !r->GetDataType()->IsScalar() ) {
            lsb = r->GetLsb();
            msb = r->GetMsb();
	}
        TREE_OPERAND(rangeSpec, 0) = ParseExpression(msb);
        TREE_OPERAND(rangeSpec, 1) = ParseExpression(lsb);
    }
    if( inFunction ) {
        spec = make_reg_spec( lsb ? rangeSpec : NULL_TREE );
    } else {
        spec = make_net_spec( default_net_type, lsb ? rangeSpec : 
						NULL_TREE, NULL_TREE );
    }
    switch( r->GetType() ) {
    case eINPUT:
        PORT_INPUT_ATTR(spec) = 1;
	break;
    case eOUTPUT:
        PORT_OUTPUT_ATTR(spec) = 1;
	break;
    case eINOUT:
        PORT_INPUT_ATTR(spec) = 1;
        PORT_OUTPUT_ATTR(spec) = 1;
	break;
    default: 
	ASSERT( FALSE );
    }
    	    
    tree  decl;
    tree  name = GetEscapedIdentifier( r->GetSymbol() );
    decl = make_decl(check_port(name), spec, NULL_TREE, NULL_TREE);
    return decl;
}

/*******************************************************
 *   DeclareReg
 *   - declare register
 *********************************************************/
static tree DeclareReg( CVar* r ) 
{
    static tree rangeSpec = make_node(RANGE_HANDLE);

    input_filename = r->GetCoord()->filename;
    lineno = r->GetCoord()->lineno;
    stmt_lineno = r->GetCoord()->lineno;
    
    NI( r->GetDataType()->GetNumberOfPackedDimensions() <= 1, 
			"packed declarations", r->GetCoord() );
    NI( r->GetDataType()->GetVarDataType() == eINTEGER || !r->GetDataType()->GetSigned(), 
		"signed register variables", r->GetCoord() );
    NI( !r->GetAutomatic(), "automatic register variables", r->GetCoord() );
    NI( !r->GetDataType()->GetTwoState(), "two state register variables", r->GetCoord() );
    
    tree spec;
    CNode* lsb = NULL;
    CNode* msb = NULL;
    if( r->GetDataType()->IsPacked() ) {
	if( !r->GetDataType()->IsScalar() ) {
            lsb = r->GetLsb();
            msb = r->GetMsb();
	}
        TREE_OPERAND(rangeSpec, 0) = ParseExpression(msb);
        TREE_OPERAND(rangeSpec, 1) = ParseExpression(lsb);
    }
    switch( r->GetDataType()->GetVarDataType() ) {
    case eTIME:
        spec = make_time_spec( rangeSpec );
        break;
    case eINTEGER:
        spec = make_integer_spec( rangeSpec );
        break;
    case eNO_VAR_TYPE:
    case eREG:
        spec = make_reg_spec( lsb ? rangeSpec : NULL_TREE );
        break;
    default:
	NI( FALSE, "variable type", r->GetCoord() );
    }
    	    
    tree  decl;
    tree  name = GetEscapedIdentifier( r->GetSymbol() );
    int dims = r->GetNumberOfUnpackedDimensions();
    if( dims == 0 ) {
        decl = make_decl(check_reg (name), spec, NULL_TREE, NULL_TREE);
    } else if( dims == 1 ) {
        tree msi = ParseExpression( r->GetUnpackedMsi(0) );
        tree lsi = ParseExpression( r->GetUnpackedLsi(0) );
        decl = make_decl (check_reg(name), spec, msi, lsi);
    } else {
        NI( FALSE, "multi-dimensional register arrays", r->GetCoord() );
    }

    return decl;
}

/*******************************************************
 *   DeclareNet
 *   - declare net
 *********************************************************/
static tree DeclareNet( CNet* n ) 
{
    static tree rangeSpec = make_node(RANGE_HANDLE);


    input_filename = n->GetCoord()->filename;
    lineno = n->GetCoord()->lineno;
    stmt_lineno = n->GetCoord()->lineno;
     
    NI( n->GetDataType()->GetNumberOfPackedDimensions() <= 1, 
		"packed declarations", n->GetCoord() );
    NI( !n->GetDataType()->GetSigned(), "signed nets", n->GetCoord() );
    
    CNode* lsb = NULL;
    CNode* msb = NULL;
    if( !n->GetDataType()->IsScalar() ) {
        lsb = n->GetLsb();
        msb = n->GetMsb();
    }
    tree spec = NULL;
    TREE_OPERAND(rangeSpec, 0) = NULL;
    TREE_OPERAND(rangeSpec, 1) = NULL;
    NET_SCALARED_ATTR(rangeSpec) = 0;
    NET_VECTORED_ATTR(rangeSpec) = 0;
    if( lsb ) {
        TREE_OPERAND(rangeSpec, 0) = ParseExpression(msb);
        TREE_OPERAND(rangeSpec, 1) = ParseExpression(lsb);
	spec = rangeSpec;
    } 
    enum tree_type nettype;
    switch( n->GetWireAttr() ) {
    case eWIRE:
	nettype = NET_WIRE_TYPE;
	break;
    case eTRI:
	nettype = NET_TRI_TYPE;
	break;
    case eWAND:
	nettype = NET_WAND_TYPE;
	break;
    case eTRIAND:
	nettype = NET_TRIAND_TYPE;
	break;
    case eWOR:
	nettype = NET_WOR_TYPE;
	break;
    case eTRIOR:
	nettype = NET_TRIOR_TYPE;
	break;
    case eTRI1:
	nettype = NET_TRI1_TYPE;
	break;
    case eTRI0:
	nettype = NET_TRI0_TYPE;
	break;
    case eTRIREG:
	nettype = NET_TRIREG_TYPE;
	break;
    case eSUPPLY0:
	nettype = NET_SUPPLY0_TYPE;
	break;
    case eSUPPLY1:
	nettype = NET_SUPPLY1_TYPE;
	break;
    default:
	ASSERT( FALSE );
    }

    if( n->GetScalared() ) {
	NET_SCALARED_ATTR(spec) = 1;
    }
    if( n->GetVectored() ) {
	NET_VECTORED_ATTR(spec) = 1;
    }
    spec = make_net_spec( nettype, spec, NULL );
    	    
    tree  decl;
    tree  name = GetEscapedIdentifier( n->GetSymbol() );
    int dims = n->GetNumberOfUnpackedDimensions();
    if( dims == 0 ) {
        decl = make_decl(check_net (name), spec, NULL_TREE, NULL_TREE);
    } else {
        NI( FALSE, "multi-dimensional net arrays", n->GetCoord() );
    }

    return decl;
}

/*******************************************************
 *   ParseBlockDecls
 *   - convert declarations
 *********************************************************/

static void ParseBlockDecls( CBlock* block, int inFunction )
{
    tree plist = NULL;
    tree list = NULL;

    NI( block->GetGenvarList()->size() == 0 , "genvar variables", block->GetCoord() );

    /*
     * declare parameters
     */
    vector<CParam*>* paramList = block->GetParamList();
    vector<CParam*>::iterator pPtr;
    for( pPtr = paramList->begin(); pPtr != paramList->end(); ++pPtr ) {
	tree param = DeclareParam( *pPtr );
        if( (*pPtr)->Specify() ) {
            MODULE_SPECDEFS(currentModule) = chainon( param,
					MODULE_SPECDEFS(currentModule));
    	} else {
           list = chainon(param,list);
    	}
    }

    vector<CPortDir*>::reverse_iterator dptr;
    vector<CPortDir*>* portdirs = block->GetPortDirList();
    for( dptr = portdirs->rbegin(); dptr != portdirs->rend(); ++dptr ) {
        tree decl = DeclarePortDir( *dptr, inFunction );    
        BLOCK_PORTS(current_scope) = chainon(decl, 
					BLOCK_PORTS(current_scope));
    }

    /*
     * declare registers
     */
    vector<CVar*>* regList = block->GetVarList();
    vector<CVar*>::iterator rPtr;
    for( rPtr = regList->begin(); rPtr != regList->end(); ++rPtr ) {
	if( (*rPtr)->GetNodeType() != eB && (*rPtr)->GetNodeType() != eS ) {
	    continue;
	}
	list = chainon( DeclareReg( *rPtr ), list );
    }

    /*
     * declare reals
     */
    vector<CVar*>* realList = block->GetVarList();
    vector<CVar*>::iterator rlPtr;
    for( rlPtr = realList->begin(); rlPtr != realList->end(); ++rlPtr ) {
	if( (*rlPtr)->GetNodeType() != eR ) {
	    continue;
	}
	list = chainon( DeclareReal( *rlPtr ), list );
    }

    /*
     * declare nets
     */
    vector<CNet*>* netList = block->GetNetList();
    vector<CNet*>::iterator nPtr;
    for( nPtr = netList->begin(); nPtr != netList->end(); ++nPtr ) {
	/*
 	 * do not declare nets undefined in source
 	 * or continuous assignments
 	 */
	if( !(*nPtr)->GetUndefined() ) {
	    list = chainon( DeclareNet( *nPtr ), list );
	}
    }

    list = chainon(plist,nreverse(list));
    BLOCK_DECL(current_scope) = chainon(list,BLOCK_DECL(current_scope));
}
/******************************************************
 *   ParseList
 *   - convert tree of op to expression list 
 *********************************************************/

static tree ParseList( NodeOp_t op, CNode* n, int createBuildTree,
		PEMode mode, int noCheck )
{
        tree   exp1;
        tree   exp2;
	
	if( n->GetOp() != op ) {
	    switch( n->GetOp() ) {
	    case eCOMMENT:
	    case ePRAGMA:
	    case eVRQ:
		return NULL;
	    default:
		if( createBuildTree ) {
	            return build_tree_list( ParseExpression( n, mode, noCheck ), NULL );
		} else {
	            return ParseExpression( n, mode, noCheck );
		}
	    }
	}

	exp1 = ParseList( op, n->Arg<CNode*>(0), createBuildTree, mode,
			noCheck );
	exp2 = ParseList( op, n->Arg<CNode*>(1), createBuildTree, mode,
			noCheck );
	return chainon( exp1, exp2 );
}


/************************************************
 	ParseExpression 
	- parse expression code
 ***********************************************/

static tree ParseExpression( CNode* n, PEMode mode, int noCheck )
{
    tree result = NULL;
    
    if( !n ) {
	goto DONE;
    }

    if( n->GetCoord() ) {
        input_filename = n->GetCoord()->filename;
        lineno = n->GetCoord()->lineno;
    }
    switch( n->GetOp() ) {
    case eELIST: {
	CNode* arg0 = n->Arg<CNode*>(0);
	tree stmt1 = ParseExpression( arg0 );
	if( arg0->GetOp() != eELIST ) {
	   stmt1 = build_tree_list( stmt1, NULL );
	}
	CNode* arg1 = n->Arg<CNode*>(1);
	tree stmt2 = ParseExpression( arg1 );
	if( arg1->GetOp() != eELIST ) {
	   stmt2 = build_tree_list( stmt2, NULL );
	}
	if( stmt1 && stmt2 ) {
	    result = chainon(stmt1,stmt2);
	} else if( stmt1 ) {
	    result = stmt1;
	} else  {
	    result = stmt2;
	}
	} break;
    case eRCONSTANT:
	result = build_real_cst( n->Arg<char*>(0) );
	break;
    case eVCONSTANT:
	result = CVectorToTree( n->Arg<CVector*>(0) );
	break;
    case eNEG: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (NEGATE_EXPR, exp1);
	} break;
    case ePLUS: {
	result = ParseExpression( n->Arg<CNode*>(0) );
	} break;
    case eNOT: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (TRUTH_NOT_EXPR, exp1);
	} break;
    case eCOM: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (BIT_NOT_EXPR, exp1);
	} break;
    case eRAND: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (AND_REDUCE_EXPR, exp1);
	} break;
    case eRNAND: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (NAND_REDUCE_EXPR, exp1);
	} break;
    case eROR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (OR_REDUCE_EXPR, exp1);
	} break;
    case eRNOR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (NOR_REDUCE_EXPR, exp1);
	} break;
    case eRXOR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (XOR_REDUCE_EXPR, exp1);
	} break;
    case eRXNOR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
        result = build_unary_op (XNOR_REDUCE_EXPR, exp1);
	} break;
    case eADD: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (PLUS_EXPR, exp1, exp2);
	} break;
    case eSUB: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (MINUS_EXPR, exp1, exp2);
	} break;
    case eMUL: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (MULT_EXPR, exp1, exp2);
	} break;
    case ePOW: {
	NI( FALSE, "** operator", n->GetCoord() );
	} break;
    case eDIV: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (DIV_EXPR, exp1, exp2);
	} break;
    case eMOD: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (MOD_EXPR, exp1, exp2);
	} break;
    case eEQ: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (EQ_EXPR, exp1, exp2);
	} break;
    case eNE: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (NE_EXPR, exp1, exp2);
	} break;
    case eCEQ: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (EQ_CASE_EXPR, exp1, exp2);
	} break;
    case eCNE: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (NE_CASE_EXPR, exp1, exp2);
	} break;
    case eLAND: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (TRUTH_ANDIF_EXPR, exp1, exp2);
	} break;
    case eLOR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (TRUTH_ORIF_EXPR, exp1, exp2);
	} break;
    case eLT: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (LT_EXPR, exp1, exp2);
	} break;
    case eLE: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (LE_EXPR, exp1, exp2);
	} break;
    case eGT: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (GT_EXPR, exp1, exp2);
	} break;
    case eGE: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (GE_EXPR, exp1, exp2);
	} break;
    case eAND: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (BIT_AND_EXPR, exp1, exp2);
	} break;
    case eOR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (BIT_OR_EXPR, exp1, exp2);
	} break;
    case eXOR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (BIT_XOR_EXPR, exp1, exp2);
	} break;
    case eXNOR: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (BIT_XNOR_EXPR, exp1, exp2);
	} break;
    case eLSH: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (LSHIFT_EXPR, exp1, exp2);
	} break;
    case eRSH: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
        result = build_binary_op (RSHIFT_EXPR, exp1, exp2);
	} break;
    case eHOOK: {
	tree exp1 = ParseExpression( n->Arg<CNode*>(0) );
	tree exp2 = ParseExpression( n->Arg<CNode*>(1) );
	tree exp3 = ParseExpression( n->Arg<CNode*>(2) );
        result = build_cond_expr(exp1, exp2, exp3);
	} break;
    case eREP: {
	result = make_node(CONCAT_REP_REF);
	CONCAT_EXPR(result) = ParseExpression( n->Arg<CNode*>(0) );
	CONCAT_LIST(result) = ParseExpList( n->Arg<CNode*>(1) );
	concat_labels(result);
        } break;
    case eUCAT: {
	result = make_node(CONCAT_REF);
	CONCAT_LIST(result) = ParseExpList( n->Arg<CNode*>(0) );
	concat_labels(result);
        } break;
    case eCAT: {
        result = make_node (CONCAT_REF);
        CONCAT_LIST(result) = nreverse(ParseList( n->GetOp(), n, TRUE,
		                mode, noCheck ));
        concat_labels(result);
        } break;
    case ePORT_REF:
    case ePARAM_REF:
    case eVAR_REF:
    case eNET_REF: {
	    tree name = GetEscapedIdentifier( n->Arg<CDecl*>(0)->GetSymbol() );
	    CDataType* dataType = n->Arg<CDecl*>(0)->GetDataType();
	    if( dataType && dataType->GetVarDataType() == eDEVENT ) {
	        result = check_named_event(name);
		break;
	    }
	    switch( mode ) {
	    case peRVAL:
		if( noCheck ) {
		    result = check_rval_nocheck(name);
		} else {
		    result = check_rval(name);
		}
		break;
	    case pePATHIN:
		result = check_input_port(name);
		break;
	    case pePATHOUT:
		result = check_output_port(name);
		break;
	    case peTIMING:
		check_input_port(name);
		result = check_rval(name);
		break;
	    case peLVAL_NET:
		if( noCheck ) {
		    result = check_lval_nocheck(name, LVAL_NET, NULL);
		} else {
		    result = check_lval(name, LVAL_NET, NULL);
		}
		break;
	    case peLVAL_REG_NET:
		if( noCheck ) {
		    result = check_lval_nocheck(name, LVAL_REG_NET, NULL);
		} else {
		    result = check_lval(name, LVAL_REG_NET, NULL);
		}
		break;
	    default:
		ASSERT( FALSE );
	    }
	} break;
    case eARRAY: {
	/*
 	 * Only single dimension reg or external arrays are supported
 	 */
	result = ParseExpression( n->Arg<CNode*>(0), mode, 1 );
	int indexCount = ListCount(n->Arg<CNode*>(1));
	int hasSelect = indexCount > 0;
	if( n->Arg<CNode*>(0)->GetOp() != eEXTERNAL_REF &&
	    n->Arg<CNode*>(0)->GetOp() != eMEMBER ) {
	    CDecl* decl = Array::ArrayDecl( n->Arg<CNode*>(0) );
	    hasSelect = (decl->GetNumberOfUnpackedDimensions() < indexCount);
	    int bad = 0;
	    switch( decl->GetClass() ) {
	    case eVAR:
	    	if( (!hasSelect && indexCount > 1) || 
		    indexCount > 2 ) {	
		    bad = 1;
		}
		break;
	    case eNET:
	    case ePARAM:
	    	if( (!hasSelect && indexCount > 0) || 
		    indexCount > 1 ) {	
		    bad = 1;
		}
		break;
	    default:
		bad = 1;
	    }
		
	    if( bad ) {
                error( "only register arrays are supported", NULL_CHAR, NULL_CHAR );
	    }
	}

	if( indexCount > 2 ||
	    (!hasSelect && indexCount == 2) ) {
            error( "multi-dimensional arrays are not supported",
                                                  NULL_CHAR, NULL_CHAR );
	    break;
	}

	// ??? mdh - multi-dimensional arrays are not supported
	vector<CNode*> indexes;
	EList2VectorExclude( n->Arg<CNode*>(1), 
			     set<NodeOp_t>(), indexes );
	if( indexes.size() ) {
	    switch( indexes.back()->GetOp() ) {
	    case eSLICE: {
	        tree msb = ParseExpression( indexes.back()->Arg<CNode*>(0) );
	        tree lsb = ParseExpression( indexes.back()->Arg<CNode*>(1) );
                result = build_part_ref(result, msb, lsb); 
		}
		break;
	    case ePSLICE:
		NI( FALSE, ":+ operator", indexes.back()->GetCoord() );
		break;
	    case eMSLICE:
		NI( FALSE, ":- operator", indexes.back()->GetCoord() );
		break;
	    default:
	        tree bit = ParseExpression( indexes.back() );
                result = build_bit_ref(result, bit); 
	    }
	    indexes.pop_back();
	}
	if( indexes.size() ) {
	    tree bit = ParseExpression( indexes.back() );
	    indexes.pop_back();
            result = build_bit_ref(result, bit); 
	}
	} break;
    case eSYSTASK_CALL: {
	tree args = ParseExpList( n->Arg<CNode*>(1) );
	result = build_nt( SYSFUNCTION_REF, n->Arg<CSymbol*>(0)->GetName(), 
				args, NULL, NULL, lineno, input_filename );
	} break;
    case eFUNCTION_CALL: {
	CSymbol* sym = n->Arg<CSymbol*>(0);
	tree name = GetEscapedIdentifier( sym );
	tree args = ParseExpList( n->Arg<CNode*>(1) );
	result = build_function_call( name, args );
        HIERARCHICAL_ATTR(result) = sym->External();
        UNKNOWN_TYPE_ATTR(result) = 1;
	} break;
   case eMEMBER:
   case eEXTERNAL_REF: {
        CSymbol* symbol = CSymbol::Lookup( HierarchicalReference2String(n).c_str() );
	int hierarchical = symbol->External();
	result = get_identifier( symbol->GetName() );			
	HIERARCHICAL_ATTR(result) = hierarchical;
        UNKNOWN_TYPE_ATTR(result) = hierarchical;
	TREE_CONSTANT_ATTR(result) = hierarchical;  
	} break;	
   case eCASEITEM: {
    	tree stmt = ParseStatementOrNull( n->Arg<CNode*>(1) );
	if( n->Arg<CNode*>(0) ) {
    	    tree elist = ParseList( eELIST, n->Arg<CNode*>(0), TRUE );
    	    result = build_tree_list( elist, stmt );
	} else {
	    STMT_CASE_DEFAULT(caseStack.top()) = stmt;
	}
	} break;
    case eMTM: {
	CNode* exp = NULL;
	switch( simulator.DelayType() ) {
	case MIN_DELAY:
	    exp = n->Arg<CNode*>(0);
	    break;
	case TYP_DELAY:
	    exp = n->Arg<CNode*>(1);
	    break;
	case MAX_DELAY:
	    exp = n->Arg<CNode*>(2);
	    break;
	default:
	    ASSERT( FALSE );
	}
	result = ParseExpression( exp );
	} break;
    case eFWD_REF:
	// this should never occur
	ASSERT( FALSE );
 	break;
    case eMSLICE:
	NI( FALSE, ":+ operator", n->GetCoord() );
	break;
    case ePSLICE:
	NI( FALSE, ":- operator", n->GetCoord() );
	break;
    case eLSHA:
	NI( FALSE, "<<< operator", n->GetCoord() );
	break;
    case eRSHA:
	NI( FALSE, ">>> operator", n->GetCoord() );
	break;
    default:
	ASSERT( FALSE );
    }

DONE:
    return result;
}

/************************************************
 	ParsePortExpression 
	- parse port expression code
 ***********************************************/

static tree ParsePortExpression( CNode* n )
{
    tree result = NULL;
    
    if( !n ) {
	goto DONE;
    }

    input_filename = n->GetCoord()->filename;
    lineno = n->GetCoord()->lineno;
    switch( n->GetOp() ) {
    case eCAT: {
	NI( FALSE, "port concatination", n->GetCoord() );
        } break;
    case ePORT_REF:
    case ePARAM_REF:
    case eVAR_REF:
    case eNET_REF: {
	    result = GetEscapedIdentifier( n->Arg<CDecl*>(0)->GetSymbol() );
	    set_decl( result, result );
	} break;
    case eARRAY: {
	tree ref = ParsePortExpression( n->Arg<CNode*>(0) );
	CNode* selectExpression = n->Arg<CNode*>(1);
	switch( selectExpression->GetOp() ) {
	case eELIST:
	    NI( FALSE, "port array", n->GetCoord() );
	    ASSERT( FALSE );
	    break;
	case eSLICE: {
	   tree msb = ParseExpression( selectExpression->Arg<CNode*>(0) );
	   tree lsb = ParseExpression( selectExpression->Arg<CNode*>(1) );
           result = build_part_ref(ref, msb, lsb); 
	   } break;
	default:
	   tree select = ParseExpression( n->Arg<CNode*>(1) );
           result = build_bit_ref(ref, select); 
	   break;
	}
	set_decl( ref, result );
	} break;
    default:
	ASSERT( FALSE );
    }

DONE:
    return result;
}

/************************************************
 	ParseEventExpression 
	- parse event expression code
 ***********************************************/
         
static tree ParseEventExpression( CNode* n )
{
    if( !n ) {
	NI( FALSE, "@* event expression", n->GetCoord() );
    }
   
    switch( n->GetOp() ) {
    case ePOSEDGE: {
	in_event = 1;
	tree exp = ParseExpression( n->Arg<CNode*>(0) );
	in_event = 0;
        return  build_tree_list
                        (build_unary_op (POSEDGE_EXPR, exp), NULL_TREE);
	} break;
    case eNEGEDGE: {
	in_event = 1;
	tree exp = ParseExpression( n->Arg<CNode*>(0) );
	in_event = 0;
        return  build_tree_list
                        (build_unary_op (NEGEDGE_EXPR, exp), NULL_TREE);
	} break;
    case eEVOR: {
	tree exp1 = ParseEventExpression( n->Arg<CNode*>(0) );	
	tree exp2 = ParseEventExpression( n->Arg<CNode*>(1) );	
	return chainon( exp1, exp2 );
	}
    case eEVENT_CONTROL:
	NI( n->Arg<CNode*>(0)!=NULL, "@* event expression", n->GetCoord() );
	return ParseEventExpression( n->Arg<CNode*>(0) );
    default: {
	if( n->GetOp() == eVAR_REF && n->Arg<CVar*>(0)->GetDataType()->GetVarDataType() == eDEVENT ) {
	    in_event = 1;
	    tree exp = ParseExpression( n );
	    in_event = 0;
            return build_tree_list(exp, NULL_TREE);
	} else {
	    in_event = 1;
	    tree exp = ParseExpression( n );
	    in_event = 0;
            return build_tree_list(build_unary_op (ANYEDGE_EXPR, exp), 
							NULL_TREE);
	}
	}
    }
} 

/************************************************
 	ParseArglist 
	- parse expression list 
 ***********************************************/

static tree ParseArgList( CNode* n )
{
    tree result = NULL;
    
    if( !n ) {
	goto DONE;
    }

    input_filename = n->GetCoord()->filename;
    lineno = n->GetCoord()->lineno;
    stmt_lineno = n->GetCoord()->lineno;
    switch( n->GetOp() ) {
    case eELIST: {
	tree exp1 = ParseArgList( n->Arg<CNode*>(0) );
	if( !exp1 ) {
	    exp1 = build_tree_list( exp1, NULL );
	}
	tree exp2 = ParseArgList( n->Arg<CNode*>(1) );
	if( !exp2 ) {
	    exp2 = build_tree_list( exp2, NULL );
	}
	result = chainon( exp1, exp2 );
	} break;
    case eARG: {
	result = ParseExpression( n->Arg<CNode*>(1) );
	if( result ) {
	    result = build_tree_list( result, NULL );
	}
	} break;
    default: 
	ASSERT( FALSE );
    }
DONE:
    return result;
}

/************************************************
 	ParseSpecifylist 
	- parse specify input/output list 
 ***********************************************/

static tree ParseSpecifyList( CNode* n, int output )
{
    tree result = NULL;
    
    if( !n ) {
	goto DONE;
    }

    input_filename = n->GetCoord()->filename;
    lineno = n->GetCoord()->lineno;
    stmt_lineno = n->GetCoord()->lineno;
    switch( n->GetOp() ) {
    case eLIST: {
	tree exp1 = ParseSpecifyList( n->Arg<CNode*>(0), output );
	if( !exp1 ) {
	    exp1 = build_tree_list( exp1, NULL );
	}
	tree exp2 = ParseSpecifyList( n->Arg<CNode*>(1), output );
	if( !exp2 ) {
	    exp2 = build_tree_list( exp2, NULL );
	}
	result = chainon( exp1, exp2 );
	} break;
    default: {
	result = ParseExpression( n, output ? pePATHOUT : pePATHIN );
	result = build_tree_list( result, NULL );
	} break;
    }
DONE:
    return result;
}

/************************************************
 	ParseExplist 
	- parse expression list 
 ***********************************************/

static tree ParseExpList( CNode* n )
{
    tree result = NULL;
    
    if( !n ) {
	goto DONE;
    }

    input_filename = n->GetCoord()->filename;
    lineno = n->GetCoord()->lineno;
    stmt_lineno = n->GetCoord()->lineno;
    switch( n->GetOp() ) {
    // ??? mdh - need to fix parser to not use eLIST
    case eLIST:
    case eELIST: {
	tree exp1 = ParseExpList( n->Arg<CNode*>(0) );
	if( !exp1 ) {
	    exp1 = build_tree_list( exp1, NULL );
	}
	tree exp2 = ParseExpList( n->Arg<CNode*>(1) );
	if( !exp2 ) {
	    exp2 = build_tree_list( exp2, NULL );
	}
	result = chainon( exp1, exp2 );
	} break;
    default: {
	result = ParseExpression( n );
	if( result ) {
	    result = build_tree_list( result, NULL );
	}
	} break;
    }
DONE:
    return result;
}
/************************************************
 	List2TreeList 
	- convert chain of objects to tree list
 ***********************************************/
static tree List2TreeList( tree t )
{
    tree head = NULL;
    tree tail = NULL;
    while( t ) {
	tree t1 = t;
	t = TREE_CHAIN(t);
	TREE_CHAIN(t1) = NULL;
	tree tnew = build_tree_list( t1, NULL_TREE );
	if( !head ) {
	    head = tnew;
	} else {
	    TREE_CHAIN(tail) = tnew;
	}
	tail = tnew;
    }

    return head;
}

/************************************************
 	ListSize
	- count number of elements in list
 ***********************************************/
static int ListSize( CNode* n ) {
    if( !n ) {
	return 0;
    }

    int len = 0;
    switch( n->GetOp() ) {
    case eELIST:
    case eLIST:
	len += ListSize( n->Arg<CNode*>(0) );
	len += ListSize( n->Arg<CNode*>(1) );
	break;
    default:
	len = 1;
	break;
    }
    return len;
}
/************************************************
 	ListElement
	- return list element index (0 based)
 ***********************************************/
static CNode* ListElement_1( CNode* n, int* count )
{
    CNode* result;
    if( !n ) {
	return NULL;
    }

    switch( n->GetOp() ) {
    case eLIST:
    case eELIST:
	result = ListElement_1( n->Arg<CNode*>(0), count );
	if( !*count ) {
	    break;
	}
	result = ListElement_1( n->Arg<CNode*>(1), count );
	break;
    default:
        (*count)--;
	if( !*count ) {
	    result = n;
	}
	break;
    }
    return result;
}

static CNode* ListElement( CNode* n, int index )
{
    int count = index+1;
    return ListElement_1(n,&count);
}

/************************************************
 	ParseDelayExpression 
	- parse delay expressions
 ***********************************************/

static tree ParseDelayExpression( CNode* delay ) 
{
    if( !delay ) {
	return NULL;
    }

    tree delayExpr = NULL;
    int listLength = ListSize(delay);
    switch( listLength ) {
    case 1: {
        tree d = ParseExpression( delay ); 
    	delayExpr = build_nt(DELAY_EXPR, d, d, NULL );
	} break;
    case 2: {
        tree d1 = ParseExpression( ListElement(delay,0) ); 
        tree d2 = ParseExpression( ListElement(delay,1) ); 
    	delayExpr = build_nt(DELAY_EXPR, d1, d2, NULL );
	} break;
    case 3: {
        tree d1 = ParseExpression( ListElement(delay,0) ); 
        tree d2 = ParseExpression( ListElement(delay,1) ); 
       tree d3 = ParseExpression( ListElement(delay,2) ); 
    	delayExpr = build_nt(DELAY_EXPR, d1, d2, d3 );
	} break;
    default:
	ASSERT( FALSE );
    }
    TREE_LABEL(delayExpr) = listLength;

    return delayExpr;
}

/************************************************
 	CreateContinuousAssignment 
	- create continuous assignment statement
 ***********************************************/
static void CreateContinuousAssignment( CNode* n, int contAssign )
{
	// ??? mdh - strength not supported
	StrengthPair_t* strength = n->Arg<StrengthPair_t*>(0);
	if( strength->s0 != eUNDEFINED || 
	    strength->s1 != eUNDEFINED ) {
	    warning ("Strength specification is not supported and is ignored", 
			NULL_CHAR, NULL_CHAR);
	}
	tree delay = NULL;
	if( n->Arg<CNode*>(1) ) {
	    ASSERT( n->Arg<CNode*>(1)->GetOp() == eDELAY_CONTROL );
	    delay = ParseDelayExpression( n->Arg<CNode*>(1)->Arg<CNode*>(0) ); 
	}
	tree lval = ParseExpression( n->Arg<CNode*>(2), peLVAL_NET ); 
	if( contAssign ) {
	    NET_ASSIGN_ATTR(lval) = 1;
	}
	tree exp = ParseExpression( n->Arg<CNode*>(3) ); 
        tree stmt = build_cont_assign ( lval, exp, lineno, delay, delay!=0);
	stmt = build_tree_list( stmt, NULL );
        MODULE_ASSIGNMENTS (currentModule) =
                        chainon(stmt,MODULE_ASSIGNMENTS(currentModule));

}

/************************************************
 	ParseModule 
	- parse module structure
 ***********************************************/

static void ParseModule( CModule* module ) 
{
	// ??? mdh lots of stuff needs to be added
        NI( !module->InlineParamDecls(), "inline parameters", module->GetCoord() );
        NI( !module->InlinePortDecls(), "inline port declarations", module->GetCoord() );
        NI( !module->Protect(), "`protect", module->GetCoord() );
        NI( module->UnconnectedDrive() == vpiHiZ, "`unconnected_drive", module->GetCoord() );

        /* keep track of the smallest precision in the model. */
	if( module->HasTimescale() ) {
    	    timescale_global.unit = -module->Unit();
            timescale_global.prec = -module->Precision();
            timescale_global.smallest_scale =
                MAX(timescale_global.smallest_scale, -module->Precision() );
            timescale_global.any = T_YES; 
	}

	/* set default net type */
	switch( module->DefaultNetType() ) {
        case vpiWire:
	    default_net_type = NET_WIRE_TYPE;
            break;
        case vpiWand:
	    default_net_type = NET_WAND_TYPE;
            break;
        case vpiWor:
	    default_net_type = NET_WOR_TYPE;
            break;
        case vpiTri:
	    default_net_type = NET_TRI_TYPE;
            break;
        case vpiTri0:
	    default_net_type = NET_TRI0_TYPE;
            break;
        case vpiTri1:
	    default_net_type = NET_TRI1_TYPE;
            break;
        case vpiTriReg:
	    default_net_type = NET_TRIREG_TYPE;
            break;
        case vpiTriAnd:
	    default_net_type = NET_TRIAND_TYPE;
            break;
        case vpiTriOr:
	    default_net_type = NET_TRIOR_TYPE;
            break;
        default:
            MASSERT( FALSE );
        }


	tree defname = GetEscapedIdentifier( module->GetSymbol() );
        currentModule = build_module(defname);
	current_scope = currentModule;
        push_scope ();
	vector<CPort*>::iterator pptr;
	vector<CPort*>* ports = module->GetPortList();
	tree tports = NULL;
 	for( pptr = ports->begin(); pptr != ports->end(); ++pptr ) {
	    tree next;
	    tree exp = ParsePortExpression( (*pptr)->GetConnection() );
	    CSymbol* sym = (*pptr)->GetSymbol();
	    tree name = NULL;
	    if( sym ) {
		name = GetEscapedIdentifier((*pptr)->GetSymbol());
	    }
	    // ??? - this matches the original parser, but
	    // doesn't appear to support (,,,a,b,c)
	    if( name || exp ) {
		 if( exp ) {
	             next = build_tree_list( exp, name );	
		 } else {
	             next = build_tree_list( name, NULL );	
		 }
	         tports = chainon( tports, next );
	    }
	}
	MODULE_PORT_LIST(currentModule) = tports; 
        module_list = tree_cons (currentModule, NULL_TREE, module_list);
	ParseBlockDecls( module, FALSE );
	ParseStatement( module->GetCodeList() );
        current_scope = pop_scope ();
        end_module (currentModule);
}

/************************************************
 	ParseUdpStatements
	- parse udp statements
 ***********************************************/

static tree currentUdpString;
static tree ParseUdpStatements( CNode* n, int sequential ) 
{
    tree result = NULL;
    if( !n ) {
	return NULL;
    }

    switch( n->GetOp() ) {
    case eLIST: {
	tree stmt1 = ParseUdpStatements( n->Arg<CNode*>(0), sequential );
	tree stmt2 = ParseUdpStatements( n->Arg<CNode*>(1), sequential );
	if( stmt1 && stmt2 ) {
	    result = chainon(stmt1,stmt2);
	} else if( stmt1 ) {
	    result = stmt1;
	} else  {
	    result = stmt2;
	}
	} break;
    case eINIT: {
	CNode* assign = n->Arg<CNode*>(0);
	ASSERT( assign->GetOp() == eASSIGN );
	CNode* lhs = assign->Arg<CNode*>(1);
	CNode* rhs = assign->Arg<CNode*>(2);
	ASSERT( lhs->GetOp() == eVAR_REF );
	ASSERT( rhs->GetOp() == eVCONSTANT );
	const char* name = lhs->Arg<CVar*>(0)->GetName();
	int   value = rhs->Arg<CVector*>(0)->GetINT32();
        if( UDP_REG_NAME(currentModule)==NULL_TREE ) {
            error( "initial statement is no allowed in combinatorial udp's",
                                      		NULL_CHAR, NULL_CHAR );
        } else if( strcmp(name,IDENT(UDP_REG_NAME(currentModule)))!=0 ) {
            error( "initial statement does not reference port output",
                                                NULL_CHAR, NULL_CHAR );
        } else {
           UDP_INITIAL_VALUE(currentModule) = build_int_cst(value);
        }
	} break;
    case eTABLE: {
	result = ParseUdpStatements( n->Arg<CNode*>(0), sequential );
	} break;
    case eTABLE_ENTRY: {
	currentUdpString = NULL;
	ParseUdpStatements( n->Arg<CNode*>(0), sequential );
	int limit = sizeof(UDP_STRING_STRING(currentUdpString));
	char* ss = UDP_STRING_STRING(currentUdpString);
  	int length = strlen(ss);
        if(!sequential && length < limit + 2) {
	    ss[length+2] = 0;
	    ss[length+1] = ss[length-1];
	    ss[length] = ss[length-2];
	    ss[length-1] = ':';
	    ss[length-2] = ':';
        } else if(sequential && length < limit + 4) {
	    ss[length+4] = 0;
	    ss[length+3] = ss[length-1];
	    ss[length+2] = ss[length-2];
	    ss[length+1] = ':';
	    ss[length] = ':';
	    ss[length-1] = ss[length-3];
	    ss[length-2] = ss[length-4];
	    ss[length-3] = ':';
	    ss[length-4] = ':';
        }
	result = currentUdpString;
        validate_udp_string( currentModule, result );
	} break;
    case eTABLE_SYMBOL: {
	char* s = n->Arg<char*>(0);
	char c1;
	char c2;
	if( strlen(s) == 1 ) {
	    c1 = c2 = s[0];
	} else {
	    /* 
 	     * ? and b must be convert different but equivilent characters
             * otherwise edges cannot be distinguished from levels 
             */
	     c1 = s[1];
	     c2 = s[2];
             if( c1=='?' ) {
                 c1 = '!';
             } else if( c1=='b' ) {
                 c1 = '%';
             }
        }
	append_udp_digits( &currentUdpString, c1, c2 );
	} break;
    default:
	break;
    }
    return result;
}

/************************************************
 	ParseUdp 
	- parse udp structure
 ***********************************************/

static void ParseUdp( CModule* module ) 
{
    // ??? mdh lots of stuff needs to be added
    NI( !module->InlineParamDecls(), "inline parameter declarations", module->GetCoord() );
    NI( !module->InlinePortDecls(), "inline port declarations", module->GetCoord() );
    // current pli calls don't support cells so ignore
    //ASSERT( !module->Cell() );
    NI( !module->Protect(), "`protect", module->GetCoord() );
    NI( module->DefaultNetType() == vpiWire, "`defaultnettype", module->GetCoord() );
    NI( module->UnconnectedDrive() == vpiHiZ, "`unconnecteddrive", module->GetCoord() );
    
    int sequential = FALSE;
    tree defname = GetEscapedIdentifier( module->GetSymbol() );
    currentModule = current_scope = build_udp( defname );
    push_scope ();
    
    /*
     * process port list
     */
    vector<CPort*>::iterator pptr;
    vector<CPort*>* ports = module->GetPortList();
    tree tports = NULL;
    for( pptr = ports->begin(); pptr != ports->end(); ++pptr ) {
        tree next;
        tree exp = ParsePortExpression( (*pptr)->GetConnection() );
        CSymbol* sym = (*pptr)->GetSymbol();
        tree name = NULL;
        if( sym ) {
    	name = GetEscapedIdentifier((*pptr)->GetSymbol());
        }
        // ??? - this matches the original parser, but
        // doesn't appear to support (,,,a,b,c)
        if( name || exp ) {
             next = build_tree_list( exp, name );	
             tports = chainon( tports, next );
        }
    }
    UDP_PORT_LIST (currentModule) = tports;

    /*
     * process port dir declaration
     */

    vector<CPortDir*>::reverse_iterator dptr;
    vector<CPortDir*>* portdirs = module->GetPortDirList();
    for( dptr = portdirs->rbegin(); dptr != portdirs->rend(); ++dptr ) {
        tree decl = DeclarePortDir( *dptr, FALSE );    
        BLOCK_PORTS(current_scope) = chainon(decl, 
					BLOCK_PORTS(current_scope));
    }
    /*
     * declare register if it exists
     * Note: we scan the whole list, but there should only be one.
     */
    vector<CVar*>* regList = module->GetVarList();
    vector<CVar*>::iterator rPtr;
    tree firstArg = DECL_NAME(TREE_PURPOSE(MODULE_PORT_LIST(currentModule)));
    for( rPtr = regList->begin(); rPtr != regList->end(); ++rPtr ) {
        if( strcmp( (*rPtr)->GetName(), IDENT(firstArg) ) != 0 ) {
            error( "only the first declared port in a udp can be a reg",
                                                  NULL_CHAR, NULL_CHAR );
        } else {
            UDP_REG_NAME(currentModule) = firstArg;
	    sequential = TRUE;
        }
        ;
    }

    UDP_STRING_LIST(currentModule) = 
		ParseUdpStatements( module->GetCodeList(), sequential );
    module_list = tree_cons (currentModule, NULL_TREE, module_list);
    current_scope = pop_scope ();
    end_primitive(currentModule);
}

/************************************************
	ParsePathAssign
	- parse path assign construct
 ***********************************************/

static void ParsePathAssign( CNode* n )
{
    ASSERT( n->GetOp() == ePATH_ASSIGN );
    tree condition = ParseExpression( n->Arg<CNode*>(0) );
    CNode* path = n->Arg<CNode*>(1);
    ASSERT( path->GetOp() == ePATH );
    // need to add semantic check for number of list items in delay
    tree delays = ParseExpression( n->Arg<CNode*>(2) );
    if( n->Arg<CNode*>(2)->GetOp() != eELIST ) {
	delays = build_tree_list(delays,NULL);
    }
    tree inputs = ParseSpecifyList( path->Arg<CNode*>(1), 0 );
    tree outputs = ParseSpecifyList( path->Arg<CNode*>(4), 1 );
    tree source = ParseExpression( path->Arg<CNode*>(6) );
    int parallel = path->Arg<int>(3);
    int edge = path->Arg<int>(0);
    // only one of the polarities is nonzero at at time
    int polarity1 = path->Arg<int>(2);
    int polarity2 = path->Arg<int>(5);
    tree tpath = build_path( inputs, outputs, source, delays,
			     condition, parallel, edge, polarity1+polarity2 );
    MODULE_SPECDEFS(currentModule) =
                chainon(MODULE_SPECDEFS(currentModule),check_path(tpath));
}
/************************************************
	ParseTimingCheckEvent
	- parse timing check event construct
 ***********************************************/

static tree ParseTimingCheckEvent( CNode* n ) 
{
    int control    = 0;
    tree terminal  = NULL;
    tree condition = NULL;

    if( n->GetOp() == eANDANDAND ) {
        condition = ParseExpression(n->Arg<CNode*>(1));
	n = n->Arg<CNode*>(0);
    } 
    ASSERT( n );
    switch( n->GetOp() ) {
    case ePOSEDGE:
	control = posedgeMask;
	terminal = ParseExpression(n->Arg<CNode*>(0), peTIMING);
	break;
    case eNEGEDGE:
	control = negedgeMask;
	terminal = ParseExpression(n->Arg<CNode*>(0), peTIMING);
	break;
    case eEDGE:
	control = n->Arg<Edge_t>(1);
	terminal = ParseExpression(n->Arg<CNode*>(0), peTIMING);
	break;
    default:
	control = alledgeMask;
	terminal = ParseExpression(n, peTIMING);
    }
    return build_timing_event( control, terminal, condition );
}
/************************************************
	ParseNotifier
	- parse notifier construct
 ***********************************************/

static tree ParseNotifier( CNode* n ) 
{
    if( !n ) {
	return NULL;
    }

    ASSERT( n->GetOp() == eVAR_REF );
    CVar* reg = n->Arg<CVar*>(0);
    return GetEscapedIdentifier( reg->GetSymbol() );
}

/************************************************
	ParseTimingCall
	- parse timing call constructs
 ***********************************************/

static void ParseTimingCall( CNode* n ) 
{
    static CSymbol* setup = CSymbol::Lookup( "$setup" ); 
    static CSymbol* hold = CSymbol::Lookup( "$hold" ); 
    static CSymbol* width = CSymbol::Lookup( "$width" ); 
    static CSymbol* period = CSymbol::Lookup( "$period" ); 
    static CSymbol* skew = CSymbol::Lookup( "$skew" ); 
    static CSymbol* recovery = CSymbol::Lookup( "$recovery" ); 
    static CSymbol* setuphold = CSymbol::Lookup( "$setuphold" ); 
    static CSymbol* nochange = CSymbol::Lookup( "$nochange" ); 

    tree arg0 = NULL;
    tree arg1 = NULL;
    tree arg2 = NULL;
    tree arg3 = NULL;
    tree arg4 = NULL;
    int tc = 0;
    CNode* elist = n->Arg<CNode*>(1);
    int llen = ListSize( elist );
    if( n->Arg<CSymbol*>(0) == setup ) {
	tc = tcSETUP;
	ASSERT( llen == 3 || llen == 4 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg1 = ParseTimingCheckEvent( ListElement( elist, 1 ) );
	arg2 = ParseExpression( ListElement( elist, 2 ) );
	arg4 = ParseNotifier( ListElement( elist, 3 ) );
    } else if( n->Arg<CSymbol*>(0) == hold ) {
	tc = tcHOLD;
	ASSERT( llen == 3 || llen == 4 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg1 = ParseTimingCheckEvent( ListElement( elist, 1 ) );
	arg2 = ParseExpression( ListElement( elist, 2 ) );
	arg4 = ParseNotifier( ListElement( elist, 3 ) );
    } else if( n->Arg<CSymbol*>(0) == period ) {
	tc = tcPERIOD;
	ASSERT( llen == 2 || llen == 3 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg2 = ParseExpression( ListElement( elist, 1 ) );
	arg4 = ParseNotifier( ListElement( elist, 2 ) );
    } else if( n->Arg<CSymbol*>(0) == width ) {
	tc = tcWIDTH;
	ASSERT( llen == 2 || llen == 4 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg2 = ParseExpression( ListElement( elist, 1 ) );
	if( ListSize( elist ) == 4 ) {
	    arg3 = ParseExpression( ListElement( elist, 2 ) );
	    arg4 = ParseNotifier( ListElement( elist, 3 ) );
	}
    } else if( n->Arg<CSymbol*>(0) == skew ) {
	tc = tcSKEW;
	ASSERT( llen == 3 || llen == 4 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg1 = ParseTimingCheckEvent( ListElement( elist, 1 ) );
	arg2 = ParseExpression( ListElement( elist, 2 ) );
	arg4 = ParseNotifier( ListElement( elist, 3 ) );
    } else if( n->Arg<CSymbol*>(0) == recovery ) {
	tc = tcRECOVERY;
	ASSERT( llen == 3 || llen == 4 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg1 = ParseTimingCheckEvent( ListElement( elist, 1 ) );
	arg2 = ParseExpression( ListElement( elist, 2 ) );
	arg4 = ParseNotifier( ListElement( elist, 3 ) );
    } else if( n->Arg<CSymbol*>(0) == setuphold ) {
	tc = tcSETUPHOLD;
	ASSERT( llen == 4 || llen == 5 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg1 = ParseTimingCheckEvent( ListElement( elist, 1 ) );
	arg2 = ParseExpression( ListElement( elist, 2 ) );
	arg3 = ParseExpression( ListElement( elist, 3 ) );
	arg4 = ParseNotifier( ListElement( elist, 4 ) );
    } else if( n->Arg<CSymbol*>(0) == nochange ) {
	//tc = tcNOCHANGE;
	ASSERT( llen == 4 );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg0 = ParseTimingCheckEvent( ListElement( elist, 0 ) );
	arg1 = ParseTimingCheckEvent( ListElement( elist, 1 ) );
	arg2 = ParseExpression( ListElement( elist, 2 ) );
	arg3 = ParseExpression( ListElement( elist, 3 ) );
	// ??? mdh - currently not supported by simulator
	NI( FALSE, "$nochange", n->GetCoord() );
    } else {
	ASSERT( FALSE );
    }

    tree stmt = build_timing_statement( tc, arg0, arg1, arg2, arg3, arg4 );
    STMT_SOURCE_LINE(stmt) = stmt_lineno;
    STMT_SOURCE_FILE(stmt) = input_filename;
    MODULE_SPECDEFS(currentModule) =
                        chainon(MODULE_SPECDEFS(currentModule),stmt);
}


/************************************************
 	ParseStatement 
	- parse code tree
 ***********************************************/

static tree ParseStatement( CNode* n ) 
{
    tree result = NULL;
    if( !n ) {
	result = build_stmt (NULL_STMT, 0);
	goto DONE;
    }

    input_filename = n->GetCoord()->filename;
    lineno = n->GetCoord()->lineno;
    stmt_lineno = n->GetCoord()->lineno;
    switch( n->GetOp() ) {
    case eMODULE_DEF: {
	CModule* module = n->Arg<CModule*>(0);
	if( module->IsPrimitive() ) {
	    ParseUdp( module );
	} else {
	    ParseModule( module );
	}
	} break; 
    case eFUNCTION_DEF: {
	CFunction* function = n->Arg<CFunction*>(0);
	CDecl* output = function->GetOutput();
	// ??? mdh lots of stuff needs to be added
        NI( !function->InlinePortDecls(), "inline port declarations", function->GetCoord() );
        NI( !function->GetAutomatic(), "automatic functions", function->GetCoord() );

	tree name = GetEscapedIdentifier( function->GetSymbol() );
	tree tmpDecl = NULL;
	if( output ) {
	    tmpDecl = currentModule;
	    current_scope = build_function( check_function( name ) );
	} else {
	    tmpDecl = current_scope;
	    current_scope = build_task( check_task( name ) );
	}
	make_block_decl( name, tmpDecl, current_scope );
	push_scope();
    
        static tree rangeSpec = make_node(RANGE_HANDLE);
	if( output ) {
            tree spec = NULL;
            CNode* lsb = NULL;
            CNode* msb = NULL;
            if( output->GetDataType()->IsPacked() ) {
		if( !output->GetDataType()->IsScalar() ) {
                    lsb = output->GetLsb();
                    msb = output->GetMsb();
	 	}
                TREE_OPERAND(rangeSpec, 0) = ParseExpression(msb);
                TREE_OPERAND(rangeSpec, 1) = ParseExpression(lsb);
            }

    	    switch( output->GetDataType()->GetVarDataType() ) {
	    case eNO_VAR_TYPE:
	    case eREG:
	        spec = make_reg_spec(lsb ? rangeSpec : NULL_TREE);
	        break;
	    case eREAL:
                spec = make_real_spec (NULL_TREE);
	        break;
	    case eINTEGER:
                spec = make_integer_spec (NULL_TREE);
	        break;
	    case eTIME:
	        NI( FALSE, "time functions", function->GetCoord() );
		break;
	    case eSHORTREAL: 
	        NI( FALSE, "shortreal functions", function->GetCoord() );
		break;
	    case eREALTIME: 
	        NI( FALSE, "realtime functions", function->GetCoord() );
		break;
	    default: 
	        NI( FALSE, "function variable type", function->GetCoord() );
	        break;;
	    }
	    FUNCT_DECL(current_scope) = make_decl( name, spec, 
						NULL_TREE, NULL_TREE );
	}

	BLOCK_DOWN( currentModule ) = chainon( current_scope, 
			BLOCK_DOWN( currentModule ) );
	BLOCK_UP( current_scope ) = currentModule;
	in_tf = 1;

	ParseBlockDecls( function, TRUE );
	in_tf = 0;
	BLOCK_BODY( current_scope ) = ParseStatementOrNull( function->GetCodeList() );
        current_scope = pop_scope ();
	} break; 
    case eLIST: {
	tree stmt1 = ParseStatement( n->Arg<CNode*>(0) );
	tree stmt2 = ParseStatement( n->Arg<CNode*>(1) );
	if( stmt1 && stmt2 ) {
	    result = chainon(stmt1,stmt2);
	} else if( stmt1 ) {
	    result = stmt1;
	} else  {
	    result = stmt2;
	}
	} break;
    case eINIT: {
	tree stmt = ParseStatementOrNull(n->Arg<CNode*>(0));
	tree initStmt = build_stmt( INITIAL_BLOCK, 
				n->GetCoord()->lineno, stmt );
	tree list = build_tree_list( initStmt, (tree)INITIAL_CODE );
	BLOCK_BODY(currentModule) = chainon( BLOCK_BODY(currentModule), list );
	} break;
    case eALWAYS: {
	tree stmt = ParseStatement(n->Arg<CNode*>(0));
	tree alwaysStmt = build_stmt( ALWAYS_BLOCK, 
				n->GetCoord()->lineno, stmt );
	tree list = build_tree_list( alwaysStmt, (tree)ALWAYS_CODE );
	BLOCK_BODY(currentModule) = chainon( BLOCK_BODY(currentModule), list );
	} break;
    case eGATE_REF: {
	static CSymbol* gCmos = CSymbol::Lookup( "cmos" );
	static CSymbol* gRcmos = CSymbol::Lookup( "rcmos" );
	static CSymbol* gAnd = CSymbol::Lookup( "and" );
	static CSymbol* gNand = CSymbol::Lookup( "nand" );
	static CSymbol* gNor = CSymbol::Lookup( "nor" );
	static CSymbol* gOr = CSymbol::Lookup( "or" );
	static CSymbol* gXor = CSymbol::Lookup( "xor" );
	static CSymbol* gXnor = CSymbol::Lookup( "xnor" );
	static CSymbol* gBuf = CSymbol::Lookup( "buf" );
	static CSymbol* gNot = CSymbol::Lookup( "not" );
	static CSymbol* gBufif0 = CSymbol::Lookup( "bufif0" );
	static CSymbol* gBufif1 = CSymbol::Lookup( "bufif1" );
	static CSymbol* gNotif0 = CSymbol::Lookup( "notif0" );
	static CSymbol* gNotif1 = CSymbol::Lookup( "notif1" );
	static CSymbol* gNmos = CSymbol::Lookup( "nmos" );
	static CSymbol* gPmos = CSymbol::Lookup( "pmos" );
	static CSymbol* gRnmos = CSymbol::Lookup( "rnmos" );
	static CSymbol* gRpmos = CSymbol::Lookup( "rpmos" );
	static CSymbol* gPullup = CSymbol::Lookup( "pullup" );
	static CSymbol* gPulldown = CSymbol::Lookup( "pulldown" );
	static CSymbol* gTranif0 = CSymbol::Lookup( "tranif0" );
	static CSymbol* gTranif1 = CSymbol::Lookup( "tranif1" );
	static CSymbol* gRtranif0 = CSymbol::Lookup( "rtranif0" );
	static CSymbol* gRtranif1 = CSymbol::Lookup( "rtranif1" );
	static CSymbol* gTran = CSymbol::Lookup( "tran" );
	static CSymbol* gRTran = CSymbol::Lookup( "rtran" );

	CGate* gate = n->Arg<CGate*>(0);
	CNode* delay = gate->GetDelayControl();
	tree  name = gate->GetSymbol() ? GetEscapedIdentifier( gate->GetSymbol() ) : NULL_TREE;
	// ??? mdh lots of stuff needs to be added
	NI( !gate->GetNumberOfUnpackedDimensions(), "gate arrays", n->GetCoord() );
	NI( !gate->Strength0() != eUNDEFINED, "gate strengths", n->GetCoord() );
	NI( !gate->Strength1() != eUNDEFINED, "gate strengths", n->GetCoord() );
	in_instantiation = 1;
	enum tree_type type;
	if( gate->GetType() == gCmos ) {
	    type = GATE_CMOS_TYPE;
	} else if( gate->GetType() ==  gRcmos ) {
	    type = GATE_RCMOS_TYPE;
	} else if( gate->GetType() == gAnd ) {
	    type = GATE_AND_TYPE;
	} else if( gate->GetType() == gNand ) {
	    type = GATE_NAND_TYPE;
	} else if( gate->GetType() == gNor ) {
	    type = GATE_NOR_TYPE;
	} else if( gate->GetType() == gOr ) {
	    type = GATE_OR_TYPE;
	} else if( gate->GetType() == gXor ) {
	    type = GATE_XOR_TYPE;
	} else if( gate->GetType() == gXnor ) {
	    type = GATE_XNOR_TYPE;
	} else if( gate->GetType() == gBuf ) {
	    type = GATE_BUF_TYPE;
	} else if( gate->GetType() == gNot ) {
	    type = GATE_NOT_TYPE;
	} else if( gate->GetType() == gBufif0 ) {
	    type = GATE_BUFIF0_TYPE;
	} else if( gate->GetType() == gBufif1 ) {
	    type = GATE_BUFIF1_TYPE;
	} else if( gate->GetType() == gNotif0 ) {
	    type = GATE_NOTIF0_TYPE;
	} else if( gate->GetType() == gNotif1 ) {
	    type = GATE_NOTIF1_TYPE;
	} else if( gate->GetType() == gNmos ) {
	    type = GATE_NMOS_TYPE;
	} else if( gate->GetType() == gPmos ) {
	    type = GATE_PMOS_TYPE;
	} else if( gate->GetType() == gRnmos ) {
	    type = GATE_RNMOS_TYPE;
	} else if( gate->GetType() == gRpmos ) {
	    type = GATE_RPMOS_TYPE;
	} else if( gate->GetType() == gPullup ) {
	    type = GATE_PULLUP_TYPE;
	} else if( gate->GetType() == gPulldown ) {
	    type = GATE_PULLDN_TYPE;
	} else if( gate->GetType() == gTranif0 ) {
	    type = GATE_TRANIF0_TYPE;
	} else if( gate->GetType() == gTranif1 ) {
	    type = GATE_TRANIF1_TYPE;
	} else if( gate->GetType() == gRtranif0 ) {
	    type = GATE_RTRANIF0_TYPE;
	} else if( gate->GetType() == gRtranif1 ) {
	    type = GATE_RTRANIF1_TYPE;
	} else if( gate->GetType() == gTran ) {
	    type = GATE_TRAN_TYPE;
	} else if( gate->GetType() == gRTran ) {
	    type = GATE_RTRAN_TYPE;
	} else {
	    ASSERT( FALSE );
	}
	tree delayExpr = ParseDelayExpression( delay );
	tree args = ParseArgList( gate->GetArguments() );
	tree g = build_gate_instance( type, n->GetCoord()->filename,
		n->GetCoord()->lineno, name, args, delayExpr, NULL_TREE );
	MODULE_INSTANCES( currentModule ) = chainon( 
				MODULE_INSTANCES( currentModule ),g );
	in_instantiation = 0;
	} break;
    case eSYSTASK_CALL: {
	in_systask = 1;
	tree args = ParseExpList( n->Arg<CNode*>(1) );
	in_systask = 0;
	tree stmt = build_stmt( SYSTASK_STMT, n->GetCoord()->lineno, 
			n->Arg<CSymbol*>(0)->GetName(), args );
	result = stmt;
	} break;
    case eTASK_ENABLE: {
	const char* name = n->Arg<CSymbol*>(0)->GetName();
	in_systask = 1;
	tree args = ParseExpList( n->Arg<CNode*>(1) );
	in_systask = 0;
	tree stmt = build_stmt( TASK_STMT, n->GetCoord()->lineno, 
			name, args );
	HIERARCHICAL_ATTR( stmt ) = n->Arg<CSymbol*>(0)->External();
	result = stmt;
	} break;
    case eSPECIFY_REF: {
	CBlock* block = n->Arg<CBlock*>(0);
	ParseStatementOrNull( n->Arg<CNode*>(1) );
	}break;
    case eBLOCK_REF: {
	CBlock* block = n->Arg<CBlock*>(0);
	int named = block && block->GetSymbol();
	int forkJoin = block && block->GetType() == ePARALLELBLOCK;
	if( !forkJoin && !named ) {
	    /* unnamed block */
	    result = ParseStatementOrNull(n->Arg<CNode*>(1));
            tree begin = build_stmt( BEGIN_STMT, n->GetCoord()->lineno );
            tree end = build_stmt( END_STMT, n->GetCoord()->lineno );
            STMT_BODY(begin) = chainon(result,end);
	    result = begin;
	} else if( !forkJoin && named ) {
	    /* named block */
            tree nb = make_node (NAMED_BLOCK);
            BLOCK_DOWN(current_scope) = chainon (nb, 
					BLOCK_DOWN(current_scope));
            BLOCK_UP(nb) = current_scope;
            BLOCK_NAME(nb) = GetEscapedIdentifier( block->GetSymbol() );
            BLOCK_BODY(nb) = NULL_TREE;
            BLOCK_CONTEXT_LIST(nb) = NULL;
            make_block_decl (check_block(BLOCK_NAME(nb)), 
					current_scope, nb);
            current_scope = nb;
            push_scope ();
	    ParseBlockDecls( block, FALSE );
	    ParseStatement( block->GetCodeList() );
	    tree stmt = ParseStatementOrNull( n->Arg<CNode*>(1) );
            tree begin = build_stmt (BEGIN_NAMED_STMT, n->GetCoord()->lineno,
                        			NULL_TREE, nb);
            tree end = build_stmt (END_NAMED_STMT, n->GetCoord()->lineno,
                        	NULL_TREE, STMT_BLOCK(begin));
            STMT_BODY(begin) = chainon(stmt,end);
            current_scope = pop_scope ();
            result = begin;
	} else if( forkJoin && !named ) {
	    /* unnamed fork/join */
            tree begin = build_stmt( FORK_STMT, n->GetCoord()->lineno );
            tree end = build_stmt( JOIN_STMT, 
				n->GetCoord()->lineno, NULL_TREE );
	    tree stmt = ParseStatementOrNull(n->Arg<CNode*>(1));
	    stmt = List2TreeList( stmt );
            STMT_BODY(begin) = chainon(end,stmt);
	    result = begin;
	} else if( forkJoin && named ) {
	    /* named fork join */
            tree nb = make_node (NAMED_BLOCK);
            BLOCK_DOWN(current_scope) = chainon (nb, 
					BLOCK_DOWN(current_scope));
            BLOCK_UP(nb) = current_scope;
            BLOCK_NAME(nb) = GetEscapedIdentifier( block->GetSymbol() );
            BLOCK_BODY(nb) = NULL_TREE;
            make_block_decl (check_block(BLOCK_NAME(nb)), 
					current_scope, nb);
            current_scope = nb;
            push_scope ();
	    ParseBlockDecls( block, FALSE );
	    tree stmt = ParseStatementOrNull( n->Arg<CNode*>(1) );
	    stmt = List2TreeList( stmt );
            tree begin = build_stmt (FORK_NAMED_STMT, n->GetCoord()->lineno,
                        			NULL_TREE, nb);
            tree end = build_stmt (JOIN_NAMED_STMT, n->GetCoord()->lineno,
                        	NULL_TREE, STMT_BLOCK(begin));
            STMT_BODY(begin) = chainon(end,stmt);
            current_scope = pop_scope ();
            result = begin;
	} else {
	    ASSERT( FALSE );
	}
	} break;
    case eIF: {
	tree exp   = ParseExpression( n->Arg<CNode*>(0) );
	tree tstmt = ParseStatementOrNull( n->Arg<CNode*>(1) );
	tree fstmt = ParseStatementOrNull( n->Arg<CNode*>(2) );
        result = build_stmt(IF_STMT, lineno, exp, tstmt, fstmt);
	} break;
    case eASSIGN: {
	CNode* ectl = n->Arg<CNode*>(0);
	tree lval = ParseExpression(n->Arg<CNode*>(1));
	tree exp  = ParseExpression(n->Arg<CNode*>(2));
	if( !ectl ) {
            result = build_stmt (ASSIGN_STMT, lineno, lval,
                        implicit_conversion (lval, exp) );
	} else if( ectl->GetOp() == eDELAY_CONTROL )  {
	    tree ctl = ParseExpression( ectl->Arg<CNode*>(0) );
            result = build_stmt (ASSIGN_DELAY_STMT, lineno, lval,
                        implicit_conversion (lval, exp),
                        NULL_TREE, NULL_TREE,
                        build_stmt (DELAY_STMT, lineno, lval, ctl ));
	} else if( ectl->GetOp() == eEVENT_CONTROL )  {
	    tree event = build_stmt (EVENT_STMT, lineno, NULL_TREE,
                       	ParseEventExpression( n->Arg<CNode*>(0) ), 
                        NULL_TREE);
            result = build_stmt(ASSIGN_EVENT_STMT, lineno, lval,
                        implicit_conversion (lval, exp),
                        NULL_TREE, NULL_TREE, event);
	}
	} break;
    case eNBASSIGN: {
	CNode* ectl = n->Arg<CNode*>(0);
	tree lval = ParseExpression(n->Arg<CNode*>(1));
	tree exp  = ParseExpression(n->Arg<CNode*>(2));
	if( !ectl ) {
            result = build_stmt (ASSIGN_NONBLK_DELAY_STMT, lineno, lval,
                        implicit_conversion (lval, exp),
			NULL_TREE, NULL_TREE, 
			build_stmt(DELAY_STMT, lineno,lval, 
				integer_zero_node ), NULL, NULL, NULL, NULL );
	} else if( ectl->GetOp() == eDELAY_CONTROL )  {
	    tree ctl = ParseExpression( ectl->Arg<CNode*>(0) );
            result = build_stmt (ASSIGN_NONBLK_DELAY_STMT, lineno, lval,
                        implicit_conversion (lval, exp),
                        NULL_TREE, NULL_TREE,
                        build_stmt (DELAY_STMT, lineno, lval, ctl ),
			NULL, NULL, NULL, NULL);
	} else if( ectl->GetOp() == eEVENT_CONTROL )  {
	    tree event = build_stmt (EVENT_STMT, lineno, NULL_TREE,
                       	ParseEventExpression( n->Arg<CNode*>(0) ), 
                        NULL_TREE);
            result = build_stmt(ASSIGN_NONBLK_EVENT_STMT, lineno, lval,
                        implicit_conversion (lval, exp),
                        NULL_TREE, NULL_TREE, event);
	}
	} break;
    case eEVENT: {
        tree stmt = build_stmt (EVENT_STMT, lineno, NULL_TREE,
                       	ParseEventExpression( n->Arg<CNode*>(0) ), 
                        NULL_TREE);
	STMT_BODY(stmt) = ParseStatementOrNull( n->Arg<CNode*>(1) );
	return stmt;
	}
    case eDELAY: {
	tree ctl = ParseExpression( n->Arg<CNode*>(0) );
	tree stmt = ParseStatementOrNull( n->Arg<CNode*>(1) );
	result = build_stmt (DELAY_STMT, lineno, stmt, ctl);
	} break;
    case eCASSIGN: 
	CreateContinuousAssignment( n, FALSE );
	break;
    case eDEFPARAM:{ 
	tree lval = ParseExpression( n->Arg<CNode*>(0) ); 
	tree exp = ParseExpression( n->Arg<CNode*>(1) ); 
        tree stmt = build_stmt( ASSIGN_STMT, lineno, lval, exp);
	MODULE_DEFPARAMS (currentModule) = chainon ( stmt,
                                     MODULE_DEFPARAMS (currentModule));
        } break;
    case eNET_DECL:
	if( n->Arg<CNode*>(1) ) {
	    CreateContinuousAssignment( n->Arg<CNode*>(1), TRUE );
	}
	break;
    case eSPECPARAM_DECL:
	// ??? mdh - need to fix decl interface to blocks
	break;
    case ePARAM_DECL:
	// ??? mdh - need to fix decl interface to blocks
	break;
    case eVAR_DECL:
	// ??? mdh - need to fix decl interface to blocks
	// ??? mdh - decl direction assignment not supported
	NI( !n->Arg<CNode*>(1), "variable initialization", n->GetCoord() );
	if( n->Arg<CDecl*>(0)->GetDataType()->GetVarDataType() == eDEVENT ) {
	    // ??? mdh - need to fix decl interface to blocks
	    DeclareEvent( n->Arg<CVar*>(0) );
	}
	break;
    case eTRIGGER: {
	tree name = ParseExpression( n->Arg<CNode*>(0) );
	result = build_stmt (ARROW_STMT, lineno, name); 
	} break;
    case ePASSIGN: {
	tree lval = ParseExpression(n->Arg<CNode*>(0));
	tree exp  = ParseExpression(n->Arg<CNode*>(1));
        result = build_stmt (ASSIGN_PROC_STMT, lineno, lval, 
						exp, NULL_TREE, NULL_TREE);
	} break;
    case eDEASSIGN: {
	tree lval = ParseExpression(n->Arg<CNode*>(0));
        result = build_stmt (DEASSIGN_STMT, lineno, lval );
	} break;
    case eFORCE: {
	tree lval = ParseExpression(n->Arg<CNode*>(0),peLVAL_REG_NET);
	tree exp  = ParseExpression(n->Arg<CNode*>(1));
        result = build_stmt (FORCE_STMT, lineno, lval, 
						exp, NULL_TREE, NULL_TREE);
	} break;
    case eRELEASE: {
	tree lval = ParseExpression(n->Arg<CNode*>(0),peLVAL_REG_NET);
        result = build_stmt (RELEASE_STMT, lineno, lval );
	} break;
    case eFOREVER: {
	tree stmt = ParseStatement(n->Arg<CNode*>(0));
        result = build_stmt (FOREVER_STMT, lineno, stmt );
	} break;
    case eREPEAT: {
	tree exp = ParseExpression(n->Arg<CNode*>(0));
	tree stmt = ParseStatement(n->Arg<CNode*>(1));
        result = build_stmt (REPEAT_INIT_STMT, lineno, stmt, exp);
        tree tmp_tree = build_stmt (REPEAT_STMT, lineno, result, exp);
        chainon (stmt, tmp_tree);
	} break;
    case eWHILE: {
	tree exp = ParseExpression(n->Arg<CNode*>(0));
	tree stmt = ParseStatement(n->Arg<CNode*>(1));
        result = build_stmt (WHILE_STMT, lineno, stmt, exp);
        } break;
    case eFOR: {
	const char* for_filename = input_filename;
	int   for_lineno = lineno;
	tree initStmt = ParseStatement(n->Arg<CNode*>(0));
	tree exp = ParseExpression(n->Arg<CNode*>(1));
	tree assStmt = ParseStatement(n->Arg<CNode*>(2));
	tree stmt = ParseStatement(n->Arg<CNode*>(3));
	input_filename = for_filename;
	lineno = for_lineno;
        tree tmp_tree = build_stmt (FOR_STMT, lineno, stmt, assStmt, exp);
        STMT_SURROGATE_ATTR (tmp_tree) = 1;
        result = build_stmt (FOR_STMT, lineno, stmt, initStmt, exp);
        TREE_CHAIN(tmp_tree) = result;  
        chainon (stmt, tmp_tree); 
	} break;
    case eWAIT: {
	tree exp = ParseExpression(n->Arg<CNode*>(0));
	tree stmt = ParseStatementOrNull(n->Arg<CNode*>(1));
        result = build_stmt(WAIT_STMT, lineno, stmt, exp);
	} break;	
    case eCASE: {
	tree exp = ParseExpression( n->Arg<CNode*>(0) );
        result = build_stmt (CASE_STMT, stmt_lineno, 
				exp, NULL_TREE, NULL_TREE);
	caseStack.push( result );
	tree caseItem = ParseList( eLIST, n->Arg<CNode*>(1), FALSE );
	STMT_CASE_LIST(result) = caseItem;
	caseStack.pop();
	} break;
    case eCASEZ: {
	tree exp = ParseExpression( n->Arg<CNode*>(0) );
        result = build_stmt (CASEZ_STMT, stmt_lineno, 
				exp, NULL_TREE, NULL_TREE);
	caseStack.push( result );
	tree caseItem = ParseList( eLIST, n->Arg<CNode*>(1), FALSE );
	STMT_CASE_LIST(result) = caseItem;
	caseStack.pop();
	} break;
    case eCASEX: {
	tree exp = ParseExpression( n->Arg<CNode*>(0) );
        result = build_stmt (CASEX_STMT, stmt_lineno, 
				exp, NULL_TREE, NULL_TREE);
	caseStack.push( result );
	tree caseItem = ParseList( eLIST, n->Arg<CNode*>(1), FALSE );
	STMT_CASE_LIST(result) = caseItem;
	caseStack.pop();
	} break;
    case eDISABLE: {
	CSymbol* sym = n->Arg<CSymbol*>(0);
	tree exp = GetEscapedIdentifier( sym );
	result = build_stmt (DISABLE_STMT, lineno, exp);
        if( sym->External() ) {
             HIERARCHICAL_ATTR (exp) = 1;
             UNKNOWN_TYPE_ATTR (exp) = 1;
             TREE_CONSTANT_ATTR (exp) = 1;
             HIERARCHICAL_ATTR (result) = 1;
	}
	} break;
    case ePATH_ASSIGN: {
	ParsePathAssign( n );
	} break;
    case eINSTANCE_REF: {
        CInstance* inst = n->Arg<CInstance*>(0);
	CModule* m = inst->GetDefinition();
	tree mname = GetEscapedIdentifier( m->GetSymbol() );
	tree args = NULL;
	tree params = NULL;
	// ??? mdh - parser needs fixed. this could be null
	tree ident = GetEscapedIdentifier( inst->GetSymbol() );
        ASSERT( m );
        if( inst->GetParameters() ) {
	    params = ParseList( eELIST, inst->GetParameters(), TRUE );
	}
	in_instantiation = 1;
	vector<CNode*>* argList = inst->GetArgumentVector();
	if( !argList->size() ) {
	    args = build_tree_list( NULL_TREE, NULL_TREE );
	} else {
	    vector<CNode*>::iterator aptr;
 	    for( aptr = argList->begin(); aptr != argList->end(); ++aptr ) {
		ASSERT( (*aptr)->GetOp() == eARG );
		CSymbol* sym = (*aptr)->Arg<CSymbol*>(0);
		CNode*   nexp = (*aptr)->Arg<CNode*>(1);
		tree exp = NULL;
		if( nexp ) {
		    exp = ParseExpression( nexp );
		}
		tree id = NULL;
		if( sym ) {
		    id = GetEscapedIdentifier( sym );
		}
		if( !sym || nexp ) {
	            args = chainon( args, build_tree_list( exp, id ) );
		}
	    }
	}
        ASSERT( !inst->IsGate() );
        NI( !inst->GetNumberOfUnpackedDimensions(), "instance arrays", n->GetCoord() );
        tree stmt = build_stmt (INSTANCE_NODE, lineno, check_non_reg (ident),
                                mname, args, params );
        MODULE_INSTANCES (currentModule) = 
			chainon ( MODULE_INSTANCES (currentModule), stmt );
	in_instantiation = 0;
	} break;
    case eTIMING_CALL: {
	ParseTimingCall( n );
	} break;
    case ePRAGMA:
    case eVRQ:
    case eCOMMENT:
    case ePORT_DECL:
	break;
    case eREPEAT_CONTROL:
        NI( FALSE, "repeat control", n->GetCoord() );
	break;
    case eGENVAR_REF:
        NI( FALSE, "genvar references", n->GetCoord() );
	break;
    case eGENVAR_DECL:
        NI( FALSE, "genvar declaration", n->GetCoord() );
	break;
    case eATTRIBUTE:
	break;
    case eGIF:
        NI( FALSE, "generate if", n->GetCoord() );
	break;
    case eGFOR:
        NI( FALSE, "generate for", n->GetCoord() );
	break;
    case eGCASE:
        NI( FALSE, "generate case", n->GetCoord() );
	break;
    case eIFNONE_PATH_ASSIGN:	
        NI( FALSE, "ifnone path", n->GetCoord() );
	break;
	break;
    case ePORT_DEF:
	// should never occur in parse tree
	ASSERT( FALSE );
	break;
    default:
	ASSERT( FALSE );
	break;
    }
DONE:
    return result;
}

/************************************************
 	ParseStatementOrNull
	- parse code tree. 
	  always return a statement node.
 ***********************************************/

static tree ParseStatementOrNull( CNode* n ) 
{
    tree result = ParseStatement( n );
    if( !result ) {
	result = build_stmt (NULL_STMT, 0);
    }
    return result;
}

/************************************************
	Finish
	- simulator finish callback
 ************************************************/

static void Finish() {
    printf_V("\nNormal exit\n");
    printf_V("Thank you for using %s\n", PLUGIN_NAME);
    throw CBackendExit();
}

/************************************************
	ReinitLex
	- reinitialize lexer
 ************************************************/

static void ReinitLex() {
}

/************************************************
	ProgParse
	- Parse program
 ************************************************/

static int ProgParse( tree& ) {
	fin = fileStack.top();
	fileStack.pop();
	error( "syntax error", NULL, NULL );
	return 1;
}

/************************************************
	PopStream
	- pop input stream
 ************************************************/

static File* PopStream() {
	if( fileStack.empty() ) {
	    return 0;
	}
	File* top = fileStack.top();
	fileStack.pop();
	return top;
}

/************************************************
	PushStream
	- push input stream
 ************************************************/

static void PushStream( File* f,int) {

	fileStack.push( f );
}

/************************************************
	Abort
	- abort plugin 
 ************************************************/

static void Abort() {
	throw CBackendAbort();
}

/************************************************
	Exit
	- exit plugin 
 ************************************************/

static void Exit( int code ) {
    if( code ) {
	throw CBackendFail();
    } else {
	throw CBackendExit();
    }
}

/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CSim::Process( list<CElement>& inputList,
			list<CElement>& outputList ) 
{
    CNode* code = NULL;
   

    /*
     * initialize simulator
     */

    simulator.finish(Finish);
    simulator.reinit_lex(ReinitLex);
    simulator.prog_parse(ProgParse);
    simulator.pop_stream(PopStream);
    simulator.push_stream(PushStream);
    simulator.abort(Abort);
    simulator.exit(Exit);
    simulator.Init(0,NULL);

    /*
     * check for switches
     */
    interactive    = GetPlusArg( "sim-interactive" ) != NULL;
    compileOnly    = GetPlusArg( "sim-compile-only" ) != NULL;
    pliLibraryList = GetPlusArg( "sim-pli" );
#ifdef BUILD_HISTOGRAMS
    veriwell::printStats     = GetPlusArg( "sim-stats" ) != NULL;
#endif

    /* Open log file */
    if (!log_available) {
        /* set if new file name given in command line */
        log_file_name = strdup( "sim.log" );
    }
    if (!strcmp(log_file_name, "nolog")) {
        log_file = 0;
        log_available = 0;
        log_enable = 0;
        warning("Proceeding without a log file", NULL_CHAR, NULL_CHAR);
    } else {
        log_file = fopen(log_file_name, "w+t");
        if (!log_file) {
            log_available = 0;
            log_enable = 0;
            warning("Cannot open log file '%s', proceeding without it",
                    log_file_name, NULL_CHAR);
        } else {
            log_available = 1;
            log_enable = 1;
        }
    }

    /* Open key file */
    if (!key_available) {
        /* set if new file name given in command line */
        key_file_name = strdup( "sim.key" );
    }
    if (!strcmp(key_file_name, "nokey")) {
        key_file = 0;
        key_available = 0;
        key_enable = 0;
        warning("Proceeding without a key file", NULL_CHAR, NULL_CHAR);
    } else {
        key_file = fopen(key_file_name, "w+t");
        if (!key_file) {
            key_available = 0;
            key_enable = 0;
            warning("Cannot open key file '%s', proceeding without it",
                    key_file_name, NULL_CHAR);
        } else {
            key_available = 1;
            key_enable = 1;
        }
    }


    traceFlag   = GetPlusArg( "sim-trace" ) != NULL;
    int delaySwitchCount = 0;
    if( GetPlusArg( "sim-mindelays" ) ) {
	delayType = MIN_DELAY;
	delaySwitchCount++;
    }
    if( GetPlusArg( "sim-typdelays" ) ) {
	delayType = TYP_DELAY;
	delaySwitchCount++;
    } 
    if( GetPlusArg( "sim-maxdelays" ) ) {
	delayType = MAX_DELAY;
	delaySwitchCount++;
    }
    if( delaySwitchCount > 1 ) {
	error( "only one delay switch may be used at a time", NULL, NULL );
    }
    simulator.DelayType( delayType );

    /*
     * initialize the plusargs
     */
    list<string> plusargList;
    GetPlusArgs( "", plusargList );
    list<string>::iterator plusargPtr;
    for( plusargPtr = plusargList.begin(); 
	 plusargPtr != plusargList.end(); ++plusargPtr ) {
	plusargs = tree_cons( (tree)strdup(plusargPtr->c_str()), NULL, plusargs );
    }

    simulator.Banner();

    /*
     * merge all input streams together
     */
    list<CElement>::iterator ptr;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
        code = cLINK( code, ptr->Code() );
    }


    /*
     * load external pli code
     */ 
    list<string> pluginPaths;
    GetPluginPaths( pluginPaths );
    list<string>::iterator pluginPath;
    for( pluginPath = pluginPaths.begin();
        pluginPath != pluginPaths.end(); ++pluginPath ) {
        veriwell::LoadPliPlugins( pluginPath->c_str() );
    }

    if( pliLibraryList ) {
	string buffer = pliLibraryList;
        int pos = 1;
	while( 1 ) {
	    int end = buffer.find( "+", pos );
	    int len = end == string::npos ? end : end-pos;
            int fail = veriwell::LoadPliLibrary(  
			buffer.substr(pos,len).c_str() );
	    if( fail ) {
		error( "cannot load pli dll %s\n", 
			buffer.substr(pos,len).c_str(), NULL );
	    }
	    if( end == string::npos ) {
	       break;
	    }
	    pos = end+1;
	}
    }

    currentModule = NULL;
    simulator.Precompile();
    if( traceFlag ) {
        trace_flag = 1;
        normal_flag = 0;
	veriwell::execAttention = 1;
    }
    ParseStatement( code );
    simulator.Compile();
    if( !compileOnly ) {
	masterStats.Start();
        simulator.Simulate(interactive);
    }
DONE: ;
}

