/*****************************************************************************
 * Copyright (C) 1997-2009, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   array.cc
 *		- array info routines
 *
 ******************************************************************************
 */


#include "array.h"
#include "member.h"
#include "cnet.h"
#include "cvar.h"
#include "cfref.h"
#include "cparam.h"
#include "cdatatype.h"

/**
 * Get width of array.
 * \param ref reference to a decl
 * \param index index list
 * \return width of expression
 */
int Array::Width( CNode* ref, CNode* index )
{ 
    CDataType* dataType = NULL;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    int result = ResolveReference( ref, index, dataType, unpackedDimensions,
		packedDimensions, residualPackedIndices );
    if( result == 0 ) {
	return 0;
    } else if( result == -1 ) {
	return 0;
    }
    if( unpackedDimensions.size() > 0 ) {
	return 0;
    }
    if( !dataType->IsVector() && !dataType->IsPacked() ) {
         return 0;
    }	

    int width = dataType->GetBaseWidth();
    for( int i = residualPackedIndices.size(); 
	     i < packedDimensions.size(); ++i ) {
	width *= packedDimensions[i]->GetWidth();
    }
    for( int i = 0; i < residualPackedIndices.size(); ++i ) {
	CNode* select = residualPackedIndices[i];
	while( select->GetOp() == eMACRO_EXPR ) {
	    select = select->Arg<CNode*>(1);
	}
	switch( select->GetOp() ) {
	case eSLICE:
	case ePSLICE:
	case eMSLICE:
	    width *= select->GetWidth();
	    break;
	}
    }
    return width;
}
/**
 * Get width of array as an expression.
 * \param ref reference to a decl
 * \param index index list
 * \return width expression.
 */
CNode* Array::WidthExp( CNode* ref, CNode* index )
{ 
    CDataType* dataType = NULL;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    int result = ResolveReference( ref, index, dataType, unpackedDimensions,
		packedDimensions, residualPackedIndices );
    if( result == 0 ) {
	return NULL;
    } else if( result == -1 ) {
	return NULL;
    }
    if( unpackedDimensions.size() > 0 ) {
	return NULL;
    }
    if( !dataType->IsVector() && !dataType->IsPacked() ) {
         return NULL;
    }	

    CNode* widthExp = NULL;
    if( !dataType->IsBaseWidthEvaluateable() || 
	dataType->GetBaseWidth() != 1 ) {
	widthExp = dataType->GetBaseWidthExp();
    }
    for( int i = residualPackedIndices.size(); 
	     i < packedDimensions.size(); ++i ) {
	CNode* width = packedDimensions[i]->GetWidthExp();
	if( !width->IsWidthEvaluateable() || width->GetWidth() != 1 ) {
	    widthExp = !widthExp ? width : cMUL( widthExp, width );
	}
    }
    for( int i = 0; i < residualPackedIndices.size(); ++i ) {
	CNode* select = residualPackedIndices[i];
	while( select->GetOp() == eMACRO_EXPR ) {
	    select = select->Arg<CNode*>(1);
	}
	switch( select->GetOp() ) {
	case eSLICE:
	case ePSLICE:
	case eMSLICE: {
	    CNode* width = select->GetWidthExp();
	    if( !width->IsWidthEvaluateable() || width->GetWidth() != 1 ) {
	        widthExp = !widthExp ? width : cMUL( widthExp, width );
	    }
	    } break;
	}
    }
    if( !widthExp ) {
	widthExp = cINT32(1);
    }
    return widthExp;
}
/**
 * Determine if width of array variable is constant.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return non-zero if width is constant. ie dependent upon only
 *         on parameters and constants.
 */
int Array::WidthConstant( CNode* ref, CNode* index ) 
{ 
    CDataType* dataType = NULL;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    int result = ResolveReference( ref, index, dataType, unpackedDimensions,
		packedDimensions, residualPackedIndices );
    if( result == 0 ) {
	return FALSE;
    } else if( result == -1 ) {
	return FALSE;
    }
    if( unpackedDimensions.size() > 0 ) {
	return FALSE;
    }
    if( !dataType->IsVector() && !dataType->IsPacked() ) {
         return NULL;
    }	

    int constant = dataType->IsBaseWidthConstant();
    for( int i = residualPackedIndices.size(); 
	     i < packedDimensions.size(); ++i ) {
	constant &= packedDimensions[i]->IsWidthConstant();
    }
    for( int i = 0; i < residualPackedIndices.size(); ++i ) {
	CNode* select = residualPackedIndices[i];
	while( select->GetOp() == eMACRO_EXPR ) {
	    select = select->Arg<CNode*>(1);
	}
	switch( select->GetOp() ) {
	case eSLICE:
	case ePSLICE:
	case eMSLICE:
	    constant &= select->IsWidthConstant();
	    break;
	}
    }
    return constant;
}
/**
 * Determine if width of array variable is volatile.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return non-zero if width is volatile. ie dependent upon only
 *         on parameters or variables.
 */
int Array::WidthVolatile( CNode* ref, CNode* index ) 
{ 
    CDataType* dataType = NULL;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    int result = ResolveReference( ref, index, dataType, unpackedDimensions,
		packedDimensions, residualPackedIndices );
    if( result == 0 ) {
	return FALSE;
    } else if( result == -1 ) {
	return FALSE;
    }
    if( unpackedDimensions.size() > 0 ) {
	return FALSE;
    }
    if( !dataType->IsVector() && !dataType->IsPacked() ) {
        return NULL;
    }	

    int vol = dataType->IsBaseWidthVolatile();
    for( int i = residualPackedIndices.size(); 
	     i < packedDimensions.size(); ++i ) {
	vol |= packedDimensions[i]->IsWidthVolatile();
    }
    for( int i = 0; i < residualPackedIndices.size(); ++i ) {
	CNode* select = residualPackedIndices[i];
	while( select->GetOp() == eMACRO_EXPR ) {
	    select = select->Arg<CNode*>(1);
	}
	switch( select->GetOp() ) {
	case eSLICE:
	case ePSLICE:
	case eMSLICE:
	    vol |= select->IsWidthVolatile();
	    break;
	}
    }
    return vol;
}
/**
 * Determine if width of array can be evaluated.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return non-zero if width can be evaluated.
 */
int Array::WidthEvaluateable( CNode* ref, CNode* index ) 
{ 
    CDataType* dataType = NULL;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    int result = ResolveReference( ref, index, dataType, unpackedDimensions,
		packedDimensions, residualPackedIndices );
    if( result == 0 ) {
	return FALSE;
    } else if( result == -1 ) {
	return FALSE;
    }
    if( unpackedDimensions.size() > 0 ) {
	return FALSE;
    }
    if( !dataType->IsVector() && !dataType->IsPacked() ) {
        return NULL;
    }	
// ??? mdh - this one needs to be changed
    int eval = dataType->IsBaseWidthEvaluateable();
    for( int i = residualPackedIndices.size(); 
	     i < packedDimensions.size(); ++i ) {
	eval &= packedDimensions[i]->IsWidthEvaluateable();
    }
    for( int i = 0; i < residualPackedIndices.size(); ++i ) {
	CNode* select = residualPackedIndices[i];
	while( select->GetOp() == eMACRO_EXPR ) {
	    select = select->Arg<CNode*>(1);
	}
	switch( select->GetOp() ) {
	case eSLICE:
	case ePSLICE:
	case eMSLICE:
	    eval &= select->IsWidthEvaluateable();
	    break;
	}
    }
    return eval;
}
/**
 * Determine type of array.
 * \todo current not fully implemented.
 * \param ref reference to a decl
 * \param index index list
 * \return type of array return value.
 */
NodeType_t Array::Type( CNode* ref, CNode* index )
{ 
    CDataType* dataType = NULL;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    int result = ResolveReference( ref, index, dataType, unpackedDimensions,
		packedDimensions, residualPackedIndices );
    if( result < 1 ) {
	return eU;
    }
    if( unpackedDimensions.size() > 0 ) {
	return eU;
    }
    if( dataType->IsVector() || dataType->IsPacked() ) {
	return dataType->GetSigned() ? eS : eB;
    }	
    return eU;
}
/**
 * Get datatype info for array reference
 * \param ref reference to a decl/subreference
 * \param index index list 
 * \param dataType resolved base data type
 * \param unpackedDimensions list of unresolved unpacked dimensions
 * \param packedDimensions list of unresolved packed dimensions 
 * \param residualPackedIndices list of unresolved packed indices
 * \return -1 if reference is valid but unresolvable
 * 	    0 if reference is invalid
 * 	    1 if reference is valid and was resolved
 */
int Array::ResolveReference( 
	CNode*          ref, 
	CNode*          index,
	CDataType*&     dataType,
	vector<CNode*>& unpackedDimensions,
	vector<CNode*>& packedDimensions,
	vector<CNode*>& residualPackedIndices )
{
    vector<CNode*>     declUnpackedDimensions;
    vector<CNode*> indexes;
    EList2VectorExclude( index, set<NodeOp_t>(), indexes );

    switch( ref->GetOp() ) {
    case eARRAY:
	// array nodes are not allowed to be cascaded
	MASSERT( FALSE );
	break;
    case eEXTERNAL_REF:
        residualPackedIndices.clear();
        for( int i = 0; i < indexes.size(); ++i ) {
            residualPackedIndices.push_back(indexes[i]);
        }
	return -1;
    case eMEMBER: {
	int result = Member::ResolveReference( ref->Arg<CNode*>(0),
		ref->Arg<CSymbol*>(1), dataType, declUnpackedDimensions,
		packedDimensions );
	if( result != 1 ) {
	    return result;
	}
	} break;
    default: {
	CDecl* decl = ref->Arg<CDecl*>(0);
	dataType = decl->GetDataType();
        for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
            declUnpackedDimensions.push_back(decl->GetUnpackedRange(i));
        }
        for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
            packedDimensions.push_back(dataType->GetPackedRange(i));
        }
	} break;
    }

    while( dataType->GetVarDataType() == ePROXY ) {
	CDecl* decl = dynamic_cast<CProxyType*>(dataType)->GetDecl();
        for( int i = 0; i < decl->GetNumberOfUnpackedDimensions(); ++i ) {
            declUnpackedDimensions.push_back(decl->GetUnpackedRange(i));
        }
	dataType = dynamic_cast<CProxyType*>(dataType)->GetDataType();
        for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
            packedDimensions.push_back(dataType->GetPackedRange(i));
        }
    }
    /*
     * handle native vector types 
     */
    int skipCount = indexes.size();
    int firstSelect = skipCount;
    for( int i = 0; i < indexes.size(); ++i ) {
	CNode* range = indexes[i];
	int isSelect = FALSE;
	while( range->GetOp() == eMACRO_EXPR ) {
	    range = range->Arg<CNode*>(1);
	}
	switch( range->GetOp() ) {
	case eSLICE:
	case ePSLICE:
	case eMSLICE:
	    isSelect = TRUE;
	}
	if( isSelect ) {
	    firstSelect = i;
	    break;
	}
    }
    if( firstSelect < declUnpackedDimensions.size() &&
	firstSelect < skipCount ) {
	skipCount = firstSelect;
    }
   
    int residualCount = 0;
    if( skipCount > declUnpackedDimensions.size() ) {
        residualCount = skipCount - declUnpackedDimensions.size();
	skipCount = declUnpackedDimensions.size();
    }
    unpackedDimensions.clear();
    for( int i = skipCount; i < declUnpackedDimensions.size(); ++i ) {
        unpackedDimensions.push_back(declUnpackedDimensions[i]);
    }

    while( residualCount > packedDimensions.size() ) {
	if( dataType->GetVarDataType() != ePROXY || 
	    !dataType->IsPacked() ) {
	    break;
	}
	dataType = dynamic_cast<CProxyType*>(dataType)->GetDataType();
        for( int i = 0; i < dataType->GetNumberOfPackedDimensions(); ++i ) {
            packedDimensions.push_back(dataType->GetPackedRange(i));
        }
	
    }

    if( dataType->IsVector() && residualCount > 1 ) {
	    return 0;
    } else if( dataType->IsPacked() &&
               residualCount > packedDimensions.size() ) {
        return 0;
    } else if( !dataType->IsVector() && 
	       !dataType->IsPacked() && residualCount > 0 ) {
	return 0;
    }
    for( int i = 0; i < residualCount; ++i ) {
	int partSelect;
	CNode* select = indexes[i+skipCount];
	while( select->GetOp() == eMACRO_EXPR ) {
	    select = select->Arg<CNode*>(1);
	}
        switch( select->GetOp() ) {
	case eSLICE:
	case ePSLICE:
	case eMSLICE:
	    partSelect = TRUE;
	    break;
	default:
	    partSelect = FALSE;
	}
	if( i+1 != residualCount && partSelect ) {
	    return 0;
	}
    }
    residualPackedIndices.clear();
    for( int i = 0; i < residualCount; ++i ) {
        residualPackedIndices.push_back(indexes[i+skipCount]);
    }
    return 1; 
}
/**
 * Get decl for array reference
 * \param ref decl reference node
 * \return decl associated with the reference
 */
CDecl* Array::ArrayDecl( CNode* ref )
{
    MASSERT( ref );
    switch( ref->GetOp() ) {
    case eNET_REF:		
	return ref->Arg<CNet*>(0);
    case eVAR_REF:		
	return ref->Arg<CVar*>(0);
    case ePARAM_REF:		
	return ref->Arg<CParam*>(0);
    case ePORT_REF:		
	return ref->Arg<CPortDir*>(0);
    case eFWD_REF:		
	return ref->Arg<CFref*>(0);
    case eGENVAR_REF:		
	return ref->Arg<CGenvar*>(0);
    default:
	MASSERT( FALSE );
    }
}


