/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cbackend.hpp
 *		- abstract class definition for code generators 
 *
 ******************************************************************************
 */

#ifndef CBACKEND_HPP
#define CBACKEND_HPP

#include <stdio.h>
#include "glue.h"
#include <list>
#include <map>
#include <string>

using namespace std;

class CNode;
class CModule;
class CInstance;
class CObstack;

/**
 * \page plugin_info How to Build a Plugin
 * Vrq plugins are independently compiled as standalone DLLs.
 * A detail description of how to create plugins can be found in
 * the following sections:
 * - \ref plugin_api
 * - \ref building_a_plugin
 * - \ref installing_a_plugin
 *
 * \section plugin_api Plugin API
 * Vrq plugins are created by creating a subclass of type \ref CBackend.
 * This class should then be compiled in a standalone dll and placed 
 * in the the plugin directory. 
 *
 * The input to a plugin is a vector of CElement. Each CElement
 * represents a unit of compilation. Each unit can be as trivial 
 * as a module, the code in a file or the entire design.
 *
 * The output of a plugin is a vector of CElement. Each CElement
 * again represents a unit of compilation. Note the number of input
 * elements need not match the number of output elements. If a plugin
 * does not modify the compilation units the plugin output may have no 
 * elements. In this case the input vector is passed to the next 
 * plugin in the processing pipeline.
 *
 * A code in a compilation unit is represented by a tree of CNode. 
 * Each node of the tree represents a statement or operation. Short 
 * cut constructors should be used to create nodes. A list of node 
 * types can be found in \ref parse_tree_node . A list of 
 * constructors can be found in \ref node_constructors . A list of 
 * utility routines for operating on CNode can be found in \ref node_utilities .
 * A list of general purpose function can be found here: \ref utilities .
 *
 * Parse tree nodes may point to auxillary objects that represent
 * elements of the parsed code such as variables, etc. See the following 
 * list of classes:
 * - CAttr
 * - CBlock
 * - CObject
 * - CDecl
 * - CFref
 * - CGenvar
 * - CVar
 * - CNet
 * - CGate
 * - CMacro
 * - CSpecify
 * - CPort
 * - CInstance
 * - CPortDir
 * - CParam
 * - CFunction
 * - CModule
 * - CVector
 * - CDataType
 * - CSimpleType
 * - CTypedef
 * - CStructType
 * - CClassType
 * - CEnumType
 * - CProxyType
 *
 * Additionally these class provide support functions for object info:
 * - CUInt
 * - CObstack
 * - CSymbol
 * - CSymtab
 * - Message
 * - Systask
 * - External
 * 
 * The plugin pipeline is specified on the command line using the 
 * on the command line. The output of one plugin is applied to the next.
 *
 * In the case that a plugin produces an output vector but there are 
 * no remaining plugins to process the output, the dump tool is invoked
 * to convert the parse trees back to textual verilog.
 * 
 * \section building_a_plugin Building a Plugin
 * Example plugins can be found here:
 * - \ref example1 
 * - \ref example2 
 *
 * \section installing_a_plugin Installing a Plugin
 * This class should then be compiled in a standalone dll and placed 
 * in the default plugin directory. The default plugin directory configured 
 * at vrq build time can be determined by executing: vrq --pkglibdir
 *
 * Alternately the plugin directory can be overriden by setting the 
 * environment variable PLUGIN_DIR with a ':' seperated list of paths
 * to search.
 */
 
/**
 * \page example1 Example Plugin
 * This example plugin searches the parsed code and finds register and net
 * declarations and references. A report of whether the variables are declared
 * and how many times they are referenced is printed. It also issues 
 * warnings for any variables that are declared but not referenced.
 *
 * The makefile for building the plugin (for linux based systems) follows (Makefile-example1):
 * \verbinclude Makefile-example1
 *
 * The header file for the plugin is below (example1.h):
 * \verbinclude example1.h
 *
 * The source for the plugin follows (example1.cc):
 * \verbinclude example1.cc
 */
 
/**
 * \page example2 Example Plugin
 * This example plugin converts all procedural if statements to case statements.
 *
 * The makefile for building the plugin (for linux based systems) follows (Makefile-example2):
 * \verbinclude Makefile-example2
 *
 * The header file for the plugin is below (example2.h):
 * \verbinclude example2.h
 *
 * The source for the plugin follows (example2.cc):
 * \verbinclude example2.cc
 */

/**
 * An element of compilation passed to and from plugins. Each element
 * contains a parse tree and optionally a filename. The filename may
 * represent the input file for the compilation unit or may have been
 * synthetically generated by a plugin. The dump plugin is configurable
 * so that all elements may be dumped to a single file (using -o \<filename>)
 * or dumped to a directory using the tagged filenames (using -dir \<directory>).
 */
class CElement
{
private:
    string filename;	   // Optional filename for compilation unit
    int    filenameValid;  // non-zero to indicate filename is valid
    CNode* code;	   // parse tree pointer for compilation unit
public:
    /**
     * Constructor for class.
     * \param filename a filename tagging the compilation unit.
     * \param filenameValid non-zero when filename is valid.
     * \param code a pointer to the parse tree for the compilation unit.
     */
    CElement( const char* filename, int filenameValid, CNode* code ) :
        filename(filename),
	filenameValid( filenameValid ),
        code(code ) {}
    /**
     * Access method for filename tagging this compilation unit.
     * \return Returns pointer to filename, NULL is returned if 
     * filename is not valid.
     */
    const char* Filename() { return filenameValid ? filename.c_str() : NULL; }
    /**
     * Access method for parse tree for this compilation unit.
     * \return Returns pointer to parse tree.
     */
    CNode* Code() { return code; }	
    /**
     * Method to set parse tree for this compilation unit.
     * \param code a pointer to parse tree.
     */
    void   Code( CNode* code ) { this->code = code; }	
};

/**
 * Base class for exceptions that can be thrown by backend plugins
 */
class CBackendException {};
 /**
 * Exception thrown by CBackend::Process when plugin aborts due to an 
 * internal error.
 */
class CBackendAbort : CBackendException {};
 /**
 * Exception thrown by CBackend::Process when plugin terminates early 
 * without an error.
 */
class CBackendExit : CBackendException {};
 /**
 * Exception thrown by CBackend::Process when plugin terminates early
 * with an error.
 */
class CBackendFail : CBackendException {};

/**
 * This is the abstract class to overload to create new backend tools.
 * Each tool should create a subclass of this and either add it to the
 * toolList in main.cc or compile it as a standalone DLL and place it in
 * a plugin directory. 
 *
 * This class and its methods will fully integrate the plugin in to
 * vrq. Note not all methods are abstract. For methods where reasonable
 * defaults make sense, a default implementation is defined. If a different
 * behavior is desired the subclass should override the default virtual
 * method.
 *
 * Note: Plugins should directly store information about the switches they
 * support in the protected instance variables: switches and switchDescription.
 */

class CBackend 
{
protected:
    list<string>       switches;  /// list of plugin command line switches
    map<string,string> switchDescription; /// map of switch to its description
public:
	/**
 	 * Access method for tool name
	 * \return pointer containing the name of the tool
	 */
	virtual char* GetToolName( void ) = 0;
	/**
	 * Access method for tool description use in vrq help
	 * \return pointer containing the description of tool
	 */
	virtual char* GetToolDescription( void ) = 0;
	/** 
 	 * Override this method if the plugin accepts arbitrary
 	 * plusargs: ie +argName[=argValue] or +argName[+argValue]. 
 	 * \return true if all plus args are allowed. Default method
 	 * disallows plusargs.
 	 */
	virtual int AcceptAllPlusArgs( void ) { return FALSE; }
	/**
 	 * Access method for list of switches. Generally this method
 	 * need not be overridden.
	 * \return list of switches.
	 */
	virtual list<string>& GetSwitches( void ) {return switches;}
	/**
 	 * Method to get the help text for a switch using the switch 
 	 * name as a key. Generally this method need not be overriden.
	 * \param sw pointer to switch.
	 * \return description of switch.
	 */
	virtual const char* GetSwitchDescription( const char* sw ) 
	{
	    MASSERT( switchDescription.find(sw) != 
				switchDescription.end() );
	    return switchDescription[sw].c_str();
	}
	/**
 	 * Method shortcut to register tool switches with vrq's help
 	 * system.
 	 * \param switchName switch string
 	 * \param description switch description used for help text.
 	 */
	virtual void RegisterSwitch( const char* switchName, 
						const char* description )
	{
	    switches.push_back( switchName );
	    switchDescription[switchName] = description;
	}
	/**
	 * Method to control if plugin requires all module definitions to be
	 * resolved.
	 * returns true if tool requires all module definitions to be
	 * resolved.
	 */
	virtual int RequireModuleResolution() { return TRUE; }
	/**
	 * Method to control if plugin enables all module definitions to be
	 * searched for
	 * returns true if tool enables module definitions to be
	 * resolved.
	 */
	virtual int ResolveModules() = 0;
	/**
	 * Method to control the resolution of individual module instances.
	 * \param module to the module that contains the instance.
	 * \param inst pointer to the instance to be resolved.
	 * \returns true if the instance should be resolved 
	 * prior to tool running.
	 */
	virtual int ResolveInstance( CModule* module, CInstance* inst ) = 0;
	/**
 	 * Method to control if plugin should be hidden, in this
 	 * case it will not appear in the help but will it
 	 * still be possible to invoke it. This is
 	 * useful for development of plugins that care currently unstable.
 	 * Note: the default behavior is to make the plugin visible.
 	 * \return TRUE if plugin should be hidden.
 	 */
	virtual int HideTool() { return FALSE; } 
	/**
 	 * Method to control how parser handles vrq comment based
 	 * pragmas: ie // vrq translate_on/off
 	 * Note: the default behavior is to not ignore vrq 
 	 * comment based pragmas.
 	 * \return TRUE if vrq translate on/off comments should 
 	 * be treated like normal comments. If FALSE code bracketed
 	 * by // vrq translate_off and //vrq translate_on will handled
 	 * as if it is a comment.
 	 */
	virtual int IgnoreVrqComments() { return FALSE; }
	/**
 	 * Method to initialize the plugin. This method will
 	 * be called once before tool is invoked.
 	 */
	virtual void Activate() = 0;
	/**
         * Method to implement plugin function. 
         * Input list is processed and output list is generated. Dimensions
	 * of input list must be > 0. Dimensions of output
	 * list must be >= 0. Input and output list dimensions
	 * need not match. This method may terminate processing early by
	 * throwing an exception that is a subclass of CBackendException.
	 * Note throwing an exeception and exiting normally is 
	 * equivilent to throwing a CBackendExit.
	 * \param inputList a list of compilation units to process.
	 * \param outputList a list of compilation units processed.
	 */
	virtual void Process( list<CElement>& inputList,
			      list<CElement>& outputList ) = 0;
};

#endif // CBACKEND_HPP
