/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cblock.hpp
 *		- class definition of blocks 
 *
 ******************************************************************************
 */

#ifndef CBLOCK_HPP
#define CBLOCK_HPP

#include <vector>
#include "glue.h"
#include "cdecl.h"
#include "csymtab.h"


class CParam;
class CVar;
class CNode;
class CNet;
class CFref;
class CInstance;
class CGenvar;
class CPortDir;


/**
 * Declaration class for block constructs
 */
class CBlock: public CDecl
{
protected:
	vector<CParam*>	  paramList;
	vector<CVar*> 	  varList;
	vector<CGenvar*>  genvarList;
	vector<CFref*>    frefList;
	vector<CNet*>	  netList;
	CBlock*		  parent;
	vector<CBlock*>   children;
	CSymtab<CDecl>	  symtab; 
private:
	CNode*		  codeList;
	int		  rebalance;
public:
	/**
 	 * Constructor for BLOCK declarations.
 	 * \param name name of block.
 	 * \param aLoc file coordinates of block.
 	 */
	CBlock( CSymbol* name, Coord_t* aLoc );
	/**
 	 * Constructor for BLOCK declarations.
 	 * \param name name of block.
 	 * \param aLoc file coordinates of block.
 	 * \param dtype block type.
 	 */
	CBlock( CSymbol* name, Coord_t* aLoc, Decl_t dtype );
	/**
 	 * Object destructor
 	 */
	virtual ~CBlock() {}
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap ) { MASSERT(FALSE); }
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add instance to block.
 	 * \param instance pointer to block object
 	 */
	virtual void Add( CInstance* instance );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add parameter to block.
 	 * \param param pointer to parameter object
 	 */
	virtual void Add( CParam* param );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add net to block.
 	 * \param net pointer to net object
 	 */
	virtual void Add( CNet* net );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add register to block.
 	 * \param var pointer to variable object
 	 */
	virtual void Add( CVar* var );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add child block to this block.
 	 * \param child pointer to child block object
 	 */
	virtual void Add( CBlock* child );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add forward reference to this block.
 	 * \param fref pointer to forward reference object
 	 */
	virtual void Add( CFref* fref );
	/**
 	 * \todo This routine will eventually be eliminated.
 	 * Currently the construct must be added with this routine as
 	 * well as by adding a node to the code tree.
 	 *
 	 * Add genvar to this block.
 	 * \param genvar pointer to genvar reference object
 	 */
	virtual void Add( CGenvar* genvar );
	/**
 	 * Add parse tree code to this block.
 	 * \param code pointer to parse tree 
 	 */
	virtual void Add( CNode* code  );
	/**
 	 * Set this block's parent block.
 	 * \param parent pointer to parent block object.
 	 */
	void	SetParent( CBlock* parent );
	/**
 	 * Get this block's parent block.
 	 * \return pointer to parent block object.
 	 */
	CBlock*	GetParent( void );
	/**
 	 * Dump information about block to a file.
 	 * \param f file descriptor to output.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * Get list of object's nets.
 	 * \return array of net objects.
 	 */
	vector<CNet*>*	GetNetList();
	/**
 	 * Get list of object's registers.
 	 * \return array of register objects.
 	 */
	vector<CVar*>*	GetVarList();
	/**
 	 * Get list of object's parameters.
 	 * \return array of parameter objects.
 	 */
	vector<CParam*>* GetParamList();
	/**
 	 * Get list of object's genvars.
 	 * \return array of genvar objects.
 	 */
	vector<CGenvar*>* GetGenvarList();
	/**
 	 * Get blocks parse tree.
 	 * \return parse tree.
 	 */
	virtual CNode*	GetCodeList();
	/**
 	 * Set block's parse tree.
 	 * \param list block's parse tree.
 	 */
	virtual void	SetCodeList( CNode* list );
	/**
 	 * Get list of block's child blocks.
 	 * \return array of child block objects.
 	 */
	vector<CBlock*>* GetChildList();
	/**
 	 * Get list of block's port declarations.
 	 * \return array of port direction objects.
 	 */
        virtual vector<CPortDir*>*  GetPortDirList() 
		{ static vector<CPortDir*> nullList; return &nullList;}
	/**
 	 * Set block's symbol table.
 	 * \param symtab reference to new symbol table.
 	 */
	virtual void 	SetSymtab( CSymtab<CDecl>& symtab );
	/**
 	 * Get block's symbol table.
 	 * \return reference to block's symbol table.
 	 */
	virtual CSymtab<CDecl>& GetSymtab() { return symtab; }
	/**
 	 * These should never be be called by plugins.
 	 * \cond
 	 */
	void    PreVisit1( int (*func)(CNode*,void*), void* data );
	void    PostVisit1( void (*func)(CNode*, void*), void* data );
	void    PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
protected:
	static void Reconnect( CNode *n );
	void RedeclarePorts();
	void LinkFrefs( CSymtab<CDecl>& portSymtab,
			CSymtab<CDecl>& moduleSymtab,
			vector<CFref*>& frefList, 
			int inferVectors );
	virtual void CodeListChanged() {}
private:
	void LinkPortFref( CSymtab<CDecl>* symtab1,
	               CSymtab<CDecl>* symtab2,
	               CFref* fref  );
	void InferFref( CSymtab<CDecl>* symtab1,
	               CFref* fref,
		       int inferVector );
	void LinkFref( CSymtab<CDecl>* symtab,
	               CFref* fref, int inferVector );
	CBlock( const CBlock& );
	/**
 	 * \endcond
 	 */
};

/**
 * Alias for CBlock
 */
typedef CBlock CScope;


#endif // CBLOCK_HPP
