/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdecl.hpp
 *		- abstract class for declaration 
 *
 *
 ******************************************************************************
 */

#ifndef CDECL_HPP
#define CDECL_HPP

#include <stdio.h>
#include <vector>
#include <list>
#include "glue.h"
#include "csymbol.h"
#include "csimpletype.h"

class CNode;
class CDecl;

/**
 * Structure to hold file coordinates
 */
struct Coord_t {
	int	    lineno;	///< file line number
	const char* filename;	///< file name
};


/**
 * Wire types
 */
enum Wire_t {
	eNO_WIRE_TYPE = 0,	///< no wire type
	eWIRE,		///< wire type
	eTRI,		///< tri type
	eWAND,		///< wand type
	eTRIAND,	///< triand type
	eWOR,		///< wor type
	eTRIOR,		///< trior type
	eTRI1,		///< tri1 type
	eTRI0,		///< tri0 type
	eTRIREG, 	///< trireg type
	eSUPPLY0,	///< supply0 type
	eSUPPLY1,	///< supply1 type
};

/**
 * Declaration types
 */
enum Decl_t {
	eVAR = 0,	///< variable declaration
	ePARAM,		///< parameter declaration
	eLOCALPARAM,	///< localparam declaration
	eBLOCK,		///< named block declaration
	ePARALLELBLOCK, ///< fork/join block declaration	
	eGENVARBLOCK,	///< generate block declaration
	eMODULE,	///< module declaration
	eMACRO,		///< macro module declaration
	eINSTANCE,	///< instance declaration
	eINPUT,		///< input declaration
	eOUTPUT,	///< output declaration
	eINOUT,		///< inout declaration
	eNODIR,		///< unspecified declaration
	ePORT,		///< port declaration
	eFREF,		///< forward reference declaration (used by parser only)
	eFUNCTION,	///< function/task declaration
	eSPECIFYBLOCK,	///< specify block declaration
	eGENVAR,	///< genvar declaration
	eTYPEDEF,	///< type declaration
	// NOTE this are class types only
	eNET,		///< declaration superclass for all net types
	ePORTDIR,	///< declaration superclass for all port dirs
	eATTR,		///< attribute declaration
	eUDP,		///< udp declaration
	eENUMVALUE,///< enum declaration
	eNONE		///< unspecified declaration used for wildcard
};

#if defined CDECL_CPP
/**
 * Array to convert Wire_t to character string
 */
const char* wireName[] = {
	"no_wire_type",
	"wire",
	"tri",
	"wand",
	"triand",
	"wor",
	"trior",
	"tri1",
	"tri0",
	"trireg",
	"supply0",
	"supply1"
};

/**
 * Array to convert Decl_t to character string
 */
const char* declName[] = {
        "variable",
	"param",
	"localparam",
	"block",
	"parallel block",
	"genvar block",
	"module",
	"macro",
	"instance",
	"input",
	"output",
	"inout",
	"nodir",
	"port",
	"fref",
	"function",
	"specify block",
	"genvar",
	"typedef",
	"net",
	"portdir",
	"attr",
	"udp",
	"enum",
	"none"
};
#else
extern char* declName[];
extern char* wireName[];
#endif



/**
 * Base class for describing declaration objects
 */
class CDecl : public CObject
{
public:
	enum Flag {
	    eFLAG_NONE = 0,	///< no support flags
	    eFLAG_PRAGMA = 1,	///< declaration accepts pragmas
	    eFLAG_ARRAY = 2,	///< declaration can be unpacked array
	    eFLAG_DATATYPE = 4	///< declaration has datatype
	};
	static Flag Or( Flag f1, Flag f2 ) 
		{ return static_cast<Flag>(((unsigned)f1 | (unsigned)f2)); }
	static Flag Or( Flag f1, Flag f2, Flag f3 ) 
		{ return static_cast<Flag>(((unsigned)f1 | (unsigned)f2) | (unsigned)f3); }
private:
	Coord_t	   loc;		  ///< file coordinates of declaration
	CSymbol*   symbol;	  ///< symbol for declaration
	Decl_t	    type;	  ///< declaration type
	int	   declStatement; ///< TRUE if decl statement was created 
				  ///< in containing block
	CNode*	   attributes;	  ///< declaration attributes
	CNode*	   pragmas;	  ///< declaration pragmas
	vector<CNode*> unpackedRange; ///< unpacked dimension ranges, 
				      ///< final dimension may be null
	int numberOfUnpackedDimensions; ///< number of unpacked dimensions
	Flag	   flags;	  ///< declaration support flags.
	int	   automatic;	  ///< declaration is a automatic type
	Wire_t	   wireAttr;	  ///< wire attribute
	int	   constAttr;	  ///< constant attribute
	int	   varAttr;	  ///< var attribute
	CDataType* dataType;	  ///< decl data type
public:
	/**
 	 * Create a deep copy of declaration
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap ) = 0;
	/**
 	 * Set declartion's data type property.
 	 * \param dt new data type property.
 	 */
	virtual void  SetDataType( CDataType* dt ) { 
	    MASSERT( flags & eFLAG_DATATYPE );
	    dataType = dt; 
	}
	/**
 	 * Get declartion's data type property.
 	 * \return value of data type property.
 	 */
	virtual CDataType*  GetDataType() { return dataType; }
	/**
 	 * Set declaration's wire attribute
 	 * \param v wire type
 	 */
	virtual void SetWireAttr( Wire_t v ) { wireAttr = v; }
	/**
 	 * Get declartion's wire type property.
 	 * \return value of wire type property.
 	 */
	virtual Wire_t  GetWireAttr() { return wireAttr; }
	/**
 	 * Determine if packed width of declaration is constant, 
 	 * ie dependent upon only constants and parameters.
 	 * \return non-zero of packed width of declaration is constant.
 	 */
	virtual int  IsWidthConstant( void ) const {
	    if( numberOfUnpackedDimensions ) {
		return FALSE;
	    }
	    MASSERT( dataType );
	    return dataType->IsPackedWidthConstant();
	}
	/**
 	 * Determine if packed width of declaration is volatile, ie depend
 	 * upon parameters or variables.
 	 * \return non-zero of packed width of declaration is volatile.
 	 */
	virtual int  IsWidthVolatile( void ) const {
	    if( numberOfUnpackedDimensions ) {
		return TRUE;
	    }
	    MASSERT( dataType );
	    return dataType->IsPackedWidthVolatile();
	}
	/**
 	 * Determine if packed width of declaration can be evaluated.
 	 * \return non-zero of packed width of declaration can be evaluated.
 	 */
	virtual int  IsWidthEvaluateable( void ) const {
	    if( numberOfUnpackedDimensions ) {
		return FALSE;
	    }
	    MASSERT( dataType );
	    return dataType->IsPackedWidthEvaluateable();
	}
	/**
 	 * Evaluate packed width of declaration.
 	 * \return packed width of declaration.
 	 */
	virtual INT32 GetWidth( void ) const {
	    if( numberOfUnpackedDimensions ) {
		return 0;
	    }
	    MASSERT( dataType );
	    return dataType->GetPackedWidth();
	}
	/**
 	 * Get expression representing width of declaration.
 	 * \return expression representing width of declaration.
	 */
	virtual CNode* GetWidthExp( void ) const {
	    if( numberOfUnpackedDimensions ) {
		return NULL;
	    }
	    MASSERT( dataType );
            return dataType->GetPackedWidthExp();
	}
        /**
         * Get expression for declaration's msb
         * \return msb expression, NULL if scalar.
         */
        virtual CNode* GetMsb() const {
	    MASSERT( dataType );
            return dataType->GetPackedMsb();
	}
        /**
         * Get expression for declaration's lsb
         * \return lsb expression, NULL if scalar.
         */
        virtual CNode* GetLsb() const {
	    MASSERT( dataType );
            return dataType->GetPackedLsb();
	}
	/**
 	 * Evaluate current decl width direction. Expression must be constant.  
	 * \return 1 - big endian, -1 - little endian, 0 - cannot be determined
	 */
	virtual int   WidthDirection( void ) const {
	    MASSERT( dataType );
            return dataType->PackedWidthDirection();
	}
	/**
 	 * Get number of unpacked dimensions of declaration.
 	 * \return number of dimensions, 0 if non-array.
 	 */
	virtual INT32 GetNumberOfUnpackedDimensions( void ) { 
	    return numberOfUnpackedDimensions;
	}
	/**
 	 * Get expression tree for upper limit of given packed array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for unpacked upper limit.
	 */
	virtual CNode* GetUnpackedMsi( INT32 dim );
	/**
 	 * Get expression tree for lower limit of unpacked array dimension.
 	 * \param dim dimension.
 	 * \return expression tree for unpacked lower limit.
	 */
	virtual CNode* GetUnpackedLsi( INT32 dim );
	/**
 	 * Set number of unpacked dimensions for declaration.
 	 * \param dim number of unpacked dimensions.
 	 */
	virtual void SetNumberOfUnpackedDimensions( INT32 dim ) { 
	    MASSERT( flags & eFLAG_ARRAY );
	    numberOfUnpackedDimensions = dim;
	    unpackedRange.reserve( dim );
	}
	/**
 	 * Get expression for range of array for dimension.
 	 * \param dim dimension to get.
 	 * \return range expression for limit.
 	 */
	virtual CNode*  GetUnpackedRange( INT32 dim ) { return unpackedRange[dim]; }
	/**
 	 * Set expression for range of unpacked array for declaration.
 	 * \param dim dimension to set.
 	 * \param v range expression for limit.
 	 */
	virtual void  SetUnpackedRange( INT32 dim, CNode* v ) 
		{ MASSERT(flags & eFLAG_ARRAY); unpackedRange[dim] = v; }
	/**
 	 * Set declaration's const attribute.
 	 * Use to indicate type is a constant variable.
 	 * \param v value to const attribute.
 	 */
	virtual void SetConstAttr( int v ) { constAttr = v; }
	/**
 	 * Get declaration's const attribute.
 	 * \return value of const attribute.
 	 */
	virtual int  GetConstAttr() { return constAttr; }
	/**
 	 * Set declaration's var attribute.
 	 * Use to indicate type was declared with var keyword.
 	 * \param v value to var attribute.
 	 */
	virtual void SetVarAttr( int v ) { varAttr = v; }
	/**
 	 * Get declaration's var attribute.
 	 * \return value of var attribute.
 	 */
	virtual int  GetVarAttr() { return varAttr; }
	/**
 	 * Set declaration's automatic property.
 	 * Use to indicate type is a automatic variable.
 	 * \param v value to automatic property.
 	 */
	virtual void SetAutomatic( int v ) { automatic = v; }
	/**
 	 * Get declaration's automatic property.
 	 * \return value of automatic property.
 	 */
	virtual int  GetAutomatic() { return automatic; }
	/**
 	 * Set declartion's vectored property.
 	 * \param v value of vectored property.
 	 */
	virtual void SetVectored( int v ) { MASSERT( FALSE ); }
	/**
 	 * Get declartion's vectored property.
 	 * \return value of vectored property.
 	 */
	virtual int  GetVectored() { MASSERT( FALSE ); return 0; }
	/**
 	 * Set declartion's scalared property.
 	 * \param v value of scalared property.
 	 */
	virtual void SetScalared( int v ) { MASSERT( FALSE ); }
	/**
 	 * Get declartion's scalared property.
 	 * \return value of scalared property.
 	 */
	virtual int  GetScalared() { MASSERT( FALSE ); return 0; }
	/**
 	 * Set declarations's attributes.
 	 * \param attr attribute list.
 	 */
	void	SetAttributes( CNode* attr ) { attributes = attr; }
	/**
 	 * Get declaration's attributes.
 	 * \return attribute list.
 	 */
	CNode*  GetAttributes() { return attributes; }
	/**
 	 * Determine if declaration has the given attribute.
 	 * \param name name of attribute to search for.
 	 * \param n don't supply this argument.
 	 * \param init don't supply this argument.
 	 */
        int     HasAttribute( char* name, CNode* n=NULL, int init = 1 );
	/**
 	 * Get node type of decl.
 	 * \return node type of decl.
 	 */
	NodeType_t GetNodeType( void ) {
	   MASSERT( dataType );
           return dataType->GetNodeType();
	}
        /**
         * Get class of declaration.
         * \return declaration class for this object.
         */
        Decl_t	GetClass( void )
        {
        	switch( type ) {
        	case ePARAM:
        	case eLOCALPARAM:
        	       	return ePARAM;
        	case eINPUT:
        	case eOUTPUT:
        	case eINOUT:
        		return ePORTDIR;
        	default:
        		return type;
        	}
        }

        /**
         * Get a list of members of the given declaration class/type.
         * \param type type/class of declaration.
         * \param result list to return members.
         */
        static void GetMembers( Decl_t type, list<Decl_t>& result )
        {
            switch( type ) {
            case ePARAM:
                result.push_back( ePARAM );
                result.push_back( eLOCALPARAM );
                break;
            case ePORTDIR:
                result.push_back( eINPUT );
                result.push_back( eOUTPUT );
                result.push_back( eINOUT );
                break;
            default:
                result.push_back( type );
                break;
            }
        }

        /**
         * Set declaration statement created attribute
         * \todo This call will be deprecated when the Block:Add(*) is
         * removed.
         */
        void SetDeclStatementCreated( void ) { declStatement = TRUE; }
        
        /**
         * Get declaration statement create attibute.
         * \return non-zero if declaration statement has been created.
         */
        int DeclStatementCreated( void ) { return declStatement; }

        /**
         * Get declaration type
         * \return declaration type.
         */
        Decl_t GetType( void ) { return type; }
        
        /**
         * Set declaration coordinate.
         * \param aLoc set file coordinates for declaration
         */
        void SetCoord( Coord_t* aLoc ) { loc = *aLoc; }
        
        /**
         * Get file coordinates for declaration.
         * \return file coordinates.
         */
        Coord_t*	GetCoord( void ) {
        	return &loc;
        }
        
        /**
         * Dump declaration info to file descriptor.
         * \param f file descriptor.
         */
        virtual void Dump( FILE* f ) 
        	{ fprintf( f, "'%s' line %d", loc.filename , loc.lineno ); }

	/**
	* Dump declaration name, type and location to file descriptor.
	* \param f file descriptor.
	*/
	virtual void DumpDeclInfo( FILE* f )
	{
	    fprintf( f, "%s: %s, defined in '%s' line %d\n", 
		declName[GetType()], GetName(), loc.filename , loc.lineno );
	}

        /**
         * Shortcut to get declaration's name
         * \return char pointer to name.
         */
        const char* GetName( void ) { return symbol->GetName(); }
        /**
         * Set declaration's symbol
         * \param aSymbol new symbol for declaration.
         */
        void SetSymbol( CSymbol* aSymbol ) { symbol = aSymbol; }
        /**
         * Get declaration's symbol
         * \return declaration's symbol.
         */
        CSymbol* GetSymbol( void ) { return symbol; }
	/**
 	 * Set declaration pragmas
 	 * \param p pragma list.
 	 */
	void    SetPragmas( CNode* p ) 
		{ MASSERT( flags & eFLAG_PRAGMA );pragmas = p; }
	/**
 	 * Get declaration pragmas
 	 * \return declaration's pragma list.
 	 */
	CNode*  GetPragmas() { return pragmas; }
protected:
        /**
         * Create instance of declaration.
         * This should never be call directly.
         * \param aSymbol declaration for symbol.
         * \param aLoc file coordinates of declaration.
         * \param aType declaration type.
         * \param dataType declaration data type
         * \param flags list of supported constructs.
         */
        CDecl( CSymbol* aSymbol, Coord_t* aLoc, 
	       Decl_t aType, CDataType* dataType, Flag flags )  {
        	loc 	 = *aLoc;
        	symbol   = aSymbol;	
        	type 	 = aType;
	        numberOfUnpackedDimensions = 0;
        	declStatement = FALSE;
        	attributes = NULL;
		pragmas = NULL;
		automatic = FALSE;
		constAttr = FALSE;
		wireAttr  = eNO_WIRE_TYPE;
		varAttr   = FALSE;
		this->flags = flags;
		this->dataType = dataType;
        }
	/**
 	 * Perform deep copy of given object to this one
 	 * This should never be call directly, only by subclasses.
 	 * \param heap heap to use for allocation.
 	 * \param o declaration to copy.
 	 */
	void Copy( CObstack* heap, const CDecl& o );
	/**
 	 * Copy constructor to do deep copy.
 	 * This should never be called directly.
 	 * \param o declaration to copy.
 	 */
private:
	/*
 	 * disable copy constructor
 	 */
	CDecl( const CDecl& o );

public:
	virtual void PreVisit1( int (*func)(CNode*,void*), void* data );
	virtual void PostVisit1( void (*func)(CNode*, void*), void* data );
	virtual void PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
};

#endif // CDECL_HPP

