/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cdumpverilog.cpp
 *		- methods for outputting verilog
 *
 ******************************************************************************
 */

/**
 * \page dump_plugin Dump Plugin
 * The dump tool emits the parsed and processed parse trees back in to 
 * verilog. If the last filter of a tool pipeline emits a parse tree the 
 * dump tool automatically appended to the end of the tool chain. The 
 * output of the dump tool may be specified in the following 3 ways:
 * - -o - This option converts all parse trees to verilog and emits the 
 *   code to stdout.
 * - -o filename This option converts all parse trees to verilog and 
 *   emits the code to file 'filename'.
 * - -dir directory This option converts all parse trees to verilog and 
 *   emits each group of parse trees to a separate file in directory 
 *   'directory'. The filenames used vary based upon the prior tools in 
 *   the pipeline.
 *
 * There are various output formating switches that effect the output 
 * code style. Please see the \ref manpage for details.
 */

/**
 * \cond PLUGIN_INTERNALS
 */

#include <ctype.h>
#include <math.h>
#include <vector>
#include <fstream>
#include "main.h"
#include "cdumpverilog.h"
#include "cmodule.h"
#include "cport.h"
#include "cfref.h"
#include "cvar.h"
#include "cnet.h"
#include "cgenvar.h"
#include "cparam.h"
#include "cnode.h"
#include "cinstance.h"
#include "cgate.h"
#include "cfunction.h"
#include "cportdir.h"
#include "cspecify.h"
#include "cattr.h"
#include "ctypedef.h"
#include "cstructtype.h"
#include "cproxytype.h"
#include "cenumtype.h"


static  INT32     	tabCount;
static  INT32		suppressTab;
static	int		skipTab;
static  const int	TAB_SPACE = 4;
extern  char*		outputPath;
extern  char*		outputDirectory;
static  int 		inFor;
static  int		suppressNullArgs = 0;
extern  int		errorCount;
static	int		inlineParams;
static	int		inlinePorts = false;
static  int		forcePartSelect = 0;
static  int		forceVectorLogicals = 0;
static  int		forceCPPComments = 0;
static  int		forceBeginAfterEvent = 0;
static  int		altRxnor = 0;
static	int		splitWireDeclAssignments = 0;
static	int		noNullStatements = 0;
static  int		spaceAroundBinaryOps = 0;
static  int		hookCount = 0;
static  int		createExplicitDeclarations = 0;
static  int		stripAttributes = 0;
static  int		simplifyExpressions = 0;
static  char*		defaultTimescale = NULL;
static  CModule*	currentModule = NULL;
static  int		defaultNetType = FALSE;

static  int DumpExpressionList( FILE* f, CNode* n );
static void DumpExpression( FILE* f, CNode* n , int sized = 1, int insertNewline= 0);
static void DumpStatement( FILE* f, CNode* n );
static int  DumpBlock( FILE* f, CNode* n, int indent, int cascade, int begin );
static void DumpModule( FILE* f, CModule* m );
static void DumpFunction( FILE* f, CFunction* func );
static void DumpNamedBlock( FILE* f, CBlock* block, CNode* statements, int noNewline );
static void DumpSpecifyBlock( FILE* f, CSpecify* block, CNode* statements );
static void ForceUnsized( CNode* n );
static void DumpPath( FILE* f, CNode* n ); 
static int DumpDeclarationsOnly( FILE* f, CNode* n );
static void DumpStrengthPair( FILE* f, StrengthPair_t* spair );
static void DumpComment( FILE* f, CNode* n );
static void DumpWireAttr( FILE* f, Wire_t t );
static void DumpDeclType( FILE* f, Decl_t dt, Wire_t wt );
static void DumpPortDirDecl( FILE* f, CPortDir* p );
static void DumpParameterDecl( FILE* f, CParam* p );
static void DumpAttributes( FILE* f, CNode* n, int last=TRUE );
static int  DumpTableSymbols( FILE* f, CNode* n, int pos1, int pos2, int pos );
static int CountPorts( CNode* n );
static int DumpInlinePortDecls( FILE* f, CNode* n, int state=-2 );
static CNode* FindPortDeclarationEnd( CNode* n, int* hasPLE = NULL );
static CNode* DumpStatementsAfter( FILE* f, CNode* cutPoint, CNode* n );
static int CountStatementsInRange( CNode* begin, CNode* end, int* found=NULL );
static void DumpRange( FILE* f, CNode* range, int collapse );
static void DumpDeclSpec( FILE* f, Decl_t type, CDecl* decl, int forceReg=FALSE );
static void DumpDataType( FILE* f, CDecl* decl, CDataType* dataType, int forceReg=FALSE );
static void DumpUnpackedDimensions( FILE* f, CDecl* d );


/************************************************
 	Constructor
 ************************************************/
CDumpVerilog::CDumpVerilog( ) 
{
    f                  = NULL;
    switches.push_back( "+dump-alt-rxnor" );
    switchDescription["+dump-alt-rxnor"] = 
    				"Use alternate rxnor op ^~";
    switches.push_back( "+dump-fps" );
    switchDescription["+dump-fps"] = 
    				"Force part selects on all vectors";
    switches.push_back( "+dump-fvl" );
    switchDescription["+dump-fvl"] = 
    				"Convert all logical ops to vector ops";
    switches.push_back( "+dump-fcc" );
    switchDescription["+dump-fcc"] = 
    				"Convert all comments to c++ style";
    switches.push_back( "+dump-swda" );
    switchDescription["+dump-swda"] = 
    				"Split wire declaration assignments";
    switches.push_back( "+dump-fbae" );
    switchDescription["+dump-fbae"] = 
    				"Force begin/end block after event statement";
    switches.push_back( "+dump-nns" );
    switchDescription["+dump-nns"] = 
    				"Replace null statements with begin/end pairs";
    switches.push_back( "+dump-sabo" );
    switchDescription["+dump-sabo"] = 
    				"Add spaces around binary operators";
    switches.push_back( "+dump-ced" );
    switchDescription["+dump-ced"] = 
    				"Create explicit declarations for implicitly declared variables";
    switches.push_back( "+dump-strip-attr" );
    switchDescription["+dump-strip-attr"] = 
    				"strip out all attributes";

    switches.push_back( "+dump-timescale=<timescale>" );
    switchDescription["+dump-timescale=<timescale>"] = 
    				"emit `timescale <timescale> statements for all modules missing `timescale declarations";

    switches.push_back( "+dump-simplify" );
    switchDescription["+dump-simplify"] = 
    				"Simplify constant expressions (EXPERIMENTAL)";
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CDumpVerilog::GetToolName( void )
{
	static char	toolName[] = "dump";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CDumpVerilog::GetToolDescription( void )
{
	static char	toolDescription[] = "Print verilog output";

	return toolDescription;
}

/*******************************************************
	Tab
	- output current tab spacing 
********************************************************/

static void Tab( FILE* f )
{
	int i;
	if( suppressTab ) {
	    suppressTab = 0;
	    return;
	}

	for( i = 0; i < tabCount*TAB_SPACE; i++ ) {
		fprintf( f, " " );
	}
}

/*******************************************************
	DumpList
	- convert to verilog code 
********************************************************/

static void DumpList( FILE* f, CNode* n )
{
	NodeOp_t op;
	CNode*   l;
	CNode*   r;

	op = n->GetOp();
	l  = n->Arg<CNode*>(0);
	r  = n->Arg<CNode*>(1);
	
	if( l->GetOp() == op ) {
		DumpList( f, l );
	} else {
		DumpExpression( f, l );
	} 
	if( r != NULL ) {
	    fputc( ',', f );
	    if( r->GetOp() == op ) {
		DumpList( f, r );
	    } else {
		DumpExpression( f, r );
	    } 
	}
}

/*******************************************************
	DumpUnary
	- convert to verilog code 
********************************************************/

static void DumpUnary( FILE* f, CNode* n, const char* opString )
{
	int precN = n->Precedence();
	int prec0 = n->Arg<CNode*>(0)->Precedence();
	int insertParen = (prec0 < precN);
	switch( n->Arg<CNode*>(0)->GetOp() ) {
	case eNOT:
	case eCOM:
	case eNEG:
	case ePLUS:
	case eRAND:
	case eROR:
	case eRNAND:
	case eRNOR:
	case eRXOR:
	case eRXNOR:
		insertParen = 1;
	default:;
	}
	fprintf( f, "%s", opString );
	DumpAttributes( f, n->GetAttributes() );
	if( insertParen ) {
	    fprintf( f, "(" );
	}
	DumpExpression( f, n->Arg<CNode*>(0) );
	if( insertParen ) {
	    fprintf( f, ")" );
	}
}

/*******************************************************
	DumpUnaryRight
	- convert to verilog code 
********************************************************/

static void DumpUnaryRight( FILE* f, CNode* n, const char* opString )
{
	int precN = n->Precedence();
	int prec0 = n->Arg<CNode*>(0)->Precedence();
	int insertParen = (prec0 < precN);
	switch( n->Arg<CNode*>(0)->GetOp() ) {
	case eNOT:
	case eCOM:
	case eNEG:
	case ePLUS:
	case eRAND:
	case eROR:
	case eRNAND:
	case eRNOR:
	case eRXOR:
	case eRXNOR:
		insertParen = 1;
	default:;
	}
	if( insertParen ) {
	    fprintf( f, "(" );
	}
	DumpExpression( f, n->Arg<CNode*>(0) );
	if( insertParen ) {
	    fprintf( f, ")" );
	}
	DumpAttributes( f, n->GetAttributes() );
	fprintf( f, "%s", opString );
}
/*******************************************************
	DumpPortList
	- dump module port list 
********************************************************/

static int MoreArgs( CNode* n ) {
	if( n == NULL ) {
		return FALSE;
	}
	switch( n->GetOp() ) {
	case eLIST:
	    {
	        int result = 0;
		while( 1 ) {
		    result |= MoreArgs(n->Arg<CNode*>(0));
		    n = n->Arg<CNode*>(1);
		    if( !n || n->GetOp() != eLIST ) {
			result |= MoreArgs(n);
			break;
		    }
		}
	    	return result;
	    }
	case ePRAGMA:
	case eCOMMENT:
	case eVRQ:
		return FALSE;
	case eARG:
		return TRUE;
	default:
		MASSERT( FALSE );
	}
	return FALSE;
}

static void DumpPortList( FILE* f, CNode* n, int moreArgs )
{
	if( n == NULL ) {
		Tab( f );
		if( moreArgs ) {
			fprintf( f, "," );
		}
		fprintf( f, "\n" );
	} else if( n->GetOp() == eLIST ) {
		DumpPortList( f, n->Arg<CNode*>(0), 
			moreArgs||MoreArgs(n->Arg<CNode*>(1)) );
		DumpPortList( f, n->Arg<CNode*>(1), moreArgs );
	} else if( n->GetOp() == eCOMMENT ||
		   n->GetOp() == eVRQ ||
		   n->GetOp() == ePRAGMA ) {
		Tab( f );
		DumpComment( f, n );
	} else {
		Tab( f );
		DumpExpression( f, n );
		if( moreArgs ) {
			fprintf( f, "," );
		}
		fprintf( f, "\n" );
	}
}
/************************************************
	StripEdges
	- remove all edge specs from event 
	  expression converting it to level sensitive
 ************************************************/
static CNode*	StripEdges( CNode* n ,void*)
{
   if( n->GetOp() == ePOSEDGE ||
       n->GetOp() == eNEGEDGE ||
       n->GetOp() == eEDGE ) {
	return n->Arg<CNode*>(0);
   }
   return n;
}

/*******************************************************
	DumpInstance
	- convert to verilog code 
********************************************************/

static void DumpInstance( FILE* f, CInstance* inst )
{
	/*
 	 * dump attributes
 	 */
	DumpAttributes( f, inst->GetAttributes() );
	CNode* params = inst->GetParameters();
	fprintf( f, "%s ", Identifier(inst->GetDefinition()->GetName()) );
	if( params != NULL ) {
               if( params->GetOp() == eLIST ) {
                   if( params->Arg<CNode*>(0)->GetOp() == ePRAGMA ) {
                       DumpExpression( f, params->Arg<CNode*>(0) );
                       params = params->Arg<CNode*>(1);
                   }
               }
               CNode* postComment = NULL;
               if( params->GetOp() == eLIST ) {
                   if( params->Arg<CNode*>(1)->GetOp() == ePRAGMA ) {
                       postComment = params->Arg<CNode*>(1);
                       params = params->Arg<CNode*>(0);
                   }
               }
       	fprintf( f, "#" );
       	fprintf( f, "(" );
       	DumpExpressionList( f, params );
       	fprintf( f, ")" ); 
               if( postComment ) {
                   DumpExpression( f, postComment );
               }
		fprintf( f, " " ); 
	}
	fprintf( f, "%s", Identifier(inst->GetName()) );
	DumpUnpackedDimensions( f, inst );
	fprintf( f, " (" );
	CNode* args = inst->GetArguments();
	if( args && (args->GetOp() != eARG || 
		     args->Arg<CNode*>(0) ||
		     args->Arg<CNode*>(1)) ) {
		fprintf( f, "\n" );
		tabCount++;
		DumpPortList( f, inst->GetArguments(), FALSE );
		tabCount--;
	}
	Tab(f);
	fprintf( f, ");\n" ); 


}

/*******************************************************
	DumpGate
	- convert to verilog code 
********************************************************/

static void DumpGate( FILE* f, CGate* inst )
{
	DumpAttributes( f, inst->GetAttributes() );
	fprintf( f, "%s ", Identifier(inst->GetType()->GetName()) );
	if( inst->Strength0() != eUNDEFINED ||
	    inst->Strength1() != eUNDEFINED ) {
	    StrengthPair_t pair = { inst->Strength0(), inst->Strength1() };
	    DumpStrengthPair( f, &pair );
	    fprintf( f, " " );
	}
	CNode* delay = inst->GetDelayControl();
	if( delay != NULL ) {
		int simple = delay->GetOp() != eELIST &&
			     delay->GetOp() != eMTM;
		fprintf( f, "#" );
		if( !simple ) {
		    fprintf( f, "(" );
		}
		DumpExpression( f, inst->GetDelayControl() );
		if( !simple ) {
		    fprintf( f, ") " ); 
		} else {
		    fprintf( f, " " ); 
		}
	}
	if( inst->GetSymbol() ) {
	    fprintf( f, "%s ", Identifier(inst->GetName()) );
	}
	DumpUnpackedDimensions( f, inst );
	if( inst->GetNumberOfUnpackedDimensions() ) {
	    fprintf( f, " " ); 
	}
	fprintf( f, "(" ); 
	if( inst->GetArguments() != NULL ) {
		DumpExpression( f, inst->GetArguments() );
	}
	fprintf( f, ");\n" ); 
}

/*******************************************************
	LeftMostOp
	- return the left most non-binary expression op
********************************************************/

static CNode* LeftMostOp( CNode* n )
{
	while( 1 ) {
	    switch( n->GetOp() ) {
	    case eRSHA:
    	    case eLSHA:
	    case eRSH:
    	    case eLSH:
    	    case eLE:
    	    case eLT:
    	    case eGE:
    	    case eGT:
    	    case eEQ:
    	    case eNE:
    	    case eLAND:
    	    case eLOR:
    	    case eCEQ:
    	    case eCNE:
    	    case eADD:
    	    case eSUB:
    	    case eDIV:
    	    case ePOW:
    	    case eMUL:
    	    case eMOD:
    	    case eOR:
    	    case eAND:
    	    case eANDANDAND:
    	    case eXOR:
    	    case eXNOR:
	    case eHOOK:
    	    	break;
    	    default:
		return n;
	    }
	    n = n->Arg<CNode*>(0);
	}
}

/*******************************************************
	DumpBinary
	- convert to verilog code 
********************************************************/

static void DumpBinary( FILE* f, CNode* n, const char* op )
{
	int precN = n->Precedence();
	int prec0 = n->Arg<CNode*>(0)->Precedence();
	int prec1 = n->Arg<CNode*>(1)->Precedence();
	if(  prec0 < precN ) {
	    fputc( '(', f );
	}
	DumpExpression( f, n->Arg<CNode*>(0) );
	if(  prec0 < precN ) {
	    fputc( ')', f );
	}
	if( spaceAroundBinaryOps ) {
	    fprintf( f, " %s ", op );
	} else {
	    fprintf( f, "%s", op );
	}
	DumpAttributes( f, n->GetAttributes() );
	CNode* leftMostOp = LeftMostOp( n->Arg<CNode*>(1) );
	NodeOp_t op1 = n->GetOp();	
	NodeOp_t op2 = leftMostOp->GetOp();	
	int insertParen = (prec1 <= precN) ||
		      (op1 == eXOR && op2 == eCOM) ||
		      (op1 == eSUB && op2 == eNEG) ||
		      (op1 == eADD && op2 == ePLUS) ||
		      ((op1 == eAND || op1 == eLAND) && op2 == eRAND) ||
		      ((op1 == eOR || op1 == eLOR) && op2 == eROR);
	if(  insertParen ) {
	    fputc( '(', f );
	}
	DumpExpression( f, n->Arg<CNode*>(1) );
	if(  insertParen ) {
	    fputc( ')', f );
	}
}

/*******************************************************
	DumpString
	- convert to verilog code 
********************************************************/

static void DumpString( FILE* f, char* s )
{
	fprintf( f, "%c", '"' );
	while( *s != 0 ) {
		switch( *s ) {
		case '\\':
			fprintf( f, "%s", "\\\\" );
			break;
		case '\t':
			fprintf( f, "%s", "\\t" );
			break;
		case '\n':
			fprintf( f, "%s", "\\n" );
			break;
		case '"':
			fprintf( f, "%s", "\\\"" );
			break;
		default:
			fprintf( f, "%c", *s );
		}
		++s;
	}
	fprintf( f, "%c", '"' );
}

/*******************************************************
	ForceUnsized
	- if node is a constant vector, force
	  it to be rendered unsized 
********************************************************/
static void ForceUnsized( CNode* n ) 
{
	if( n == NULL ) {
		return;
	}
	if( n->GetOp() != eVCONSTANT ) {
		return;
	}
	MASSERT( n->GetNodeType() != eR );
	n->Arg<CVector*>(0)->Sized(FALSE);
	n->Arg<CVector*>(0)->SetPreferredBase(10);
}

/*******************************************************
	DumpIndexes
	- convert indexes to verilog code 
********************************************************/

static void DumpIndexes( FILE* f, vector<CNode*>& indexes )
{
	vector<CNode*>::iterator ptr;
	for( ptr = indexes.begin(); ptr != indexes.end(); ++ptr ) {
	    fputc( '[', f );
	    DumpExpression( f, *ptr );
	    fputc( ']', f );
	}
} 
/*******************************************************
	DumpRange
	- convert range to verilog code 
*******************************************************/
static void DumpRange( FILE* f, CNode* range, int collapse ) 
{
    if( !range ) {
	fprintf( f, "[]" );
	return;
    }
    if( range->GetOp() != eRANGE ) {
        fprintf( f, "[" );
	DumpExpression( f, range );
        fprintf( f, "]" );
	return;
    }
    CNode* msi = range->Arg<CNode*>(0);
    CNode* lsi = range->Arg<CNode*>(1);
// ??? mdh - remove this check and always execute this code
    if( !collapse && msi ) {
        ForceUnsized( msi );
    }
    if( !collapse && lsi ) {
        ForceUnsized( lsi );
    }
    fprintf( f, "[" );
    if( lsi ) {
        DumpExpression( f, msi );
        if( !collapse || msi != lsi ) {
            fprintf( f, ":" );
            DumpExpression( f, lsi );	
        }
    } else {
	DumpExpression( f, msi );
    }
    fprintf( f, "]" );
}

/*******************************************************
	DumpExpressionList
	- convert expression lists to verilog code 
********************************************************/

static int DumpExpressionList( FILE* f, CNode* n )
{
    int dumped = 0;

    if( !n ) {
	return 1;
    }
    if( n->GetOp() != eLIST && 
	n->GetOp() != eELIST ) {
	DumpExpression( f, n );
	return n->GetOp() != eCOMMENT && n->GetOp() != ePRAGMA;
    }
    while( 1 ) {
        dumped = DumpExpressionList( f, n->Arg<CNode*>(0) );
        n = n->Arg<CNode*>(1);
        if( (!suppressNullArgs || n) && dumped ) {
           fputc( ',', f );
        }
        if( !n || (n->GetOp() != eLIST &&
           n->GetOp() != eELIST) ) {
           dumped = DumpExpressionList( f, n );
           break;
        }
    }
    return dumped;
}
/*******************************************************
	DumpExpression
	- convert to verilog code 
********************************************************/
static void DumpExpression( FILE* f, CNode* n, int sized, int insertNewline )
{
	CDecl*   decl;
	CVector* v;
	CFref*   fref;
	CSymbol* sym;
        CVar*    reg;
        CNet*    net;
        CNode*	 msb;
        CNode*	 lsb;

	if( n == NULL ) {
		return;
	}

	switch( n->GetOp() ) {
	case eREPEAT_CONTROL:
		fprintf( f, "repeat(" );
		DumpExpression( f, n->Arg<CNode*>(0), 0 );		
		fprintf( f, ") " );
		DumpExpression( f, n->Arg<CNode*>(1), 0 );		
		break;
	case eEVENT_CONTROL:
		fprintf( f, "@" );
		if( n->Arg<CNode*>(0) == NULL ) {
			fprintf( f, "*" );
		} else {
			fprintf( f, "(" );
			DumpExpression( f, n->Arg<CNode*>(0), 0 );		
			fprintf( f, ")" );
		}
		break;
	case eDELAY_CONTROL:
		fprintf( f, "#(" );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ")" );
		break;
	case eFUNCTION_CALL:
	case eSYSTASK_CALL: {
		fprintf( f, "%s", n->Arg<CSymbol*>(0)->GetName() );
		CNode* attr = n->GetAttributes();
		if( attr ) {
		    MASSERT( n->GetOp() == eFUNCTION_CALL );
		    printf( " " );
		    DumpAttributes( f, attr );
		    printf( " " );
		}
		if( n->Arg<CNode*>(1) != NULL ) {
			fprintf( f, "(" );
			DumpExpression( f, n->Arg<CNode*>(1) );
			fprintf( f, ")" );
		}
		} break;
	case eARG:
		DumpAttributes(f, n->GetAttributes());
		if( n->Arg<CSymbol*>(0) != NULL ) {
                    if( !strcmp(n->Arg<CSymbol*>(0)->GetName(), "*") ) {
			fprintf( f, ".*");
                        break;
                    } else {
			fprintf( f, ".%s(", Identifier(n->Arg<CSymbol*>(0)->GetName()) );
                    }
		}
		DumpExpression( f, n->Arg<CNode*>(1) );
		if( n->Arg<CNode*>(0) != NULL ) {
			fprintf( f, ")" );
		}
		break;
	case eRCONSTANT:
		{ 
		   char* s = n->Arg<char*>(0);
		   fprintf( f, "%s", s);
		}
		break;
	case eVCONSTANT:
		v = n->Arg<CVector*>(0);
//	        MASSERT( v->GetWidth() > 0 );
		if( v->GetPreferredBase() == 0 ) {
		    DumpString( f, v->GetVString() );
		} else {
		    fprintf( f, "%s", v->GetVString() );
		}
		break;
	case eCAT:
		fputc( '{', f );	
		DumpList( f, n );
		fputc( '}', f );	
		break;
	case eUCAT:
		fputc( '{', f );	
		DumpExpression( f, n->Arg<CNode*>(0) );
		fputc( '}', f );	
		break;
	case eENUM_SPEC: {
		CNode* slice = n->Arg<CNode*>(1);
		CNode* exp = n->Arg<CNode*>(2);
	        fprintf( f, "%s", n->Arg<CSymbol*>(0)->GetName() );
		if( slice ) {
		    fputc( '[', f );	
		    DumpExpression( f, slice );
		    fputc( ']', f );	
		}
		if( exp ) {
		    fprintf( f, " = " );
		    DumpExpression( f, exp );
		}
		} break;
	case eSLICE:
		ForceUnsized( n->Arg<CNode*>(0) );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fputc( ':', f );
		ForceUnsized( n->Arg<CNode*>(1) );
		DumpExpression( f, n->Arg<CNode*>(1) );
		break;
	case ePSLICE:
		ForceUnsized( n->Arg<CNode*>(0) );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, "+:" );
		ForceUnsized( n->Arg<CNode*>(1) );
		DumpExpression( f, n->Arg<CNode*>(1) );
		break;
	case eMSLICE:
		ForceUnsized( n->Arg<CNode*>(0) );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, "-:" );
		ForceUnsized( n->Arg<CNode*>(1) );
		DumpExpression( f, n->Arg<CNode*>(1) );
		break;
	case eNET_REF: {
                net = n->Arg<CNet*>(0);
		fprintf( f, "%s", Identifier(net->GetName()));
		CDataType* dataType = net->GetDataType();
                if( sized && forcePartSelect && dataType->IsPacked() &&
		    !dataType->IsVector() &&
 		    (net->GetDataType()->GetVarDataType() == eREG || 
 		    net->GetDataType()->GetVarDataType() == eLOGIC || 
 		    net->GetDataType()->GetVarDataType() == eBIT || 
 		    net->GetDataType()->GetVarDataType() == eNO_VAR_TYPE) &&
		    dataType->GetPackedRange() ) {
			DumpRange( f, dataType->GetPackedRange(), TRUE );
                }
		}break;
	case eVAR_REF: {
                reg = n->Arg<CVar*>(0);
		fprintf( f, "%s", Identifier(reg->GetName()));
		CDataType* dataType = reg->GetDataType();
                if( sized && forcePartSelect && dataType->IsPacked() &&
		    !dataType->IsVector() &&
 		    (reg->GetDataType()->GetVarDataType() == eREG || 
 		    reg->GetDataType()->GetVarDataType() == eLOGIC || 
 		    reg->GetDataType()->GetVarDataType() == eBIT || 
 		    reg->GetDataType()->GetVarDataType() == eNO_VAR_TYPE) &&
		    dataType->GetPackedRange() ) {

		    DumpRange( f, dataType->GetPackedRange(), TRUE );
                }
		} break;
	case eMACRO_EXPR:
		fprintf( f, "%s", (const char*)n->Arg<const char*>(0) );
		break;
	case eGENVAR_REF:
	case ePARAM_REF:
	case eENUM_REF:
        case ePORT_REF:
		decl = n->Arg<CDecl*>(0);
		fprintf( f, "%s", Identifier(decl->GetName()) );
		break;
	case eTYPE_REF: {
		CTypedef* td = n->Arg<CTypedef*>(0);
		DumpDataType( f, td, td->GetDataType() );
		} break;
	case eEXTERNAL_REF:
		sym = n->Arg<CSymbol*>(0);
		fprintf( f, "%s", sym->GetName() );
		break;
	case eARRAY:
		{ 
		    int saveForcePartSelect = forcePartSelect;
		    forcePartSelect = 0;
		    DumpExpression( f, n->Arg<CNode*>(0) );
		    forcePartSelect = saveForcePartSelect;
		    vector<CNode*> indexes;
		    EList2VectorExclude( n->Arg<CNode*>(1), 
					 set<NodeOp_t>(), indexes );
		    DumpIndexes( f, indexes );
		    if( n->Arg<CNode*>(0)->GetOp() == eVAR_REF ) {
			CVar* reg = n->Arg<CNode*>(0)->Arg<CVar*>(0);
			int hasSelect = 
				indexes.size() > reg->GetNumberOfUnpackedDimensions();
			CDataType* dataType = reg->GetDataType();
                 	if( sized && forcePartSelect && 
			    !hasSelect && dataType->IsPacked() &&
			    !dataType->IsVector() &&
			    dataType->GetPackedRange() ) { 
		            DumpRange( f, dataType->GetPackedRange(), FALSE );
                        }
		    }
		}
		break;
	case eMEMBER:
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ".%s", n->Arg<CSymbol*>(1)->GetName() );
		break;
	case eFWD_REF:
		fref = n->Arg<CFref*>(0);
		MASSERT( fref->GetDecl() );
		fprintf( f, "%s", Identifier((fref->GetDecl())->GetName()) );
		break;
	case eRSHA:
		DumpBinary( f, n, ">>>" );
		break;
	case eLSHA:
		DumpBinary( f, n, "<<<" );
		break;
	case eRSH:
		DumpBinary( f, n, ">>" );
		break;
	case eLSH:
		DumpBinary( f, n, "<<" );
		break;
	case eLE:
		DumpBinary( f, n, "<=" );
		break;
	case eLT:
		DumpBinary( f, n, "<" );
		break;
	case eGE:
		DumpBinary( f, n, ">=" );
		break;
	case eGT:
		DumpBinary( f, n, ">" );
		break;
	case eEQ:
		DumpBinary( f, n, "==" );
		break;
	case eNE:
		DumpBinary( f, n, "!=" );
		break;
	case eNOT:
                DumpUnary( f, n, "!" );
		break;
	case eLAND:
                DumpBinary( f, n, "&&" );
		break;
	case eLOR:
                DumpBinary( f, n, "||" );
		break;
	case eCOM:
		DumpUnary( f, n, "~" );
		break;
	case ePREDEC:
		DumpUnary( f, n, "--" );
		break;
	case ePREINC:
		DumpUnary( f, n, "++" );
		break;
	case ePOSTDEC:
		DumpUnaryRight( f, n, "--" );
		break;
	case ePOSTINC:
		DumpUnaryRight( f, n, "++" );
		break;
	case eCEQ:
		DumpBinary( f, n, "===" );
		break;
	case eCNE:
		DumpBinary( f, n, "!==" );
		break;
	case eADD:
		DumpBinary( f, n, "+" );
		break;
	case eSUB:
		DumpBinary( f, n, "-" );
		break;
	case eDIV:
		DumpBinary( f, n, "/" );
		break;
	case ePOW:
		DumpBinary( f, n, "**" );
		break;
	case eMUL:
		DumpBinary( f, n, "*" );
		break;
	case eMOD:
		DumpBinary( f, n, "%" );
		break;
	case eOR:
		DumpBinary( f, n, "|" );
		break;
	case eAND:
		DumpBinary( f, n, "&" );
		break;
	case eANDANDAND:
		DumpBinary( f, n, "&&&" );
		break;
	case eXOR:
		DumpBinary( f, n, "^" );
		break;
	case eXNOR:
		DumpBinary( f, n, (altRxnor ? "^~" : "~^") );
		break;
	case eNEG:
		DumpUnary( f, n, "-" );
		break;
	case ePLUS:
		DumpUnary( f, n, "+" );
		break;
	case eEVOR:
		DumpExpression( f, n->Arg<CNode*>(0), 0 );
		fprintf( f, " or " );
		DumpExpression( f, n->Arg<CNode*>(1), 0 );
		break;
	case eELIST:
	case eLIST:
	    while( 1 ) {
		DumpExpression( f, n->Arg<CNode*>(0), sized, insertNewline );
		n = n->Arg<CNode*>(1);
		if( !suppressNullArgs || n ) {
		    fputc( ',', f );
		    if( insertNewline ) {
			fputc( '\n', f );
			Tab( f );
		    }
		}
		if( !n || (n->GetOp() != eLIST &&
			   n->GetOp() != eELIST) ) {
		    DumpExpression( f, n );
		    break;
		}
	    }
	    break;	
	case eWIDTH:
	        DumpExpression( f, n->Arg<CNode*>(1) );
		break;
	case eCVRI:
	case eCVIR:
		DumpExpression( f, n->Arg<CNode*>(0) );
		break;
	case eNEGEDGE:
		fprintf( f, "negedge " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		break;
	case ePOSEDGE:
		fprintf( f, "posedge " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		break;
	case eEDGE:
		{
		    Edge_t edge = (Edge_t)(int)n->Arg<int>(1);
		    int first = 1;
		    fprintf( f, "edge [" );
		    if( edge & eEDGE01 ) {
			if( !first ) {
			    fprintf( f, "," );
			}
		        fprintf( f, "01" ); 	
			first = 0;
		    }
		    if( edge & eEDGE10 ) {
			if( !first ) {
			    fprintf( f, "," );
			}
		        fprintf( f, "10" ); 	
			first = 0;
		    }
		    if( edge & eEDGE0x ) {
			if( !first ) {
			    fprintf( f, "," );
			}
		        fprintf( f, "0x" ); 	
			first = 0;
		    }
		    if( edge & eEDGE1x ) {
			if( !first ) {
			    fprintf( f, "," );
			}
		        fprintf( f, "1x" ); 	
			first = 0;
		    }
		    if( edge & eEDGEx0 ) {
			if( !first ) {
			    fprintf( f, "," );
			}
		        fprintf( f, "x0" ); 	
			first = 0;
		    }
		    if( edge & eEDGEx1 ) {
			if( !first ) {
			    fprintf( f, "," );
			}
		        fprintf( f, "x1" ); 	
			first = 0;
		    }
		    fprintf( f, "] " );
		    DumpExpression( f, n->Arg<CNode*>(0) );
		}
		break;
	case eRAND:
		DumpUnary( f, n, "&" );
		break;
	case eROR:
		DumpUnary( f, n, "|" );
		break;
	case eRNAND:
		DumpUnary( f, n, "~&" );
		break;
	case eRNOR:
		DumpUnary( f, n, "~|" );
		break;
	case eRXOR:
		DumpUnary( f, n, "^" );
		break;
	case eRXNOR:
		DumpUnary( f, n, "~^" );
		break;
	case eREP:
		fputc( '{', f );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fputc( '{', f );
		if( n->Arg<CNode*>(1)->GetOp() == eCAT ) {
		    DumpList( f, n->Arg<CNode*>(1) );
		} else {
		    DumpExpression( f, n->Arg<CNode*>(1) );
		}
		fprintf( f, "}}" );
		break;	
	case eMTM:
		DumpExpression( f, n->Arg<CNode*>(0) );
		fputc( ':', f );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fputc( ':', f );
		DumpExpression( f, n->Arg<CNode*>(2) );
		break;
	case eCAST:
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, "\'(" );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, ")" );
		break;
	case eHOOK:
		hookCount++;
		if( hookCount == 1 ) {
		    tabCount++;
		}
		{ 
		    int precN = n->Precedence();
	            int prec0 = n->Arg<CNode*>(0)->Precedence();
	            if(  prec0  <= precN ) {
	                fputc( '(', f );
		    }
		    DumpExpression( f, n->Arg<CNode*>(0) );
		    if(  prec0 <= precN ) {
	                fputc( ')', f );
	            }
		}
		fprintf( f, " ? " );
		DumpAttributes( f, n->GetAttributes() );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, " : " );
		if( n->Arg<CNode*>(2)->GetOp() == eHOOK ) {
		    fprintf( f, "\n" );
		    Tab(f);
		}
		DumpExpression( f, n->Arg<CNode*>(2) );
		if( hookCount == 1 ) {
		    tabCount--;
		}
		hookCount--;
		break;
	case eCOMMENT:
	case ePRAGMA:
	case eVRQ:
		DumpComment( f, n );
		break;
	default:
		MASSERT( FALSE );
		break;
	}
}


/*******************************************************
	DumpStrengthPair
	- convert to verilog code 
********************************************************/

static void DumpStrengthPair( FILE* f, StrengthPair_t* spair )
{
	int i;

	if( spair->s0 == eUNDEFINED ) {
		return;
	}
	fputc( '(', f );
	for( i = 0; i < 2; i++ ) {
		Strength_t* p;
		if( i == 1 ) {
			fputc( ',', f );
			p = &spair->s1;	
		} else {
			p = &spair->s0;	
		}
		switch( *p ) {
		case eSUPPLY:
			fprintf( f, "supply%d", i );
			break;
		case eSTRONG:
			fprintf( f, "strong%d", i );
			break;
		case ePULL:
			fprintf( f, "pull%d", i );
			break;
		case eLARGE:
			fprintf( f, "large" );
			i++;
			break;
		case eWEAK:
			fprintf( f, "weak%d", i );
			break;
		case eMEDIUM:
			fprintf( f, "medium" );
			i++;
			break;
		case eSMALL:
			fprintf( f, "small" );
			i++;
			break;
		case eHIGHZ:
			fprintf( f, "highz%d", i );
			break;
		default:
			MASSERT( FALSE );
		}
	}
	fputc( ')', f );
}

/*******************************************************
	DumpCaseItem
	- convert to verilog code 
********************************************************/

static int DumpCaseItem( FILE* f, CNode* n, int initialItem, int needNewline )
{
	if( needNewline ) {
	    fprintf( f, "\n" );
	    needNewline = 0;
	}
	if( n->GetOp() == eCOMMENT ||
	    n->GetOp() == eVRQ ) { 
		if( initialItem ) {
		    fprintf( f, "\n" );
		}
		Tab(f);
                DumpComment( f, n );
		return 0;
	}
	if( n->GetOp() == ePRAGMA ) {
		if( initialItem ) {
		    fprintf( f, " " );
		} else {
		    Tab(f);
		}
                DumpComment( f, n );
		return 0;
	}

	if( n->GetOp() == eLIST ) {
	    while( 1 ) {
	        needNewline = DumpCaseItem( f, n->Arg<CNode*>(0), 
					initialItem, needNewline );
		initialItem = 0;
		n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eLIST ) {
	    	    needNewline = DumpCaseItem( f, n, initialItem, needNewline );
		    break;
		}
	    }
	    return needNewline;
	}
	
	if( initialItem ) {
	    fprintf( f, "\n" );
	}

	Tab( f );
	if( n->Arg<CNode*>(0) != NULL ) {
		DumpExpression( f, n->Arg<CNode*>(0), 1, 1 );
	} else {
		fprintf( f, "default" );
	}
	fprintf( f, ": " );
	CNode* b = n->Arg<CNode*>(1);
	DumpBlock( f, b, FALSE, FALSE, 
		   b && b->GetOp() != eBLOCK_REF );
	return 0;
}
/*******************************************************
	CountStatements
	- return number of statements (including comments) in list
	  counting current node
********************************************************/

static int CountStatements( CNode* n )
{
	int count = 0;
	if( n == NULL ) {
		return 0;
	}
	switch( n->GetOp() ) {
	case eLIST:
	    while( 1 ) {
		count += CountStatements( n->Arg<CNode*>(0) );
		n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eLIST ) {
		    count += CountStatements( n );
		    break;
		}
	    }
	    return count;
	default:
		return 1;
	}
}

/*******************************************************
	CountRealStatements
	- return number of real statements in list
	  counting current node
********************************************************/

static int CountRealStatements( CNode* n )
{
	int count = 0;
	if( n == NULL ) {
		return 0;
	}
	switch( n->GetOp() ) {
	case ePRAGMA:
	case eCOMMENT:
		return 0;
	case eLIST:
	    while( 1 ) {
		count += CountRealStatements( n->Arg<CNode*>(0) );
		n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eLIST ) {
		    count += CountRealStatements( n );
		    break;
		}
	    }
	    return count;
	default:
		return 1;
	}
}

/*******************************************************
	CountPragmaStatements
	- return number of pragma statements in list
	  counting current node
********************************************************/

static int CountPragmaStatements( CNode* n )
{
	int count = 0;
	if( n == NULL ) {
		return 0;
	}
	switch( n->GetOp() ) {
	case ePRAGMA:
		return 1;
	case eLIST:
	    while( 1 ) {
		count += CountPragmaStatements( n->Arg<CNode*>(0) );
		n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eLIST ) {
		    count += CountPragmaStatements( n );
		    break;
		}
	    }
	    return count;
	default:
		return 0;
	}
}

/*******************************************************
	FirstStatement
	- return the first statement in list
	  counting current node
********************************************************/

static CNode* FirstStatement( CNode* n )
{
	if( n == NULL ) {
		return 0;
	}
	switch( n->GetOp() ) {
	case eLIST:
	    while( 1 ) {
		CNode* rn = FirstStatement( n->Arg<CNode*>(0) );
		if( rn ) {
		    return rn;
		}
		n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eLIST ) {
		    return FirstStatement( n );
		}
	    }
	default:
		return n;
	}
}

/*******************************************************
	FirstRealStatement
	- return the first real statement in list
	  counting current node
********************************************************/

static CNode* FirstRealStatement( CNode* n )
{
	if( n == NULL ) {
		return 0;
	}
	switch( n->GetOp() ) {
	case ePRAGMA:
	case eCOMMENT:
		return NULL;
	case eLIST:
	    while( 1 ) {
		CNode* rn = FirstRealStatement( n->Arg<CNode*>(0) );
		if( rn ) {
		    return rn;
		}
		n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eLIST ) {
		    return FirstRealStatement( n );
		}
	    }
	default:
		return n;
	}
}

/*******************************************************
	DumpWireAttr
	- output wire attribute
********************************************************/
static void DumpWireAttr( FILE* f, Wire_t t )
{
    if( t != eNO_WIRE_TYPE ) {
	fprintf( f, "%s ", wireName[t] );
    }
}

/*******************************************************
	DumpDeclType
	- output decl type
********************************************************/
static void DumpDeclType( FILE* f, Decl_t dt, Wire_t wt )
{
    if( dt == eNONE ) {
    } else if( dt == eNET ) {
	fprintf( f, "%s", wireName[wt] );
    } else {
	fprintf( f, "%s", declName[dt] );
    }
}

/*******************************************************
	DumpBlock
	- convert to verilog code 
	  This is called for a new block of statements
	  return TRUE if output is cascading
********************************************************/

static int DumpBlock( FILE* f, CNode* n, int indent, int cascade, int begin ) 
{
	int cascadeOut = FALSE;
	if( indent ) {
		Tab( f );
	}
	if( n == NULL ) {
	    if( noNullStatements ) {
		fprintf( f, "begin end\n" );
	    } else {
		fprintf( f, ";\n" );
	    }
	    return cascadeOut;
	}
	int s = CountStatements(n);
	int r = CountRealStatements(n);
	int p = CountPragmaStatements(n);
	CNode* firstRealStatement = FirstRealStatement(n);
	/*
 	 * this is special to deal with:
	 *
	 *	if( ) 
	 *	    // pragma ....
	 *	    a = b;
	 *	    // pragma ....
	 *	the last line is not part of the if block
	 *	and creating a begin end pair will mess things up
 	 */
	if( r == 1 && firstRealStatement->GetOp() == eBLOCK_REF ) {
		CNode* firstStatement = FirstStatement(n);
		if( firstStatement->GetOp() != eBLOCK_REF ) {
		    tabCount++;
		}
    		DumpStatement( f, n );
		if( firstStatement->GetOp() != eBLOCK_REF ) {
		    tabCount--;
		}
	} else if(  (begin && p==0) || r == 0 || r > 1 ) {
		fprintf( f, "begin\n" );
		tabCount++;
		Tab( f );
		DumpStatement( f, n );
		tabCount--;
		Tab( f );
		fprintf( f, "end" );	
		if( !cascade ) {
		    fprintf( f, "\n" );	
		} else { 
		    cascadeOut = TRUE;
		}
	} else if( s > 1 ) {
		if( !indent ) {
		    fprintf( f, "\n" );
		}
		tabCount++;
		Tab( f );
		DumpStatement( f, n );
		tabCount--;
	} else {
		DumpStatement( f, n );
	}
	return cascadeOut;
}
	
/*******************************************************
	DumpPortDirDecl
	- Dump Portdir declaration
********************************************************/
static void DumpPortDirDecl( FILE* f, CPortDir* p )
{
    /*
     * dump attributes
     */
    DumpAttributes( f, p->GetAttributes() );
    switch( p->GetType() ) {
    case eINPUT:
	fprintf( f, "input" );
	break;
    case eOUTPUT:
	fprintf( f, "output" );
	break;
    case eINOUT:
	fprintf( f, "inout" );
	break;
    case eNODIR:
	break;
    default:
	MASSERT( FALSE );
    }
    fprintf( f, " " );
    DumpWireAttr( f, p->GetWireAttr() );
    DumpDataType( f, p, p->GetDataType() );
    fprintf( f, "%s", Identifier(p->GetName()) );
    DumpUnpackedDimensions( f, p );
}   

/*******************************************************
	DumpAttributes
	- dump attributes
********************************************************/

static void DumpAttributes( FILE* f, CNode* n, int last ) 
{
    if( !n ) {
	return;
    }
    switch( n->GetOp() ) {
    case eLIST:
	while(1) {
	    DumpAttributes( f, n->Arg<CNode*>(0), FALSE );
	    if( n->Arg<CNode*>(0) != NULL &&
	        n->Arg<CNode*>(1)->GetOp() != eATTRIBUTE ) {
		Tab( f );
  	    }
	    n = n->Arg<CNode*>(1);
	    if( !n || n->GetOp() !=eLIST ) {
	   	DumpAttributes( f, n, last );
		break;
	    }
	}
	return;	
    case eATTRIBUTE:
	break;
    case eVRQ:
    case ePRAGMA:
    case eCOMMENT:
	DumpComment( f, n) ;
	if( last ) {
	    Tab( f );
	}
	return;
    default:
	MASSERT( FALSE );
    }
   
    if( stripAttributes ) {
	return;
    }

    CAttr* attr = n->Arg<CAttr*>(0);
    int terminateAttribute = 0;
    for( int i = 0; i < attr->GetAttributeCount(); ++i ) {
	const char* name = attr->GetAttributeName(i);
	CNode*      exp  = attr->GetAttributeExpression(i);
	/*
  	 * hide any attributes that start with . 
  	 * as they are used for internal communication 
  	 * between tools
  	 */
	if( !strncmp(".", name, 1 ) ) {
	    continue;
	}
	if( i != 0 ) {
	    fprintf( f, ", " );
	} else {
    	    fprintf( f, "(* " );
	    terminateAttribute = 1;
	}
	if( !exp ) {
	    fprintf( f, "%s", name );
	} else {
	    fprintf( f, "%s=", name );
	    DumpExpression( f, exp );
	}
    }
    if( terminateAttribute ) {
        fprintf( f, " *) " );
    }
}
/*******************************************************
	DumpUnpackedDimensions
	- convert to verilog code 
********************************************************/
static void DumpUnpackedDimensions( FILE* f, CDecl* d )
{
    for( int i = 0; i < d->GetNumberOfUnpackedDimensions(); ++i ) {
        DumpRange( f, d->GetUnpackedRange(i), FALSE );
    }
}

/*******************************************************
	DumpStatement
	- convert to verilog code 
********************************************************/

static void DumpStatement( FILE* f, CNode* n ) 
{
	if( n == NULL ) {
		return;
	}

	switch( n->GetOp() ) {
	case ePREDEC:
		DumpUnary( f, n, "--" );
		if( !inFor ) {
	            fprintf( f, ";\n" );
		}
		break;
	case ePREINC:
		DumpUnary( f, n, "++" );
		if( !inFor ) {
	            fprintf( f, ";\n" );
		}
		break;
	case ePOSTDEC:
		DumpUnaryRight( f, n, "--" );
		if( !inFor ) {
	            fprintf( f, ";\n" );
		}
		break;
	case ePOSTINC:
		DumpUnaryRight( f, n, "++" );
		if( !inFor ) {
	            fprintf( f, ";\n" );
		}
		break;
	case ePORTLIST_END:
	    suppressTab = 1;
	    break;
	case eATTRIBUTE:
	    DumpAttributes( f, n );
	    fprintf( f, ";\n" );
	    break;
	case eSPECPARAM_DECL:
	    {
		CParam* p = n->Arg<CParam*>(0);
		DumpAttributes( f, p->GetAttributes() );
		fprintf( f, "specparam %s", Identifier(p->GetName()) );
		CNode* expr = p->GetExpression();
		MASSERT( expr != NULL ); 
		fprintf( f, "=" );
		DumpExpression( f, expr );
		fprintf( f, ";\n" );
	    }
	    break;
	case ePARAM_DECL: {
	    CParam* p = n->Arg<CParam*>(0);
	    if( !inlineParams || !p->GetInlined() ) {
		DumpParameterDecl( f, p );
		fprintf( f, ";\n" );
	    } else {
		suppressTab = 1;
	    }
	    } break;
	case eTRIGGER:
	    {
		fprintf( f, "-> " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ";\n" );
	    }
	    break;
	case eRETURN:
	    {
		if( n->Arg<CNode*>(0) ) {
		    fprintf( f, "return " );
		    DumpExpression( f, n->Arg<CNode*>(0) );
		} else {
		    fprintf( f, "return" );
		}
		fprintf( f, ";\n" );
	    }
	    break;
	case ePORT_DECL:
	    {
		    CPortDir* p = n->Arg<CPortDir*>(0);
		    DumpPortDirDecl( f, p );
		    MASSERT( !inlinePorts );
		    fprintf( f, ";\n" );
	    }
	    break;
	case eNET_DECL:
	    {
		CNet*  net = n->Arg<CNet*>(0);
		CNode* delay = net->GetDelay();
		DumpAttributes( f, net->GetAttributes() );
		DumpWireAttr( f, net->GetWireAttr() );
		if( n->Arg<CNode*>(1) ) {
		    CNode* c = n->Arg<CNode*>(1);
		    StrengthPair_t* st = c->Arg<StrengthPair_t*>(0);
		    DumpStrengthPair( f, st );
		    if( st->s0 != eUNDEFINED ) {
			fputc( ' ', f );
		    }
		} else {
		    StrengthPair_t st = net->GetStrength();
		    DumpStrengthPair( f, &st );
		    if( st.s0 != eUNDEFINED ) {
			fputc( ' ', f );
		    }
		}
		if( net->GetVectored() ) {
		    fprintf( f, "vectored " );
		}
		if( net->GetScalared() ) {
		    fprintf( f, "scalared " );
		}
    	        DumpDataType( f, net, net->GetDataType() );
                if( delay ) {
		    DumpExpression( f, delay );
		    fprintf( f, " " );
		}
		if( n->Arg<CNode*>(1) ) {
			CNode* c = n->Arg<CNode*>(1);
                        if (splitWireDeclAssignments) {
        			DumpExpression( f, c->Arg<CNode*>(2), 0 );
        			fprintf( f, "; assign " );
        			DumpExpression( f, c->Arg<CNode*>(2) );
        			fputc( '=', f );
        			DumpExpression( f, c->Arg<CNode*>(3) );
        			fprintf( f, ";\n" );
                        }
                        else
                        {
        			DumpExpression( f, c->Arg<CNode*>(1) );
				if( c->Arg<CNode*>(1) ) {
				    fputc( ' ', f );
				}
        			DumpExpression( f, c->Arg<CNode*>(2), 0 );
        			fputc( '=', f );
        			DumpExpression( f, c->Arg<CNode*>(3) );
        			fprintf( f, ";\n" );
                        }
		} else {
			fprintf( f, "%s", Identifier(net->GetName()) );
			DumpUnpackedDimensions( f, net );
			fprintf( f, ";\n" );
		}
	    }
	    break;
	case eGENVAR_DECL:
		{
			CGenvar* g = n->Arg<CGenvar*>(0);
		        DumpAttributes( f, g->GetAttributes() );
			fprintf( f, "%s %s;\n", 
				declName[g->GetType()],
				Identifier(g->GetName()) );
		}
		break;
	case eVAR_DECL:
	    { 
		CVar* r = n->Arg<CVar*>(0);
		DumpAttributes( f, r->GetAttributes() );
		DumpDeclSpec( f, r->GetType(), r );
		if( n->Arg<CNode*>(1) ) {
		    CNode* c = n->Arg<CNode*>(1);
            	    fprintf( f, " = ");
        	    DumpExpression( f, c->Arg<CNode*>(2) );
		}
		if( !inFor ) {
		    fprintf( f, ";\n" );
		}
	    }
	    break;
	case eTYPEDEF_DECL:
	    { 
		CTypedef* td = n->Arg<CTypedef*>(0);
		DumpAttributes( f, td->GetAttributes() );
		fprintf( f, "typedef " );
		if( td->GetDataType() ) {
		    DumpDataType( f, td, td->GetDataType() );
		}
		fprintf( f, "%s", td->GetName() );
    	 	DumpUnpackedDimensions( f, td );
		fprintf( f, ";\n", td->GetName() );
	    }
	    break;
	case eTASK_ENABLE:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "%s", n->Arg<CSymbol*>(0)->GetName() );
		if( n->Arg<CNode*>(1) != NULL ) {
			fprintf( f, "(" );
			DumpExpression( f, n->Arg<CNode*>(1) );
			fprintf( f, ")" );
		}
		fprintf( f, ";\n" );
		break;
	case eFUNCTION_CALL:
		fprintf( f, "%s(", Identifier(n->Arg<CSymbol*>(0)->GetName()) );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, ");\n" );
		break;
	case eSYSTASK_CALL:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "%s", n->Arg<CSymbol*>(0)->GetName() );
                if( n->Arg<CNode*>(1) != NULL ) {
			fprintf( f, "(" );
			DumpExpression( f, n->Arg<CNode*>(1) );
			fprintf( f, ")" );
		}
		fprintf( f, ";\n" );
		break;
	case eTIMING_CALL:
		fprintf( f, "%s(", n->Arg<CSymbol*>(0)->GetName() );
		suppressNullArgs = 1;
		DumpExpression( f, n->Arg<CNode*>(1) );
		suppressNullArgs = 0;
		fprintf( f, ");\n" );
		break;
	case eINSTANCE_REF:
		DumpInstance( f, n->Arg<CInstance*>(0) );
		break;
	case eGATE_REF:
		DumpGate( f, n->Arg<CGate*>(0) );
		break;
	case eMODULE_DEF:
		DumpModule( f, n->Arg<CModule*>(0) );
		break;
	case eFUNCTION_DEF:
		DumpFunction( f, n->Arg<CFunction*>(0) );
		break;
	case eLIST:
		while(1) {
		    DumpStatement( f, n->Arg<CNode*>(0) );
		    if( n->Arg<CNode*>(0) != NULL &&
		        n->Arg<CNode*>(1)->GetOp() != eATTRIBUTE ) {
		        Tab( f );
		    }
		    n = n->Arg<CNode*>(1);
		    if( !n || n->GetOp() !=eLIST ) {
		    	DumpStatement( f, n );
			break;
		    }
		}
		break;	
	case eALWAYS:
	case eALWAYS_FF:
	case eALWAYS_COMB:
	case eALWAYS_LATCH:
		DumpAttributes( f, n->GetAttributes() );
		switch( n->GetOp() ) {
		case eALWAYS:
		    fprintf( f, "always " );
		    break;
		case eALWAYS_FF:
		    fprintf( f, "always_ff " );
		    break;
		case eALWAYS_COMB:
		    fprintf( f, "always_comb " );
		    break;
		case eALWAYS_LATCH:
		    fprintf( f, "always_latch " );
		    break;
		default:
		    MASSERT( FALSE );
		}
		DumpBlock( f, n->Arg<CNode*>(0), FALSE, FALSE, FALSE );
		break;
	case eINIT:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "initial " );
		DumpBlock( f, n->Arg<CNode*>(0), FALSE, FALSE, FALSE );
		break;
	case eASSIGN:
	case eADD_ASSIGN:
        case eSUB_ASSIGN:
        case eMUL_ASSIGN:
        case eDIV_ASSIGN:
        case eMOD_ASSIGN:
        case eAND_ASSIGN:
        case eOR_ASSIGN:
        case eXOR_ASSIGN:
        case eLSH_ASSIGN:
        case eRSH_ASSIGN:
        case eLSHA_ASSIGN:
        case eRSHA_ASSIGN:		
		DumpAttributes( f, n->GetAttributes() );
		DumpExpression( f, n->Arg<CNode*>(1) );
	        switch( n->GetOp() ) {
	        case eASSIGN:
		    fprintf( f, " = " );
		    break;
	        case eADD_ASSIGN:
		    fprintf( f, " += " );
		    break;
                case eSUB_ASSIGN:
		    fprintf( f, " -= " );
		    break;
                case eMUL_ASSIGN:
		    fprintf( f, " *= " );
		    break;
                case eDIV_ASSIGN:
		    fprintf( f, " /= " );
		    break;
                case eMOD_ASSIGN:
		    fprintf( f, " %= " );
		    break;
                case eAND_ASSIGN:
		    fprintf( f, " &= " );
		    break;
                case eOR_ASSIGN:
		    fprintf( f, " |= " );
		    break;
                case eXOR_ASSIGN:
		    fprintf( f, " ^= " );
		    break;
                case eLSH_ASSIGN:
		    fprintf( f, " <<= " );
		    break;
                case eRSH_ASSIGN:
		    fprintf( f, " >>= " );
		    break;
                case eLSHA_ASSIGN:
		    fprintf( f, " <<<= " );
		    break;
                case eRSHA_ASSIGN: 
		    fprintf( f, " >>>= " );
		    break;
		default: 
		    MASSERT( FALSE );
		}
		   
		if( n->Arg<CNode*>(0) != NULL ) {
			DumpExpression( f, n->Arg<CNode*>(0) );
			fprintf( f, " " );
		}
		DumpExpression( f, n->Arg<CNode*>(2) );
		if( !inFor ) {
			fprintf( f, ";\n" );
		}
		break;
	case eRELEASE:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "release " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ";\n" );
		break;
	case eFORCE:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "force " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, " = " );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, ";\n" );
		break;
	case eDISABLE:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "disable " );
		fprintf( f, "%s", n->Arg<CSymbol*>(0)->GetName() );
		fprintf( f, ";\n" );
		break;
	case eDEASSIGN:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "deassign " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ";\n" );
		break;
	case ePASSIGN:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "assign " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, " = " );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, ";\n" );
		break;
	case eDEFPARAM:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "defparam " );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, " = " );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, ";\n" );
		break;
	case eNBASSIGN:
		DumpAttributes( f, n->GetAttributes() );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, " <= " );
		if( n->Arg<CNode*>(0) != NULL ) {
			DumpExpression( f, n->Arg<CNode*>(0) );
			fprintf( f, " " );
		}
		DumpExpression( f, n->Arg<CNode*>(2) );
		fprintf( f, ";\n" );
		break;
	case eCASSIGN:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "assign " );
		DumpStrengthPair( f, n->Arg<StrengthPair_t*>(0) );
		fputc( ' ', f );
		DumpExpression( f, n->Arg<CNode*>(1) );
		fputc( ' ', f );
		DumpExpression( f, n->Arg<CNode*>(2) );
		fputc( '=', f );
		DumpExpression( f, n->Arg<CNode*>(3) );
		fprintf( f, ";\n" );
		break;
	case eEVENT: {
		DumpAttributes( f, n->GetAttributes() );
		/*
 		 * The attribute .dump_strip_edges_ifdef=<name> will cause
 		 * the following transformation:
 		 * `ifdef name 
 		 *     @(event_expression with edges stripped)
 		 * `else
 		 *     @(event_expression)
 		 * `endif
 		 *
 		 * The attribute .dump_strip_edges_ifndef=<name> will cause
 		 * the following transformation:
 		 * `ifndef name 
 		 *     @(event_expression with edges stripped)
 		 * `else
 		 *     @(event_expression)
 		 * `endif
 		 */
		if( n->Arg<CNode*>(0) == NULL ) {
		    fprintf( f, "@* " );
		} else {
		    int ifdef = n->HasAttribute( ".dump_strip_edges_ifdef" );
		    int ifndef = n->HasAttribute( ".dump_strip_edges_ifndef" );
		    CAttr* attr = NULL;
		    const char* iftype = NULL;
		    CNode* condExpr = NULL;
		    CNode* expr1 = n->Arg<CNode*>(0);
		    CNode* expr2 = NULL;
		    if( ifdef ) {
			attr = n->GetAttribute( ".dump_strip_edges_ifdef" );
			condExpr = attr->GetAttributeExpression( ".dump_strip_edges_ifdef" );
		        expr2 = expr1->Clone();
		        expr1 = expr1->PostSubVisit1( StripEdges, NULL );
			iftype = "ifdef";
		    } else if( ifndef ) {
			attr = n->GetAttribute( ".dump_strip_edges_ifndef" );
			condExpr = attr->GetAttributeExpression( ".dump_strip_edges_ifndef" );
		        expr2 = expr1->Clone();
		        expr1 = expr1->PostSubVisit1( StripEdges, NULL );
			iftype = "ifndef";
		    }
		    if( attr ) {
		       MASSERT( condExpr->GetOp() == eVCONSTANT );
		       char* condString = condExpr->Arg<CVector*>(0)->GetString();
		       fprintf( f, "\n" );
		       Tab( f );
		       fprintf( f, "`%s %s\n", iftype, condString );
		       tabCount++;
		       Tab( f );
		    }
		    fprintf( f, "@(" );
		    DumpExpression( f, expr1, 0 );
		    fprintf( f, ") " );
		    if( attr ) {
		       fprintf( f, "\n" );
		       tabCount--;
		       Tab( f );
		       tabCount++;
		       fprintf( f, "`else\n" );
		       Tab( f );
		       fprintf( f, "@" );
		       fprintf( f, "(" );
		       DumpExpression( f, expr2, 0 );
		       fprintf( f, ") " );
		       fprintf( f, "\n" );
		       tabCount--;
		       Tab( f );
		       fprintf( f, "`endif\n" );
		       Tab( f );
		    }
		}
		DumpBlock( f, n->Arg<CNode*>(1), FALSE, FALSE, forceBeginAfterEvent );
		} break;
	case eGCASE:
	case eCASE:
	case eCASEX:
	case eCASEZ:
		DumpAttributes( f, n->GetAttributes() );
		switch( n->GetOp() ) {
		case eGCASE:
		case eCASE:
			fprintf( f, "case(" );
			break;
		case eCASEX:
			fprintf( f, "casex(" );
			break;
		case eCASEZ:
			fprintf( f, "casez(" );
			break;
		default:
			MASSERT( FALSE );
		}
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ")" );
		tabCount++;
		DumpCaseItem( f, n->Arg<CNode*>(1),1,0 );
		tabCount--;
		fprintf( f, "\n" );
		Tab( f );
		fprintf( f, "endcase\n" );
		break;
	case ePRAGMA:
	case eCOMMENT:
	case eVRQ:
		DumpComment( f, n) ;
		break;
	case eGIF:
	case eIF:
		{
			CNode* tStmt = n->Arg<CNode*>(1);
			CNode* fStmt = n->Arg<CNode*>(2);
		        DumpAttributes( f, n->GetAttributes() );
			fprintf( f, "if(" );
			DumpExpression( f, n->Arg<CNode*>(0) );
			fprintf( f, ") " ); 
			int cascade = DumpBlock( f, tStmt, 
					FALSE, fStmt!=NULL, 
					tStmt && tStmt->GetOp() != eBLOCK_REF );
			if( fStmt == NULL ) {
				if( cascade ) {
				     fprintf( f, "\n" );
				}
				break;
			}
			if( !cascade ) {
			    Tab( f );
			} else {
			    fprintf( f, " " ); 
			}
			fprintf( f, "else " ); 
			DumpBlock( f, fStmt, FALSE, FALSE, 
			        fStmt->GetOp() != eIF &&
			        fStmt->GetOp() != eBLOCK_REF &&
				fStmt->GetOp() != eGIF );
		}
		break;	
	case eWHILE:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "while(" );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ") " ); 
		DumpBlock( f, n->Arg<CNode*>(1), FALSE, FALSE, FALSE);
		break;	
	case eWAIT:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "wait(" );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ") " ); 
		DumpBlock( f, n->Arg<CNode*>(1), FALSE, FALSE, FALSE);
		break;	
	case eREPEAT:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "repeat(" );
		DumpExpression( f, n->Arg<CNode*>(0) );
		fprintf( f, ") " ); 
		DumpBlock( f, n->Arg<CNode*>(1), FALSE, FALSE, FALSE  );
		break;	
	case eFOREVER:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "forever " );
		DumpBlock( f, n->Arg<CNode*>(0), FALSE, FALSE, FALSE );
		break;	
	case eDELAY:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "#" );
		if( n->Arg<CNode*>(0)->GetOp() != eVCONSTANT &&
		    n->Arg<CNode*>(0)->GetOp() != eRCONSTANT ) {
		    fprintf( f, "(" );
		}
		DumpExpression( f, n->Arg<CNode*>(0) );
		if( n->Arg<CNode*>(0)->GetOp() != eVCONSTANT &&
		    n->Arg<CNode*>(0)->GetOp() != eRCONSTANT ) {
		    fprintf( f, ")" );
		}
		fprintf( f, " " );
		DumpBlock( f, n->Arg<CNode*>(1), FALSE, FALSE, FALSE);
		break;	
	case eGFOR:
	case eFOR:
		DumpAttributes( f, n->GetAttributes() );
		fprintf( f, "for(" );
		inFor = TRUE;
		DumpStatement( f, n->Arg<CNode*>(0) );
		fprintf( f, ";" ); 
		DumpExpression( f, n->Arg<CNode*>(1) );
		fprintf( f, ";" ); 
		DumpStatement( f, n->Arg<CNode*>(2) );
		inFor = FALSE;
		fprintf( f, ") " );
		{ 
		   CNode* b = n->Arg<CNode*>(3);
		   DumpBlock( f, n->Arg<CNode*>(3), 
			      FALSE, FALSE, 
			      b && b->GetOp() != eBLOCK_REF );
		}
		break;	
	case eBLOCK_REF:
		DumpAttributes( f, n->GetAttributes() );
		DumpNamedBlock( f, n->Arg<CBlock*>(0), n->Arg<CNode*>(1), FALSE );
		break;
	case eSPECIFY_REF:
		DumpAttributes( f, n->GetAttributes() );
		DumpSpecifyBlock( f, n->Arg<CSpecify*>(0), n->Arg<CNode*>(1) );
		break;
	case ePATH_ASSIGN:
		{
		    CNode* cond = n->Arg<CNode*>(0);
		    if( cond ) {
			fprintf( f, "if( " );
			DumpExpression( f, cond );
			fprintf( f, " ) " );
		    }
		    DumpPath( f, n->Arg<CNode*>(1) );
		    fprintf( f, " = " );
		    if( n->Arg<CNode*>(2)->GetOp() == eELIST ) {
		        fprintf( f, "(" );
	            }
		    DumpExpression( f, n->Arg<CNode*>(2) );
		    if( n->Arg<CNode*>(2)->GetOp() == eELIST ) {
		        fprintf( f, ")" );
	            }
		    fprintf( f, ";\n" );
		}
		break;
	case eIFNONE_PATH_ASSIGN:
		{
		    fprintf( f, "ifnone " );
		    DumpPath( f, n->Arg<CNode*>(0) );
		    fprintf( f, " = " );
		    if( n->Arg<CNode*>(1)->GetOp() == eELIST ) {
		        fprintf( f, "(" );
	            }
		    DumpExpression( f, n->Arg<CNode*>(1) );
		    if( n->Arg<CNode*>(1)->GetOp() == eELIST ) {
		        fprintf( f, ")" );
	            }
		    fprintf( f, ";\n" );
		}
		break;
	case eTABLE:
    		fprintf( f, "table\n" );
    		Tab( f );
    		DumpStatement( f, n->Arg<CNode*>(0) );
		Tab( f );
    		fprintf( f, "endtable\n" );
		break;
	case eTABLE_ENTRY:
		{
		    int p = CountPorts(currentModule->GetPorts())-1;
    		    DumpTableSymbols( f, n->Arg<CNode*>(0), p, p+1, 0 );
    		    fprintf( f, ";\n" );
		}
		break;
	default:
		MASSERT( FALSE );
	}
	
}
/*******************************************************
	CountPorts
	- return number of ports in list
********************************************************/
static int CountPorts( CNode* n )
{
    if( !n ) {
	return 0;
    }
    int sum;
    switch( n->GetOp() ) {
    case eLIST:
	sum = CountPorts(n->Arg<CNode*>(0));
	sum += CountPorts(n->Arg<CNode*>(1));
	return sum;
    case ePORT_DEF:
	return 1;
    default:
	return 0;
    }
}

/*******************************************************
	DumpTableSymbols
	- dump upd table symbols
********************************************************/

static int DumpTableSymbols( FILE* f, CNode* n, 
			      int pos1, int pos2, int pos )
{
    switch( n->GetOp() ) {
    case eTABLE_SYMBOL:
	if( pos == pos1 || pos == pos2 ) {
	    fprintf( f, ":    " );
	}   
	fprintf( f, n->Arg<char*>(0) );
	for( int i = strlen(n->Arg<char*>(0)); i < 5; ++i ) {
	    fprintf( f, " " );
        }
	pos++;
	break;
    case eLIST:
	pos = DumpTableSymbols( f, n->Arg<CNode*>(0), pos1, pos2, pos);
	pos = DumpTableSymbols( f, n->Arg<CNode*>(1), pos1, pos2, pos );
	break;
    default:
	MASSERT( FALSE );
    }
    return pos;
}

/*******************************************************
	DumpPath
	- convert path to verilog code 
********************************************************/

static void DumpPath( FILE* f, CNode* n ) 
{
	fprintf( f, "(" );
	switch( n->Arg<int>(0) ) {
	case 0:
	    break;
	case 1: 
	    fprintf( f, "posedge " );	
	    break;
	case -1:
	    fprintf( f, "negedge " );	
	    break;
	default:
	    MASSERT( FALSE );
	}

	DumpExpression( f, n->Arg<CNode*>(1) );

	switch( n->Arg<int>(2) ) {
	case 0:
	    break;
	case 1: 
	    fprintf( f, " +" );	
	    break;
	case -1:
	    fprintf( f, " -" );	
	    break;
	default:
	    MASSERT( FALSE );
	}

	switch( n->Arg<int>(3) ) {
	case 0:
	    fprintf( f, " *> " );	
	    break;
	case 1: 
	    fprintf( f, " => " );	
	    break;
	default:
	    MASSERT( FALSE );
	}

	if( n->Arg<CNode*>(6) ) {
	    fprintf( f, " (" );
	}

	DumpExpression( f, n->Arg<CNode*>(4) );

	switch( n->Arg<int>(5) ) {
	case 0:
	    break;
	case 1: 
	    fprintf( f, " +" );	
	    break;
	case -1:
	    fprintf( f, " -" );	
	    break;
	default:
	    MASSERT( FALSE );
	}

	if( n->Arg<CNode*>(6) ) {
		fprintf( f, " : " );
		DumpExpression( f, n->Arg<CNode*>(6) );
		fprintf( f, ") " );
	}

	fprintf( f, ") " );
 

}

/*******************************************************
	DumpPortNames
	- convert to verilog code 
********************************************************/

static int MorePorts( CNode* n ) {
	int result = 0;
	if( n == NULL ) {
		return FALSE;
	}
	switch( n->GetOp() ) {
	case eLIST:
	    while(1) {
	        result |= MorePorts(n->Arg<CNode*>(0));
	        n = n->Arg<CNode*>(1);
		if( !n || n->GetOp() != eLIST ) {
	    	    result |= MorePorts(n);
		    break;
		}
	    }
	    return result;
	case ePRAGMA:
	case eCOMMENT:
	case eVRQ:
		return FALSE;
	case ePORT_DEF:
		return TRUE;
	default:
		MASSERT( FALSE );
	}
	return FALSE;
}

static void DumpPortNames( FILE* f, CNode* n, int morePorts ) 
{
	if( n == NULL ) {
		return;
	}
	switch( n->GetOp() ) {
	case eLIST:
		DumpPortNames( f, n->Arg<CNode*>(0), 
				morePorts||MorePorts(n->Arg<CNode*>(1)) );
		DumpPortNames( f, n->Arg<CNode*>(1), morePorts );
		return;
	case ePORT_DEF:
		break;
	case ePRAGMA:
	case eCOMMENT:
	case eVRQ:
		Tab( f );
                DumpComment( f, n );
		return;
	default:
		MASSERT( FALSE );
	}

	CPort* port = n->Arg<CPort*>(0);
	CNode* connection;
	connection = port->GetConnection();

	/*
	 * emit pre port comment if present
	 */
	CNode* comment = port->GetPragmas();
	Tab( f );
	if( comment ) {
		DumpStatement( f, comment );
		Tab(f);
	}

	/*
	 * check if dotted
	 */
	if( port->IsDotted() ) {
		fprintf( f, ".%s(", Identifier(port->GetName()) );
	}
	
	/*
	 * if unconnected, its a null port
	 */
	if( connection != NULL ) {
		switch( connection->GetOp() ) {
		case eCAT:
	        case eUCAT:
		case eARRAY:
		case eNET_REF:
		case eVAR_REF:
		case eFWD_REF:
			DumpExpression( f, connection, 0 );
			break;
		default:	
//			MASSERT( FALSE );
			break;
		}  
	}
	if( port->IsDotted() ) {
		fprintf( f, ")" );
	}
	if( morePorts ) {
		fprintf( f, ",\n" );
	} else {
		fprintf( f, "\n" );
	}
}


/*******************************************************
	DumpSpecifyBlock
	- convert to verilog code 
********************************************************/

static void DumpSpecifyBlock( FILE* f, CSpecify* block, CNode* statements )
{
	/*	
	 * dump block header
	 */
	fprintf( f, "specify\n" );
	tabCount++;

	if( statements ) {
		Tab( f );
		DumpStatement( f, statements );
	}
	tabCount--;
	Tab( f );
	fprintf( f, "endspecify\n" ); 

}

/*******************************************************
	DumpDataType
	- dump out specification for data type
********************************************************/
static void DumpDataType( FILE* f, CDecl* decl, CDataType* dataType, int forceReg )
{
    CNode* members = NULL;
    CNode* enumList = NULL;
    int	   vector = dataType && dataType->IsVector();;
    if( dataType ) {
	int signedAttr   = FALSE;
	int unsignedAttr = FALSE;
	int packedAttr = FALSE;
	int dumpName   = TRUE;
	CDataType* baseType = NULL;
	switch( dataType->GetVarDataType() ) {
        case eREAL:
        case eSHORTREAL:
        case eREALTIME:
        case eINT:
        case eINTEGER:        
        case eBYTE:           
        case eSHORTINT:       
        case eLONGINT:    
        case eTIME:
	    signedAttr   = static_cast<CSimpleType*>(dataType)->GetSignedAttr();
	    unsignedAttr = static_cast<CSimpleType*>(dataType)->GetUnsignedAttr();
	    break;
        case eBIT:            
        case eREG:
        case eLOGIC:          
	    signedAttr   = static_cast<CSimpleType*>(dataType)->GetSignedAttr();
	    unsignedAttr = static_cast<CSimpleType*>(dataType)->GetUnsignedAttr();
	    break;
	case eNO_VAR_TYPE:
	    if( forceReg && decl && decl->GetType() == eVAR && 
		!decl->GetVarAttr() ) {
	        fprintf( f, "reg " );
	    }
	    signedAttr   = static_cast<CSimpleType*>(dataType)->GetSignedAttr();
	    unsignedAttr = static_cast<CSimpleType*>(dataType)->GetUnsignedAttr();
	    dumpName   = FALSE;
	    break;
	case eDEVENT:
	case eCLASS:
	    break;
	case eENUM: {
	    baseType = static_cast<CEnumType*>(dataType)->GetBaseType();
	    enumList = static_cast<CEnumType*>(dataType)->GetEnumList();
	    } break;
	case eSTRUCT:
	case eUNION:
	case eUNION_TAGGED:
	    members = static_cast<CStructType*>(dataType)->GetMembers();
	    packedAttr = static_cast<CStructType*>(dataType)->GetPackedAttr();
	    signedAttr   = static_cast<CStructType*>(dataType)->GetSignedAttr();
	    unsignedAttr = static_cast<CStructType*>(dataType)->GetUnsignedAttr();
	    break;
	case ePROXY:
	    dumpName = FALSE;
	    fprintf( f, "%s ", static_cast<CProxyType*>(dataType)->GetSymbol()->GetName() );
	    break;
	default:
	    MASSERT( FALSE );
	}
	if( dumpName ) {
	    fprintf( f, "%s ", varDataTypeName[dataType->GetVarDataType()] );
	}
	if( baseType ) {
	    DumpDataType( f, NULL, baseType );
	}
	if( packedAttr ) {
	    fprintf( f, "packed " );
	}
	if( signedAttr ) {
	    fprintf( f, "signed " );
	}
	if( unsignedAttr ) {
	    fprintf( f, "unsigned " );
	}
    } else {
	MASSERT( decl );
	fprintf( f, "%s ", declName[decl->GetType()] );
    }

    if( members ) {
        fprintf( f, "{\n" );
	tabCount++;
	Tab(f);
	DumpStatement( f, members );
	tabCount--;
	Tab(f);
        fprintf( f, "} " );
    }
    if( enumList ) {
        fprintf( f, "{" );
        DumpExpression( f, enumList );
        fprintf( f, "} " );
    }
    if( !vector ) {
        int packedDims = dataType->GetNumberOfPackedDimensions();
        for( int i = 0; i < packedDims; ++i ) {
            CNode* range = dataType->GetPackedRange(i);
            DumpRange( f, range, FALSE );
        }
        if( packedDims ) {
            fprintf( f, " " );
        }
    }
}

/*******************************************************
	DumpDeclSpec
	- dump out specification for declarations
********************************************************/
static void DumpDeclSpec( FILE* f, Decl_t t, CDecl* decl, int forceReg )
{
    if( decl->GetVarAttr() ) {
	fprintf( f, "var " );
    }
    if( decl->GetConstAttr() ) {
	 fprintf( f, "const " );
    }
    if( decl->GetAutomatic() ) {
	 fprintf( f, "automatic " );
    }
    DumpDataType( f, decl, decl->GetDataType(), forceReg );

    if( decl->GetPragmas() ) {
        DumpExpression( f, decl->GetPragmas() );
    }	
    fprintf( f, "%s", Identifier(decl->GetName()) );
    DumpUnpackedDimensions( f, decl );
}

/*******************************************************
	DumpNamedBlock
	- convert to verilog code 
********************************************************/

static void DumpNamedBlock( FILE* f, CBlock* block, CNode* statements, int noNewline )
{
	int parallel = block && (block->GetType() == ePARALLELBLOCK);
	int generate = block && (block->GetType() == eGENVARBLOCK);
	int nameless = !block || (block->GetSymbol() == NULL);
	/*	
	 * dump block header
	 */
	if( parallel ) {
	    if( nameless ) {
	        fprintf( f, "fork\n" );
	    } else {
	        fprintf( f, "fork : %s\n", Identifier(block->GetName()) );
	    }
	} else if( generate ) {
	    fprintf( f, "generate\n" );
	} else {
	    if( nameless ) {
	        fprintf( f, "begin\n" );
	    } else {
	        fprintf( f, "begin : %s\n", Identifier(block->GetName()) );
	    }
	}
	tabCount++;

	/*
	 * dump parameters
	 */	
	if( block ) {
	    vector<CParam*>* paramList = block->GetParamList();
	    vector<CParam*>::iterator pptr;
	    for( pptr = paramList->begin(); pptr != paramList->end(); ++pptr ) {
	    	if( (*pptr)->DeclStatementCreated() ) {
	    		continue;
	    	}
	    	Tab( f );
		DumpParameterDecl( f, *pptr );
	    	fprintf( f, ";\n" );
	    }	

	    /*
	     * dump local variable declarations
	     */
	    vector<CVar*>* varList = block->GetVarList();
	    vector<CVar*>::iterator vptr;
	    for( vptr = varList->begin(); vptr != varList->end(); ++vptr ) {
	    	if( (*vptr)->DeclStatementCreated() ) {
	    		continue;
	    	}
	    	if( !createExplicitDeclarations && (*vptr)->GetUndefined() ) {
	    		continue;
	    	}
	    	Tab( f );
		DumpAttributes( f, (*vptr)->GetAttributes() );
		DumpDeclSpec( f, (*vptr)->GetType(), *vptr, (*vptr)->GetUndefined() );
	        fprintf( f, ";\n" );
	    }
	    
	    vector<CGenvar*>* genvarList = block->GetGenvarList();
	    vector<CGenvar*>::iterator glptr;
	    for( glptr = genvarList->begin(); glptr != genvarList->end(); ++glptr ) {
	    	if( (*glptr)->DeclStatementCreated() ) {
	    		continue;
	    	}
	    	Tab( f );
	    	fprintf( f, "genvar %s;\n ", Identifier((*glptr)->GetName()) );
	    }	

	    CNode* declarations = block->GetCodeList();
	    if( declarations ) {
	        Tab( f );
	        DumpStatement( f, declarations );
	    }
	}
	if( statements ) {
	    Tab( f );
	    DumpStatement( f, statements );
	}
	tabCount--;
	Tab( f );
	if( parallel ) {
	    fprintf( f, "join" ); 
	} else if( generate ) {
	    fprintf( f, "endgenerate" ); 
	} else {
	    fprintf( f, "end" ); 
	}
	
	if( !noNewline ) {
	    fprintf( f, "\n" ); 
	}

}
/*******************************************************
	DumpDeclarationsOnly
	- dump only declaration statements
	  returns TRUE if a declaration was dumped
 ******************************************************/
static int DumpDeclarationsOnly( FILE* f, CNode* n )
{
	switch( n->GetOp() ) {
	case eLIST:
		if( DumpDeclarationsOnly( f, n->Arg<CNode*>(0) ) ) {
		    Tab(f);
		}
		return DumpDeclarationsOnly( f, n->Arg<CNode*>(1) );
	case ePARAM_DECL:
	case ePORT_DECL:
	case eNET_DECL:
	case eVAR_DECL:
	case eGENVAR_DECL:
	case ePRAGMA:
	case eCOMMENT:
	case eVRQ:
		DumpStatement( f, n );
		return TRUE;
	default:
		return FALSE;
	}
}
/*******************************************************
	DumpStatementsOnly
	- dump only non-declaration statements
	  returns TRUE if a non-declaration was dumped
 ******************************************************/
static int DumpStatementsOnly( FILE* f, CNode* n )
{
	switch( n->GetOp() ) {
	case eLIST:
		if( DumpStatementsOnly( f, n->Arg<CNode*>(0) ) ) {
			Tab(f);
		}
		return DumpStatementsOnly( f, n->Arg<CNode*>(1) );
	case ePARAM_DECL:
	case ePORT_DECL:
	case eNET_DECL:
	case eVAR_DECL:
		return FALSE;
	default:
		DumpStatement( f, n );
		return TRUE;
	}
}

/*******************************************************
	DumpFunction
	- convert to verilog code 
********************************************************/

static void DumpFunction( FILE* f, CFunction* function )
{
        int saveInlinePorts = inlinePorts;
	inlinePorts = function->InlinePortDecls();
	/*
	 * dump attributes
	 */
	DumpAttributes( f, function->GetAttributes() );
	/*	
	 * dump module header
	 */
	CDecl* output = function->GetOutput();
	if( output ) {
	    fprintf( f, "function " );
	} else {
	    fprintf( f, "task " );
	}

	if( function->GetAutomatic() ) {
	    fprintf( f, "automatic " );
	}

	if( output ) {
	    DumpDataType( f, output, output->GetDataType() );
	}
        fprintf( f, "%s", Identifier(function->GetName()) );
	
	if( !inlinePorts ) {
    	    fprintf( f, ";\n" );
	}
	
	tabCount++;

	
	/*
	 * find inline port declaration boundry
	 */
	CNode* statements =  function->GetCodeList();
	CNode* cutPoint = NULL;
	int count = CountStatementsInRange( statements, NULL );
	if( inlinePorts ) {
	    cutPoint = FindPortDeclarationEnd( statements );
	    count = CountStatementsInRange( statements, cutPoint );
	}

	/*
	 * declare ports
	 */
	if( inlinePorts ) {
	    fprintf( f, "(\n" );
	    DumpInlinePortDecls( f, statements );
	    tabCount--;
	    Tab( f );
	    tabCount++;
	    fprintf( f, ");\n" );

	    fprintf( f, "\n" );
	}

	/*
	 * dump local variable declarations
	 */
	vector<CVar*>* varList = function->GetVarList();
	vector<CVar*>::iterator vptr;
	for( vptr = varList->begin(); vptr != varList->end(); ++vptr ) {
		if( (*vptr)->DeclStatementCreated() ) {
			continue;
		}
		if( !createExplicitDeclarations && (*vptr)->GetUndefined() ) {
			continue;
		}
		DumpAttributes( f, (*vptr)->GetAttributes() );
		DumpDeclSpec( f, (*vptr)->GetType(), *vptr, (*vptr)->GetUndefined() );
		fprintf( f, ";\n" );
	}

	/*
	 * dump function statements
	 */
	Tab( f );
	if( count ) {
	    DumpStatementsAfter( f, cutPoint, statements );
	} else {
	    fprintf( f, "begin\n" );
	    Tab( f );
	    fprintf( f, "end\n" );
	}
	    
	
	/*
	 * end function
	 */
	tabCount--;
	Tab( f );
	if( output ) {
	    fprintf( f, "endfunction" );
	} else {
	    fprintf( f, "endtask" );
	}
	if( function->GetTrailingLabel() ) {
	    fprintf( f, " : %s", function->GetTrailingLabel()->GetName() );
	}
	fprintf( f, "\n" );
        inlinePorts = saveInlinePorts;
}

/*******************************************************
	TsMagnitude
	- convert timescale to magnitude
********************************************************/

static int TsMagnitude( int ts )
{
    if( ts > 0 ) {
	return (int)pow(10.,ts);
    }
    return (int)pow(10.,(ts+15)%3);
}

/*******************************************************
	TsUnit
	- convert timescale to unit
********************************************************/

static const char* TsUnit( int ts )
{
    if( ts >= 0 ){
	return "s";
    } 
    if( ts >= -3 ) {
	return "ms";
    }
    if( ts >= -6 ) {
	return "us";
    }
    if( ts >= -9 ) {
	return "ns";
    } 
    if( ts >= -12 ) {
	return "ps";
    }
    if( ts >= -15 ) {
	return "fs";
    }
    return "???";
}
/*******************************************************
	DumpParameterDecl
	- dump parameter decl
********************************************************/
static void DumpParameterDecl( FILE* f, CParam* p ) 
{
    DumpAttributes( f, p->GetAttributes() );
    switch( p->GetType() ) {
    case eLOCALPARAM:
    	fprintf( f, "localparam " );
	break;
    case ePARAM:
    	fprintf( f, "parameter " );
	break;
    default:
	MASSERT( FALSE );
	break;
    }
    DumpDataType( f, p, p->GetDataType() );
    if( p->GetPragmas() ) {
        DumpExpression( f, p->GetPragmas() );
    }
    fprintf( f, "%s", Identifier(p->GetName()) );

    CNode* expr = p->GetExpression();
    MASSERT( expr != NULL ); 
    fprintf( f, " = " );
    DumpExpression( f, expr );
}

/*******************************************************
	FindPortDeclarationEnd
	- return statement following PORTLIST_END
	  return n if none exists in list
********************************************************/
static CNode* FindPortDeclarationEnd( CNode* n, int* hasPLE )
{
    int ple = 0;
    CNode* plen = NULL;

    if( !n ) {
	goto done;
    }
    switch( n->GetOp() ) {
    case ePORTLIST_END:
	ple = 1;
	break;
   case eLIST:
	{
	    int hasPLE0 = 0;
	    int hasPLE1 = 0;
	    CNode* n0 = FindPortDeclarationEnd( n->Arg<CNode*>(0), &hasPLE0 );
	    CNode* n1 = FindPortDeclarationEnd( n->Arg<CNode*>(1), &hasPLE1 );
	    if( hasPLE1 ) {
		ple = 1;
		plen = n1;
	    } else if( hasPLE0 ) {
		ple = 1;
		plen = n0 ? n0 : n->Arg<CNode*>(1);
	    } else {
		ple = 0;
		plen = NULL;
	    }
	}
	break;
    default:
	break;
    }

done:
    if( !hasPLE && !ple ) {
	return n;
    }

    if( hasPLE ) {
	*hasPLE = ple;
    }	
    return plen;
}
    

/*******************************************************
	CountStatementsInRange
	- return number of statements in tree rooted
	  at begin to end(exclusive)
********************************************************/
static int CountStatementsInRange( CNode* begin, CNode* end, int* found )
{
   int count = 0;
   int fd = 0;

   if( begin == end ) {
	count = 0;
	fd = 1;
	goto DONE;
   }
   if( !begin ) {
	goto DONE;
   }
   switch( begin->GetOp() ) {
   case eLIST:
	{
	    int found0 = 0;
	    int found1 = 0;
	    int c0 = CountStatementsInRange( begin->Arg<CNode*>(0), end, &found0 );
	    int c1 = CountStatementsInRange( begin->Arg<CNode*>(1), end, &found1 );
	    if( found1 ) {
		fd = 1;
	 	count = c0 + c1;	
	    } else if( found0 ) {
		fd = 1;
		count = c0;
	    } else {
		fd = 0;
		count = c0 + c1;
	    }
	}
	break;
   default: ;
	count = 1;
	break;
   }
   
DONE:
    if( found ) {
    	*found = fd;
    }
    return count;
}
/*******************************************************
	DumpInlinePortDecls
	- Dump inline port decls
********************************************************/
static int DumpInlinePortDecls( FILE* f, CNode* n, int state )
{
   /*
    * state:
    * 	-2 uninitialized
    * 	-1  done
    * 	>=0 remaining count
    */
   /*
    * initiailize
    */
   if( state == -2 ) {
	state = ListCount(n,ePORT_DECL);
   }
   /*
    * done
    */
   if( state == -1 ) {
	return state;
   }
   if( !n ) {
	return state;
   }
   switch( n->GetOp() ) {
   case ePORTLIST_END:
	return -2;
   case eLIST:
	while(1) {
	    state = DumpInlinePortDecls( f, n->Arg<CNode*>(0), state );
	    if( state < 0 ) {
		break;
	    }
	    n = n->Arg<CNode*>(1);
	    if( !n || n->GetOp() !=eLIST ) {
	        state = DumpInlinePortDecls( f, n, state );
		break;
	    }
	}
	break;	
   default: 
	switch( n->GetOp() ) {
	case ePORT_DECL:
	    MASSERT( state > 0 );
	    state--;
	    Tab( f );
	    DumpPortDirDecl( f, n->Arg<CPortDir*>(0) );
	    if( state > 0 ) {
	        fprintf( f, "," );
	    }
	    fprintf( f, "\n" );
	    break;
	default:
	      Tab( f );
	      DumpStatement( f, n );
	      break;
	}
	break;
   }
   return state;
}
/*******************************************************
	DumpStatementsBefore
	- Dump all statements before cutpoint
********************************************************/
static int DumpStatementsBefore( FILE* f, CNode* cutPoint, CNode* n, int* tab=NULL )
{
   int tabOut = !tab || *tab;
   int found = FALSE;
   if( !n ) {
	return FALSE;
   }
   if( n == cutPoint ) {
	found = TRUE;	
   }
   switch( n->GetOp() ) {
   case eLIST:
	while(1) {
	    found = DumpStatementsBefore( f, cutPoint, n->Arg<CNode*>(0), &tabOut );
	    if( found ) {
		break;
	    }
	    n = n->Arg<CNode*>(1);
	    if( !n || n->GetOp() !=eLIST ) {
	        found = DumpStatementsBefore( f, cutPoint, n, &tabOut );
		break;
	    }
	}
	break;	
   default: ;
	if( !found ) {
	    if( !tabOut ) {
	        Tab( f );
	    }
	    DumpStatement( f, n );
	    if( n->GetOp() != ePORTLIST_END ) {
	        tabOut = 0;
	    }
	}
	break;
   }
   if( tab ) {
        *tab = tabOut;
   }
   return found;
}
/*******************************************************
	DumpStatementsAfter
	- Dump all statements after and including cutpoint
********************************************************/
static CNode* DumpStatementsAfter( FILE* f, CNode* cutPoint, CNode* n )
{
   if( !n ) {
	return cutPoint;
   }
   if( n == cutPoint ) {
	cutPoint = NULL;
   }
   switch( n->GetOp() ) {
   case eLIST:
	while(1) {
	    cutPoint = DumpStatementsAfter( f, cutPoint, n->Arg<CNode*>(0) );
	    if( !cutPoint && n->Arg<CNode*>(0) != NULL &&
	        n->Arg<CNode*>(0)->GetOp() != ePORTLIST_END &&
	        n->Arg<CNode*>(0)->GetOp() != eATTRIBUTE ) { 
	        Tab( f );
	    }
	    n = n->Arg<CNode*>(1);
	    if( n == cutPoint ) {
		cutPoint = NULL;
	    }
	    if( !n || n->GetOp() !=eLIST ) {
	        cutPoint = DumpStatementsAfter( f, cutPoint, n );
		break;
	    }
	}
	break;	
   default: ;
	if( !cutPoint ) {
	    DumpStatement( f, n );
 	}
	break;
   }
   return cutPoint;
}

/*******************************************************
	DumpModule
	- convert module/primitive to verilog code 
********************************************************/

static void DumpModule( FILE* f, CModule* module )
{

	currentModule = module;
	inlineParams = module->InlineParamDecls();
        int saveInlinePorts = inlinePorts;
	inlinePorts  = module->InlinePortDecls();
	/*
	 * dump timescale primitve
	 */
	if( module->HasTimescale() ) {
	    fprintf( f, "`timescale %d%s/%d%s\n", 
		TsMagnitude( module->Unit() ),
		TsUnit( module->Unit() ),
		TsMagnitude( module->Precision() ),
		TsUnit( module->Precision() ) );
	} else if( defaultTimescale ) {
	    fprintf( f, "`timescale %s\n", defaultTimescale );
	}
	/*
 	 * Dump default net type if defined
 	 */
	if( defaultNetType != module->DefaultNetType() ) {
	    defaultNetType = module->DefaultNetType();
	    fprintf( f, "`default_nettype " );
	    switch( module->DefaultNetType() ) {
	    case vpiWire:
		fprintf( f, "wire" );
		break;
	    case vpiWand:
		fprintf( f, "wand" );
		break;
	    case vpiWor:
		fprintf( f, "wor" );
		break;
	    case vpiTri:
		fprintf( f, "tri" );
		break;
	    case vpiTri0:
		fprintf( f, "tri0" );
		break;
	    case vpiTri1:
		fprintf( f, "tri1" );
		break;
	    case vpiTriReg:
		fprintf( f, "trireg" );
		break;
	    case vpiTriAnd:
		fprintf( f, "triand" );
		break;
	    case vpiTriOr:
		fprintf( f, "trior" );
		break;
	    default:
		MASSERT( FALSE );
	    }
	    fprintf( f, "\n" );
	}
	/*
	 * dump attributes
	 */
	DumpAttributes( f, module->GetAttributes() );
	/*	
	 * dump module header
	 */
	if( module->IsPrimitive() ) {
	    fprintf( f, "primitive %s", Identifier(module->GetName()) );
	} else {
	    fprintf( f, "module %s", Identifier(module->GetName()) );
	}

	/*
	 * find inline port declaration boundry
	 */
	CNode* statements =  module->GetCodeList();
	CNode* cutPoint = NULL;
	int    count = 0;
	int    total = CountStatementsInRange( statements, NULL );
	if( inlinePorts ) {
	    cutPoint = FindPortDeclarationEnd( statements );
	    count = CountStatementsInRange( statements, cutPoint );
	}

	if( inlineParams ) {
	    fprintf( f, " #(\n" );
	    vector<CParam*>* paramList = module->GetParamList();
	    vector<CParam*>::iterator pptr;
	    tabCount++;
	    int needNewline = 0;
	    for( pptr = paramList->begin(); pptr != paramList->end(); ++pptr ) {
		    if( !(*pptr)->GetInlined() ) {
		        continue;
		    }
		    if( needNewline ) {
		        fprintf( f, ",\n" );
			needNewline = 0;
		    }
		    Tab( f );
		    DumpParameterDecl( f, *pptr );
		    needNewline = 1;
	    }	
	    if( needNewline ) {
		fprintf( f, "\n" );
	    }
	    tabCount--;
	    fprintf( f, ")" );
	}
	fprintf( f, "(" );

	tabCount++;
	if( !inlinePorts ) {
	    CNode* ports = module->GetPorts();
	    if( MorePorts( ports ) ) {
		    fprintf( f, "\n" );
	    }
	    DumpPortNames( f, ports, FALSE );
	} else {
	    fprintf( f, "\n" );
	    DumpInlinePortDecls( f, statements );
	}	

	fprintf( f, ");\n" );

	
	if( !inlineParams ) {
	    vector<CParam*>* paramList = module->GetParamList();
	    vector<CParam*>::iterator pptr;
	    for( pptr = paramList->begin(); pptr != paramList->end(); ++pptr ) {
		    if( (*pptr)->DeclStatementCreated() ) {
			    continue;
		    }
		    Tab( f );
		    DumpParameterDecl( f, *pptr );
		    fprintf( f, ";\n" );
	    }	
	}

	/*
	 * declare ports
	 */
	if( !inlinePorts ) {
	    vector<CPortDir*>* dirList = module->GetPortDirList();
	    vector<CPortDir*>::iterator pdptr;
	    for( pdptr = dirList->begin(); pdptr != dirList->end(); ++pdptr ) {
		    if( (*pdptr)->DeclStatementCreated() ) {
			    continue;
		    }
		    Tab( f );
		    DumpPortDirDecl( f, *pdptr );
		    fprintf( f, ";\n" );
	    }
	}	

	/*
	 * dump local variable declarations
	 */
	vector<CVar*>* varList = module->GetVarList();
	vector<CVar*>::iterator vptr;
	for( vptr = varList->begin(); vptr != varList->end(); ++vptr ) {
		if( (*vptr)->DeclStatementCreated() ) {
			continue;
		}
		if( !createExplicitDeclarations && (*vptr)->GetUndefined() ) {
			continue;
		}
		Tab( f );
		DumpAttributes( f, (*vptr)->GetAttributes() );
		DumpDeclSpec( f, (*vptr)->GetType(), *vptr, (*vptr)->GetUndefined() );
		fprintf( f, ";\n" );
	}
	
	vector<CNet*>* netList = module->GetNetList();
	vector<CNet*>::iterator nptr;
	for( nptr = netList->begin(); nptr != netList->end(); ++nptr ) {
		if( (*nptr)->DeclStatementCreated() ) {
			continue;
		}
		if( !createExplicitDeclarations && (*nptr)->GetUndefined() ) {
			continue;
		}
		if( (*nptr)->GetUndefined() && (*nptr)->PortDir() ) {
			continue;
		}
		Tab( f );
		DumpWireAttr( f, (*nptr)->GetWireAttr() );
    	        DumpDataType( f, *nptr, (*nptr)->GetDataType() );
		fprintf( f, "%s", Identifier((*nptr)->GetName()) );
		fprintf( f, ";\n" );
	}


	/*
	 * dump module statements
	 */
	if( total-count ) {
	    Tab( f );
	    DumpStatementsAfter( f, cutPoint, statements );
	}
	
	/*
	 * end module
	 */
	tabCount--;
	Tab( f );
	if( module->IsPrimitive() ) {
	    fprintf( f, "\nendprimitive\n" );
	} else {
	    fprintf( f, "\nendmodule\n" );
	}
        inlinePorts = saveInlinePorts;
}
                
/*******************************************************
	DumpComment
	- output comment
********************************************************/

static void DumpComment( FILE* f, CNode* n )
{
        char* comment = n->Arg<char*>(0);
        char* ptr = comment;
        int comment_style = 0;
	if( !strncmp( comment, "`protect", 8 ) ) {
	    fprintf( f, "\n" );
	    fprintf( f, comment );
	    return;
	}
        while( *ptr != 0 ) {
                if( *ptr == '\n' ||
                    *ptr == '\r' ) {
                        fprintf( f, "\n" ); 
                        if( *(ptr+1) != 0 ) {
                                Tab( f );
                                if (comment_style) fprintf( f, "// " );
                        }
                } else if (forceCPPComments && !comment_style && (ptr[0] == '/') && (ptr[1] == '/')) {
			while( *ptr && *ptr != '\n' && *ptr != '\r' ) {
			    fprintf( f, "%c", *ptr++ );
			}
			ptr--;
                } else if (forceCPPComments && !comment_style && (ptr[0] == '/') && (ptr[1] == '*')) {
                        // convert opening /* comment if in cpp mode
                        fprintf( f, "// " ); 
                        ptr += 1;
                        comment_style = 1;
                } else if (comment_style && (ptr[0] == '*') && (ptr[1] == '/')) {
                        // convert closing */ comment if cpp mode and inside comment
                        fprintf( f, "\n" ); 
                        ptr += 1;
                        comment_style = 0;
                } else {
                        fprintf( f, "%c", *ptr );
                }
                ptr++;
        }
        if( comment[strlen(comment)-1] != '\n' ) {
                fprintf( f, "\n" );
        }
}
/************************************************
	SimplifyExpression
	- simplify constant expressions
 ************************************************/
CNode*	SimplifyExpression( CNode* n ,void*)
{
    switch( n->GetOp() ) {
    case eASSIGN: 
    case eADD_ASSIGN:
    case eSUB_ASSIGN:
    case eMUL_ASSIGN:
    case eDIV_ASSIGN:
    case eMOD_ASSIGN:
    case eAND_ASSIGN:
    case eOR_ASSIGN:
    case eXOR_ASSIGN:
    case eLSH_ASSIGN:
    case eRSH_ASSIGN:
    case eLSHA_ASSIGN:
    case eRSHA_ASSIGN: {
	NodeType_t type = n->Arg<CNode*>(1)->GetNodeType();
	if( type != eR && 
		(n->Arg<CNode*>(1)->IsWidthVolatile() ||
	         !n->Arg<CNode*>(1)->IsWidthEvaluateable()) ) {
	    break;
	}
	n->Arg<CNode*>(2) = n->Arg<CNode*>(2)->Simplify(
		n->Arg<CNode*>(1)->GetWidth(), 
		type==eS ? eB : type );
	} break;
    default:
	break;
    }
    return n;
}

/************************************************
	ForceVectorLogicals
	- convert Logicals (for scalar expressions)
	  to bitwise ops.
 ************************************************/
static void _ForceVectorLogicals( CNode* n, void* )
{
    if( !n ) {
	return;
    }
    NodeOp_t op;

    switch( n->GetOp() ) {
    case eLAND:
	op = eAND;
	break;
    case eLOR:
	op = eOR;
	break;
    case eNOT:
	op = eCOM;
	break;
    default: ;
	return;
    }
    
    for( int i = 0; i < n->ArgCount(); ++i ) {
	if( !n->Arg<CNode*>(i)->IsWidthConstant() || 
	    !n->Arg<CNode*>(i)->IsWidthEvaluateable() ||
	    n->Arg<CNode*>(i)->IsVector() ) { 
	    return;
	}
    }
    n->SetOp( op );
}

/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CDumpVerilog::Process( list<CElement>& inputList,
	                    list<CElement>& outputList )
{
    int count;
    /*
     * parse command line args
     */
    altRxnor = GetPlusArg( "dump-alt-rxnor" ) != NULL;
    forceCPPComments = GetPlusArg("dump-fcc") != NULL;
    forceBeginAfterEvent = GetPlusArg("dump-fbae") != NULL;
    splitWireDeclAssignments = GetPlusArg("dump-swda") != NULL;
    spaceAroundBinaryOps = GetPlusArg("dump-sabo") != NULL;
    noNullStatements = GetPlusArg("dump-nns") != NULL;
    createExplicitDeclarations = GetPlusArg("dump-ced") != NULL;
    stripAttributes = GetPlusArg("dump-strip-attr") != NULL;
    simplifyExpressions = GetPlusArg( "dump-simplify" ) != NULL;
    forcePartSelect = GetPlusArg("dump-fps") != NULL;
    forceVectorLogicals = GetPlusArg("dump-fvl") != NULL;
    defaultTimescale = GetPlusArg("dump-timescale" );
    if( defaultTimescale ) {
	defaultTimescale++;
    }

    /*
     * check elements for valid filenames
     */
    list<CElement>::iterator ptr;
    int validFilenames = 1;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
	validFilenames &= (ptr->Filename() != NULL);
    }

    /*
     * dump trees
     */
    int split = (validFilenames && outputDirectory != NULL);
    if( !split ) {
	/*
	 * fail if output path cannot be opened
	 */
	if( outputPath == NULL ) {
	    error( (Coord_t*)NULL, "output filename was not specified" );
	    goto DONE;
	}
	if( !strcmp( outputPath, "-" ) ) {
	    f = stdout;
	} else {
	    f = fopen( outputPath, "w" );
	    if( f == NULL ) {
		error( (Coord_t*)NULL, "output file '%s' could not be openned", 
								outputPath );
		goto DONE;
	    }
	}
    }	

    count = 0;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
	/* 
	 * if dumping to a directory, open each file
	 */
	if( split ) {
	    string filename;
	    filename = outputDirectory;
	    filename += "/";
	    filename += ptr->Filename();
	    f = fopen( filename.c_str(), "w" );
	    if( f == NULL ) {
		error( (Coord_t*)NULL, "output file '%s' could not be openned", 
							filename.c_str() );
		goto DONE;
	    }
	}
	/*
	 * output header to file
	 */
	if( split || count == 0 ) {
    	    tabCount = 0;
	    suppressTab = 0;
    	    skipTab  = FALSE;
    	    inFor    = FALSE;
    	    fprintf( f, "//****************************************\n" );
    	    fprintf( f, "// Warning! This file is machine generated\n" );
    	    fprintf( f, "//****************************************\n" );
    	    fprintf( f, "\n\n" );
	}

	/*
	 * process each module
	 */
	defaultNetType = vpiWire; 
	Tab( f );
	CNode* code = ptr->Code();
        /*
         * simplify expressions
         */
        if( simplifyExpressions && code ) {
	    code = code->PostSubVisit1( SimplifyExpression, NULL );
        }
	/*
 	 * if requested rewrite logical ops as bitwise ops
 	 */
    	if( forceVectorLogicals && code ) {
	    code->PostVisit1( _ForceVectorLogicals, NULL );
	}
	DumpStatement( f, code );

	/*
	 * print trailer
	 */
	if( split || count+1 == inputList.size() ) {
	    fprintf( f, "\n\n" );
	    fprintf( f, "//****************************************\n" );
	    fprintf( f, "// Warning! This file is machine generated\n" );
	    fprintf( f, "//****************************************\n" );
	}
	/* 
	 * if dumping to a directory, close each file
	 */
	if( split ) {
	    fclose( f );
	}
	count++;
    }
    
    if( !split ) {
	fclose( f );
    }

DONE:;
}

/**
 * \endcond 
 */
