/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cfref.cpp
 *		- class definition of forward reference 
 *		  definition nodes
 *
 ******************************************************************************
 */

#include "cfref.h"
#include "cnode.h"




/****************************************************
	Constructor
*****************************************************/
CFref::CFref( CSymbol* name, Coord_t* aLoc ) :
	CDecl( name, aLoc, eFREF, NULL, eFLAG_NONE ) {
	    decl = NULL;
	    direction = eNONE;
	    constrained = FALSE;
	    lsb = NULL;
	    msb = NULL;
	    lval = FALSE;
	    refCount = 0;
	    lsbVolatile = 0;
	    msbVolatile = 0;
	    lsbConstant = 1;
	    msbConstant = 1;
	    isArray = 0;
	    isInconsistant = 0;
	    currentDirection = 0;
}


/****************************************************
	GetDecl
*****************************************************/
CDecl*	CFref::GetDecl( )
{
	return decl;
}

/****************************************************
	SetDecl
*****************************************************/
void	CFref::SetDecl( CDecl* aDecl )
{
	decl = aDecl;
}

/****************************************************
	SetRangeInfo
*****************************************************/
void 	CFref::SetRangeInfo( int isArray, CNode* range )
{
    MASSERT( !range || range->GetOp() == eRANGE || 
	     range->GetOp() == eMACRO_EXPR );
    if( range && range->GetOp() == eMACRO_EXPR ) {
	range = range->Arg<CNode*>(1);
	MASSERT( range );
    }
    CNode* msb = range ? (CNode*)range->Arg<CNode*>(0) : NULL;
    CNode* lsb = range ? (CNode*)range->Arg<CNode*>(1) : NULL;
    /*
     * keep track of number of assignments
     * and cummulative range properties
     */
    refCount++;
    if( msb ) {
	msbVolatile |= msb->IsVolatile();
	msbConstant &= msb->IsConstant();
    }
    if( lsb ) {
	lsbVolatile |= lsb->IsVolatile();
	lsbConstant &= lsb->IsConstant();
    }
    /*
     * accumulate array attribute
     */
    this->isArray |= isArray;
    if( !this->msb ) {
        /*
         * keep the first range spec for msb/lsb
         */
        this->msb = msb;
        this->lsb = lsb;
        if( msb && lsb && !msbVolatile && !lsbVolatile &&
            msb->IsEvaluateable() && lsb->IsEvaluateable() ) {
   	    int msb32 = msb->EvalINT32(); 
   	    int lsb32 = lsb->EvalINT32(); 
   	    currentDirection = msb32==lsb32 ? 0 : msb32>lsb32  ? 1 : -1;
        }
	return;
    }
    /*
     * if possible calculate new range bounds
     */
   if( !msb || !lsb ) {
       return;
   }
   if( msbVolatile || lsbVolatile || 
       !lsb->IsEvaluateable() || !msb->IsEvaluateable() ) {
	return;
   }
   int oldMsb32 = this->msb->EvalINT32(); 
   int oldLsb32 = this->lsb->EvalINT32(); 
   int msb32 = msb->EvalINT32(); 
   int lsb32 = lsb->EvalINT32(); 
   int oldDirection = oldMsb32==oldLsb32 ? 0 : 
		      oldMsb32>oldLsb32  ? 1 : -1;
   int direction = msb32==lsb32 ? 0 : 
		   msb32>lsb32  ? 1 : -1;

   if( !currentDirection && direction ) {
	currentDirection = direction;
   }

   if( direction != currentDirection && direction && currentDirection ) {
	isInconsistant = 1;
	return;
   }

   if( direction >= 0 ) {
      if( msb32 > oldMsb32 ) {
	this->msb = msb;
      }
      if( lsb32 < oldLsb32 ) {
	this->lsb = lsb;
      }
   } else {
      if( msb32 < oldMsb32 ) {
	this->msb = msb;
      }
      if( lsb32 > oldLsb32 ) {
	this->lsb = lsb;
      }
   }

}

/****************************************************
	GetRange
*****************************************************/
CNode*	CFref::GetRange( void ) {
// ??? mdh - this needs to be rewritten
	if( !msb ) {
	    return NULL;
	}
	return cRANGE(msb,lsb);
}	

/****************************************************
	GetMsb
*****************************************************/
CNode*	CFref::GetMsb( void ) {
	return msb;
}	

/****************************************************
	GetLsb
*****************************************************/
CNode*	CFref::GetLsb( void ) {
	return lsb;
}	

/****************************************************
 	IsWidthConstant
*****************************************************/
int  CFref::IsWidthConstant( void )
{
    if( !msb ) {
	return TRUE;
    } 
    return msb->IsConstant() && lsb->IsConstant();
}

/****************************************************
 	IsWidthVolatile
*****************************************************/
int  CFref::IsWidthVolatile( void )
{
    if( !msb ) {
	return FALSE;
    } 
    return msb->IsVolatile() || lsb->IsVolatile();
}
/****************************************************
 	IsWidthEvaluateable
*****************************************************/
int  CFref::IsWidthEvaluateable( void )
{
    if( !msb ) {
	return TRUE;
    } 
    return msb->IsEvaluateable() && lsb->IsEvaluateable();
}

/*****************************************************
   Evaluate current fref width. 
   Expression must be constant.  
  \return evaluated with expression.
******************************************************/ 
INT32 CFref::GetWidth( void ) 
{
	if( !msb ) {
	    return 1;
	}
	MASSERT( msb && lsb );
        if( Equivalent(msb,lsb) ) {
                return 1;
        }
 	INT32 l = lsb->EvalINT32();
 	INT32 m = msb->EvalINT32();
 	return m>=l ? m-l+1 : l-m+1;

}

/****************************************************
	IndicesRangeValid
	- returns true if the min and max indexs are 
	  valid; 
*****************************************************/
int CFref::IndicesRangeValid()
{
    /* 
     * there are a number cases:
     * undimensioned variable - inferencing ok
     */
    if( !isArray && !msb ) {
	MASSERT( !lsb );
	return TRUE;
    }
    /*
     * array		      - inferencing not ok
     */
    if( isArray ) {
	return FALSE;
    }
    /*
     * inconsistant range assignment - inferencing not ok
     */
    if( isInconsistant ) {
	return FALSE;
    }
    /*
     * multiple references with at least one volatile index - inferencing not ok
     */
    if( refCount > 1 && (lsbVolatile || msbVolatile ) ) {
	return FALSE;
    }
    
    /*
     * non-constant indices - inferencing not ok
     */
   if( !lsbConstant || !msbConstant ) {
       return FALSE;
   }
     
   /*
    * otherwise - inferencing ok
    */
    return TRUE;
}

/****************************************************
	Dump
*****************************************************/
void	CFref::Dump( FILE* f )
{
	fprintf( f, "Fref: %s, defined in ", GetName() );
	CDecl::Dump( f );
	fprintf( f, "\n" );
	set<Decl_t>::iterator ptr;
	for( ptr = typeConstraints.begin();
	     ptr != typeConstraints.end(); ++ptr ) {
		fprintf( f, "    Constraint: %s\n", declName[*ptr] );
	}
}	

