/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cfref.hpp
 *		- class definition of forward reference 
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CFREF_HPP
#define CFREF_HPP

#include <stdio.h>
#include <set>
#include <list>
#include <algorithm>
#include "glue.h"
#include "cnode.h"
#include "cdecl.h"


class CNode;

/**
 * Forward reference declaration.
 * This class should only be used by the parser
 * and should never appear in or be created by 
 * plugins.
 */
class CFref : public CDecl
{
private:
	CDecl*	    decl;	      ///< declaration assigned to this decl
	Decl_t      direction;	      ///< direction assigned to reference
	int	    constrained;      ///< non-zero if decl has been constrained
	set<Decl_t> typeConstraints;  ///< list of valid types for decl.
	CNode*      msb;	      ///< msb of vector
	CNode*      lsb;	      ///< lsb of vector
	int 	    lval;	      ///< non-zero if decl has been used as 
				      ///< an lval
	int	    isArray;	      ///< non-zero if decl has been referenced
        int         lsbVolatile;      ///< non-zero if lsb is volatile
        int         msbVolatile;      ///< non-zero if msb is volatile
        int         lsbConstant;      ///< non-zero if lsb is constant
        int         msbConstant;      ///< non-zero if msb is constant
        int         isInconsistant;   ///< non-zero if decl constraints 
				      ///< are conflicting
        int         currentDirection; ///< -1 little endian, 0 none, 
				      ///< +1 big endian
	int	    refCount;	      ///< number of references to decl
public:
	/**
 	 * Create a forward reference.
 	 * \param symbol declarations symbol.
 	 * \param aLoc file coordinates.
 	 */
	CFref( CSymbol* symbol, Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap ) { MASSERT(FALSE); }
	/**
 	 * Set forward declaration's actual declaration.
 	 * \param aDecl actual declaration for forward reference.
 	 */
	void SetDecl( CDecl* aDecl );
	/**
 	 * Get forward reference's actual declaration
 	 * \return actual declaration, NULL if not assigned.
 	 */
	CDecl* GetDecl( void );
	/**
 	 * Set range info for declaration.
 	 * \param isArray non-zero if declaration is an array.
 	 * \param range vector range expression.
	 */
	void 	SetRangeInfo( int isArray, CNode* range );
	/**
 	 * Get msb of vector range.
 	 * \return msb expression, NULL if scalar.
 	 */
	CNode*	GetMsb( void );
	/**
 	 * Get lsb of vector range.
 	 * \return lsb expression, NULL if scalar.
 	 */
	CNode*	GetLsb( void );
	/**
 	 * Disable width inferencing
 	 **/
        void DisableWidthInferencing() { isInconsistant = TRUE; } 
        /**
         * Determine if packed width of declaration is constant, 
         * ie dependent upon only constants and parameters.
         * \return non-zero of packed width of declaration is constant.
         */
         virtual int  IsWidthConstant( void );
        /**
         * Determine if packed width of declaration is volatile, ie depend
         * upon parameters or variables.
         * \return non-zero of packed width of declaration is volatile.
         */
         virtual int  IsWidthVolatile( void );
        /**
         * Determine if packed width of declaration can be evaluated.
         * \return non-zero of packed width of declaration can be evaluated.
         */
        virtual int  IsWidthEvaluateable( void );
	/**
 	 * Evaluate packed width of declaration.
 	 * \return packed width of declaration.
 	 */
	virtual INT32 GetWidth( void );
        /**
         * Get expression for declaration's range (msb/lsb)
         * \return range expression, NULL if scalar.
         */
        virtual CNode* GetRange();
	/**
 	 * Get direction of declaration
 	 * \return d direction type.
 	 */
	Decl_t Direction() { return direction; }
	/**
 	 * Set direction of declaration
 	 * \param d direction type.
 	 */
	void  Direction( Decl_t d ) { direction = d; }
	/**
 	 * Set lval flag
 	 * \param aLval non-zero means decl has been used as an lval.
 	 */
	void    Lval( int aLval ) { lval = aLval; }
	/**
 	 * Get lval flag
 	 * \return non-zero if decl has been used as an lval.
 	 */
	int     Lval() { return lval; }
	/**
 	 * Set array flag.
 	 * \param v non-zero if decl is an array.
 	 */
	void	IsArray( int v ) { isArray = v; }
	/**
 	 * Get array flag.
 	 * \return non-zero if decl is an array.
 	 */
	int	IsArray() { return isArray; }
	/**
 	 * Determine if range can be inferred.
 	 * \return non-zero if range can be inferred.
 	 */
	int	IndicesRangeValid();
	/**
 	 * Add a constraint on valid types for this declaration.
	 * Note eNONE must always be in type list.
	 * \param aTypes list of valid types (must incude eNONE).
	 */
	void   ConstrainTypes( const list<Decl_t>& aTypes ) {
	    list<Decl_t> types(aTypes);
	    
	    list<Decl_t>::const_iterator ptr;
	    for( ptr = aTypes.begin(); ptr != aTypes.end(); ++ptr ) {
		CDecl::GetMembers( *ptr, types );
	    }
	    if( typeConstraints.size() == 0 ) {
		MASSERT( !constrained );
		set_union( types.begin(), types.end(),
			    typeConstraints.begin(), typeConstraints.end(),
			    inserter( typeConstraints, 
				      typeConstraints.begin() ) );
	    } else {
		set_intersection( types.begin(), types.end(),
			    typeConstraints.begin(), typeConstraints.end(),
			    inserter( typeConstraints, 
				      typeConstraints.begin() ) );
	    }
	    constrained = TRUE;
	} 
	/**
 	 * Test if a give declaration type is valid for this reference.
 	 * \param t declaration type to test for.
 	 * \return non-zero if type is compatable with constraints.
 	 */
	int    IsTypeValid( Decl_t t ) {
		if( !constrained ) {
			MASSERT( typeConstraints.size() == 0 );
			return TRUE;
		}
		list<Decl_t> searchList;
		CDecl::GetMembers( t, searchList );
	        list<Decl_t>::iterator ptr;
		if( typeConstraints.count( t ) != 0 ) {
		    return TRUE;
		}
	        for( ptr = searchList.begin(); 
					ptr != searchList.end(); ++ptr ) {
		    if( typeConstraints.count( *ptr ) != 0 ) {
			return TRUE;
		    }
		}
		return FALSE;	
	}
	/**
 	 * Dump declaration info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
private:
	/**
	 * Ensure these are never called.
	 */
	void SetRange( CNode* ) { MASSERT( FALSE ); }
	CFref( const CFref& );
};

#endif // CFREF_HPP
