/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */

/******************************************************************************
 *
 *
 *	   cmodule.cpp
 *		- class for managing module, macromodule and	
 *		  primitive definitions 
 *
 *
 ******************************************************************************
 */


#include <stdio.h>
#include <list>
#include "main.h"
#include "cmodule.h"
#include "csymbol.h"
#include "cinstance.h"
#include "cparam.h"
#include "cvar.h"
#include "cfunction.h"
#include "cnet.h"
#include "cnode.h"
#include "cportdir.h"
#include "cport.h"
#include "cgenvar.h"
#include "cfref.h"
#include "csymtab.h"

extern CSymtab<CDecl> 	cuSymtab;
extern CObstack*	unitStack; 

/*******************************************
 	LookupModule
 *******************************************/

CModule* CModule::LookupModule( CSymtab<CDecl>& symtab, char* aName )
{
	CSymbol* symbol = CSymbol::Lookup( aName );
	CDecl*   decl = symtab.Lookup( symbol ); 
	switch( decl->GetType() ) {
	case eUDP:
	case eMODULE:
	case eMACRO:
	    return dynamic_cast<CModule*>(decl);    
	}
	return NULL;
}

/*******************************************
	Constructor
 *******************************************/
	CModule::CModule( CSymbol* aName, Coord_t* aLoc ) : 
		CBlock( aName, aLoc, eMODULE ),
		moduleDirectlyDefined(FALSE),
		moduleIncluded(FALSE)
{

	defined		 = FALSE;
	isMacroModule    = FALSE;
	isPrimitive      = FALSE;
	cell             = 0;
	protect          = 0;
	hasTimescale	 = FALSE;
	precision        = 0;
	unit             = 0;
	defaultNetType   = vpiWire;
	unconnectedDrive = vpiHiZ;
	inlinePortDecls  = 0;
	inlineParamDecls = 0;
	portListValid    = FALSE;
	portDirListValid    = FALSE;
}
/*******************************************
	SetFrefSymtab
	- Set modules fref symbol table 
 *******************************************/
void	CModule::SetFrefSymtab( CSymtab<CDecl>& symtab )
{
	frefSymtab = symtab;
}
/*******************************************
	SetPortSymtab
	- Set modules port symbol table 
 *******************************************/
void	CModule::SetPortSymtab( CSymtab<CDecl>& symtab )
{
	portSymtab = symtab;
}
/*******************************************
	IsMacroModule
	- return TRUE if module is a macromodule 
 *******************************************/
int	CModule::IsMacroModule( void )
{
	return isMacroModule;
}
/*******************************************
	SetMacroModule
	- set macro module flag 
 *******************************************/
void	CModule::SetMacroModule( int flag )
{
	isMacroModule = flag;
}
/*******************************************
	IsPrimitive
	- return TRUE if module is a primitive 
 *******************************************/
int	CModule::IsPrimitive( void )
{
	return isPrimitive;
}
/*******************************************
	SetPrimitive
	- set primitive flag 
 *******************************************/
void	CModule::SetPrimitive( int flag )
{
	isPrimitive = flag;
}
/*******************************************
	GetPortDirList
	- return module's port dir list 
 *******************************************/

vector<CPortDir*>* CModule::GetPortDirList()
{
	if( !portDirListValid ) {
	    portDirList.erase( portDirList.begin(), portDirList.end() );
	    ArgList2Vector<CPortDir*>(GetCodeList(),ePORT_DECL,0, portDirList);
	    portDirListValid = TRUE;
	}
	return &portDirList;
}
/*******************************************
	GetPortList
	- return module's port list 
 *******************************************/

vector<CPort*>*	CModule::GetPortList()
{
	if( !portListValid ) {
	    portList.erase( portList.begin(), portList.end() );
	    ArgList2Vector<CPort*>(ports,ePORT_DEF,0,portList);
	    portListValid = TRUE;
	}
	return &portList;
}
/*******************************************
	SetPorts
	- set module's ports
 *******************************************/

void  CModule::SetPorts( CNode* ports )
{
	portListValid = FALSE;
 	this->ports = ports; 
}
/*******************************************
	GetFunctionList
	- return module's function/task list 
 *******************************************/

vector<CFunction*>* CModule::GetFunctionList()
{
	return &functionList;
}
/*******************************************
	GetInstanceList
	- return module's instance list 
 *******************************************/

vector<CInstance*>* CModule::GetInstanceList()
{
	return &instanceList;
}

/*******************************************
	PreVisit
	- Support routine for PreVisit 
 	  tree transveral routine
 *******************************************/

void CModule::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	codeList->PreVisit1( func, data );	
    }
}

/*******************************************
	PostVisit1
	- Support routine for PostVisit 
 	  tree transveral routine
 *******************************************/

void CModule::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	codeList->PostVisit1( func, data );	
    }
}

/*******************************************
	PostVisit1
	- Support routine for PostVisit 
 	  tree transveral routine
 *******************************************/

void CModule::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	SetCodeList( codeList->PostSubVisit1( func, data ) );	
    }
}

/*******************************************
	Add
	- add function/tasks to module 
 *******************************************/
void	CModule::Add( CFunction* function )
{
	functionList.push_back( function );
}

/*******************************************
	Add
	- add instance to module 
 *******************************************/
void	CModule::Add( CInstance* instance )
{
	instanceList.push_back( instance );
}
/*******************************************
	Add
	- add fref to module 
 *******************************************/
void	CModule::Add( CFref* fref )
{
	frefList.push_back( fref );
}
/*******************************************
	Add
	- add code to module 
 *******************************************/
void	CModule::Add( CNode* n )
{
	portDirListValid = FALSE;
	CBlock::Add(n); 
}
/*******************************************
	FindDecl
	- find decl by name(symbol) 
 *******************************************/
CDecl*	CModule::FindDecl( CSymbol* sym )
{
    return symtab.Lookup( sym );
}
/*******************************************
	FixupPorts	
	- perform post parsing fixup 
 *******************************************/

static Decl_t __ValidateDirection( CPort* port, CNode* n ){
    CFref* fref;

    if( !n ) {
	return eNONE;
    }
    switch( n->GetOp() ) {
    case eFWD_REF:
	fref = n->Arg<CFref*>(0);	
	return fref->Direction();	
    case eARRAY:
	return __ValidateDirection( port, n->Arg<CNode*>(0) );
    case eCAT:
	{
	    Decl_t ldir = __ValidateDirection( port, n->Arg<CNode*>(0) );
	    Decl_t rdir = __ValidateDirection( port, n->Arg<CNode*>(1) );
	    if( ldir != rdir ) {
		error( port->GetCoord(),
		    "Inconsistant direction specified for port\n" );
	    }
	    return ldir;
	}
    default:
	MASSERT( FALSE );
    }
}

void CModule::FixupPorts( ) 
{
    int count = 0;
    int depth = 0;
    CNode* codeList = GetCodeList();
	vector<CPort*>::iterator itr;
	/*
	 * declare all untyped ports as wires
	 * connect all port and forward references
	 * to their real decl
	 */
	RedeclarePorts();
	LinkFrefs( portSymtab, cuSymtab, frefList, inferVectors );
	for( itr = GetPortList()->begin(); 
				itr != GetPortList()->end(); ++itr ) {
		CPort*	port;
		port = (*itr);
		Decl_t dir = __ValidateDirection( port, port->GetConnection() );
		MASSERT( dir == eINPUT ||
			 dir == eOUTPUT ||
			 dir == eINOUT ||
			 dir == eNODIR ||
			 dir == eNONE );
		port->Direction( dir );
	}
	Reconnect( codeList );
	for( itr = GetPortList()->begin(); 
			itr != GetPortList()->end(); ++itr ) {
		CPort*	port;
		CNode*	connect;
		INT32	width;
		port = (*itr);
		connect = port->GetConnection();
		Reconnect( connect );
	}

}
/*******************************************
	Dump	
	- print out module info 
 *******************************************/

class Dump_ {
	FILE* f;
public:
	Dump_(FILE* f): f(f){};
	void operator()( CDecl* r ) { r->Dump(f); }
};

void	CModule::Dump( FILE* f )
{
	
	if( isMacroModule ) {
		fprintf( f, "MACROMODULE: " );
	} else if( isPrimitive ) {
		fprintf( f, "PRIMITIVE: " );
	} else {
		fprintf( f, "MODULE: " );
	}
	fprintf( f, "%s, defined in  ", GetName() );
	CDecl::Dump( f );
	fprintf( f, "\n" );
	fprintf( f, "symbol table:\n" );
	GetSymtab().Dump(f,0);
	fprintf( f, "definitions:\n" );

	for_each( GetPortList()->begin(), GetPortList()->end(), Dump_(f) );
	for_each( GetPortDirList()->begin(), GetPortDirList()->end(), Dump_(f) );
	for_each( paramList.begin(), paramList.end(), Dump_(f) );
	for_each( varList.begin(), varList.end(), Dump_(f) );
	for_each( netList.begin(), netList.end(), Dump_(f) );
	for_each( genvarList.begin(), genvarList.end(), Dump_(f) );
	for_each( functionList.begin(), functionList.end(), Dump_(f) );
	vector<CBlock*>* blocks = GetChildList();
	for_each( blocks->begin(), blocks->end(), Dump_(f) );
}

