/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cxprop.cpp
 *		- methods for verilog design rule checker 
 *
 ******************************************************************************
 */

/**
 *
 * \page xprop_plugin Xprop Tool
 * The xprop tool is specified using the -tool xprop switch on the 
 * command line. This filter will cause the verilog code (ASTs) to be 
 * instrumented such that X's are properly propagated by verilog 
 * simulation tools. This is used to reduce the simlulated differences 
 * between gatelevels representations and RTL. I a properly constructed 
 * design flow this greatly reduces the need for gate level simulation 
 * (which is much more expensive than RTL simulation). For a detailed 
 * description of the transformations done by vrq and the reasoning 
 * behind it see the \ref xpropRational page.
 *
 * \section xprop_switches Xprop Switches
 * Vrq adds instrumentation all constructs except clock gating. The 
 * switches below may be used to alter the default behavior:
 * - \b +xprop-disable-if Disable if instrumentation
 * - \b +xprop-disable-case Disable case instrumentation
 * - \b +xprop-disable-index Disable reference assignment instrumentation
 * - \b +xprop-disable-ternary Disable ’?’ instrumentation
 * - \b +xprop-enable-clock Enable instrumentation of clocks 
 *   (to check clock gaters)
 *
 * The following switches alter how variables and constructs are treated. 
 * By default all variables are assume to be able to be 0,1,x,z. You can 
 * selectively indicate certain variables can never be x using these switches:
 * - \b +xprop-nox-attr=<name>: Allow an attribute on a variable declaration 
 *   to specify the variable will never be x. 'name' is the attribute name 
 *   that will be used to indicate the variable will never be x. For 
 *   instance if the switch +xprop-nox-attr=NOX is used then a variable 
 *   tag as non-x by:
 * \code
 * (* NOX *) reg v1;          // indicate v1 is never x
 * (* NOX *) wire w1, w2, w3; // indicate w1, w2 and w3 will never be x</pre>
 * \endcode
 *
 * - \b +xprop-int-nox Declare that all variables specified as integer will 
 *   never be x. This is useful when the local design rules only use 
 *   integers as index in loops, etc.
 *
 * The \b +xprop-nox-attr=<name> switch may also be used to disable
 * x-propagation on selected constructs:
 *
 * - \b if
 * \code
 * // In the case below the attribute indicates that this if statement
 * // should not be xprop'd.
 * (* NOX *) if( condition ) ...
 * \endcode
 *
 * - \b case
 * \code
 * // In the case below the attribute indicates that this case statement
 * // should not be xprop'd.
 * (* NOX *) case( condition ) ...
 * \endcode
 *
 * - \b procedural blocks
 * \code
 * // In the case below the attribute indicates that all statements
 * // between the begin/end pair should not be xprop'd.
 * (* NOX *) begin
 * 	...
 * end
 * \endcode
 *
 * - \b structural blocks
 * \code
 * // In the case below the attribute indicates that all statements
 * // between the generate/endgenerate pair should not be xprop'd.
 * (* NOX *) generate
 * 	...
 * endgenerate
 * \endcode
 *
 * - \b functions
 * \code
 * // In the case below the attribute indicates that the function 
 * // should not be xprop'd.
 * (* NOX *) function funcName;
 * 	...
 * endfunction
 * \endcode
 *
 * - \b task
 * \code
 * // In the case below the attribute indicates that the task 
 * // should not be xprop'd.
 * (* NOX *) task funcName;
 * 	...
 * endtask
 * \endcode
 *
 * - \b module
 * \code
 * // In the case below the attribute indicates that the module 
 * // should not be xprop'd.
 * (* NOX *) task funcName;
 * 	...
 * endmodule
 * \endcode
 *
 * - \b ternary operator
 * \code
 * // In the case below the attribute indicates that the ternary operator
 * // should not be xprop'd.
 * initial out = cond ? (* NOX *) true_expression : false_expression;
 * \endcode
 *
 * - \b clock blocks 
 * \code
 * // In the case below the attribute indicates that the clock 
 * // should not be xprop'd.
 * if (* NOX *) @(posedge clock) ...
 * \endcode
 *
 *
 * These switches alter other miscellaneous behaviors:
 * - \b +xprop-allow-casex Allow casex statements in instrumented code. 
 *   Normally casex statements are disallowed due to the fact that an 
 *   'x' is treated as a true don't care in the matching criteria, 
 *   making it no possible to instrument properly.
 * - \b +xprop-begin=<pragma> Supply a comment to append before the insertion 
 *   of xprop instrumentation
 * - \b +xprop-end=<pragma> Supply a comment to post pend after the insertion 
 *   of xprop instrumentation
 * - \b +xprop-clk-begin=<pragma> Supply a comment to append before 
 *   the insertion of clock xprop instrumentation
 * - \b +xprop-clk-end=<pragma> Supply a comment to post pend after 
 *   the insertion of clock xprop instrumentation
 * - \b +xprop-clk-edge-control=[!]<tickdefine> Supply a preprocessor 
 *   tickdefine to enable instrumentation on both edges of the clock. 
 *   ! indicates tickdefine disables both edges. If switch isn't supplied 
 *   the behavior is single edge unless XPROP_BOTH_EDGES is defined.  
 *
 * Note these constructs are useful for either visually flagging xprop code 
 * or to enable other tools like coverage or synthesis to recognize the 
 * instrumentation. For example when the following switches 
 * +xprop-begin="coverage off" +xprop-end="coverage on" are used on 
 * this code snippet:
 * \code
 * if( cond ) begin
 *     a = b;
 * end else begin
 *     a = b+1;
 * end
 * \endcode
 *
 * This is the resultant output:
 * \code
 * if(cond) begin
 *    a = b;
 * end else if(~cond) begin
 *    a = b+1;
 * end
 * // coverage off
 * else begin
 *    a = 1'hx;
 * end
 * // coverage on
 * \endcode
 *
 *
 * \page xpropRational Xprop Rational
 * Xprop is a tool that instruments verilog code such that x's are 
 * propagated and not squashed. This is desirable as it reduces the 
 * simulation differences between RTL and gate level simulations. This 
 * sometimes masks bugs in RTL simulations and thereby requires a much 
 * greater degree of gate level simulation be performed. It is the 
 * experience of many who have used this tool that most of the discovered 
 * bug at the gate level would have been caught at the RTL simulation 
 * level if an xprop tool had been applied prior to RTL simulation.
 *
 * X's occur in simulation for a number of reasons:
 * - Use in RTL to indicated don't care degrees of freedom to synthesis.
 * - Uninitialized flops.
 *
 * \section Terms
 * - <b>X-propagation</b> is the process of outputing an X from a block of 
 *   logic when it's inputs are X.
 * - <b>X-pessimism</b> Is the condition when output of a block of logic 
 *   is X in cases where it never could be based upon the inputs.
 * - <b>X-optimism or X-squashing</b> Is the condition when a block of 
 *   logic outputs a non-X value when one or more of it's inputs are X. 
 *   Note this  naturally occurs in verilog due to the fact that X's are 
 *   interpreted as 0's in conditional expressions.
 *
 *
 * \section theory_of_operation Theory of Operation
 * Here is an example:
 * \code
 * if( select )
 *     // this branch is taken when cond is 1
 *     out = in1;
 * else
 *     // this branch is taken when cond is either 0 or X
 *     out = in2;
 * \endcode
 *
 * In this case an X on signal select does not result in out being X. 
 * Hence X's are squashed. Synthesis may result in code that looks 
 * like this:
 * \code
 * out = (in1 & select) | (in2 & select)
 * \endcode
 *
 * The simulation of these statements will not match in all cases.
 * Another example where gate level simulation may not match RTL is 
 * for a 2-to-1 mux expressed as:
 * \code
 * out = select ? in1 : in2;
 * \endcode
 *
 * However synthesis may result in the following:
 * \code
 * out = (in1 & select) | (in2 & select)
 * \endcode
 *
 * When select is X the simulation of these statements will differ.
 *
 * Similar issues arise with case statements and assignments to variables 
 * that use non-constant indices. The solution is to instrument the 
 * verilog code to add pessimism to such statements.
 * 
 * One way to model this is whenever an ambitious (relative to X) construct 
 * is used there are multiple possible outcomes. These multiple 
 * possible outcomes need to be modeled resolving the ambiguity create by the X.
 * This can be done using a XResolution() function. It takes two inputs, 
 * one for each possible expression output. The resolution function combines 
 * the two possiblities and creates a pessimistic output. For instances 
 * the code:
 * \code
 * if( select )
 *     out = in1;
 * else
 *     out = in2;</pre>
 * \endcode
 * Can be abstractly represented as:
 * \code
 * if( ^select === 1'bX )
 *     out = XResolution( in1, in2 );
 * else if( select )
 *     out = in1;
 * else
 *     out = in2;
 * \endcode
 *
 * Vrq uses a very simplistic XResolution function that results in less 
 * code bloat and overhead at the expense of increased pessimism. The 
 * XResolution function ignores the inputs and always returns X.
 *
 *
 * \section transformations Transformations
 *
 * \subsection if_transformation If Transformation
 * The following code snippet:
 * \code
 * if( cond ) being
 *     r1 = r2;
 * end
 * \endcode
 * Is transformed into:
 * \code
 * if(cond) begin
 *     r1 = r2;
 * end else if(~cond) ;
 * else begin
 *     r1 = 1'hx;
 * end
 * \endcode
 *
 * When an else clause is included, this code snippet:
 * \code
 * if( cond ) begin
 *     r1 = r2;
 * end else begin
 *    r1 = r3;
 * end
 * \endcode
 * is transformed into:
 * \code
 * if(cond) begin
 *     r1 = r2;
 * end else if(~cond) begin
 *     r1 = r3;
 * end else begin
 *     r1 = 1'hx;
 * end
 * \endcode
 *
 * \section case_transformation Case Constructs
 * A case statement without a default clause, like this:
 * \code
 * case( cond )
 * 0: out = in0;
 * 1: out = in1;
 * 2: out = in2;
 * endcase
 * \endcode
 *
 * Is transformed into this:
 * \code
 * case(cond)
 * 0: begin
 *     out = in0;
 *    end
 * 1: begin
 *     out = in1;
 *    end
 * 2: begin
 *     out = in2;
 *    end
 * default: begin
 *    case(^cond)
 *        1'b0,
 *        1'b1: ;
 *           default: begin
 *              out = 1'hx;
 *           end
 *     endcase
 *   end
 * endcase
 * \endcode
 * 
 * A case statement with a default clause is transformed from:
 * \code
 * case( cond )
 * 0: out = in0;
 * 1: out = in1;
 * 2: out = in2;
 * default: out = in3;
 * endcase
 * \endcode
 * to this:
 * \code
 * case(cond)
 * 0: begin
 *     out = in0;
 * end
 * 1: begin
 *     out = in1;
 * end
 * 2: begin
 *     out = in2;
 * end
 * default: begin
 *     case(^cond)
 *         1'b0,
 *         1'b1: begin
 *     out = in3;
 *     end
 *     default: begin
 *             out = 1'hx;
 *         end
 *     endcase
 * end
 * endcase
 * \endcode
 *
 *
 * A case statement with a constant case expression is transformed from:
 * \code
 * case( 1'b1 )
 * cond1: out = in0;
 * cond2: out = in1;
 * cond3: out = in2;
 * default: out = in3;
 * endcase
 * \endcode
 * to this:
 * \code
 * case(^{cond1,cond2,cond3})
 * 1'b0,
 * 1'b1: begin
 *     case(1'b0)
 *         cond1: begin
 *             out = in0;
 *         end
 *         cond2: begin
 *             out = in1;
 *         end
 *         cond3: begin
 *             out = in2;
 *         end
 *         default: begin
 *             out = in3;
 *         end
 *     endcase
 * end
 * default: begin
 *     out = 1'hx;
 * end
 * endcase
 * \endcode
 * \section bit_select_transformation Assignment with Bit Select
 * An assignment with a bit select is transformed from:
 * \code
 * reg [3:0] out;
 * ...
 * out[i] = in;
 * \endcode
 * to:
 * \code
 * case(^i)
 * 1'b0,
 * 1'b1: begin
 *      out[i] = in;
 * end
 * default: begin
 *      out = 4'hx;
 * end
 * endcase
 * \endcode
 *
 * \section part_select_transformation Assignment with Part Select
 * An assignment with a part select is transformed from:
 * \code
 * reg [1:0] in;
 * reg [3:0] out;
 * ...
 * out[i+:2] = in;
 * \endcode
 * to:
 * \code
 * case(^i)
 * 1'b0,
 * 1'b1: begin
 *     out[i+:2] = in;
 * end
 * default: begin
 *     out = 4'hx;
 * end
 * endcase
 * \endcode
 *
 * \section array_with_index_transformation Assignment to Array with Index
 * The following code gets transformed:
 * \code
 * reg [3:0] in;
 * reg [3:0] array[8:0][2:0];
 * ...
 * array[w][y] = in;
 * array[w][1] = in;
 * array[w][y][1:0] = in[1:0];
 * \endcode
 * to:
 * \code
 * // array[w][y] = in;
 * case(^{w,y})
 * 1'b0,
 * 1'b1: begin
 *     array[w][y] = in;
 * end
 * default: begin : ___block0002
 *     integer ___index0000;
 *     integer ___index0001;
 *     for(___index0001 = 2&gt;=0 ? 0 : 2;___index0001&lt;=(2&gt;=0 ? 2 : 0);___index0001 = ___index0001+1) begin
 *         for(___index0000 = 8&gt;=0 ? 0 : 8;___index0000&lt;=(8&gt;=0 ? 8 : 0);___index0000 = ___index0000+1) begin
 *              array[___index0000][___index0001] = 4'hx;
 *         end
 *     end
 * end
 * endcase
 *
 * // array[w][1] = in;
 * case(^w)
 * 1'b0,
 * 1'b1: begin
 *      array[w][1] = in;
 * end
 * default: begin : ___block0004
 *     integer ___index0003;
 *     for(___index0003 = 8&gt;=0 ? 0 : 8;___index0003&lt;=(8&gt;=0 ? 8 : 0);___index0003 = ___index0003+1) begin
 *          array[___index0003][1] = 4'hx;
 *     end
 * end
 * endcase
 *
 * // array[w][y][1:0] = in[1:0];
 * case(^{w,y})
 * 1'b0,
 * 1'b1: begin
 *      array[w][y][1:0] = in[1:0];
 * end
 * default: begin : ___block0007
 *     integer ___index0005;
 *     integer ___index0006;
 *     for(___index0006 = 2&gt;=0 ? 0 : 2;___index0006&lt;=(2&gt;=0 ? 2 : 0);___index0006 = ___index0006+1) begin
 *          for(___index0005 = 8&gt;=0 ? 0 : 8;___index0005&lt;=(8&gt;=0 ? 8 : 0);___index0005 = ___index0005+1) begin
 *              array[___index0005][___index0006][1:0] = 2'hx;
 *          end
 *      end
 *  end
 *  endcase
 * \endcode
 * 
 * \section hook_transformation ? Operator
 * This code is transformed:
 * \code`
 * out = cond ? in1 : in2;
 * \endcode
 * to:
 * \code
 * out = cond ? in1 : ~cond ? in2 : 1'hx;
 * \endcode
 * 
 * \section clock_enable_t Clock Enables
 * This code is transformed:
 * \code
 * always @(posedge clk)
 *     r <= r_P;
 * \endcode
 * to:
 * \code
 * `ifdef XPROP_BOTH_EDGES
 * always @(clk)
 * `else
 * always @(posedge clk)
 * `endif
 * if( clk ) begin
 *     r <= r_P;
 * end else if( ~clk ) ; 
 * else begin
 *     r <= 1'bx;
 * end
 * \endcode
 */

/**
 * \cond PLUGIN_INTERNALS
 */

#include <vector>
#include <fstream>
#include <set>
#include "glue.h"
#include "main.h"
#include "cxprop.h"
#include "cdumpverilog.h"
#include "cproxytype.h"
#include "cmodule.h"
#include "csymbol.h"
#include "cport.h"
#include "cfref.h"
#include "cvar.h"
#include "cnet.h"
#include "cparam.h"
#include "cnode.h"
#include "cinstance.h"
#include "cfunction.h"
#include "cportdir.h"
#include "array.h"
#include "compile.h"
#include "cdatatype.h"
#include "cstructtype.h"

enum    TranslateState { eTRANSLATE_DEFAULT, eTRANSLATE_OFF, eTRANSLATE_ON } ;
static  TranslateState	translateState;
static  CNode* prependToModule;
static  CNode* insertionBeginComment = NULL;
static  CNode* insertionEndComment = NULL;
static  CNode* insertionClkBeginComment = NULL;
static  CNode* insertionClkEndComment = NULL;
static  int disableIf = 0;
static  int disableCase = 0;
static  int disableIndex = 0;
static  int disableTernary = 0;
static  char* noxAttribute = NULL;
static  int integerIsNoX = 0;
static  int enableClock  = 0;
static  const char* edgeControl  = NULL;
static  const char* edgeAttrName  = NULL;
static  Message* mCCBMS;
static  Message* mCDWOO;
static  Message* mNSTOS;
static  Message* mUESCI;
static  Message* mXBADE;
static  Message* mXVREF;

extern  char*		outputPath;

static CNode* XPropStatement( CBlock* module, CBlock* block, CNode* n );
static CNode* CreateXAssignment( 
	    CBlock* module, 
	    CBlock* block, 
	    NodeOp_t op, 
	    CNode* delay, 
	    CNode* reference,
	    vector<CNode*>& indices,
	    int conditional,
	    const set<CNode*>& xOutIndices,
	    const set<CNode*>& noXOutIndices );
static CNode* GetReferenceSelect( CNode* reference );
static CNode* GetReferenceVarRef( CNode* reference );
static CNode* GetReferenceRoot( CNode* reference, 
				vector<CNode*>& residualIndices );
static int IsReferenceUnpackedArray( CNode* reference );

extern CDumpVerilog*	dumpVerilog;
extern int		errorCount;

inline CNode* Clone( CNode* n ) {
	if( n == NULL ) {
	    return NULL;
	}
	return n->Clone();
}

/**
 * LHS (Left Hand Side) record
 */
class LHS {
public:
    CNode*         reference;  ///< full reference to lHS
    vector<CNode*> indices;    ///< reference indices
    CNode*         delay;      ///< assignment delay
    NodeOp_t       op;	       ///< assigment op type
    Coord_t*       loc;	       ///< assignment coordinate
    // precalculated info on LHS
    CNode*   varRef;		///< reference to base variable
    CNode*   root;		///< reference to root for comparisions
    vector<CNode*> residualRootIndices;  ///< residual root indices
    CNode*   select;		///< part or bit select expression
    CNode*   msb;		///< msb expression of part select
    CNode*   lsb;		///< lsb expression of part select
    int	     psType;		///< part select type 
    int      hasBitSelect;	///< bit select present flag
    int      hasPartSelect;	///< part select present flag
    int      isArray;		///< reference is an unpacked array
    // temporaries for analysis
    set<CNode*> xOutIndices;
    set<int> coveredBitVector;
    int	     coveringInfoValid;
    int	     fullyCovered;
    int	     drop;
public:
    /**
     * Constructor
     * \param reference lhs reference tree
     * \param indices vector of reference indices
     * \param delay assignment delay construct
     * \param op assignement node opcode
     * \param loc coordinate of assignment
     */
    LHS( CNode*         reference,
         vector<CNode*> indices,
         CNode*	        delay,
         NodeOp_t	op,
         Coord_t*	loc ) 
      : reference(reference),
        indices(indices),
        delay(delay),
        op(op),
        loc(loc)
    {
	varRef = GetReferenceVarRef( reference );
	root = GetReferenceRoot( reference, residualRootIndices );
	isArray = IsReferenceUnpackedArray( reference );
    
        hasBitSelect = FALSE;
        hasPartSelect = FALSE;
        msb     = NULL;
        lsb     = NULL;
        psType     = 0;
        select =  GetReferenceSelect( reference );
        if( select ) {
	    residualRootIndices.pop_back();
            CNode* rselect = select;
            while( rselect && rselect->GetOp() == eMACRO_EXPR ) {
               rselect = rselect->Arg<CNode*>(1);
            }
            switch( rselect->GetOp() ) {
            case eSLICE:
                    psType = 0;
                    hasPartSelect = TRUE;
                    msb = rselect->Arg<CNode*>(0);
                    lsb = rselect->Arg<CNode*>(1);
                    break;
            case ePSLICE:
                    psType = 1;
                    hasPartSelect = TRUE;
                    msb = rselect->Arg<CNode*>(0);
                    lsb = rselect->Arg<CNode*>(1);
                    break;
            case eMSLICE:
                    psType = -1;
                    hasPartSelect = TRUE;
                    msb = rselect->Arg<CNode*>(0);
                    lsb = rselect->Arg<CNode*>(1);
                    break;
            default:
                    hasBitSelect = TRUE;
                    break;
            }
        }
    }
    /**
     * Calculate LHS hash ID
     * \return hash index.
     */
    unsigned Hash() const {
        unsigned result = 0; 
        int s1 = residualRootIndices.size();
        if( root ) { result ^= root->Hash(); }
        if( delay ) { result ^= delay->Hash(); }
        result ^= s1;
        for( int i = 0; i < s1; ++i ) {
    	    if( residualRootIndices[i] ) { 
		result ^= residualRootIndices[i]->Hash(); 
	    }
        }
        return result;
    }
     /**
     * Dump LHS info for debug
     */
    void Dump() const {
        printf( "LHS: " );
        printf( "ref=%p,delay=%p,op=%d,msb=%p,lsb=%p"
    	    ",psType=%d,hasBitSelect=%d,isize=%d,ciValid=%d,fully=%d"
    	    ",covered={", 
    	    reference, delay, op, msb, lsb, psType, 
	    hasBitSelect, indices.size(),
    	    coveringInfoValid, fullyCovered  );
        set<int>::iterator ptr;
        for( ptr = coveredBitVector.begin(); 
    	 ptr != coveredBitVector.end(); ++ptr ) {
    	if( ptr != coveredBitVector.begin() ) {
    	    printf( "," );
    	}
    	printf( "%d", *ptr );
        }
        printf( "}\n" );
    }
    /**
     * Check if left hand sides are compatible.
     * Compatible LHS are either equivilent of differ
     * only in the select info.
     * \param lhs left hand side to compare to.
     * \return TRUE if LHS's are compatible.
     */
    int Compatible( const LHS* lhs ) const
    {
        /*
         * reference roots info must match for LHS's to 
         * compatible
         */
        int s1 = residualRootIndices.size();
        int s2 = lhs->residualRootIndices.size();
        if( !Equivalent( root, lhs->root ) || 
            !Equivalent( delay, lhs->delay ) || 
            op != lhs->op || 
            s1 != s2 ) { 
            return FALSE;
        }
        for( int d = 0; d < s1; ++d ) {
            if( !Equivalent( residualRootIndices[d], lhs->residualRootIndices[d] )  ) {
        	return FALSE;
            }
        }
        /*
         * if covering info is available and vector is fully
         * covered, then entries are compatible.
         */
        if( (coveringInfoValid && fullyCovered) ||
            (lhs->coveringInfoValid && lhs->fullyCovered) ) {
    	return TRUE;
        }
        /*
         * if covering info is avalable for both entries
         * they are compatable and can match. Otherwise
         * the selects must match exactly.
         */
        if( coveringInfoValid != lhs->coveringInfoValid ) {
            return FALSE;
        }
        if( !coveringInfoValid && !lhs->coveringInfoValid ) {
            if( hasBitSelect != lhs->hasBitSelect ) {
        	    return FALSE;
            }
            if( hasBitSelect && lhs->hasBitSelect &&
                !Equivalent( select, lhs->select ) ) {
                return FALSE;
            }
            if( psType != lhs->psType ) {
        	    return FALSE;
            }
            if( msb && !Equivalent( msb, lhs->msb) ) {
                return FALSE;
            }
            if( lsb && !Equivalent( lsb, lhs->lsb) ) {
                return FALSE;
            }
        }
        
        return TRUE;
    }
    /**
     * Determine if this lhs is a subset of parent LHS
     * \param parent left handside to check against.
     * \return TRUE if this lhs is a subset of the parent LHS.
     */
    int IsSubsetOf( const LHS* parent ) const
    {
        /*
         * LHS must be compatible to be subset
         */
        if( !Compatible( parent ) ) {
    	return FALSE;
        }
        /*
         * if covering info is not available then 
         * match is exact subset
         */
        if( !coveringInfoValid ) {
                return TRUE;	   
        }
    
        /*
         * if parent is fully covered, exact match
         */
        if( parent->fullyCovered ) {
            return TRUE;
        }
    
        /*
         * check if range is a subset of existing range
         */
        if( !fullyCovered ) {
            set<int> difference;
            set_difference( coveredBitVector.begin(),
            		coveredBitVector.end(),
            		parent->coveredBitVector.begin(),
            		parent->coveredBitVector.end(),
            		inserter(difference, difference.begin() ) );
            if( difference.empty() ) {
                return TRUE;
            }
        }
        return FALSE;
    }
};

/**
 * Hash table for managing LHS (left hand sides)
 */
class LHSHash {
    map<unsigned,vector<LHS> > id2LHS;	///< hash map
public:
    /**
     * Constructor
     */
    LHSHash() {}
    /**
     * Destructor
     */
    ~LHSHash() {}
    /**
     * find compatable LHS
     * \param lhs left hand side to search for.
     * \return compatable LHS, NULL if none found.
     */
    LHS* Find( const LHS& lhs )
    {
        LHS* match = NULL;
        vector<LHS>::iterator lhsPtr;
        map<unsigned,vector<LHS> >::iterator mapPtr;
        unsigned id = lhs.Hash();
        mapPtr = id2LHS.find(id);
        if( mapPtr == id2LHS.end() ) {
    	    return NULL;
        }	
        for( lhsPtr = mapPtr->second.begin(); 
	     lhsPtr != mapPtr->second.end(); ++lhsPtr ) {
    	    if( lhs.Compatible( &*lhsPtr ) ) {
                return &*lhsPtr;
    	    }
            return NULL;
        }
    }
    /**
     * insert LHS into hash table
     * \param lhs left hand side to insert.
     * \return return inserted LHS
     */
    LHS* Insert( const LHS& lhs )
    {
        unsigned id = lhs.Hash();
        id2LHS[id].push_back( lhs );
        return &(id2LHS[id].back());
    }
};

/************************************************
 	CreateToolInstance
	- entry point for dll
 ************************************************/
extern "C" {
CBackend* CreateToolInstance()
{
    return new CXProp();
}
}

/************************************************
 	Constructor
 ************************************************/
CXProp::CXProp( ) 
{
	translateState = eTRANSLATE_DEFAULT;
	switches.push_back( "+xprop-begin=<pragma>" );
	switchDescription["+xprop-begin=<pragma>"] = 
		"insertion begin pragma";
	switches.push_back( "+xprop-end=<pragma>" );
	switchDescription[ "+xprop-end=<pragma>" ] =
		"insertion end pragma";
	switches.push_back( "+xprop-clk-begin=<pragma>" );
	switchDescription["+xprop-clk-begin=<pragma>"] = 
		"insertion clock begin pragma";
	switches.push_back( "+xprop-clk-end=<pragma>" );
	switchDescription[ "+xprop-clk-end=<pragma>" ] =
		"insertion clock end pragma";
	switches.push_back( "+xprop-disable-if" );
	switchDescription[ "+xprop-disable-if" ] = "disable if instrumentation";
	switches.push_back( "+xprop-disable-case" );
	switchDescription[ "+xprop-disable-case" ] = 
						"disable case instrumentation";
	switches.push_back( "+xprop-disable-index" );
	switchDescription[ "+xprop-disable-index" ] = 
					"disable reference instrumentation";
	switches.push_back( "+xprop-disable-ternary" );
	switchDescription[ "+xprop-disable-ternary" ] = 
					"disable '?' instrumentation";
	switches.push_back( "+xprop-nox-attr=<name>" );
	switchDescription[ "+xprop-nox-attr=<name>" ] = 
					"attribute specifing variable will never be x";
	switches.push_back( "+xprop-int-nox" );
	switchDescription[ "+xprop-int-nox" ] = 
					"declare integers will never be x";

	switches.push_back( "+xprop-enable-clock" );
	switchDescription[ "+xprop-enable-clock" ] = 
					"instrument register clocks";

 	switches.push_back( "+xprop-clk-edge-control=[!]<tickdefine>" );
 	switchDescription[ "+xprop-clk-edge-control=[!]<tickdefine>" ] = 
 		"Supply a preprocessor tick define to enable instrumentation"
		" on both edges of the clock. ! indicates tickdefine"
		" disables both edges. If switch isn't supplied the behavior is"
		" single edge unless XPROP_BOTH_EDGES is defined";

	/*
 	 * register messages
 	 */
	mCCBMS = Message::RegisterWarning( this, Message::eERROR,
		"CCBMS", "casex cannot be made safe, please use casez" );
	mCDWOO = Message::RegisterWarning( this, Message::eERROR,
		"CDWOO", "cannot determine width of ?"
			 " operands, cannot instrument expression" );
	mNSTOS = Message::RegisterError( this, "NSTOS", 
			"nested synopsys translate_off statements" );
	mUESCI = Message::RegisterWarning( this, Message::eWARNING, "UESCI", 
			"unsupported edge sensitive construct for clock"
			" instrumentation" );
	mXBADE = Message::RegisterWarning( this, Message::eERROR, "XBADE", 
			"xprop on blocking assigns with delay events is"
			" not supported" );
	mXVREF = Message::RegisterWarning( this, Message::eWARNING, "XVREF", 
			"xprop on this variable reference is"
			" not supported" );
}

/************************************************
	GetToolName
 	- return name of tool
 ************************************************/
char* CXProp::GetToolName( void )
{
	static char	toolName[] = "xprop";

	return toolName;
}

/************************************************
	GetToolDescription
 	- return description of tool
 ************************************************/
char* CXProp::GetToolDescription( void )
{
	static char	toolDescription[] = "Add X propagation instrumentation";

	return toolDescription;
}
/************************************************
 * Bracket inserted code 
 ************************************************/
static CNode* BIC( CNode* n, int forceSingleStatement )
{
    if( forceSingleStatement &&
	( insertionBeginComment || insertionEndComment ) ) {
        n = (!n || n->GetOp() != eBLOCK_REF) ? 
	    cBLOCK_REF( NULL, n ) : n;	
    }
    if( insertionBeginComment ) {
        n = cLINK( insertionBeginComment->Clone(), n );
    }
    if( insertionEndComment ) {
        n = cLINK( n, insertionEndComment->Clone() );
    }
    return n;
}
/************************************************
 * Bracket original code 
 ************************************************/
static CNode* BOC( CNode* n, int forceSingleStatement )
{
    if( forceSingleStatement &&
	( insertionBeginComment || insertionEndComment ) ) {
        n = (!n || n->GetOp() != eBLOCK_REF) ? 
	    cBLOCK_REF( NULL, n ) : n;	
    }
    if( insertionEndComment ) {
        n = cLINK( insertionEndComment->Clone(), n );
    }
    if( insertionBeginComment ) {
        n = cLINK( n, insertionBeginComment->Clone() );
    }
    return n;
}
/************************************************
 * Bracket if with original true and
 * inserted false clause
 ************************************************/
static CNode* BIF_OTIF( CNode* condition, CNode* t, CNode* f )
{
    if( insertionBeginComment ) {
        t = (!t || t->GetOp() != eBLOCK_REF) ? 
	    cBLOCK_REF( NULL, t ) : t;	
        t = cLINK( t, insertionBeginComment->Clone() );
    }
    CNode* n = cIF( condition, t, f );
    if( insertionEndComment ) {
        n = cLINK( n, insertionEndComment->Clone() );
    }

    return n;
}


/*******************************************************
	IsTranslateOn
	- return TRUE if PRAGMA node is synopsys translate on 
********************************************************/

static int IsTranslateOn( CNode* n )
{
	char* comment = n->Arg<char*>(0);
	char* ptr = strstr( comment, "synopsys" );
	if( ptr == NULL ) {
		return FALSE;
	}
	ptr = strstr( ptr, "translate_on" );
	return (ptr != NULL );
}

/*******************************************************
	IsTranslateOff
	- return TRUE if PRAGMA node is synopsys translate off 
********************************************************/

static int IsTranslateOff( CNode* n )
{
	char* comment = n->Arg<char*>(0);
	char* ptr = strstr( comment, "synopsys" );
	if( ptr == NULL ) {
		return FALSE;
	}
	ptr = strstr( ptr, "translate_off" );
	return (ptr != NULL );
}
/*******************************************************
 	IsNoxNode
	- return TRUE if xprop is disabled for node 
	  and it's subtrees.
 ******************************************************/
static int IsNoxNode( CNode* n ) 	
{
    switch( n->GetOp() ) {
    case eBLOCK_REF:
	return  n->HasAttribute( noxAttribute );
    case eMODULE_DEF:
    case eFUNCTION_DEF: 
	return n->Arg<CBlock*>(0)->HasAttribute( noxAttribute );
    } 
    return FALSE;
}

/*******************************************************
	StringList2Comments
	- Convert list of strings to a tree of COMMENT
********************************************************/
static CNode* StringList2Comments( list<string>& args )
{
    CNode* result = NULL;
    list<string>::iterator ptr;
    for( ptr = args.begin(); ptr != args.end(); ++ptr ) {
	result = cLINK( result, cCOMMENT( strdup(ptr->c_str()) ) );	
    }
    return result;
}
/**
 * Determine if reference resolve into an unpacked array
 * \param reference reference to anaylse
 * \return returns TRUE if reference resolves into an unpacked array.
 */
static int IsReferenceUnpackedArray( CNode* reference )
{
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    CDataType*     dataType;

    if( reference->GetOp() == eARRAY ) {
	reference = reference->Arg<CNode*>(0);
    }
    ResolveReference( reference, 
		dataType, unpackedDimensions,
	        packedDimensions, residualPackedIndices );

    return unpackedDimensions.size() != 0;
} 

/**
 * return reference's select
 * \param reference reference to anaylse
 * \return returns select if reference has a part or bit select, returns
 *         NULL if reference does not have a select.
 */
static CNode* GetReferenceSelect( CNode* reference )
{
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    CDataType*     dataType;

    ResolveReference( reference, dataType, unpackedDimensions,
		      packedDimensions, residualPackedIndices );

    int hasSelect = (dataType->IsPacked() && 
	    (residualPackedIndices.size() > packedDimensions.size()) ||
	    (residualPackedIndices.size() > 0 && packedDimensions.size() == 1));
    if( hasSelect ) {
	return residualPackedIndices.back();
    } else {
	return NULL;
    }
} 

/**
 * return reference's variable ref node
 * \param reference reference to anaylse
 * \return returns references variable ref node
 */
static CNode* GetReferenceVarRef( CNode* reference )
{
    MASSERT( reference );
    while( 1 ) {
	switch( reference->GetOp() ) {
	case eARRAY:
	case eMEMBER:
	    reference = reference->Arg<CNode*>(0);
	case eVAR_REF:
	case eNET_REF:
	    return reference;
	default:
	    MASSERT( FALSE );
	}
    }
} 

/**
 * return root reference and residual indices
 * \param reference reference to anaylse
 * \param residualIndices residual root indices
 * \return root reference
 */
static CNode* GetReferenceRoot( CNode* reference, 
				vector<CNode*>& residualRootIndices )
{
    switch( reference->GetOp() ) {
    case eMEMBER:
	return reference;
    case eARRAY: {
        vector<CNode*> unpackedDimensions;
        vector<CNode*> packedDimensions;
	vector<CNode*> residualIndices;
        CDataType*     dataType;
        ResolveReference( reference, dataType, 
		      unpackedDimensions,
		      packedDimensions, residualIndices );
	if( residualIndices.size() ) {
            EList2VectorExclude( reference->Arg<CNode*>(1), 
			         set<NodeOp_t>(), residualRootIndices );
	    return reference->Arg<CNode*>(0);
	} else {
	    return reference;
	}
	} break;
    default:
	return reference;
    }
}

/**
 * return mapping of reference indices to corresponding ranges
 * \param reference reference to var
 * \param indices2Range mapping to populate with indices to range info
 */
static void CalculateReferenceRanges( CNode* reference, 
				map<CNode*,CNode*>& indices2Range )
{
    switch( reference->GetOp() ) {
    case eMEMBER:
	break;
    case eARRAY: {
        vector<CNode*> indices;
        EList2VectorExclude( reference->Arg<CNode*>(1), 
			     set<NodeOp_t>(), indices );
        vector<CNode*> unpackedDimensions;
        vector<CNode*> packedDimensions;
        vector<CNode*> residualPackedIndices;
        CDataType*     dataType;
        ResolveReference( reference->Arg<CNode*>(0), dataType, 
		      unpackedDimensions,
		      packedDimensions, residualPackedIndices );
	
	for( int i = 0; i < indices.size(); ++i ) {
	    CNode* range;
	    if( i < unpackedDimensions.size() ) {
		range = unpackedDimensions[i];
	    } else {
		range = packedDimensions[i-unpackedDimensions.size()];
	    }
	    indices2Range[indices[i]] = range;
	}
	} break;
    default:
	return;
    }

    CalculateReferenceRanges( reference->Arg<CNode*>(0), indices2Range );
}

/**
 * Clone reference substituting index expression. Also delete final
 * index if requested
 * \param reference reference to clone
 * \param indiceMap map of old to new indices
 * \param deleteSelect delete final select index
 * \param appendList list of indices to append to final array
 * \return cloned and modified reference
 */
static CNode*  CloneReferenceWithSubstitution( CNode* reference, 
		map<CNode*,CVar*>& indiceMap, int deleteSelect, 
		const vector<CNode*>& appendList )
{
    CNode* newRef = NULL;
    switch( reference->GetOp() ) {
    case eARRAY: {
        vector<CNode*> indiceList;
        EList2VectorExclude( reference->Arg<CNode*>(1), 
					set<NodeOp_t>(), indiceList );
	if( deleteSelect ) {
	    indiceList.pop_back();
	}
	indiceList.insert( indiceList.end(), 
			appendList.begin(), appendList.end() );
	
	vector<CNode*>::iterator ptr;
	for( ptr = indiceList.begin(); ptr != indiceList.end(); ++ptr ) {
	    if( indiceMap.count(*ptr) ) {
		*ptr = cVAR_REF(indiceMap[*ptr]);
	    }
	}
	CNode* indices = Vector2EList( indiceList );
	newRef = cARRAY( CloneReferenceWithSubstitution(
		reference->Arg<CNode*>(0), indiceMap, FALSE, vector<CNode*>() ), indices );
	} break;
    case eMEMBER:
	newRef = cMEMBER( CloneReferenceWithSubstitution(
		reference->Arg<CNode*>(0), indiceMap, FALSE, vector<CNode*>() ),
			  reference->Arg<CSymbol*>(1) );
	if( appendList.size() ) {
	    CNode* indices = Vector2EList( appendList );
	    newRef = cARRAY( newRef, indices );
	}
	break;
    default:
	newRef = reference->Clone();
	if( appendList.size() ) {
	    CNode* indices = Vector2EList( appendList );
	    newRef = cARRAY( newRef, indices );
	}
	break;
    }
    return newRef;
}

/*******************************************************
	XPropCaseItem
	- add x propagation logic 
********************************************************/

static CNode* XPropCaseItem( CBlock* module, CBlock* block, CNode* n ) 
{
        if( n->GetOp() == eCOMMENT ||
	    n->GetOp() == eVRQ ||
	    n->GetOp() == ePRAGMA ) {
                return n;
        }
	CNode* orig = n;
        while( n->GetOp() == eLIST ) {
            n->Arg<CNode*>(0) = XPropCaseItem( module, block, 
						n->Arg<CNode*>(0) ); 
		
	    if( n->Arg<CNode*>(1)->GetOp() != eLIST ) {
                n->Arg<CNode*>(1) = 
			XPropCaseItem( module, block, n->Arg<CNode*>(1) );
                return orig;
	    }
	    n = n->Arg<CNode*>(1);
        }

        n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
	return n;

}
/*******************************************************
	_CollectIndex
	- collect index expression from bit selects
	  in vector 
********************************************************/

struct IndexInfo {
    vector<CNode*> indices;    // list of indexes
    CNode*         reference;  // lval reference
    CNode*	   declRef;    // decl reference
};

//#define WORKING
#ifdef WORKING
static int _CollectIndex( CNode* n, void* arg )
{
	vector<IndexInfo>* list = (vector<IndexInfo>*)arg;
	switch( n->GetOp() ) {
	case eVAR_REF:
	    {
		if( n->Arg<CVar*>(0)->GetNodeType() != eS &&
		    n->Arg<CVar*>(0)->GetNodeType() != eB ) {
		    return TRUE;
		}
		IndexInfo p;
		p.reference = n;
		list->push_back( p );
	    }
	    break;
	// ??? mdh - structs not supported
	case eMEMBER:
	    return TRUE;
	case eARRAY:
	    // ??? mdh - structs not supported
	    if( n->Arg<CNode*>(0)->GetOp() != eVAR_REF ) { 
		return TRUE;
	    }
	    {
		IndexInfo p;
		CDecl* decl = n->Arg<CNode*>(0)->Arg<CDecl*>(0);
		p.reference = n;
		set<NodeOp_t> excludeOps;
		excludeOps.insert( eCOMMENT );
		excludeOps.insert( eVRQ );
		excludeOps.insert( ePRAGMA );
		EList2VectorExclude( n->Arg<CNode*>(1), excludeOps, p.indices );
		CNode* varRef = GetReferenceVarRef( n );
		// ??? - mdh temporary to disable xprop on complex data types
		switch( varRef->Arg<CVar*>(0)->GetDataType()->GetVarDataType() ) {
		case eSTRUCT:
		case eDEVENT:
		case eUNION:
		case eUNION_TAGGED:
		case eENUM:
		case eCLASS:
		case ePROXY:	
		    return TRUE;
		}
		list->push_back( p );
	    }
	    return FALSE;
	default: ;
	}
	return TRUE;
}
#else
static void _CollectIndex1( CNode* n, IndexInfo& info )
{
	switch( n->GetOp() ) {
	case eEXTERNAL_REF:
	// ??? mdh - currently these routines are only called for
	// procedural assignments and NET_REF should never happen
	case eNET_REF:
	case eVAR_REF:
	    info.declRef = n;
	    break;
	case eMEMBER:
	   _CollectIndex1( n->Arg<CNode*>(0), info ); 
	   break;
	case eARRAY: {
		set<NodeOp_t> excludeOps;
		excludeOps.insert( eCOMMENT );
		excludeOps.insert( eVRQ );
		excludeOps.insert( ePRAGMA );
		EList2VectorExclude( n->Arg<CNode*>(1), excludeOps, info.indices );
	        _CollectIndex1( n->Arg<CNode*>(0), info ); 
	    } break;
	default: 
	    MASSERT( FALSE );
	    break;
	}
}
static int _CollectIndex( CNode* n, void* arg )
{
	vector<IndexInfo>* list = (vector<IndexInfo>*)arg;
	switch( n->GetOp() ) {
	// ??? mdh - currently these routines are only called for
	// procedural assignments and NET_REF should never happen
	case eNET_REF:
	case eVAR_REF:
	case eMEMBER:
	case eARRAY: {
	    IndexInfo info;
	    info.reference = n;
	    _CollectIndex1( n, info );
	    if( n->GetNodeType() == eR ) {
		return FALSE;
	    }
	    if( info.declRef->GetOp() != eVAR_REF &&
	        info.declRef->GetOp() != eNET_REF ) {
		return FALSE;
	    }
	    // ??? - mdh temporary to disable xprop on complex data types
	    switch( info.declRef->Arg<CVar*>(0)->GetDataType()->GetVarDataType() ) {
	    case eDEVENT:
	    case eCLASS:
//	    case eUNION:
//	    case eUNION_TAGGED:
//	    case eENUM:
//	    case ePROXY:	
//	    case eSTRUCT:
	        return FALSE;
	    }
	    list->push_back( info );
	    } return FALSE;
	default: ;
	}
	return TRUE;
}
#endif

/**
 * Collect information about the indices of the given reference
 * \param ref reference to anaylse
 * \param indexList vector to return result within
 */
static void CollectIndices( CNode* ref, vector<IndexInfo>& indexList )
{
    ref->PreVisit1( _CollectIndex, &indexList );
}

/*******************************************************
	IsNonX
	- return TRUE if expression cannot resolve to X
********************************************************/
static int IsNonX( CNode* n )
{
    switch( n->GetOp() ) {
    case eLIST:
	return IsNonX( n->Arg<CNode*>(0) ) && IsNonX( n->Arg<CNode*>(1) );
    default:
	break;
    }
    return n->IsNonX(integerIsNoX,noxAttribute) || n->GetNodeType() == eR;
}
/*******************************************************
	_CollectRHD
	- collect right hand directions
********************************************************/

static TranslateState lhsTranslateState;

static int _CollectRHD( CNode* n, void* arg )
{
	list<CDecl*>* l = (list<CDecl*>*) arg;

	switch( n->GetOp() ) {
	case eVAR_REF:
	case ePARAM_REF:
	case eNET_REF:
	case ePORT_REF:
	case eEXTERNAL_REF:
	case eGENVAR_REF:
		l->push_back(n->Arg<CDecl*>(0) );
		break;
	default:
		return TRUE;
	}
	return FALSE;
}

/*******************************************************
	_CollectLHS
	- collect lhs of statements 
********************************************************/

static int _CollectLHS( CNode* n, void* arg )
{
	vector<LHS>* lhsList = (vector<LHS>*) arg;
	switch( n->GetOp() ) {
	case ePRAGMA:
		if( IsTranslateOn(n) ) {
			lhsTranslateState = eTRANSLATE_ON;
		} else if( IsTranslateOff(n) ) {
			if( lhsTranslateState == eTRANSLATE_OFF ) {
				message( n->GetCoord(), mNSTOS );
			}
			lhsTranslateState = eTRANSLATE_OFF;
		}
		break;
	case eADD_ASSIGN:
        case eSUB_ASSIGN:
        case eMUL_ASSIGN:
        case eDIV_ASSIGN:
        case eMOD_ASSIGN:
        case eAND_ASSIGN:
        case eOR_ASSIGN:
        case eXOR_ASSIGN:
        case eLSH_ASSIGN:
        case eRSH_ASSIGN:
        case eLSHA_ASSIGN:
        case eRSHA_ASSIGN:		
	case eASSIGN:
	case eNBASSIGN:
		{
		    if( lhsTranslateState != eTRANSLATE_OFF ) {
			if( n->GetOp() != eNBASSIGN && 
			    n->Arg<CNode*>(0) ) {
			    message( n->GetCoord(), mXBADE );
			} else {
			    CNode* l = n->Arg<CNode*>(0);
			    CNode* rl = n->Arg<CNode*>(1);
		            vector<IndexInfo> indexList;
		            CollectIndices( n->Arg<CNode*>(1), indexList );
			    vector<IndexInfo>::iterator ptr;
			    for( ptr = indexList.begin(); ptr != indexList.end(); ++ptr ) {
				LHS lhs(
				    ptr->reference,
				    ptr->indices,
				    l,
				    n->GetOp(),
				    n->GetCoord() );
			        lhsList->push_back( lhs );
			    }
			}
		    }
		}
		return FALSE;
	default:
		return TRUE;
	}
}
/**
 * Collect info about all LHS (left hand sides) in given code tree
 * \param tree code tree to analyse
 * \param lhsList vector of LHS to populate
 */
static void CollectLHS( CNode* tree, vector<LHS>& lhsList )
{
    if( tree != NULL ) {
        tree->PreVisit1( _CollectLHS, &lhsList );
    }
}
/*******************************************************
	_CountEdges
	- count the number of edge constructs in expression
********************************************************/

static int _CountEdges( CNode* n, void* arg )
{
    int* count = (int*)arg;

    switch( n->GetOp() ) {
    case ePOSEDGE:
    case eNEGEDGE:
    case eEDGE:
	(*count)++;
	return FALSE;
    }
    return TRUE;
}
/*******************************************************
	_CountEventTerms
	- count the number of event terms in expression
********************************************************/

static int _CountEventTerms( CNode* n, void* arg )
{
    int* count = (int*)arg;

    switch( n->GetOp() ) {
    case eEVOR:
	break;
    default:
	(*count)++;
	return FALSE;
    }
    return TRUE;
}

/*******************************************************
	FindParentOfDefaultCase
	- find default CASEITEM
	  if found return pointer to point to node,
	  else return NULL
********************************************************/
static CNode** FindParentOfDefaultCase( CNode** parent )
{
    CNode** result = NULL;
    CNode* n = *parent;
    if( !n ) {
	return NULL;
    }
    switch( n->GetOp() ) {
    case eCASEITEM:
	if( n->Arg<CNode*>(0) == NULL ) {
	    return parent;
	}
	break;
    case eLIST:
	result = FindParentOfDefaultCase( n->Arg<CNode*>(0).Ptr() );
	if( result ) {
	    break;
	}
	result = FindParentOfDefaultCase( n->Arg<CNode*>(1).Ptr() );
	break;
    }	
    return result;
}

/*******************************************************
	_CollectCases
	- collect expression from case item in case statements
********************************************************/

static int _CollectCases( CNode* n, void* arg )
{
	list<CNode*>& caseList = *(list<CNode*>*)arg;
	switch( n->GetOp() ) {
	case eCASEITEM:
	    if( n->Arg<CNode*>(0) && 
	        !IsNonX(n->Arg<CNode*>(0)) && 
	        !n->Arg<CNode*>(0)->IsConstant() ) {
	        caseList.push_back( n->Arg<CNode*>(0) );
	    }
	    return FALSE;
	default:
	    return TRUE;
	}
}
/*******************************************************
  	InScope
	- return 1 if decl is in symtab
 *******************************************************/
static int InScope( CSymtab<CDecl>& symtab, CDecl* d ) 
{
    CDecl* ld = symtab.Lookup( d->GetSymbol() );
    return ld && ld == d;

}

/*******************************************************
  	ExpressionInScope
	- return 1 if all decls in expression are in scope
 *******************************************************/
static int ExpressionInScope( CSymtab<CDecl>& symtab, CNode* n ) 
{
    list<CDecl*> RHD;
    n->PreVisit1( _CollectRHD, &RHD );		
    list<CDecl*>::iterator dptr;
    int inScope = 1;
    for( dptr = RHD.begin(); dptr != RHD.end(); ++dptr ) {
	if( !InScope( symtab, *dptr ) ) {
	    return 0;	
	}
    }
    return 1;
}
/*******************************************************
	EList2Cat
	- create cat of from list of expression
********************************************************/
static CNode* EList2Cat( CNode* n )
{
	if( !n ) {
	    return NULL;
	}
	if( n->GetOp() != eELIST ) {
	    return n->Clone();
	}
	CNode* result = NULL;
	CNode* last   = NULL;
	while(1) {
	    /*
	     * Note: n is put in to the right fork to keep
	     *       _Cat func from creating a eUCAT. It
	     *       will be overridden
	     */
	    CNode* next = cCAT( EList2Cat(n->Arg<CNode*>(0)), n );
	    if( !result ) {
		result = next;
	    }
	    if( last ) {
		last->Arg<CNode*>(1) = next;
	    }
	    n = n->Arg<CNode*>(1);
	    if( n && n->GetOp() != eELIST ) {
	        next->Arg<CNode*>(1) = EList2Cat( n );
	        break;
	    }
	    last = next;
	}
	return result;
}


/**
 * Create a loop variable for using in looping constructs
 * \param module module in which to create var
 * \param stmts stmt tree into which to link var declaration
 * \return loop variable
 */
static CVar* CreateLoopVar( CBlock* module, CNode*& stmts )
{
    CVar* loopVar;
    Coord_t coord = { 0, "???" };
    CSimpleType* stype = new(CNode::CurrentHeap()) 
		CSimpleType(eINTEGER);
    loopVar = new(CNode::CurrentHeap()) CVar( 
    		CSymbol::GenSymbol("___index"), 
    		&coord, stype, FALSE );
    loopVar->SetDeclStatementCreated();
    stmts = cLINK( stmts, cVAR_DECL(loopVar, NULL) );
    module->Add( loopVar );
    return loopVar;
}

/*
 * Create for loop wrapping statement using given index and range.
 * \param loopVar loop variable to use in for loop.
 * \param rangeSpec range or slices node to define loop extent.
 * \param body loop body.
 * \return for loop code
 */
CNode* CreateForLoop( CVar* loopVar, CNode* rangeSpec, CNode* body )
{
	// create msb/lsb for for loop range
        while( rangeSpec && rangeSpec->GetOp() == eMACRO_EXPR ) {
            rangeSpec = rangeSpec->Arg<CNode*>(1);
        }
	CNode* msb = rangeSpec->Arg<CNode*>(0)->Clone(); 
	CNode* lsb = rangeSpec->Arg<CNode*>(1) ?
	    	     rangeSpec->Arg<CNode*>(1)->Clone() : NULL;
	if( !lsb ) {
	    msb = cSUB(msb, cINT32(1)); 
	    lsb = cINT32(0);
	}

    	Coord_t coord = { 0, "???" };
    	CNode* lb =  cHOOK( cGE(msb, lsb), lsb->Clone(), 
			    msb->Clone(), &coord );
    	CNode* mb = cHOOK( cGE(msb->Clone(), lsb->Clone()), 
			   msb->Clone(), lsb->Clone(), &coord);

        return cFOR( 
             cASSIGN( NULL, cVAR_REF(loopVar), lb ),
        	    cLE( cVAR_REF(loopVar), mb ),
        	    cASSIGN( NULL, cVAR_REF(loopVar),
        		cADD(cVAR_REF(loopVar),cINT32(1)) 
        	    ),
        	    body 
        	);
}

/**
 * Create a code fragment to assign X's to given reference.
 * The decision to x out an index is made in the following priority order:
 * - Constants indices are not x'd out.
 * - NonX indices are not x'd out.
 * - Indices in xOutIndices are x'd out.
 * - Indices in noXOutIndices are not x'd out.
 * \param module module to contain assignment
 * \param block immediate block to contain assignment
 * \param op assignment node type
 * \param delay delay assignment expression
 * \param reference reference expression to assign to X.
 * \param indices list of all reference indices
 * \param conditional if TRUE create conditional assignment
 * \param xOutIndices set containing indices that should be x'd out
 * \param noXOutIndices set containing indices that should not be x'd out
 * \return code tree with X assignments
 */ 
static CNode* CreateXAssignment( 
	    CBlock* module, 
	    CBlock* block, 
	    NodeOp_t op, 
	    CNode* delay, 
	    CNode* reference,
	    vector<CNode*>& indices,
	    int conditional,
	    const set<CNode*>& xOutIndices,
	    const set<CNode*>& noXOutIndices )
{
    CNode* declStmts = NULL;
    /*
     * analyse reference to assign to x
     */
    CDataType* dataType;
    vector<CNode*> unpackedDimensions;
    vector<CNode*> packedDimensions;
    vector<CNode*> residualPackedIndices;
    int result = ResolveReference( reference, dataType,
       		unpackedDimensions, packedDimensions, 
       		residualPackedIndices );
    MASSERT( result == 1 );
    while( dataType->GetVarDataType() == ePROXY ) {
	dataType = dynamic_cast<CProxyType*>(dataType)->GetDataType();
    }
    VarDataType_t vdt = dataType->GetVarDataType();

    /*
     * if unpacked structure, create and X assignment for
     * each member, taking into account unresolved unpacked 
     * indices.
     */
    if( !dataType->IsPacked() && 
	(vdt == eSTRUCT || vdt == eUNION || vdt == eUNION_TAGGED) ) {
	CNode* members = dynamic_cast<CStructType*>(dataType)->GetMembers();
	vector<CNode*> memberList;
	List2VectorExclude( members, set<NodeOp_t>(), memberList );
	vector<CNode*>::iterator ptr;
	CNode* s = NULL;
	for( ptr = memberList.begin(); ptr != memberList.end(); ++ptr ) {
	    MASSERT( (*ptr)->GetOp() == eVAR_DECL );
	    CVar* var = (*ptr)->Arg<CVar*>(0);
	    vector<CNode*> newIndices = indices; 
	    set<CNode*> newNoXOutIndices = noXOutIndices;
	    CNode* newRef = reference;
	    vector<CNode*> prependIndices;
	    map<CNode*,CNode*> prependIndice2Range;
	    CNode* firstDimension = NULL;
	    if( unpackedDimensions.size() ) {
		if( reference->GetOp() == eARRAY ) {
		    // ??? mdh - note this is overly conservative
		    //           if select is SLICE range could be
		    //           restricted
		    EList2VectorExclude( reference->Arg<CNode*>(1), 
				set<NodeOp_t>(), prependIndices );
		    newRef = newRef->Arg<CNode*>(0);	
		    CNode* lastIndex = prependIndices.back();
		    while( lastIndex->GetOp() == eMACRO_EXPR ) {
			lastIndex = lastIndex->Arg<CNode*>(1);
		    }
		    switch( lastIndex->GetOp() ) {
		    case eSLICE:
		    case ePSLICE:
		    case eMSLICE:
			firstDimension = prependIndices.back();
			prependIndices.pop_back();
			break;
		    }
		}
		for( int i = 0; i < unpackedDimensions.size(); ++i ) {
		    CVar* loopVar = CreateLoopVar( module, declStmts );
		    CNode* loopVarRef = cVAR_REF(loopVar);
		    newIndices.push_back( loopVarRef );
		    if( firstDimension ) {
		        prependIndice2Range[loopVarRef] = firstDimension;
			firstDimension = NULL;
		    } else {
		        prependIndice2Range[loopVarRef] = unpackedDimensions[i];
		    }
		    newNoXOutIndices.insert( loopVarRef );
		    prependIndices.push_back( loopVarRef );
		}
	        newRef = cARRAY( newRef,
				Vector2EList( prependIndices ),
				reference->GetCoord() );
	    }
	    newRef = cMEMBER( newRef, var->GetSymbol(), 
				     reference->GetCoord() );
	    CNode* n = CreateXAssignment( module, block, op, 
	    			   delay, newRef, newIndices,
	    			   conditional,xOutIndices,
	    			   newNoXOutIndices );
	    if( n ) {
		vector<CNode*>::iterator ptr;
		for( ptr = prependIndices.begin();
		     ptr != prependIndices.end(); ++ptr ) {
		     if( prependIndice2Range.count(*ptr ) ) {
			 CVar* loopVar = (*ptr)->Arg<CVar*>(0);
		         CNode* range = prependIndice2Range[*ptr];
		         n = CreateForLoop( loopVar, range, n );
		     }
		}
	    }
	    s = cLINK(s,n);
	}
	return s;
    }

    /*
     * reals LHS's do not require x prop
     */
    if( dataType->GetNodeType() == eR ) {
	return NULL;
    }

    /*
     * attempts to xprop unpacked LHS's is not supported
     */
    if( !dataType->IsPacked() ) { 
	// xprop on this variable reference is not supported;
        message( reference->GetCoord(), mXVREF );
        return NULL;
    }

    /*
     * if integers are declared to non-x, integer LHS's do not
     * require x prop.
     */
    if( integerIsNoX && 
        dataType->GetVarDataType() == eINTEGER ) {
	return NULL;
    }
     
    /*
     * resolve whether to x out each indices
     */
    set<CNode*> resolvedXIndices = xOutIndices;
    vector<CNode*>::iterator ptr;
    for( ptr = indices.begin(); ptr != indices.end(); ++ptr ) {
       if( !noXOutIndices.count(*ptr) && 
	   !(*ptr)->IsConstant() && 
	   !IsNonX(*ptr) ) {
           resolvedXIndices.insert(*ptr);
       }
    }

    /*
     * Part and bit select require special attention.
     * Precalculate select related values.
     */
    CNode* xExp; 
    CNode* select         = NULL;
    int    resolvedXSelect = FALSE;
    int    hasPartSelect   = FALSE;
    int    psType          = 0;
    if( residualPackedIndices.size() > 0 ) {
       select = residualPackedIndices.back();
       if( resolvedXIndices.count(select) ) {
           resolvedXSelect = TRUE;
       }
    }
    if( select ) {
       CNode* rselect = select;
       while( rselect->GetOp() == eMACRO_EXPR ) {
           rselect = rselect->Arg<CNode*>(1);
       }
       switch( rselect->GetOp() ) {
       case eSLICE:
           psType = 0;
           hasPartSelect = TRUE;
           break;
       case ePSLICE:
           psType = 1;
           hasPartSelect = TRUE;
           break;
       case eMSLICE:
           psType = -1;
           hasPartSelect = TRUE;
           break;
       }
    }


    /*
     * Calculate assignment X RHS.
     * Width is dependent on unspecified packed dimensions,
     * base data type and part select, if present.
     */
    int rpi = residualPackedIndices.size() - (resolvedXSelect ? 1 : 0);
    int widthEval = !dataType->IsBaseWidthVolatile() && 
       	            dataType->IsBaseWidthEvaluateable();
    if( !resolvedXSelect && hasPartSelect ) {
        widthEval &= !select->IsWidthVolatile() &&
       	             select->IsWidthEvaluateable();
    }
    for( int i = rpi; i < packedDimensions.size(); ++i ) {
        widthEval &= !packedDimensions[i]->IsWidthVolatile() &&
       	             packedDimensions[i]->IsWidthEvaluateable();
    }

    if( widthEval ) {
       int width = dataType->GetBaseWidth();
       if( !resolvedXSelect && hasPartSelect ) {
           width *= select->GetWidth();
       }
       for( int i = rpi; i < packedDimensions.size(); ++i ) {
           width *= packedDimensions[i]->GetWidth();
       }
       CVector xw(width);
       xw.LoadHex( "x" );
       xw.SetPreferredBase( 16 );
       xExp = cVECTOR(xw);
     } else {
       CNode* widthExp = NULL;
	/*
 	 * eliminate symbolic multiplies by 1
 	 */
       if( dataType->IsBaseWidthVolatile() || 
           !dataType->IsBaseWidthEvaluateable() || 
           dataType->GetBaseWidth() != 1 ) {
           widthExp = dataType->GetBaseWidthExp();
       }

       if( !resolvedXSelect && hasPartSelect ) {
	   /*
 	    * eliminate symbolic multiplies by 1
 	    */
           if( !select->IsWidthVolatile() && 
	       select->IsWidthEvaluateable() && 
       	       select->GetWidth() == 1 ) {
	       // do nothing
           } else if( widthExp ) {
               widthExp = cMUL(widthExp, select->GetWidthExp() );
           } else {
               widthExp = select->GetWidthExp();
           }
       }
       for( int i = rpi; i < packedDimensions.size(); ++i ) {
	   /*
 	    * eliminate symbolic multiplies by 1
 	    */
           if( !packedDimensions[i]->IsWidthVolatile() && 
       	       packedDimensions[i]->IsWidthEvaluateable() &&
       	       packedDimensions[i]->GetWidth() == 1 ) {
           } else if( widthExp ) {
               widthExp = cMUL( widthExp, packedDimensions[i]->GetWidthExp() );
           } else {
               widthExp = packedDimensions[i]->GetWidthExp();
           }
       }
       if( !widthExp ) {
           widthExp = cINT32(1);
       }
       CVector x1(1);
       x1.LoadHex( "x" );
       x1.SetPreferredBase( 16 );
       xExp= cREP( widthExp,cVECTOR(x1) );
     }

    /*
     * map each index in the given reference to the corresponding
     * dimensions range specification.
     */
    map<CNode*,CNode*> indices2Range; 
    CalculateReferenceRanges( reference, indices2Range );

    vector<CNode*> lb;
    vector<CNode*> mb;
    vector<CVar*> loopVarList;
    map<CNode*,CVar*> index2LoopVar;
    map<CVar*,CNode*> loopVar2Range;
    /*
     * iterate through each index, creating for loop if needed for each
     * x'd out dimensions.
     */
    map<CNode*,CNode*>::iterator rptr;
    for( rptr = indices2Range.begin(); rptr != indices2Range.end(); ++rptr ) {
        if( resolvedXSelect && rptr->first == select ) { 
           continue;
        }
        if( resolvedXIndices.count(rptr->first) ) {
            CNode* rangeSpec = rptr->second;
	    CNode* resolvedRangeSpec = rangeSpec;
            while( resolvedRangeSpec && 
		   resolvedRangeSpec->GetOp() == eMACRO_EXPR ) {
                resolvedRangeSpec = resolvedRangeSpec->Arg<CNode*>(1);
            }
 	    if( !resolvedRangeSpec ) {
	        // xprop on this variable reference is not supported;
                message( reference->GetCoord(), mXVREF );
	        return NULL;
	    }

	    CVar* loopVar = CreateLoopVar( module, declStmts );
	    loopVarList.push_back( loopVar );
	    loopVar2Range[loopVar] = rptr->second;
            index2LoopVar[rptr->first] = loopVar;
        }
    }
    /*
     * handle unspecified unpacked dimensions
     */
    int xOutSlice = 0;
    vector<CNode*> appendList;
    for( int i = 0; i < unpackedDimensions.size(); ++i ) {
            CNode* rangeSpec = unpackedDimensions[i];
	    int skip = 0;
            for( rptr = indices2Range.begin(); 
		 rptr != indices2Range.end(); ++rptr ) {
		if( rptr->second == unpackedDimensions[i] ) {
		    CNode* coveredIndex = rptr->first;
                    while( coveredIndex->GetOp() == eMACRO_EXPR ) {
                        coveredIndex = coveredIndex->Arg<CNode*>(1);
                    }
	            switch( coveredIndex->GetOp() ) {
	            case eSLICE:
		        rangeSpec = coveredIndex;
	                xOutSlice = 1;
	                break;
	            case ePSLICE:
	            case eMSLICE:
	                xOutSlice = 1;
	                break;
		    default:
			skip = 1;
	            }
		    break;
		}
	    }
	    if( skip ) {
	        continue;
	    }

	    CNode* resolvedRangeSpec = rangeSpec;
            while( resolvedRangeSpec && 
		   resolvedRangeSpec->GetOp() == eMACRO_EXPR ) {
                resolvedRangeSpec = resolvedRangeSpec->Arg<CNode*>(1);
            }
 	    if( !resolvedRangeSpec ) {
	        // xprop on this variable reference is not supported;
                message( reference->GetCoord(), mXVREF );
	        return NULL;
	    }

	    CVar* loopVar = CreateLoopVar( module, declStmts );
	    loopVarList.push_back( loopVar );
	    loopVar2Range[loopVar] = rangeSpec;
            appendList.push_back(cVAR_REF(loopVar));
    }
     /*
      * create new reference subsituting new indices
      */
    CNode* newReference = CloneReferenceWithSubstitution(
       				reference, 
				index2LoopVar, 
				resolvedXSelect|xOutSlice, 
				appendList );
     /*
      * create assignment statement
      */
    CNode* s;
    switch( op ) {
    case eADD_ASSIGN:
    case eSUB_ASSIGN:
    case eMUL_ASSIGN:
    case eDIV_ASSIGN:
    case eMOD_ASSIGN:
    case eAND_ASSIGN:
    case eOR_ASSIGN:
    case eXOR_ASSIGN:
    case eLSH_ASSIGN:
    case eRSH_ASSIGN:
    case eLSHA_ASSIGN:
    case eRSHA_ASSIGN:		
    case eASSIGN:
        s = cASSIGN( Clone(delay), newReference->Clone(), xExp );
	break;
    default:
        s = cNBASSIGN( Clone(delay), newReference->Clone(), xExp );
	break;
    }

//static CNode* CreateForLoop( CNode* loopVar, CNode* range );
    /*
     * create forloops for each loop variable.
     */
    for( int i = 0; i < loopVarList.size(); ++i ) {
	CVar* loopVar = loopVarList[i];
	CNode* rangeSpec = loopVar2Range[loopVar];
	s = CreateForLoop( loopVar, rangeSpec, s );

    }



    /*
     * if conditional x prop was specified, create an condition 
     * expression, consisting of indices, to check for X's.
     */
    if( conditional ) {
        CNode* cond = NULL;
        int limit = indices.size() - (hasPartSelect ? 1 : 0);
        
	/*
 	 * build conditional expression, skipping:
 	 * - part select
 	 * - constant expressions
 	 * - expressions that cannot be x
 	 * - indices that are explicitly either x'd out or not x'd out
 	 */
        for( int i = 0; i < limit; ++i ) {
            if( !indices[i]->IsConstant() && 
                !xOutIndices.count(indices[i]) && 
                !noXOutIndices.count(indices[i]) && 
        	!IsNonX(indices[i]) ) {
                cond = cond ?  cCAT( cond, indices[i]->Clone() ) : 
        	               indices[i]->Clone();
            }
        }
       if( hasPartSelect && psType != 0 ) { 
           CNode* msb = select->Arg<CNode*>(0);
           if( !msb->IsConstant() && 
       	       !xOutIndices.count(select) &&
	       !IsNonX(msb) ) {
               cond = cond ?  cCAT( cond, msb->Clone() ) : msb->Clone();
           }
       }
       /*
 	* create conditional expression wrapping less pessimistic
 	* x prop expression.
 	*/
       if( cond ) {
           CVector zero(1);
           zero.LoadBinary( "0" );
           CVector one(1);
           one.LoadBinary( "1" );
           CNode* alt = NULL;
           CNode* indexes1 = NULL;
           if( cond->IsWidthVolatile() || 
       	       !cond->IsWidthEvaluateable() || cond->GetWidth() != 1 ) {
       	       cond = cRXOR(cond);
           }
	   /*
 	    * conditional assignment should not x out indices as they
 	    * are checked for X's in the wrapping case statement.
 	    */
	   set<CNode*> noXSet;
	   noXSet.insert( indices.begin(), indices.end() );
           alt = CreateXAssignment( module, block, op, delay, 
       		reference, indices,  FALSE, 
       	    	xOutIndices, noXSet );

	   /*
 	    * wrap assignment in case statement to check for no X indices.
 	    */
           s = cCASE( cond, cLIST(
       	          cCASEITEM( cELIST(cVECTOR(zero),cVECTOR(one)), alt ),
       	          cCASEITEM( NULL, s ) ) );
       }
    }

    /*
     * add block to contain declaration statements for 
     * forloop indices if needed.
     */
    if( declStmts ) {
    	Coord_t coord = { 0, "???" };
        CBlock* block = new(CNode::CurrentHeap()) CBlock( 
    		CSymbol::GenSymbol("___block"), &coord );
        module->Add( block );
        block->SetParent( module );
        s = cBLOCK_REF( block, cLINK( declStmts, s) );
    }

    return s;
} 
     
/**
 * Create X assignment for LHS in list.
 * \param module module containing assigments
 * \param block block containing assignements
 * \return tree containing code to assign X's.
 */
static CNode* CreateLHSAssignments( CBlock* module, CBlock* block, 
				  vector<LHS>& lhsList )
{
    /*
     * check each LHS to see if its indices or selects are dependent
     * upon other LHS in the block or out of scope variabls. If so, 
     * be pessimistic and X the whole structure.
     */
    // ??? mdh - need to make this less pessimistic. 
    // 1. do a deeper analyse to eliminate false temporal dependencies.
    // 2. do a deeper analyse to eliminate false subrange dependencies.
    set<CDecl*> lhsSet;
    vector<LHS>::iterator ptr;
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {
	CDecl* decl = ptr->varRef->Arg<CDecl*>(0);
	lhsSet.insert( decl );	
    }
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {
	for( int i = 0; i < ptr->indices.size(); ++i ) {
            list<CDecl*> RHD;
	    ptr->indices[i]->PreVisit1( _CollectRHD, &RHD );
	    list<CDecl*>::iterator rptr;
	    for( rptr = RHD.begin(); rptr != RHD.end(); ++rptr ) {
	        if( lhsSet.find(*rptr) != lhsSet.end() ) {
		    ptr->xOutIndices.insert(ptr->indices[i]);
		    break;
	        }
	    } 
	}
	for( int i = 0; i < ptr->indices.size(); ++i ) {
	    if( !ExpressionInScope( block->GetSymtab(), ptr->indices[i] ) ) {
		ptr->xOutIndices.insert(ptr->indices[i]);
	    }
	}
    }

    /*
     * Fill in covering (vector range) info for each entry
     */
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {
	/*
 	 * initialized default covering info
 	 */
	ptr->fullyCovered      = 0;
	ptr->coveringInfoValid = 0;
	ptr->drop              = 0;

	/*
 	 * non part and bit references are fully covered
 	 */
	if( !ptr->hasBitSelect && !ptr->hasPartSelect ) {
	    ptr->fullyCovered      = 1;
	    ptr->coveringInfoValid = 1;
	    continue;
	}	
	/*
 	 * add covering info for bit vectors with non-volatile selects
 	 */
	if( ptr->hasBitSelect && !ptr->select->IsVolatile() &&
	    ptr->select->IsEvaluateable() ) {
	    int lsb = ptr->select->EvalINT32();
	    ptr->coveredBitVector.insert( lsb );
	    ptr->coveringInfoValid = 1;
	    continue;
	}

	/*
 	 * non const bit selects in loops are fully covered.
 	 * Note: this is because a range analysis would require
 	 * the for loop to be unrolled and the include procedural code
 	 * to be executed.
 	 */
	if( ptr->hasBitSelect && !ptr->select->IsConstant() &&
	    ptr->xOutIndices.count(ptr->select) ) {
	    ptr->fullyCovered      = 1;
	    ptr->coveringInfoValid = 1;
	}

	/*
 	 * add covering info for part selects
 	 */
	if( ptr->hasPartSelect && ptr->psType == 0 &&
	    !ptr->lsb->IsVolatile() && !ptr->msb->IsVolatile() &&
	    ptr->lsb->IsEvaluateable() && ptr->msb->IsEvaluateable() ) {
	    int lsb = ptr->lsb->EvalINT32();
	    int msb = ptr->msb->EvalINT32();
	    if( msb < lsb ) {
		int tmp = lsb;
		lsb = msb;
		msb = tmp;
	    }
	    for( int i = lsb; i <= msb; ++i ) {
	       ptr->coveredBitVector.insert( i );
	    }
	    ptr->coveringInfoValid = 1;
	    continue;
	}

	/*
 	 * add covering info for range selects
 	 */
	if( ptr->hasPartSelect && ptr->psType != 0 &&
	    (!ptr->lsb->IsVolatile() && !ptr->msb->IsVolatile() &&
	     ptr->lsb->IsEvaluateable() && ptr->msb->IsEvaluateable() ) ) {
	    int lsb = ptr->lsb->EvalINT32();
	    int msb = ptr->msb->EvalINT32();
	    CDecl* decl = ptr->varRef->Arg<CDecl*>(0);
	    if( ptr->psType > 0 ) {
	        int temp = msb+lsb-1;
	        lsb = msb;
	        msb = temp;
	    } else {
	        lsb = msb-lsb+1;
	    }
		
	    if( msb < lsb ) {
		int tmp = lsb;
		lsb = msb;
		msb = tmp;
	    }
	    for( int i = lsb; i <= msb; ++i ) {
	        ptr->coveredBitVector.insert( i );
	    }
	    ptr->coveringInfoValid = 1;
	    continue;
	}
    }


    /*
     * mark all duplicates that are subsets of 
     * other entries to be dropped.
     */
    vector<LHS>::reverse_iterator rptr;
    vector<LHS>::reverse_iterator rptr1;
    int index;
    for( index = 0, rptr = lhsList.rbegin(); rptr != lhsList.rend(); ++rptr, ++index ) {
	int index1;
        for( index1 = 0, rptr1 = lhsList.rbegin(); rptr1 != lhsList.rend(); ++rptr1, ++index1 ) {
	    /*
  	     * if this is a subset of a another entry drop it
  	     */
	    if( index != index1 && !rptr1->drop && rptr->IsSubsetOf( &*rptr1 ) ) {
		rptr->drop = 1;
		break;
	    }
	}
    }

    /*
     * process each lhs and build covering set of assignments.
     * Note the set will not be the minimal set, only a sufficient set.
     */
    LHSHash	lhsHash;
    vector<LHS> lhsKeep;
    for( ptr = lhsList.begin(); ptr != lhsList.end(); ++ptr ) {

	
	/*
 	 * if a subset of another entry, drop it
 	 */
	if( ptr->drop ) {
	    continue;
	}

// ??? - mdh this needs to be rethought
	if( disableIndex && ptr->isArray ) {
	    continue;
	}
	
	/*
 	 * drop out of scope assignments
 	 */
	if( !ExpressionInScope( block->GetSymtab(), ptr->varRef ) ) {
	    continue;
	}

	/*
         * Check to see if we've seen compatible a lhs before
         * If not allocate add this one to hash table and 
         * lhs list.
         */
	LHS* lhsPtr;
	lhsPtr = lhsHash.Find(*ptr);
	if( !lhsPtr ) {
	    lhsPtr = lhsHash.Insert(*ptr);
	    goto KEEP;
	}

	
	/*
  	 * if this is a subset of a previous entry,
  	 * drop it.
  	 */
	if( ptr->IsSubsetOf( lhsPtr ) ) {
            continue;	   
	}

	/*
 	 * merge range info with existing entry and add to list
 	 */
	lhsPtr->fullyCovered |= ptr->fullyCovered;
	lhsPtr->coveredBitVector.insert( ptr->coveredBitVector.begin(),
					 ptr->coveredBitVector.end() );

KEEP:
	lhsKeep.push_back( *ptr );
    }

    /*
     * Create an assignment statement for each LHS
     */
    CNode* t = NULL;
    for( ptr = lhsKeep.begin(); ptr != lhsKeep.end(); ++ptr ) {
        /* This is a conservative way to figure out if any indices was assigned
         * as part of a looping construct. To keep loop variables from being 
         * referenced in instrumentation outside the loop, make x assignment
         * unconditional.
         */
        int inForLoopHint = (ptr->xOutIndices.size()!=0);
	t = cLINK( CreateXAssignment( module, block, ptr->op, ptr->delay, 
		   ptr->reference, ptr->indices, !inForLoopHint, 
		   ptr->xOutIndices, set<CNode*>() ), t );	
    }

    return t;
}

/*******************************************************
 	XPropClockEvent
	- add xprop clock instrumentation to clock event 
 *******************************************************/
static void XPropClockEvent( CNode* event )
{
    /*
     * Add hidden attribute to create conditional code around
     * event expression to handle multi-edge intrumentation.
     */
    CNode* eventAttr = event->GetAttributes();
    if( !eventAttr ) {
        eventAttr = cATTRIBUTE(new(CNode::CurrentHeap()) 
    				CAttr(event->GetCoord()));
        event->SetAttributes( eventAttr );
    }
    eventAttr->Arg<CAttr*>(0)->AddAttribute( edgeControl, 
    				cSTRING(edgeAttrName) );
}

/*******************************************************
 	XPropClockBody
	- add xprop clock instrumentation to clock body
 *******************************************************/
static void XPropClockBody( CBlock* module, CBlock* block, 
			   NodeOp_t clockEdge,
			   CNode*   clockExp,
		           CNode_sp<CNode*> insertionPoint,
			   int dropTrailingComment )
{
    CNode* body = insertionPoint;
    /*
     * create xprop assignments for state elements
     */
    vector<LHS> lhsList;	
    lhsTranslateState = eTRANSLATE_ON;
    CollectLHS( body, lhsList );
    body = XPropStatement( module, block, body );
    CNode* xStmt = CreateLHSAssignments( module, block, lhsList );

    /*
     * create clock intrumentation
     */
    if( insertionClkBeginComment ) {
        CNode* b = (!body || body->GetOp() != eBLOCK_REF) ?
    			cBLOCK_REF( NULL, body) : body;	
        body = cLINK( b, insertionClkBeginComment->Clone() );
    }
    if( !dropTrailingComment && insertionClkEndComment ) {
        xStmt = cLINK( cBLOCK_REF( NULL, xStmt ), 
    			insertionClkEndComment->Clone() );
    }
    if( insertionClkEndComment ) {
        body = cLINK( insertionClkEndComment->Clone(), body );
    }
    if( clockEdge == ePOSEDGE ) {
        body = cIF( clockExp, body, cIF( cCOM(clockExp->Clone()), NULL, xStmt ) );
    } else {
        body = cIF( cCOM(clockExp->Clone()), body, cIF( clockExp, NULL, xStmt ) );
    }
    if( insertionClkBeginComment ) {
        body = cLINK( insertionClkBeginComment->Clone(), body );
    }
    insertionPoint = body;

}

/*******************************************************
 	XPropSimpleStateElement
	- xprop simple state element (clock, but no async controls).
	  module is block pointer to containing module.
	  block is block pointer to containing block.
	  event is event expression for element.
 *******************************************************/
static int XPropSimpleStateElement( CBlock* module, CBlock* block, CNode* event )
{
    MASSERT( event->GetOp() == eEVENT );
    /* 
     *  instrument simple clocking constructs 
     *  1) always @(posedge expression)
     *  2) always @(negedge expression)
     */
    CNode*   condition = event->Arg<CNode*>(0);
    CNode*   body = event->Arg<CNode*>(1);
    MASSERT( condition && (condition->GetOp() == ePOSEDGE ||
	                   condition->GetOp() == eNEGEDGE) );

    CNode_sp<CNode*>   insertionPoint = event->Arg<CNode*>(1);
    CNode*   clockExp = condition->Arg<CNode*>(0);
    NodeOp_t clockEdge = condition->GetOp();


    /*
     * transform:
     *    always @(POSEDGE exp)
     *	stmts
     * to:
     *    always @(POSEDGE exp)
     * 	if( expr ) 
     * 	    stmt
     *	else 
     *	    xprop'd LHS
     *
     * transform:
     *    always @(NEGEDGE exp)
     *	stmts
     * to:
     *    always @(NEGEDGE exp)
     * 	if( ~expr ) 
     * 	    stmt
     *	else 
     *	    xprop'd LHS
     *
     */

    XPropClockEvent( event );
    XPropClockBody( module, block, clockEdge, clockExp, 
		    insertionPoint, FALSE );
}

/*******************************************************
 	XPropStateElement
	- xprop state element. 
	  module is block pointer to containing module.
	  block is block pointer to containing block.
	  event is event expression for element.
	  return TRUE if element was sucessfully intrumented.
 *******************************************************/

static int _CollectPosedge( CNode* n, void* arg ) 
{
    set<CNode*>* s = static_cast<set<CNode*>*>(arg);
    switch( n->GetOp() ) {
    case ePOSEDGE:
	s->insert(n->Arg<CNode*>(0));
	break;
    default:
        return TRUE;
    }
}

static int _CollectNegedge( CNode* n, void* arg ) 
{
    set<CNode*>* s = static_cast<set<CNode*>*>(arg);
    switch( n->GetOp() ) {
    case eNEGEDGE:
	s->insert(cCOM(n->Arg<CNode*>(0)));
	break;
    default:
        return TRUE;
    }
}

static int _CollectIfExp( CNode* n, void* arg ) 
{
    set<CNode*>* s = static_cast<set<CNode*>*>(arg);
    switch( n->GetOp() ) {
    case eIF:
	s->insert(n->Arg<CNode*>(0));
	return FALSE;
    default:
        return TRUE;
    }
}
static void Elements2Vector(CNode_sp<CNode*> n, vector<CNode_sp<CNode*> >& v)
{
    if( static_cast<CNode*>(n) == NULL ) {
	return;
    }
    switch( n->GetOp() ) {
    case eLIST:
	Elements2Vector(n->Arg<CNode*>(0),v);
	Elements2Vector(n->Arg<CNode*>(1),v);
	break;
    case ePRAGMA:
    case eVRQ:
    case eCOMMENT:
	break;
    case eBLOCK_REF:
	Elements2Vector(n->Arg<CNode*>(1),v);
	break;
	break;
    default:
	v.push_back(n);
	break;
    }
}

static CNode* StripNot( CNode* n ) 
{
    MASSERT( n );
    switch( n->GetOp() ) {
    case eCOM:
	if( n->IsWidthEvaluateable() && n->GetWidth() == 1 ) {
	    return n->Arg<CNode*>(0);
	} else {
	    return n;
	}
	MASSERT( FALSE );
    case eNOT:
	return n->Arg<CNode*>(0);
    default: 
	return n; 
    }
}

static int XPropStateElement( CBlock* module, CBlock* block, CNode* event )
{
    CNode*   body = event->Arg<CNode*>(1);
    CNode*   exp = NULL;
    NodeOp_t edge = ePOSEDGE;
    CNode*   condition = event->Arg<CNode*>(0);

    /*
     * create list of all state element specifications
     */
     vector<CNode_sp<CNode*> > elements;
     Elements2Vector(event->Arg<CNode*>(1), elements);

    /*
     * inspect event expression
     */
    set<CNode*> posedgeList;
    set<CNode*> negedgeList;
    condition->PreVisit1( _CollectPosedge, &posedgeList );
    condition->PreVisit1( _CollectNegedge, &negedgeList );

    /*
     * validate each state element
     */
    vector<CNode_sp<CNode*> >::iterator ePtr;
    for( ePtr = elements.begin(); ePtr != elements.end(); ++ePtr ) {
	CNode_sp<CNode*> insertionPoint = *ePtr;
        set<CNode*> bothEdges;
	bothEdges.insert( posedgeList.begin(), posedgeList.end() );
	bothEdges.insert( negedgeList.begin(), negedgeList.end() );
    	set<CNode*> ifExpList;
	while( static_cast<CNode*>(insertionPoint) != NULL && 
	       insertionPoint->GetOp() == eIF ) {
	    CNode* cond = insertionPoint->Arg<CNode*>(0);
	    ifExpList.insert( cond );
            set<CNode*>::iterator ptr;
	    int found = 0;
	    for( ptr = posedgeList.begin(); ptr != posedgeList.end(); ++ptr ) {
		if( Equivalent( *ptr, cond ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    for( ptr = negedgeList.begin(); ptr != negedgeList.end(); ++ptr ) {
		if( Equivalent( StripNot(*ptr), StripNot(cond) ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    if( !found ) {
		break;
	    }
	    insertionPoint = insertionPoint->Arg<CNode*>(2);
	}
	if( bothEdges.size() != 1 ) {
	    return FALSE;
	}
    }

    /*
     * add instrumentation to clock event
     */
    XPropClockEvent( event );
	
    /*
     * instrument each state element
     */
    for( ePtr = elements.begin(); ePtr != elements.end(); ++ePtr ) {
	CNode_sp<CNode*> insertionPoint = *ePtr;
        set<CNode*> bothEdges;
	bothEdges.insert( posedgeList.begin(), posedgeList.end() );
	bothEdges.insert( negedgeList.begin(), negedgeList.end() );
    	set<CNode*> ifExpList;
	int count = 0;
	while( static_cast<CNode*>(insertionPoint) != NULL && 
	       insertionPoint->GetOp() == eIF ) {
	    CNode* cond = insertionPoint->Arg<CNode*>(0);
	    ifExpList.insert( cond );
            set<CNode*>::iterator ptr;
	    int found = 0;
	    for( ptr = posedgeList.begin(); ptr != posedgeList.end(); ++ptr ) {
		if( Equivalent( *ptr, cond ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    for( ptr = negedgeList.begin(); ptr != negedgeList.end(); ++ptr ) {
		if( Equivalent( StripNot(*ptr), StripNot(cond) ) ) {
		    bothEdges.erase( *ptr );
		    found = 1;
		    break;
		}
	    }
	    if( !found ) {
		break;
	    }

    	    /*
      	     * create xprop assignments for sync controls
             */
    	    vector<LHS> lhsList;	
            lhsTranslateState = eTRANSLATE_ON;
	    CNode* tStmt = insertionPoint->Arg<CNode*>(1);
	    CNode* fStmt = insertionPoint->Arg<CNode*>(2);
	    CollectLHS( tStmt, lhsList );
	    CollectLHS( fStmt, lhsList );
            tStmt = XPropStatement( module, block, tStmt );
    	    CNode* xStmt = CreateLHSAssignments( module, block, lhsList );

             CNode* n1 = cIF( cCOM(cond->Clone()), fStmt, xStmt );
	     CNode_sp<CNode*> newInsertionPointer = n1->Arg<CNode*>(1);

             if( (insertionClkBeginComment || insertionClkEndComment) &&
                 (!fStmt/* || fStmt->GetOp() != eBLOCK_REF */) ) {
       		 CNode* block = cBLOCK_REF( NULL, fStmt);	
		 newInsertionPointer = block;
	         newInsertionPointer = block->Arg<CNode*>(1);
	     }
             if( insertionClkEndComment ) {
                 CNode* comment = cLIST( insertionClkEndComment->Clone(), fStmt );
                 newInsertionPointer = comment;
		 if( comment->GetOp() == eLIST ) {
                     newInsertionPointer = comment->Arg<CNode*>(1);
		 }
             }

             if( insertionClkBeginComment || (count==0 && insertionClkEndComment) ) {
                 n1 = cBLOCK_REF( NULL, n1 );
	     }
             if( insertionClkBeginComment ) {
                 n1 = cLINK( insertionClkBeginComment->Clone(), n1 );
             }
             if( count == 0 && insertionClkEndComment ) {
                 n1 = cLINK( n1, insertionClkEndComment->Clone() );
             }
             CNode* body = cIF( cond, tStmt,  n1 );
            insertionPoint = body;
	    insertionPoint = newInsertionPointer;
	    count++;
	}
	CNode*   clockExp = *bothEdges.begin();
	/*
 	 * edge is always specified as POSEDGE as clockExp is already inverted
 	 */
        XPropClockBody( module, block, ePOSEDGE, clockExp, 
			insertionPoint, TRUE );
    }

    return TRUE;
}

/*******************************************************
	XPropStatement
	- add x propagation logic 
********************************************************/

static CNode* XPropStatement( CBlock* module, CBlock* block, CNode* n ) 
{
	if( n == NULL ) {
		return n;
	}

        if( IsNoxNode( n ) ) {
	    return n;
        }

	switch( n->GetOp() ) {
	case eDISABLE:
	case ePASSIGN:
	case eDEASSIGN:
	case eTRIGGER:
	case ePORT_DECL:
	case eNET_DECL:
	case eDEFPARAM:
	case ePARAM_DECL:
	case eVAR_DECL:
	case eGENVAR_DECL:
	case eTYPEDEF_DECL:
	case eFUNCTION_CALL:
	case eTASK_ENABLE:
	case eSYSTASK_CALL:
	case eINSTANCE_REF:
	case eGATE_REF:
	case ePORTLIST_END:
		break;
	case ePRAGMA:
		if( IsTranslateOn(n) ) {
			translateState = eTRANSLATE_ON;
		} else if( IsTranslateOff(n) ) {
			if( translateState == eTRANSLATE_OFF ) {
				message( n->GetCoord(), mNSTOS );
			}
			translateState = eTRANSLATE_OFF;
		}
		break;
	case eCOMMENT:
	case eVRQ:
		break;
	case eLIST:
		{
		  CNode* c = n;
		  while( 1 ) {
		      c->Arg<CNode*>(0) = 
			   XPropStatement( module, block, c->Arg<CNode*>(0) );
		      if( !c->Arg<CNode*>(1) || 
			   c->Arg<CNode*>(1)->GetOp() != eLIST ) {
		  	  c->Arg<CNode*>(1) = 
			      XPropStatement( module, block, c->Arg<CNode*>(1) );
			  break;
		      }
		      c = c->Arg<CNode*>(1);
		  }
	   	}
		break;	
	case eINIT:
		n->Arg<CNode*>(0) = XPropStatement( module, block, n->Arg<CNode*>(0) );
		break;
	case eEVENT:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;
	case eWHILE:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;	
	case eREPEAT:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;	
	case eFOREVER:
		n->Arg<CNode*>(0) = XPropStatement( module, block, n->Arg<CNode*>(0) );
		break;	
	case eGFOR:
	case eFOR:
		n->Arg<CNode*>(3) = XPropStatement( module, block, n->Arg<CNode*>(3) );
		break;	
	case eBLOCK_REF:
		{
		    CBlock* nameblock = n->Arg<CBlock*>(0);
		    if( nameblock ) {
			n->Arg<CNode*>(1) = XPropStatement( module, 
						nameblock, n->Arg<CNode*>(1) );
		    } else {
			n->Arg<CNode*>(1) = XPropStatement( module, 
						block, n->Arg<CNode*>(1) );
		    }
		}
		break;
	case eSPECIFY_REF:
		break;
	case eDELAY:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		break;	
	case eFUNCTION_DEF:
		{
			CBlock* block = n->Arg<CBlock*>(0);
			prependToModule = NULL;
			CNode* newBlock = XPropStatement( module, block, 
						block->GetCodeList() ); 
			block->SetCodeList( cLINK( prependToModule, newBlock) );	
		}
		break;
	case eMODULE_DEF:
		{
			CBlock* block = n->Arg<CBlock*>(0);
			prependToModule = NULL;
			CNode* newBlock = XPropStatement( block, block, 
						block->GetCodeList() ); 
			block->SetCodeList( cLINK( prependToModule, newBlock) );	
		}
		break;
	case eRETURN:
	case ePOSTDEC:
	case ePOSTINC:
	case ePREDEC:
	case ePREINC:
	case eWAIT:
	case eCASSIGN:
	case eFORCE:
	case eRELEASE:
	case eTABLE:
		break;
	case eATTRIBUTE:
		break;
	case eGCASE:
		XPropCaseItem( module, block, n->Arg<CNode*>(1) );
		break;
	case eCASEX:
		message( n->GetCoord(), mCCBMS );
		XPropCaseItem( module, block, n->Arg<CNode*>(1) );
		break;
	case eGIF:
		n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
		n->Arg<CNode*>(2) = XPropStatement( module, block, n->Arg<CNode*>(2) );
		break;
/********************************************************
	these are the case that need special processing
 ********************************************************/

        case eADD_ASSIGN:
        case eSUB_ASSIGN:
        case eMUL_ASSIGN:
        case eDIV_ASSIGN:
        case eMOD_ASSIGN:
        case eAND_ASSIGN:
        case eOR_ASSIGN:
        case eXOR_ASSIGN:
        case eLSH_ASSIGN:
        case eRSH_ASSIGN:
        case eLSHA_ASSIGN:
        case eRSHA_ASSIGN:		
	case eASSIGN:
	case eNBASSIGN:
		if( translateState == eTRANSLATE_OFF || 
		     n->HasAttribute( noxAttribute ) ) {
			break;
		}
		if( disableIndex ) {
		    break;
		}
		{
		  vector<IndexInfo> indexList;
		  CollectIndices( n->Arg<CNode*>(1), indexList );
	          if( indexList.size() == 0 ) {
			break;
		  }
		  vector<IndexInfo>::iterator ptr;
		  CVector one(1);
		  one.LoadBinary( "1" );
		  CVector zero(1);
		  zero.LoadBinary( "0" );
		  CNode* cond = NULL;
		  CNode* t    = NULL;
		  set<CDecl*> seen;
		  map<CDecl*,NodeOp_t> decl2Op;
		  map<CDecl*,CNode*> decl2Delay;
		  for( ptr = indexList.begin(); 
					ptr != indexList.end(); ++ptr ) {
		      vector<CNode*>::iterator ptr1;
		      for( ptr1 = ptr->indices.begin(); 
					ptr1 != ptr->indices.end(); ++ptr1 ) {
			CNode* index = *ptr1;
			if( index ) {
			    while( index && index->GetOp() == eMACRO_EXPR ) {
				index = index->Arg<CNode*>(1);
			    }
			    CNode* scond;
			    switch( index->GetOp() ) {
			    case eSLICE:
				continue;
			    case ePSLICE:
			    case eMSLICE:
				scond = index->Arg<CNode*>(0);
				break;
			    default:
			        scond = index;
				break;
			    }
	    		    if( IsNonX(scond) ) {
			       continue; 
			    }
			    if( cond == NULL ) {
				cond = scond->Clone();
			    } else {
				cond = cCAT( cond, scond->Clone() );
			    }
			}
		      }
		   }
		   if( cond ) {
		       if( cond->IsWidthVolatile() || 
		           !cond->IsWidthEvaluateable() || cond->GetWidth() != 1 ) {
		           cond = cRXOR(cond);
		      }
		      for( ptr = indexList.begin(); ptr != indexList.end(); ++ptr ) {
			      vector<int> noXDim(ptr->indices.size(),0);
			      set<CNode*> noXIndices;
			      CNode* s = CreateXAssignment( module, 
			      	  block, n->GetOp(), 
			      	  n->Arg<CNode*>(0), 
			      	  ptr->reference, ptr->indices, 
			      	  FALSE, 
				  noXIndices,
			      	  set<CNode*>() );	
			      t = cLINK( s, t );
		     }
		     CNode* def = cCASEITEM( NULL, t );
		     n = cCASE( cond, cLIST( 
			    cCASEITEM( cELIST(cVECTOR(zero),cVECTOR(one)), BOC(n, true) ),
			    def 
			)
		      );
		     n = BIC( n, false );
		  }
	       }		  
	       break;
	case eCASEZ:
	case eCASE:
		if( translateState == eTRANSLATE_OFF || disableCase ||
		    n->HasAttribute( noxAttribute ) ) {
			n->Arg<CNode*>(1) = XPropCaseItem( module, block, n->Arg<CNode*>(1) );
			break;
		}
		{

			CNode* r = n->Arg<CNode*>(1);
			CNode* l = n->Arg<CNode*>(0);
			vector<LHS> lhsList;	
			lhsTranslateState = translateState;
			CollectLHS( r, lhsList );
			r = XPropCaseItem( module, block, r );
			n->Arg<CNode*>(1) = r;
			int instrumentExpression = !IsNonX(l);
			if( lhsList.size() == 0 ) {
				break;
			}
			list<CNode*> caseList;
			r->PreVisit1( _CollectCases, &caseList );

			if( caseList.size() == 0 && !instrumentExpression ) {
			    break;
			}

			/*
			 * find default case item
			 */
		        CNode** defltParent = FindParentOfDefaultCase( n->Arg<CNode*>(1).Ptr() );
		        CNode* deflt = defltParent ? *defltParent : NULL;


			CNode* innerCond = NULL;
			CNode* outerCond = NULL;
		        CVector one(1);
		        one.LoadBinary( "1" );
		        CVector zero(1);
		        zero.LoadBinary( "0" );
			if( instrumentExpression ) {
			    innerCond = l->Clone();
			}
			if( caseList.size() != 0 ) {
			    list<CNode*>::iterator ptr;  
			    for( ptr = caseList.begin(); ptr != caseList.end(); ++ptr ) {
				CNode* next = EList2Cat( *ptr );
				outerCond = outerCond ? 
						cCAT( outerCond, next ) : next;
			    }
			}
			if( innerCond ) {
			    if( innerCond->IsWidthVolatile() || 
				!innerCond->IsWidthEvaluateable() ||
		                innerCond->GetWidth() != 1 ) {
			        innerCond = cRXOR(innerCond);
			    }
			}
			if( outerCond ) {
			    if( outerCond->IsWidthVolatile() || 
				!outerCond->IsWidthEvaluateable() ||
			        outerCond->GetWidth() != 1 ) {
			        outerCond = cRXOR(outerCond);
			    }
			}


			if( innerCond ) {
			    CNode* defaultCaseItem = deflt;
			    int noDefault = !deflt;
		            if( !deflt ) {
			        deflt = cCASEITEM( NULL, NULL );
				defaultCaseItem = deflt;
				n->Arg<CNode*>(1) = cLIST( n->Arg<CNode*>(1), deflt );
				defltParent = FindParentOfDefaultCase( n->Arg<CNode*>(1).Ptr() );
			    }
			    CNode* t = CreateLHSAssignments( module, block, lhsList );
			    CNode* deflt2 = cCASEITEM( NULL, t );
			    CNode* n2 = defaultCaseItem->Arg<CNode*>(1);
			    if( !noDefault ) {
				n2 = BOC(n2,true);
			    }
		            CNode* newDeflt = cCASE( innerCond,
			        cLIST( 
			            cCASEITEM( 
			    	    cELIST(cVECTOR(zero),cVECTOR(one)), 
			    	    n2 ),
			            deflt2 ) );
			    if( !noDefault ) {
			        newDeflt = BIC(newDeflt,true);
			    }
			    defaultCaseItem->Arg<CNode*>(1) = newDeflt;
			    if( noDefault ) {
			        *defltParent = BIC(defaultCaseItem,false);
			    }
			}
			if( outerCond ) {
			    CNode* t = CreateLHSAssignments( module, block, lhsList );
			    CNode* def = cCASEITEM( NULL, t );
			    n = BOC(n,true);
		            CNode* newN = cCASE( outerCond,
			        cLIST( 
			            cCASEITEM( 
			    	    cELIST(cVECTOR(zero),cVECTOR(one)), 
			    	    n ),
			            def ) );
			    newN->SetAttributes( n->GetAttributes() );
			    newN = BIC(newN, false);
			    n = newN;
			}
		}
		break;
	case eIF:
		if( translateState == eTRANSLATE_OFF || disableIf ||
		    n->HasAttribute( noxAttribute ) ) {
			n->Arg<CNode*>(1) = XPropStatement( module, block, n->Arg<CNode*>(1) );
			n->Arg<CNode*>(2) = XPropStatement( module, block, n->Arg<CNode*>(2) );
			break;
		}
		{
			CNode* condition  = n->Arg<CNode*>(0);
			CNode* tStmt = n->Arg<CNode*>(1);
			CNode* fStmt = n->Arg<CNode*>(2);
			vector<LHS> lhsList;	
			lhsTranslateState = translateState;
			CollectLHS( tStmt, lhsList );
			CollectLHS( fStmt, lhsList );
			tStmt = XPropStatement( module, block, tStmt );
			fStmt = XPropStatement( module, block, fStmt );
			n->Arg<CNode*>(1) = tStmt;
			n->Arg<CNode*>(2) = fStmt;
			if( IsNonX(condition) ) {
				break;
			}
			if( lhsList.size() == 0 ) {
				break;
			}
			CNode* xStmt = CreateLHSAssignments( module, block, lhsList );
			CNode* n_new = NULL;
			if( !fStmt ) {
			    CNode* n1 = cIF( cCOM(condition), fStmt, xStmt );
			    n_new = BIF_OTIF( condition, tStmt, n1 );
			} else {
			    CNode* n1 = BIF_OTIF( cCOM(condition), fStmt, xStmt );
			    n_new = cIF( condition, tStmt, n1 );
			}
			n_new->SetAttributes( n->GetAttributes() );
			n = n_new;
		}
		break;	
	case eALWAYS: 
	case eALWAYS_FF: 
	case eALWAYS_COMB: 
	case eALWAYS_LATCH: {
		int disabled = translateState == eTRANSLATE_OFF || !enableClock;
		int unsupported = FALSE;
		/* 
		 * check to see if always statement is a state element 
		 */
		CNode* event = n->Arg<CNode*>(0);
		int    edgeCount = 0;
		int    eventCount = 0;
		if( event && event->GetOp() == eEVENT ) {
		    if( event->HasAttribute( noxAttribute ) ) {
			disabled = 1;
		    }
    		    if( event->Arg<CNode*>(0) && 
			IsNonX(event->Arg<CNode*>(0)) ) {
			disabled = 1;
		    }
		    CNode* condition = event->Arg<CNode*>(0);
		    if( condition ) {
	                condition->PreVisit1( _CountEdges, &edgeCount );
	                condition->PreVisit1( _CountEventTerms, &eventCount );
			if( edgeCount != eventCount ) {
			    disabled = 1;
			}
		    }
		}	
		/*
 		 * handle simple sync clock blocks as special case
 		 */
		if( !disabled && edgeCount == 1 ) {
			XPropSimpleStateElement( module, block, event );
		/*
 		 * handle complex state element case
 		 */
		} else if( !disabled && edgeCount > 1 ) {
		    unsupported = !XPropStateElement( module, block, event );
		    if( unsupported ) {
        	        message( event->GetCoord(), mUESCI );
		    }
		}
		/*
 		 * handle all combinational, unsupported state elements,
 		 * and blocks with clock instrumentation disabled
 		 */
		if( disabled || unsupported || edgeCount == 0 ) {
		    n->Arg<CNode*>(0) = XPropStatement( module, 
					block, n->Arg<CNode*>(0) );
		}
		} break;
	default:
		MASSERT( FALSE );
	}

	return n;
	
}

/************************************************
	XPropTernary
	- Instrument ? operator
	  s ? a : b => out = s ? a : (~s ? b : n'bx)
 ************************************************/
int XPropTernary( CNode* n, void* )
{
    MASSERT( n );
    if( IsNoxNode( n ) ) {
	return 0;
    }
    switch( n->GetOp() ) {
    case eHOOK: 
	{
	   CNode* fexp = n->Arg<CNode*>(2);
	   CNode* texp = n->Arg<CNode*>(1);
	   CNode* xexp = NULL;
	   CNode* cond = n->Arg<CNode*>(0);
	   if( n->HasAttribute( noxAttribute ) ) {
		break;
	   }
	   /*
 	    * don't instrument if condition cannot be X
 	    */
	   if( IsNonX( cond ) ) {
		break;
	   }
	   NodeType_t etype = n->GetNodeType();
	   NodeType_t ttype = texp->GetNodeType();
	   if( ttype != eS && ttype != eB ) {
	       break;
	   }
	   NodeType_t ftype = fexp->GetNodeType();
	   if( ftype != eS && ftype != eB ) {
		break;
	   }
	   if( fexp->IsWidthVolatile() || !fexp->IsWidthEvaluateable() ) {
		CNode* w = fexp->GetWidthExp();
		if( !w ) {
		    message( n->GetCoord(), mCDWOO );
		} else {
	           CVector x(1);
	           x.LoadBinary( "x" );
	           x.SetPreferredBase(2);
		   xexp = cREP(w,cVECTOR(x));
		}
	   }
	   if( !xexp ) {
	       CVector x(fexp->GetWidth());
	       x.SetToX();
	       x.SetPreferredBase( 16 );
	       xexp = cVECTOR(x);
	   }
	   if( etype == eS ) {
	       xexp = cSYSTASK_CALL( CSymbol::Lookup( "$signed" ), xexp );
	   }
	   n->Arg<CNode*>(0)->PreVisit1( XPropTernary, NULL );
	   n->Arg<CNode*>(1)->PreVisit1( XPropTernary, NULL );
	   n->Arg<CNode*>(2)->PreVisit1( XPropTernary, NULL );
	   CNode* newHook = cHOOK( cCOM(n->Arg<CNode*>(0)->Clone()),
				   n->Arg<CNode*>(2), xexp );
	   n->Arg<CNode*>(2) = newHook;
	}
	return 0;
    default:;
    }
    return 1;
}

/************************************************
	Process
	- process a compilation unit 
 ************************************************/
void CXProp::Process( list<CElement>& inputList,
			list<CElement>& outputList ) 
{ 

    /* 
     * allow synth on/off comments to be overridden
     */
    list<string> commentList1;
    GetPlusArgs( "xprop-begin=", commentList1 );
    insertionBeginComment = StringList2Comments( commentList1 );

    list<string> commentList2;
    GetPlusArgs( "xprop-end=", commentList2 );
    insertionEndComment = StringList2Comments( commentList2 );

    list<string> commentList3;
    GetPlusArgs( "xprop-clk-begin=", commentList3 );
    insertionClkBeginComment = StringList2Comments( commentList3 );

    list<string> commentList4;
    GetPlusArgs( "xprop-clk-end=", commentList4 );
    insertionClkEndComment = StringList2Comments( commentList4 );

    disableIf       = GetPlusArg( "xprop-disable-if" ) != NULL;
    disableCase     = GetPlusArg( "xprop-disable-case" ) != NULL;
    disableIndex    = GetPlusArg( "xprop-disable-index" ) != NULL;
    disableTernary  = GetPlusArg( "xprop-disable-ternary" ) != NULL;
    integerIsNoX    = GetPlusArg( "xprop-int-nox" ) != NULL;
    noxAttribute    = GetPlusArg( "xprop-nox-attr" );
    if( noxAttribute ) {
	// skip over "="
	noxAttribute++;
    }
    enableClock  = GetPlusArg( "xprop-enable-clock" ) != NULL;
    edgeControl	 = GetPlusArg( "xprop-clk-edge-control" );
    if( edgeControl ) {
	if( edgeControl[1] == '!' ) {
           edgeAttrName = strdup( &edgeControl[2] );
           edgeControl = ".dump_strip_edges_ifndef";
	} else {
           edgeAttrName = strdup( &edgeControl[1] );
           edgeControl = ".dump_strip_edges_ifdef";
	}
    } else {
       edgeControl = ".dump_strip_edges_ifdef";
       edgeAttrName = "XPROP_BOTH_EDGES";
    }



    list<CElement>::iterator ptr;
    for( ptr = inputList.begin(); ptr != inputList.end(); ++ptr ) {
	/*
	 * scan though code list adding
	 * code to propagate x's
	 */
	CNode* code = ptr->Code();
	if( code && !disableTernary ) {
	    /*
 	     * Accellerate label operations to speed up
 	     * processing. Node label info should not change
 	     * due to this processing stage.
 	     */
	    CNode::EnableLabelCache();
	    code->PreVisit1( XPropTernary, NULL);
	    CNode::DisableAndClearLabelCache();
	}
	CNode::EnableLabelCache();
	code = XPropStatement( NULL, NULL, code );
	CNode::DisableAndClearLabelCache();
	outputList.push_back( CElement( ptr->Filename(), ptr->Filename()!=NULL, code ) );
    }
}

/**
 * \endcond
 */
