/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */

#ifndef GLUE_HPP
#define GLUE_HPP

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <list>
#include <map>
#include <iostream>
#include <stdarg.h>

using std::string;
using std::map;
using std::list;

struct Coord_t;
class  CBackend;

/**
 * Short cut for signed 32 bit integer.
 */
typedef long		   INT32;
/**
 * Short cut for unsigned 32 bit integer.
 */
typedef unsigned long	   UINT32;
/**
 * Short cut for signed 64 bit integer.
 */
typedef signed long long   INT64;
/**
 * Short cut for unsigned 64 bit integer.
 */
typedef unsigned long long UINT64;

/**
 * Plugin's should need these.
 * \cond
 */
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define PATH_SEPARATOR "/"

extern int	warningCount;
extern int	errorCount;
extern FILE*	logfile;

/**
 * \todo these should be cleanedup
 */
void shell_assert( const char*, int );
void shell_xfree( void*, int );
void* shell_xmalloc( int );
/**
 * \endcond
 */



/**
 * Class for standardizing error/warning/info messages.
 */
class Message {
public:
    /**
     * Message behaviors.
     */
    enum Action_t {
        eIGNORE,	///< ignore message, don't print or increment counts.
        eERROR,		///< error message, print and increment error count
        eWARNING,	///< warning message, print and increment warning count
	eINFO		///< informative message, print only
    };
private:
    static map<string,Message*> abbrev2Message;	///< abbreviation to object map
    static list<Message*> messages;		///< list of all messages
    CBackend* tool;				///< tool associated with message
    Action_t  action;				///< current message action
    int	      locked;				///< non-zero if action cannot be changed
    const string abbrev;			///< message abbreviation
    const string helpText;			///< help text for message
    const string format;			///< format for message
public:
    /**
     * Register a warning message.
     * \ingroup utilities
     * \param tool backend tool owning message, NULL for unowned.
     * \param defaultAction default message behavior.
     * \param abbrev message abbreviation. Must be unique per tool.
     * \param format format string for printing message.
     * \return message object.
     */
    static Message* RegisterWarning( CBackend* tool, 
		      Action_t   defaultAction,
		      const char* abbrev, 
		      const char* format, ... );
    /**
     * Register an error message. Policy is set to eERROR and cannot be changed.
     * \ingroup utilities
     * \param tool backend tool owning message, NULL for unowned.
     * \param abbrev message abbreviation. Must be unique per tool.
     * \param format format string for printing message.
     * \return message object.
     */
    static Message* RegisterError( CBackend* tool, 
		      const char* abbrev, 
		      const char* format, ... );
    /**
     * These routines should not be called directly by plugins.
     * \cond
     */
    /**
     * Find message object from abbreviation.
     * \param abbrev abbreviation to search for. Must include tool name.
     * \return message object.
     */
    static Message* Find( const char* abbrev );
    /**
     * Get a list of all defined messages.
     * \return list of all message objects.
     */
    static list<Message*>& MessageList() { return messages; }
    Message( CBackend* tool,
   	     Action_t action,
	     int    locked,
	     const char* abbrev,
	     const char* helpText,
	     const char* format );
    
    const char* Format() { return format.c_str(); }
    const char* Abbreviation() { return abbrev.c_str(); }
    const char* HelpText() { return helpText.c_str(); }
    Action_t    Action() { return action; }
    void        Action( Action_t a ) { if( !locked ) { action = a; } }
    CBackend*   Tool() { return tool; }
    int	        Locked() { return locked; }
    /**
     * \endcond
     */
};

/**
 * Print message.
 * \ingroup utilities
 * \param location file coordinates to use in message. NULL if none.
 * \param message message object to use for message format.
 * \param ... additional arguments used by format.
 */
void message( struct Coord_t* location, Message* message, ... );
/**
 * Print to logfile.
 * \ingroup utilities
 * \param format format to use for output.
 * \param args variable argument list for values.
 */
void vlogprintf( const char* format, va_list args );
/**
 * Print to logfile.
 * \ingroup utilities
 * \param format format to use for output.
 * \param ... additional arguments used by format.
 */
void logprintf( const char* format, ... );


/**
 * This routine should not be used by plugins.
 * \todo Cleanup internal and remove the use of these.
 */
void warning( struct Coord_t* location, const char* format, ... );
/**
 * This routine should not be used by plugins.
 * \todo Cleanup internal and remove the use of these.
 */
void error( struct Coord_t* location, const char* format, ... );
/**
 * This routine should not be used by plugins.
 * \todo Cleanup internal and remove the use of these.
 */
void info( struct Coord_t* location, const char* format, ... );
/**
 * This routine should not be used by plugins.
 * \todo Cleanup internal and remove the use of these.
 */
void fatal( struct Coord_t* location, const char* format, ... );
/**
 * This routine should not be used by plugins.
 * \todo Cleanup internal and remove the use of these.
 */
void trace( struct Coord_t* location, const char* format, ... );


/**
 * Add invariant checks to code. Failure results in a fatal internal error.
 * Condition \a c must be true.
 * \ingroup utilities
 */
#define MASSERT(c) if(!(c)) { shell_assert( __FILE__, __LINE__ ); abort(); }
/**
 * Add invariant checks to code. Failure results in a fatal internal error.
 * Pointer \a p must not be NULL.
 * \ingroup utilities
 */
#define MTHROW_NIL(p) MASSERT(p!=NULL);

#endif // GLUE_HPP

