%{
/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
#include <map>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <regex.h>
#include <string>
#include <set>
#include <stack>
#include "main.h"
#include "glue.h"
#include "cnode.h"
#include "cobstack.h"
#include "csymbol.h"
#include "cmacro.h"
#include "cvector.h"
#include "lex.h"
#include "glue.h"
#include "parse.h"
#include "csymtab.h"
extern "C" {
    #include "vpi_user.h"
}

Message* mCOIDF = Message::RegisterWarning( NULL, Message::eERROR,
		"COIDF", "cannot open include file '%s'", "<filename>" );

char* lex_copyStringWithSubstitution( char *ptr, int size );
extern "C" {
   char* strndup(const char* s, size_t n);
}

/*
A.8.7 Numbers
    decimal_number ::=
	unsigned_number
	| [ size ] decimal_base unsigned_number
	| [ size ] decimal_base x_digit { _ }
	| [ size ] decimal_base z_digit { _ }
    binary_number ::= [ size ] binary_base binary_value
    octal_number ::= [ size ] octal_base octal_value
    hex_number ::= [ size ] hex_base hex_value
    sign ::= + | -
    size ::= non_zero_unsigned_number
    non_zero_unsigned_number2 ::= non_zero_decimal_digit { _ | decimal_digit}
    binary_value2 ::= binary_digit { _ | binary_digit }
    octal_value2 ::= octal_digit { _ | octal_digit }
    hex_value2 ::= hex_digit { _ | hex_digit }
    decimal_base2 ::= '[s|S]d | '[s|S]D
    binary_base2 ::= '[s|S]b | '[s|S]B
    octal_base2 ::= '[s|S]o | '[s|S]O
    hex_base2 ::= '[s|S]h | '[s|S]H
    non_zero_decimal_digit ::= 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
    decimal_digit ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
    binary_digit ::= x_digit | z_digit | 0 | 1
    octal_digit ::= x_digit | z_digit | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7
    hex_digit ::=
	x_digit | z_digit | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
	| a | b | c | d | e | f | A | B | C | D | E | F
    x_digit ::= x | X
    z_digit ::= z | Z | ?

A.9.3 Identifiers
    hierarchical_block_identifier ::= hierarchical_identifier
    hierarchical_event_identifier ::= hierarchical_identifier
    hierarchical_function_identifier ::= hierarchical_identifier
    hierarchical_identifier ::= { identifier [ [ constant_expression ] ] 
			        . } identifier
    hierarchical_net_identifier ::= hierarchical_identifier
    hierarchical_parameter_identifier ::= hierarchical_identifier
    hierarchical_variable_identifier ::= hierarchical_identifier
    hierarchical_task_identifier ::= hierarchical_identifier
*/
void  clean_identifier( char* s );
void  processString( char* sPtr );
void  doIfdef( const char* macro );
void  doIfndef( const char* macro );
void  doElse( void );
int   doEndif( void );
void  doElsif( const char* macro );
void  doInclude( const char* filename );
void  doUndef( const char* macro );
void terminateMacro();
void startMacro();
char*	lex_ifdefPragma( char* s, int sense );
int   yy_input( char* b, int ms ); 
int   doSimpleMacro( char* macroName );
void  doComplexMacro( char** argList );
FILE* fopenSearch( const char* name, const char* options, 
		   CSymbol** searchList, int isInclude, string* resolvedPath );
void  CompilerDirective( char* string );
int lex_appendComment( char* ptr );
void ConvertTimescale( char* arg, int& unit, int& precision );
int Ts2Value( char* mult, int multLen, char* unit, int unitLen );
MacroToken AllocateMacroToken( const char* text, int leng, int consumed=-1 );
void DumpMacroInfo( const MacroToken& a );
int 	IsMacroDefined( char* macroName );
char* Macro2String( CMacro* macro );
int SVKeyword( int token, YYSTYPE* yystype, char* yytext  );
int SVToken( int token );

//#define TRACE	logprintf
inline void TRACE( const char*,... ) {};
#define GOT(t) TRACE( "got a " #t " -> %s\n", yytext )
#define NEW_LINE	if( fstack && fstack->filename ) loc.lineno++
#define YY_DECL int yylex_( YYSTYPE* yylval )
#define SVTOKEN(token) yylval->macro = AllocateMacroToken( yytext, yyleng ); return SVToken(token);
#define SVEQ(svtoken) if(svEnable) { BEGIN SV_EAT_EQ; return svtoken; } else { yylval->macro = AllocateMacroToken( yytext, yyleng ); return yytext[0]; }

/*
 * Assign each entry into lexer with a unique timestamp.
 * This will be used to associate `macros with the corresponding
 * substituted text. The purpose of this is to allow numbers
 * that are generated from macros have dual representations
 * in the parser so tick define variables can be maintained if
 * desired.
 */
int yylex_( YYSTYPE* yylval);
unsigned	tokenTimestamp = 0;
unsigned	lastPragmaTimestamp = 0xffffffff;
int yylex( YYSTYPE* yylval ) { tokenTimestamp++; return yylex_(yylval); }

/*
 * Stack for keeping track of macros and include files.
 * An entry is create for every `include and macro that is
 * resolved. 
 *
 * Complex macro are processed in two passes. In the first
 * pass the macro arguments are subsituted. In the second
 * pass the processed text is treated as a simple macro
 * an another entry is created.
 */
typedef struct fstack_t {
	struct fstack_t*	next;	  // previous stack entry, NULL if none
	int			serialNumber; // number unique to this entry
	int			complex;  // non-zero if first pass of complex macro
	int 			eof;	  // set when eof input has been reached
	char*			s;	  // pointer to start of macro text
					  // NULL if a file entry.
	char*			c;	  // pointer to next char to be 
					  // consumed of macro text
					  // NULL if a file entry.
	char*			e;	  // pointer to last character of macro
					  // text, NULL if a file entry.
	FILE*			f;	  // FILE pointer, NULL if macro entry
	const char*		filename; // filename for entries context
	int			lineno;
	char**			argList;  // complex macro argument list
	YY_BUFFER_STATE		state;	  // state for this processing level
	int			timestamp; // token timestamp when text was pushed
	int			macroBytesConsumed;
	const char*		macroName; // macro name
	int			minDependentZone; // zone number of definition
	int			insertionZone; // zone number of insertion point
} fstack_t;

int		fstackSerialNumber = 0;
int		initialStateSaved = FALSE;
YY_BUFFER_STATE	initialState    = NULL;
char*		macroText    = NULL;
Coord_t		loc;
int		ifPragma     = 0;
int		ifPragmaSense = 0;
int		ifDepth      = 0;
int		disableDepth = 0;
int		endifCount   = 0;
stack<int>	endifCountStack;
fstack_t*	fstack       = NULL;
int		regex_init   = 0;
list<regex_t>   regex;
int             undefinedMacrosAreTokens = 0;
int 		keepMacroDefinitions = 0;
// zone 0 is reserved for command line defines
int		preprocessorZone = 1;

extern CObstack* permStack; 
extern CObstack* unitStack;
extern CObstack* tokenStack;
CObstack	commentStack( "commentStack" );
CObstack	tempStack( "tempStack", 8*1024); 
CObstack	temp1Stack( "temp1Stack", 1024); 
CObstack	macroStack( "macroStack", 1024); 
char**		macroArgList;
char**		realArgList;
int		interactive;
int		curlyCount;
int		parenCount;
int		inMacroDef = FALSE;
CMacro*		currentMacro = NULL;
extern CSymbol** incSearchList;
static int	eofCount;
static int	macroSaveState;
extern list<string> pragmas;
extern set<string>  pragmaTickDefines;
extern int	macrocheck;
extern int	keepMacros;
/*
 * maintain list of include files
 */
map<char*,int> includeFiles;

/*
 * global lexer state for compiler directives
 */
int compilerDirectiveCell	      = 0;
int compilerDirectiveProtect	      = 0;
int compilerDirectivePrecision	      = 0; // 1 sec
int compilerDirectiveUnit	      = 0; // 1 sec
int compilerDirectiveTimescaleValid    = FALSE;
int compilerDirectiveDefaultNetType   = vpiWire;
int compilerDirectiveUnconnectedDrive = vpiHighZ;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#undef YY_INPUT
#define YY_INPUT( b, r, ms )	(r = yy_input( b, ms ))


%}

ws	[ \t\f\r]+
ident	([a-zA-Z_][a-zA-Z_$0-9]*)
eident	{ident}|(\\[^ \t\f\n]+[ \t\f\n])
sysident \$[a-zA-Z0-9_$]+
notident [^a-zA-Z_]
sign [+-]
nettype wire|tri|tri0|wand|triand|tri1|wor|trior|trireg
drivetype pull0|pull1
tunit	s|ms|us|ns|ps|fs
tnum	1|10|100
exp	e|E



real_number ({unsigned_number}\.{unsigned_number})|({unsigned_number}(\.{unsigned_number})?{exp}{sign}?{unsigned_number})
decimal_number ({size}?{decimal_base}{unsigned_number})|({size}?{decimal_base}{x_digit}[_]*)|({size}?{decimal_base}{z_digit}[_]*) 
binary_number {size}?{binary_base}{binary_value}
octal_number {size}?{octal_base}{octal_value}
hex_number  {size}?{hex_base}{hex_value}
size {non_zero_unsigned_number}
non_zero_unsigned_number {non_zero_decimal_digit}(_|{decimal_digit})*
unsigned_number {decimal_digit}("_"|{decimal_digit})*
binary_value {binary_digit}(_|{binary_digit})*
octal_value {octal_digit}(_|{octal_digit})*
hex_value {hex_digit}(_|{hex_digit})*
decimal_base '[sS]?[dD]
binary_base '[sS]?[bB]
octal_base '[sS]?[oO]
hex_base '[sS]?[hH]
non_zero_decimal_digit [1-9]
decimal_digit [0-9]
binary_digit ({x_digit}|{z_digit}|[01])
octal_digit  ({x_digit}|{z_digit}|[0-7]) 
hex_digit    ({x_digit}|{z_digit}|[0-9a-fA-F])
x_digit      [xX]
z_digit      [zZ?]
literal_digit [01xXzZ]

%x EAT
%x IFPRAGMA
%x IFPRAGMACOMMENT
%x IFSCAN
%x IFSCANCOMMENT
%x IFDEFNAME
%x IFNDEFNAME
%x ELSEIFNAME
%x COMMENT
%x UDP
%x UDP_COMMENT
%x VCOMMENT
%x VCOMMENT1
%x VCOMMENT2
%x INCLUDE
%x DMSTEXT
%x DEF
%x DEF1
%x UNDEF
%x DEFCOM1
%x DEFCOM2
%x DEFCOM3
%x MACRO1
%x MACRO2
%x SUB
%x DRIVETYPE
%x NETTYPE
%x TIME1
%x TIME2
%x TIME3
%x TIME4
%x TIME5
%x EDGE_SPEC
%x DBASE
%x HBASE
%x OBASE
%x BBASE
%x PROTECT
%x SV_EAT_EQ

%%
	/*
	 * compiler directives
	 */
`uselib.*	      |
`celldefine	      |
`endcelldefine        |
`resetall 	      |
`nounconnected_drive  { GOT(compiler directive); 
			CompilerDirective( yytext );
			BEGIN EAT; }

`timescale[ \t]+	{ yymore(); BEGIN TIME1; }
<TIME1>{tnum} 		{ yymore(); BEGIN TIME2; }	   
<TIME1>[^ \t\n\f]	{ yyerror( "bad time specification" ); BEGIN EAT; } 
<TIME2>[ \t]+		{ yymore(); }
<TIME2>{tunit}		{ yymore(); BEGIN TIME3; }
<TIME2>[^ \t\n\f]	{ yyerror( "bad time specification" ); BEGIN EAT; } 
<TIME3>[ \t]+		{ yymore(); }
<TIME3>"/"[ \t]*	{ yymore(); BEGIN TIME4; }
<TIME3>[^ \t\n\f]	{ yyerror( "bad time specification" ); BEGIN EAT; } 
<TIME4>{tnum} 		{ yymore(); BEGIN TIME5; }	   
<TIME4>[^ \t\n\f]	{ yyerror( "bad time specification" ); BEGIN EAT; } 
<TIME5>[ \t]+		{ yymore(); }
<TIME5>{tunit}		{ yymore(); GOT(timespec); 
			  CompilerDirective( yytext );
			  BEGIN EAT; } 
<TIME5>[^ \t\n\f]	{ yyerror( "bad time specification" ); BEGIN EAT; } 

	/*
	 * unconnected_drive processing
	 */
`unconnected_drive[ \t]+  { yymore(); BEGIN DRIVETYPE; }
<DRIVETYPE>{drivetype}    { GOT(drivetype); 
			    CompilerDirective( yytext );
			    BEGIN EAT; } 
<DRIVETYPE>[^ \t\n\f]+	  { yyerror( "illegal drive type" ); BEGIN INITIAL; }

	/*
	 * default_nettype processing
	 */
`default_nettype[ \t]+  { yymore(); BEGIN NETTYPE; }
<NETTYPE>{nettype}      { GOT(nettype); 
			  CompilerDirective( yytext );
			  BEGIN EAT; } 
<NETTYPE>[^ \t\n\f]+	{ yyerror( "illegal net type" ); BEGIN INITIAL; }

	/*
	 * define processing
	 */

`define[ \t]+		{ BEGIN DEF; macroStack.Grow( yytext, yyleng ); }	
<DEF1>"("[ \t]*		{ BEGIN DEFCOM2; macroStack.Grow( yytext, yyleng ); }
	/*
	 * I'm not sure the correct way to distinguish
	 * a these cases:
	 * 	define mc(a) (a+1)
	 * 	define mc (a) (a+1) Is this one a simple macro?
	 *	define mc (a+1)  This is simple but doesn't look it
	 */
<DEF1>[ \t]+		{ startMacro(); macroStack.Grow( yytext, yyleng ); }
<DEF1>[^(]		{ startMacro(); yyless(0); macroStack.Grow( yytext, yyleng ); }
<DEFCOM2>[ \t]*		;
<DEFCOM2>{ident}	{
				void*	ptr;
				ptr = permStack->Copy0(yytext,yyleng); 
				tempStack.PtrGrow( ptr );
				macroStack.Grow( yytext, yyleng );
				BEGIN DEFCOM3; 
			}
<DEFCOM2>{notident}	{ 
				macroStack.Grow( yytext, yyleng );
				yyerror( "bad argument in macro definition" ); 
				if( yytext[yyleng-1] == '\n' ) {
					yyless(yyleng-1);
				}
				BEGIN EAT;
			}
<DEFCOM3>","[ \t]*	{ BEGIN DEFCOM2; macroStack.Grow( yytext, yyleng ); }
<DEFCOM3>[ \t]*		{  macroStack.Grow( yytext, yyleng ); }
<DEFCOM3>\n		{ yyerror( "missing comma between macro arguments" ); BEGIN INITIAL; }
<DEFCOM3>")"		{ 
			char**	argList;
			char**	tempArgList;

			macroStack.Grow( yytext, yyleng );
			startMacro();	 
			currentMacro->SetComplex(TRUE); 
			tempStack.PtrGrow(NULL);
			tempArgList = (char**) tempStack.GetBase();
			argList = (char**) permStack->Copy( tempArgList, tempStack.GetObjectSize() );
			tempStack.Free( NULL /*tempArgList*/ ); 
			currentMacro->SetArgList( argList );	
			}
<DEFCOM3>[^,]		{ yyerror( "missing comma between macro arguments" ); BEGIN EAT; }	

<DEF>{ident}		{ 
			 	macroStack.Grow( yytext, yyleng );
				CSymbol* symbol;
				currentMacro = CMacro::LookupMacro( yytext ); 
				if( currentMacro == NULL ) {
					symbol = CSymbol::Lookup( yytext ); 	
					currentMacro = new(permStack) CMacro( symbol, &loc );
				} else {
					currentMacro->SetCoord( &loc );
				}
				currentMacro->SetComplex(FALSE);
				currentMacro->SetDefined(TRUE);
				BEGIN DEF1;
			 }
<DEF>[^a-zA-Z_].*	{ yyerror( "illegal macro name" ); BEGIN INITIAL; }
<DMSTEXT>[^/\\\n]*	{ permStack->Grow( yytext, yyleng); macroStack.Grow( yytext, yyleng); }
<DMSTEXT>\n		{ 
			macroStack.Grow( yytext, yyleng );
			terminateMacro();
			NEW_LINE; 
			if( keepMacroDefinitions ) {
			    char* text = Macro2String( currentMacro );
			    yylval->macro = AllocateMacroToken( text, 
				strlen(text), macroStack.GetObjectSize() );
			    macroStack.Free(NULL);
	  		    lastPragmaTimestamp = tokenTimestamp;
			    return PRAGMA;
			}
			}
<DMSTEXT>"//"[^\\\n]*	{ macroStack.Grow( yytext, yyleng ); }	
<DMSTEXT>\\\n		{ macroStack.Grow( "\n", 1 );
			   permStack->Grow( "\n", 1 ); NEW_LINE; }
<DMSTEXT>.		{ macroStack.Grow( yytext, yyleng );
			  permStack->Grow( yytext, yyleng ); }

	/*
	 * undef processing
	 */

`undef[ \t]+		BEGIN UNDEF;	
<UNDEF>{ident}		{ BEGIN EAT; doUndef( yytext ); }
<UNDEF>[^a-zA-Z_].*	{ yyerror( "illegal macro name" ); BEGIN INITIAL; }

	/*
	 * ifdef processing
	 */

<INITIAL,IFSCAN>`ifdef[ \t]+	{ BEGIN IFDEFNAME; }
<IFPRAGMA>`ifdef[ \t].*	{ doIfdef( "" ); }
<IFDEFNAME>{ident}	{ doIfdef( yytext ); }
<IFDEFNAME>[^a-zA-Z_].*	{ yyerror( "illegal macro name" ); doIfdef( "" ); }
<INITIAL,IFSCAN>`ifndef[ \t]+	{ BEGIN IFNDEFNAME; }
<IFPRAGMA>`ifndef[ \t]+	{ doIfndef(""); }
<IFNDEFNAME>{ident}	{ doIfndef( yytext ); }
<IFNDEFNAME>[^a-zA-Z_].* { yyerror( "illegal macro name" ); doIfndef( "" ); }
<INITIAL,IFSCAN>`else[ \t]*	{ doElse(); }	
<IFPRAGMA>`else[ \t]*	{ doElse(); }	
<INITIAL,IFSCAN>`elsif[ \t]+	{ BEGIN ELSEIFNAME; }	
<IFPRAGMA>`elsif[ \t]+	{ doElsif( "" ); }	
<ELSEIFNAME>{ident}	 { doElsif( yytext ); }
<ELSEIFNAME>[^a-zA-Z_].* { yyerror( "illegal macro name" ); doElsif( "" ); }
<INITIAL,IFSCAN>`endif[ \t]*	{ doEndif(); }	
<IFPRAGMA>`endif[ \t]*	{ if( doEndif() ) {
			     char* text = lex_ifdefPragma( yytext, ifPragmaSense ); 
			     yylval->macro = AllocateMacroToken( text, strlen(text) );
	  		     lastPragmaTimestamp = tokenTimestamp;
			     return PRAGMA;
			  } 
			}	

	/*	
	 * handle encrypted regions as a PRAGMA
	 */
`protect	{
		    yymore();
		    BEGIN PROTECT;		    
		} 
<PROTECT>`endprotect {
		    unitStack->Grow( yytext, yyleng ); 
		    unitStack->Grow( "\n", 2 );
		    char* text = (char*)unitStack->Finish();
		    yylval->macro = AllocateMacroToken( text, strlen(text) );
	            BEGIN INITIAL;
	  	    lastPragmaTimestamp = tokenTimestamp;
		    return PRAGMA;
		}
<PROTECT>\n	{ 
		    yymore();
		    NEW_LINE; 
		}
<PROTECT>. 	{
				yymore();	
			}	

	/*
	 * include processing
	 */
`include[ \t]+			BEGIN INCLUDE;	
<INCLUDE>\"[^\"\n]*\" 	{ 
			yytext[strlen(yytext)-1] = 0;
	  		BEGIN INITIAL;
			doInclude( CSymbol::Lookup(yytext+1)->GetName() );
	       	}
<INCLUDE>\"[^\"\n]*$	{ yyerror( "filename is missing closing quote" );
			  BEGIN INITIAL; }
<INCLUDE>[^\"].*	{ yyerror( "bad filename argument" ); BEGIN INITIAL; }
	
	/*
	 * macros
	 */

<DBASE,OBASE,HBASE,BBASE,INITIAL>`{ident}	{ 
			macroSaveState = YYSTATE;
			if( undefinedMacrosAreTokens &&
			    !IsMacroDefined( yytext+1 ) ) {
    			    yylval->string = (char*)unitStack->Copy0( yytext+1, yyleng-1 );
			    return UNDEF_MACRO;
			}
			if( !doSimpleMacro( yytext+1 ) ) {
				BEGIN MACRO1; 
			}
		}
<MACRO1>"("		{ BEGIN MACRO2; curlyCount = 0; parenCount = 0; }
<MACRO1>[ \t]+		;	
<MACRO1>\n		{ NEW_LINE; }
<MACRO1>[^(]		{ yyerror( "missing macro argument list" ); BEGIN macroSaveState; }

<MACRO2>\n		{ NEW_LINE; tempStack.Grow( yytext, yyleng ); }
<MACRO2>"("		{ parenCount++; tempStack.Grow( yytext, yyleng ); }
<MACRO2>"{"		{ curlyCount++; tempStack.Grow( yytext, yyleng ); }	
<MACRO2>")"		{ if( parenCount == 0 && curlyCount != 0 ) {
			  	yyerror( "unmatched ')' in expression" );
				BEGIN macroSaveState;
			  } else  if( parenCount == 0 && curlyCount == 0 ) {
				char* ptr;
				tempStack.Grow( "", 1 );
				ptr = (char*) tempStack.Finish();
				temp1Stack.PtrGrow( ptr );
				temp1Stack.PtrGrow( NULL );
				TRACE( "Got argument -> %s\n", ptr );
				BEGIN macroSaveState;
				TRACE( "ending complex macro\n" );
				doComplexMacro( (char**) temp1Stack.Finish() );
			  } else {
				tempStack.Grow( yytext, yyleng );
				parenCount--;
			  }
			}	
<MACRO2>"}"		{ 
			  if( curlyCount == 0 ) {
				yyerror( "unmatched '}' in expression" );
				BEGIN macroSaveState;
			  } else {
				tempStack.Grow( yytext, yyleng );
				curlyCount--; 
			  }
			}	
<MACRO2>","		{ 	
			  if( parenCount == 0 && curlyCount == 0 ) {
				char* ptr;
				tempStack.Grow( "", 1 );
				ptr = (char*) tempStack.Finish();
				temp1Stack.PtrGrow( ptr );
				TRACE( "Got argument -> %s\n", ptr );
			  } else {
				tempStack.Grow( yytext, yyleng );
			  }
			}
<MACRO2>.		{ tempStack.Grow( yytext, yyleng ); }

<SUB>`{ident}		{ tempStack.Grow( yytext, yyleng ); }
<SUB>{ident}		{ 
			    char** p1;
			    char** p2;
			    p1 = realArgList;
			    p2 = macroArgList;
			    while( *p1 ) {
			 	if( !strncmp( yytext, *p1, yyleng ) ) {
					tempStack.Grow( *p2, strlen(*p2) ); 
					break;
				}
				p1++;
				p2++;
			    }
			    if( !*p1 ) {
				tempStack.Grow( yytext, yyleng );
			    }
			}
<SUB>.|\n		{ tempStack.Grow( yytext, yyleng ); }

	/*
	 * numbers
 	 */

{real_number}		{
			    yylval->macro = AllocateMacroToken( yytext, yyleng );
			    return REAL_NUMBER;
			}
{unsigned_number}	{ 
			    yylval->macro = AllocateMacroToken( yytext, yyleng );
			    return NUMBER;
			} 
			    
<DBASE>{unsigned_number} |
<DBASE>{x_digit}[_]*	 |
<DBASE>{z_digit}[_]*	 |
<BBASE>{binary_value}    |	
<OBASE>{octal_value}     |	
<HBASE>{hex_value}       { 
			    BEGIN INITIAL;
			    yylval->macro = AllocateMacroToken( yytext, yyleng );
			    return BNUMBER;
			} 

{binary_base}	{ 
		    BEGIN BBASE; 
		    yylval->macro = AllocateMacroToken( yytext, yyleng );
		    return BASE;
		} 
{octal_base}	{ 
		    BEGIN OBASE; 
		    yylval->macro = AllocateMacroToken( yytext, yyleng );
		    return BASE;
		} 
{hex_base}	{ 
		    BEGIN HBASE; 
		    yylval->macro = AllocateMacroToken( yytext, yyleng );
		    return BASE;
		} 
{decimal_base}	{ 
		    BEGIN DBASE; 
		    yylval->macro = AllocateMacroToken( yytext, yyleng );
		    return BASE;
		} 
\'{literal_digit} {
		    yylval->macro = AllocateMacroToken( yytext, yyleng );
		    return UNBASED_LITERAL;
		}
			    
	/*
	 * keywords
	 */
always		{ return ALWAYS; }	
and		{ return GAND; }	
assign		{ return ASSIGN; }	
automatic	{ return AUTOMATIC; }
begin		{ return BEGIN_; }	
buf		{ return GBUF; }	
bufif0		{ return GBUFIF0; }	
bufif1		{ return GBUFIF1; }	
case		{ return CASE; }	
casex		{ return CASEX; }	
casez		{ return CASEZ; }	
cell		{ return CELL; }	
cmos		{ return GCMOS; }	
config		{ return CONFIG; }	
deassign	{ return DEASSIGN; }	
default		{ return DEFAULT; }	
defparam	{ return DEFPARAM; }	
design		{ return DESIGN; }	
disable		{ return DISABLE; }	
edge		{ 
		  BEGIN EDGE_SPEC;
		  return EDGE; 
		}	
else	 	{ return ELSE; }	
end		{ return END; }	
endattribute	{ return UNIMPLEMENTED; }	
endcase		{ return ENDCASE; }	
endconfig	{ return ENDCONFIG; }	
endfunction	{ return ENDFUNCTION; }	
endgenerate	{ return ENDGENERATE; }
endmodule	{ return ENDMODULE; }	
endprimitive	{ return ENDPRIMITIVE; }	
endspecify	{ return ENDSPECIFY; }
endtask		{ return ENDTASK; }	
event		{ return EVENT; }	
for		{ return FOR; }	
force		{ return FORCE; }	
forever		{ return FOREVER; }	
fork		{ return FORK; }	
function	{ return FUNCTION; }	
generate	{ return GENERATE; }
genvar		{ return GENVAR; }
highz0		{ return HIGHZ0; }	
highz1		{ return HIGHZ1; }	
if		{ return IF; }	
ifnone		{ return IFNONE; }
incdir		{ return INCDIR; }
include		{ return INCLUDE; }
initial		{ return INITIAL_; }	
inout		{ return INOUT; }	
input		{ return INPUT; }	
instance	{ return INSTANCE; }	
integer		{ return INTEGER; }	
join		{ return JOIN; }	
large		{ return LARGE; }	
library		{ return LIBRARY; }
localparam	{ return LOCALPARAM; }
liblist		{ return LIBLIST; }	
macromodule	{ return MACROMODULE; }	
medium		{ return MEDIUM; }	
module		{ return MODULE; }	
nand		{ return GNAND; }	
negedge		{ return NEGEDGE; }	
nmos		{ return GNMOS; }	
nor		{ return GNOR; }	
noshowcancelled { return NOSHOWCANCELLED; }
not		{ return GNOT; }	
notif0		{ return GNOTIF0; }	
notif1		{ return GNOTIF1; }	
or		{ return OR; }	
output		{ return OUTPUT; }	
parameter	{ return PARAMETER; }	
pmos		{ return GPMOS; }	
posedge		{ return POSEDGE; }	
primitive	{ return PRIMITIVE; }
pull0		{ return PULL0; }	
pull1		{ return PULL1; }	
pullup		{ return GPULLUP; }	
pulldown	{ return GPULLDOWN; }	
pulsestyle_onevent   { return PULSESTYLE_ONEVENT; }
pulsestyle_ondetect  { return PULSESTYLE_ONDETECT; }
rcmos		{ return GRCMOS; }	
real		{ return REAL; }	
realtime	{ return REALTIME; }	
reg		{ return REG; }	
release		{ return RELEASE; }	
repeat		{ return REPEAT; }	
rnmos		{ return GRNMOS; }	
rpmos		{ return GRPMOS; }	
rtran		{ return GRTRAN; }
rtranif0	{ return GRTRANIF0; }
rtranif1	{ return GRTRANIF1; }
scalared	{ return SCALARED; }
showcancelled	{ return SHOWCANCELLED; }
signed		{ return SIGNED; }
small		{ return SMALL; }
specify		{ return SPECIFY; }
specparam	{ return SPECPARAM; }
strong0		{ return STRONG0; }
strong1		{ return STRONG1; }
supply0		{ return SUPPLY0; }
supply1		{ return SUPPLY1; }
table		{ return TABLE; }
task		{ return TASK; }
time		{ return TIME; }
tran		{ return GTRAN; }
tranif0		{ return GTRANIF0; }
tranif1		{ return GTRANIF1; }
tri		{ return TRI; }
tri1		{ return TRI1; }
tri0		{ return TRI0; }
triand		{ return TRIAND; }
trior		{ return TRIOR; }
trireg		{ return TRIREG; }
unsigned	{ return UNSIGNED; }
use		{ return USE; }
uwire		{ return UWIRE; }
vectored	{ return VECTORED; }
wait		{ return WAIT; }
wand		{ return WAND; }
weak0		{ return WEAK0; }
weak1		{ return WEAK1; }
wire		{ return WIRE; }
while		{ return WHILE; }
wor		{ return WOR; }
xnor		{ return GXNOR; }	
xor		{ return GXOR; }	

   /*
    * system verilog keywords
    */
alias		{ return SVKeyword(ALIAS,yylval,yytext); }
always_comb	{ return SVKeyword(ALWAYS_COMB,yylval,yytext); }
always_ff	{ return SVKeyword(ALWAYS_FF,yylval,yytext); }
always_latch	{ return SVKeyword(ALWAYS_LATCH,yylval,yytext); }
assert		{ return SVKeyword(ASSERT,yylval,yytext); }
assume		{ return SVKeyword(ASSUME,yylval,yytext); }
before		{ return SVKeyword(BEFORE,yylval,yytext); }
bind		{ return SVKeyword(BIND,yylval,yytext); }
bins		{ return SVKeyword(BINS,yylval,yytext); }
binsof		{ return SVKeyword(BINSOF,yylval,yytext); }
bit		{ return SVKeyword(BIT,yylval,yytext); }
break		{ return SVKeyword(BREAK,yylval,yytext); }
byte		{ return SVKeyword(BYTE,yylval,yytext); }
chandle		{ return SVKeyword(CHANDLE,yylval,yytext); }
class		{ return SVKeyword(CLASS,yylval,yytext); }
clocking	{ return SVKeyword(CLOCKING,yylval,yytext); }
const		{ return SVKeyword(CONST,yylval,yytext); }
constraint	{ return SVKeyword(CONSTRAINT,yylval,yytext); }
context		{ return SVKeyword(CONTEXT,yylval,yytext); }
continue	{ return SVKeyword(CONTINUE,yylval,yytext); }
cover		{ return SVKeyword(COVER,yylval,yytext); }
covergroup	{ return SVKeyword(COVERGROUP,yylval,yytext); }
coverpoint	{ return SVKeyword(COVERPOINT,yylval,yytext); }
cross		{ return SVKeyword(CROSS,yylval,yytext); }
dist		{ return SVKeyword(DIST,yylval,yytext); }
do		{ return SVKeyword(DO,yylval,yytext); }
endclass	{ return SVKeyword(ENDCLASS,yylval,yytext); }
endclocking	{ return SVKeyword(ENDCLOCKING,yylval,yytext); }
endgroup	{ return SVKeyword(ENDGROUP,yylval,yytext); }
endinterface	{ return SVKeyword(ENDINTERFACE,yylval,yytext); }
endpackage	{ return SVKeyword(ENDPACKAGE,yylval,yytext); }
endprogram	{ return SVKeyword(ENDPROGRAM,yylval,yytext); }
endproperty	{ return SVKeyword(ENDPROPERTY,yylval,yytext); }
endsequence	{ return SVKeyword(ENDSEQUENCE,yylval,yytext); }
enum		{ return SVKeyword(ENUM,yylval,yytext); }
expect		{ return SVKeyword(EXPECT,yylval,yytext); }
export		{ return SVKeyword(EXPORT,yylval,yytext); }
extends		{ return SVKeyword(EXTENDS,yylval,yytext); }
extern		{ return SVKeyword(_EXTERN,yylval,yytext); }
final		{ return SVKeyword(FINAL,yylval,yytext); }
first_match	{ return SVKeyword(FIRST_MATCH,yylval,yytext); }
foreach		{ return SVKeyword(FOREACH,yylval,yytext); }
forkjoin	{ return SVKeyword(FORKJOIN,yylval,yytext); }
iff		{ return SVKeyword(IFF,yylval,yytext); }
ignore_bins	{ return SVKeyword(IGNORE_BINS,yylval,yytext); }
illegal_bins	{ return SVKeyword(ILLEGAL_BINS,yylval,yytext); }
import		{ return SVKeyword(IMPORT,yylval,yytext); }
inside		{ return SVKeyword(INSIDE,yylval,yytext); }
int		{ return SVKeyword(INT,yylval,yytext); }
interface	{ return SVKeyword(INTERFACE,yylval,yytext); }
intersect	{ return SVKeyword(INTERSECT,yylval,yytext); }
join_any	{ return SVKeyword(JOIN_ANY,yylval,yytext); }
join_none	{ return SVKeyword(JOIN_NONE,yylval,yytext); }
local		{ return SVKeyword(LOCAL,yylval,yytext); }
logic		{ return SVKeyword(LOGIC,yylval,yytext); }
longint		{ return SVKeyword(LONGINT,yylval,yytext); }
matches		{ return SVKeyword(MATCHES,yylval,yytext); }
modport		{ return SVKeyword(MODPORT,yylval,yytext); }
new		{ return SVKeyword(NEW,yylval,yytext); }
null		{ return SVKeyword(_NULL,yylval,yytext); }
package		{ return SVKeyword(PACKAGE,yylval,yytext); }
packed		{ return SVKeyword(PACKED,yylval,yytext); }
priority	{ return SVKeyword(PRIORITY,yylval,yytext); }
program		{ return SVKeyword(PROGRAM,yylval,yytext); }
property	{ return SVKeyword(PROPERTY,yylval,yytext); }
protected	{ return SVKeyword(PROTECTED,yylval,yytext); }
pure		{ return SVKeyword(PURE,yylval,yytext); }
rand		{ return SVKeyword(RAND,yylval,yytext); }
randc		{ return SVKeyword(RANDC,yylval,yytext); }
randcase	{ return SVKeyword(RANDCASE,yylval,yytext); }
randsequence	{ return SVKeyword(RANDSEQUENCE,yylval,yytext); }
ref		{ return SVKeyword(REF,yylval,yytext); }
return		{ return SVKeyword(RETURN,yylval,yytext); }
sequence	{ return SVKeyword(SEQUENCE,yylval,yytext); }
shortint	{ return SVKeyword(SHORTINT,yylval,yytext); }
shortreal	{ return SVKeyword(SHORTREAL,yylval,yytext); }
solve		{ return SVKeyword(SOLVE,yylval,yytext); }
static		{ return SVKeyword(STATIC,yylval,yytext); }
string		{ return SVKeyword(STRING,yylval,yytext); }
struct		{ return SVKeyword(STRUCT,yylval,yytext); }
super		{ return SVKeyword(SUPER,yylval,yytext); }
tagged		{ return SVKeyword(TAGGED,yylval,yytext); }
this		{ return SVKeyword(THIS,yylval,yytext); }
throughout	{ return SVKeyword(THROUGHOUT,yylval,yytext); }
timeprecision	{ return SVKeyword(TIMEPRECISION,yylval,yytext); }
timeunit	{ return SVKeyword(TIMEUNIT,yylval,yytext); }
type		{ return SVKeyword(TYPE,yylval,yytext); }
typedef		{ return SVKeyword(TYPEDEF,yylval,yytext); }
union		{ return SVKeyword(UNION,yylval,yytext); }
unique		{ return SVKeyword(UNIQUE,yylval,yytext); }
var		{ return SVKeyword(VAR,yylval,yytext); }
virtual		{ return SVKeyword(VIRTUAL,yylval,yytext); }
void		{ return SVKeyword(VOID,yylval,yytext); }
wait_order	{ return SVKeyword(WAIT_ORDER,yylval,yytext); }
wildcard	{ return SVKeyword(WILDCARD,yylval,yytext); }
with		{ return SVKeyword(WITH,yylval,yytext); }

"==?"  { SVTOKEN(WILDCARD_EQUALITY); }
"!=?"  { SVTOKEN(WILDCARD_INEQUALITY); }
"##"   { SVTOKEN(CYCLE_DELAY); }
".*"   { SVTOKEN(WILDCARD); }
"--"   { SVTOKEN(DECREMENT); }
"++"   { SVTOKEN(INCREMENT); }
\+/=  { SVEQ(PLUS_EQ); }
\-/=  { SVEQ(MINUS_EQ); }
"*="   { SVTOKEN(MUL_EQ); }
"/="   { SVTOKEN(DIV_EQ); }
"%="   { SVTOKEN(MOD_EQ); }
"&="   { SVTOKEN(AND_EQ); }
"|="   { SVTOKEN(OR_EQ); }
"^="   { SVTOKEN(XOR_EQ); }
">>="  { SVTOKEN(RIGHT_SHIFT_EQ); }
"<<="  { SVTOKEN(LEFT_SHIFT_EQ); }
">>>=" { SVTOKEN(RRR_EQ); }
"<<<=" { SVTOKEN(LLL_EQ); }
<SV_EAT_EQ>"=" { BEGIN INITIAL; }

	/*
	 * operators
	 */
"~&"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return REDUCTION_NAND; }
"~|"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return REDUCTION_NOR; }
"~^"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return XNOR; }
"^~"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return XNOR; }
"=="	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return LOGICAL_EQUALITY; }	
"!="	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return LOGICAL_INEQUALITY; }	
"==="	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return CASE_EQUALITY; }	
"!=="	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return CASE_INEQUALITY; }	
"||"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return OROR; }	
"<="	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return LE; }	
">="	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return GE; }	
">>"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return RIGHT_SHIFT; }	
"<<"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return LEFT_SHIFT; }	
"&&"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return ANDAND; }
"&&&"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return ANDANDAND; }
"<"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return LT; }
">"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return GT; }
"*>"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return STARARROW; }
"=>"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return EQUALARROW; }
"->"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return ARROW_; }
"(*"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return ATTRIBUTE_START; }
"*)"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return ATTRIBUTE_END; }
"**"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return STARSTAR; }
">>>"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return RRR; }
"<<<"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return LLL; }
"+:"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return PCOLON; }
"-:"	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return MCOLON; }

[:\^+\-=!~&*/|%\?,\(\)\{\}\[\]]	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return yytext[0]; }
\'	{ yylval->macro = AllocateMacroToken( yytext, yyleng ); return TICK; }
    /*
     * this is a special marker returned by yy_input to indicate the end 
     * of an macro or include file. It should only be returned to the parser
     * if the previous token was a PRAGMA.
     */
\001	{ 
	  yylval->macro = AllocateMacroToken( "EOI", 3 ); 
	  if( yylval->macro.history ) {
	      yylval->macro.history[0].final = TRUE;
	  }
	  // ??? mdh - keepMacros isn't really needed here.
	  // 	       I added it to reduce the chance the EOI
	  //	       causes issues when it isn't needed.
	  if( keepMacros && 
	      lastPragmaTimestamp == tokenTimestamp-1 ) {
	      return EOI; 
	  }
	}

	/*
	 * systask
	 */
\$setup		{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return SETUP; 
		}
\$hold		{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return HOLD; 
		}
\$period	{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return PERIOD; 
		}
\$width		{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return WIDTH; 
		}
\$skew		{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return SKEW; 
		}
\$recovery	{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return RECOVERY; 
		}
\$setuphold	{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return SETUPHOLD; 
		}
\$nochange	{ 
		    yylval->symbol = CSymbol::Lookup( yytext ); 
		    return NOCHANGE; 
		}

{sysident}		{ 
			yylval->symbol = CSymbol::Lookup( yytext ); 
			return SYSTEM_IDENTIFIER; 
			}

	/*
	 * identifiers
	 */
{eident}			{ 
			clean_identifier( yytext );
			yylval->symbol = CSymbol::Lookup( Identifier(yytext) ); 
			CDecl*	decl = cuSymtab.Lookup( yylval->symbol );
			if( decl && decl->GetType() == eTYPEDEF ) {
			    return TYPE_IDENTIFIER;
			}
			return IDENTIFIER; 
			}


	/*
	 * eat rest of line
	 */

<EAT>.*$	BEGIN INITIAL;

	/*
	 * comments
	 */

"//"[ \t]*vrq[ \t]+translate_off.*	{ 
				if( ignoreVrqComments ) {
				    unitStack->Grow( yytext, yyleng ); 
				    unitStack->Grow( "\n", 2 );
				    char* ptr = (char*)unitStack->Finish();
				    if( !lex_appendComment( ptr ) ) {
				        char* text = lex_copyStringWithSubstitution( ptr, strlen(ptr) );
		                        yylval->macro = AllocateMacroToken( text, strlen(text) );
	  	                        lastPragmaTimestamp = tokenTimestamp;
				        return PRAGMA;
				    }
				} else { 
				    unitStack->Grow( yytext, yyleng ); 
				    unitStack->Grow( "\n", 1 );
				    BEGIN VCOMMENT;
				}
			}
<INITIAL,UDP>"//".*			{ 
				unitStack->Grow( yytext, yyleng ); 
				unitStack->Grow( "\n", 2 );
				char* ptr = (char*)unitStack->Finish();
				if( !lex_appendComment( ptr ) ) {
				    char* text = lex_copyStringWithSubstitution( ptr, strlen(ptr) );
		                    yylval->macro = AllocateMacroToken( text, strlen(text) );
	  	                    lastPragmaTimestamp = tokenTimestamp;
				    return PRAGMA;
				}
			}
<IFSCAN>"//".*		{}
<IFPRAGMA>"//".*	{ yymore(); }
		
"/*"[ \t]*vrq[ \t]+translate_off	{
				if( ignoreVrqComments ) {
				    unitStack->Grow( yytext, yyleng ); 
				    BEGIN COMMENT;
				} else {
				    yymore();	
				    BEGIN VCOMMENT2;
				}
			}
"/*"			{
				unitStack->Grow( yytext, yyleng ); 
				BEGIN COMMENT;
			}
<VCOMMENT2>[^*\n]* 	{
				yymore();	
			}	
<VCOMMENT2>"*"+[^*/\n]* 	{
				yymore();	
			}
<VCOMMENT2>\n		{ 
				yymore();
				NEW_LINE; 
			}
<VCOMMENT2>"*"+"/"	{ 
				yymore();
				BEGIN VCOMMENT; 
			}
<COMMENT>[^*\n]* 	{
				unitStack->Grow( yytext, yyleng );
			}	
<COMMENT>"*"+[^*/\n]* 	{
				unitStack->Grow( yytext, yyleng );
			}
<COMMENT>\n		{ 
				unitStack->Grow( yytext, yyleng ); 
				NEW_LINE; 
			}
<COMMENT>"*"+"/"	{ 
				unitStack->Grow( yytext, yyleng ); 
				unitStack->Grow( "\n", 2 );
				char* ptr = (char*)unitStack->Finish();
				BEGIN INITIAL; 
				if( !lex_appendComment( ptr ) ) {
				    char* text = lex_copyStringWithSubstitution( ptr, strlen(ptr) );
		                    yylval->macro = AllocateMacroToken( text, strlen(text) );
	  	                    lastPragmaTimestamp = tokenTimestamp;
				    return PRAGMA;
				}
			}
<VCOMMENT>[ \t]*"//"[ \t]*vrq[ \t]+translate_on.*	{ 
				unitStack->Grow( yytext, yyleng ); 
				unitStack->Grow( "\n", 2 );
				char* text = (char*)unitStack->Finish();
		                yylval->macro = AllocateMacroToken( text, strlen(text) );
				BEGIN INITIAL;
				return VRQ;
			}
<VCOMMENT>"/*"[ \t]*vrq[ \t]+translate_on	{
				unitStack->Grow( yytext, yyleng );	
				BEGIN VCOMMENT1;
			}
<VCOMMENT>.		{
				unitStack->Grow( yytext, yyleng );	
			}
<VCOMMENT>\n		{
				unitStack->Grow( yytext, yyleng );	
				NEW_LINE; 
			}
<VCOMMENT1>[^*\n]* 	{
				unitStack->Grow( yytext, yyleng );	
			}	
<VCOMMENT1>"*"+[^*/\n]* 	{
				unitStack->Grow( yytext, yyleng );	
			}
<VCOMMENT1>\n		{ 
				unitStack->Grow( yytext, yyleng );	
				NEW_LINE; 
			}
<VCOMMENT1>"*"+"/"	{ 
				unitStack->Grow( yytext, yyleng ); 
				unitStack->Grow( "\n", 2 );
				char* text = (char*)unitStack->Finish();
		                yylval->macro = AllocateMacroToken( text, strlen(text) );
				BEGIN INITIAL; 
				return VRQ;
			}

<IFPRAGMA>"/*"		{ yymore(); BEGIN IFPRAGMACOMMENT; }
<IFPRAGMACOMMENT>[^*\n]* 	{ yymore(); }	
<IFPRAGMACOMMENT>"*"+[^*/\n]* 	{ yymore(); }
<IFPRAGMACOMMENT>\n	{ yymore(); NEW_LINE; }
<IFPRAGMACOMMENT>"*"+"/"	{ yymore(); BEGIN IFPRAGMA; }

<IFSCAN>"/*"		{ BEGIN IFSCANCOMMENT; }
<IFSCANCOMMENT>[^*\n]* 	{}	
<IFSCANCOMMENT>"*"+[^*/\n]* 	{}
<IFSCANCOMMENT>\n	{ NEW_LINE; }
<IFSCANCOMMENT>"*"+"/"	{ BEGIN IFSCAN; }

<EDGE_SPEC>"]"		{
			        BEGIN INITIAL;
				return yytext[0];
			}

<EDGE_SPEC>01     	{
				yylval->edge = eEDGE01;
				return DOUBLE_DIGIT;
			}
<EDGE_SPEC>10     	{
				yylval->edge = eEDGE10;
				return DOUBLE_DIGIT;
			}
<EDGE_SPEC>0[xX]     	{
				yylval->edge = eEDGE0x;
				return DOUBLE_DIGIT;
			}
<EDGE_SPEC>1[xX]     	{
				yylval->edge = eEDGE1x;
				return DOUBLE_DIGIT;
			}
<EDGE_SPEC>[xX]0     	{
				yylval->edge = eEDGEx0;
				return DOUBLE_DIGIT;
			}
<EDGE_SPEC>[xX]1     	{
				yylval->edge = eEDGEx1;
				return DOUBLE_DIGIT;
			}
<EDGE_SPEC>,|"["	{ return yytext[0]; }

	/*
	 * strings
	 */

\"[^\"\n]*(\"|\n)  	{ 
		if( yytext[yyleng-1] != '"' ) {
			yyerror( "unterminated character string" );
			/* is the char befor the quote a \ ? */
		} else if( yytext[yyleng-2] == '\\' ) {
			yyless(yyleng-1);	/* return last quote */
			yymore();		/* append next string */
		  } else {
			yytext[strlen(yytext)-1] = 0;
			processString( yytext+1 );
	   		yylval->macro = AllocateMacroToken( yytext+1, yyleng-1, yyleng );
			return STRING_LITERAL;
		  }
	     }
	/*
	 * special mode for udp specification
	 */
<UDP>"/*"		{
				unitStack->Grow( yytext, yyleng ); 
				BEGIN UDP_COMMENT;
			}
<UDP_COMMENT>[^*\n]* 	{
				unitStack->Grow( yytext, yyleng );
			}	
<UDP_COMMENT>"*"+[^*/\n]* 	{
				unitStack->Grow( yytext, yyleng );
			}
<UDP_COMMENT>\n		{ 
				unitStack->Grow( yytext, yyleng ); 
				NEW_LINE; 
			}
<UDP_COMMENT>"*"+"/"	{ 
				unitStack->Grow( yytext, yyleng ); 
				unitStack->Grow( "\n", 2 );
				char* ptr = (char*)unitStack->Finish();
				BEGIN UDP; 
				if( !lex_appendComment( ptr ) ) {
				    char* text = lex_copyStringWithSubstitution( ptr, strlen(ptr) );
		                    yylval->macro = AllocateMacroToken( text, strlen(text) );
	  	                    lastPragmaTimestamp = tokenTimestamp;
				    return PRAGMA;
				}
			}

<UDP>endtable			{ BEGIN INITIAL; return ENDTABLE; }	
<UDP>[01xXbBrRFfpPnN():;?*-]     	{return yytext[0];}
<UDP>.				;

	/*
	 * white space
	 */
<INITIAL,HBASE,BBASE,OBASE,DBASE>{ws}		{ fstack->macroBytesConsumed += yyleng; }
<IFSCAN,UDP>{ws}	;
<IFPRAGMA>{ws}		{ yymore(); }
<INITIAL>\n		{ NEW_LINE; fstack->macroBytesConsumed++; }
<IFSCAN,UDP>\n		{ NEW_LINE; }
<IFPRAGMA>\n		{ yymore(); NEW_LINE; }

	/*
	 * end of file
	 */
<<EOF>>		{ 
		  if( eofCount != 0 ) {
			return 0;
		  } else {
			eofCount++;
			return EOF_;
		  }
		 }
	/*
	 * single character catch all
	 */

<IFSCAN>.	; 
<IFPRAGMA>.	{ yymore(); } 
.		{ return yytext[0]; } 



%%
const char* whs = " \t\n\f\r";
const char* whss = " \t\n\f\r/";
const char* digits = "01";

/****************************************************
	SVToken
	- sv enablement for token
 ***************************************************/
int SVToken( int token )
{
    /*
     * if System Verilog support is not enabled
     * flag error
     */
    if( !svEnable ) {
	yyerror( "System Verilog construct, add -sv to command line to allow" );
    }
    return token;
}

/****************************************************
	SVKeyword
	- sv enablement for keywords
 ***************************************************/
int SVKeyword( int token, YYSTYPE* yylval, char* yytext  )
{
    if( svEnable ) {
       return token;
    } else {
	/*
	 * if System Verilog support is not enabled
	 * return keyword as an IDENTIFIER
	 */
        clean_identifier( yytext );
        yylval->symbol = CSymbol::Lookup( Identifier(yytext) ); 
        return IDENTIFIER;
    }
}

/*****************************************************
   ConvertTimescale 
    - convert timescale specification to internal representation
 ****************************************************/
void ConvertTimescale( char* arg, int& unit, int& precision )
{
	char* arg1 = arg;
	int   argLen1 = strspn( arg1, digits );
	char* arg2 = arg1+argLen1;
	arg2 = arg2 + strspn( arg2, whs );
	int argLen2 = strcspn( arg2, whss );
	char* arg3 = arg2+argLen2;
	arg3 = arg3 + strspn( arg3, whs );
	int argLen3 = strspn( arg3, "/" );
	char* arg4 = arg3+argLen3;
	arg4 = arg4 + strspn( arg4, whs );
	int argLen4 = strspn( arg4, digits );
	char* arg5 = arg4+argLen4;
	arg5 = arg5 + strspn( arg5, whs );
	int argLen5 = strcspn( arg5, whs );
	unit      = Ts2Value( arg1, argLen1, arg2, argLen2 );
	precision = Ts2Value( arg4, argLen4, arg5, argLen5 );
} 

/*****************************************************
    Ts2Value
    - convert timescale strings to internal representation
 ****************************************************/

int Ts2Value( char* mult, int multLen, char* unit, int unitLen )
{
    int result;
    
    if( !strncmp( unit, "s", unitLen ) ) {
	result = 0;
    } else if( !strncmp( unit, "ms", unitLen ) ) {
	result = -3;
    } else if( !strncmp( unit, "us", unitLen ) ) {
	result = -6;
    } else if( !strncmp( unit, "ns", unitLen ) ) {
	result = -9;
    } else if( !strncmp( unit, "ps", unitLen ) ) {
	result = -12;
    } else if( !strncmp( unit, "fs", unitLen ) ) {
	result = -15;
    }
	
    if( !strncmp( mult, "1", multLen ) ) {
    } else if( !strncmp( mult, "10", multLen ) ) {
	result += 1;
    } else if( !strncmp( mult, "100", multLen ) ) {
	result += 2;
    }
	
    return result;
}

/*****************************************************
    CompilerDirective
    - handle compiler directive processing
 ****************************************************/

void CompilerDirective( char* line )
{
    char* directive = line;
    int   directiveLen = strcspn( directive, whs );
    char* arg = line+directiveLen;
    arg += strspn( arg, whs );
    int   argLen = strcspn( arg, whs );

    if( !strncmp( "`celldefine", line, directiveLen ) ) {
	compilerDirectiveCell = 1;
    } else if( !strncmp( "`endcelldefine", line, directiveLen ) ) {
	compilerDirectiveCell = 1;
    } else if( !strncmp( "`resetall", line, directiveLen ) ) {
	compilerDirectiveCell		  = 0;
	compilerDirectiveProtect	  = 0;
	compilerDirectivePrecision	  = 0; // 1 sec
	compilerDirectiveUnit		  = 0; // 1 sec
	compilerDirectiveTimescaleValid   = FALSE;
	compilerDirectiveDefaultNetType   = vpiWire;
	compilerDirectiveUnconnectedDrive = vpiHighZ;
    } else if( !strncmp( "`protect", line, directiveLen ) ) {
	compilerDirectiveProtect = 1;
    } else if( !strncmp( "`endprotect", line, directiveLen ) ) {
	compilerDirectiveProtect = 0;
    } else if( !strncmp( "`nounconnected_drive", line, directiveLen ) ) {
	compilerDirectiveUnconnectedDrive = vpiHighZ;
    } else if( !strncmp( "`unconnected_drive", line, directiveLen ) ) {
	if( !strncmp( "pull0", arg, argLen ) ) {
	    compilerDirectiveUnconnectedDrive = vpiPull0;
	} else if( !strncmp( "pull1", arg, argLen ) ) {
	    compilerDirectiveUnconnectedDrive = vpiPull1;
	}	
    } else if( !strncmp( "`timescale", line, directiveLen ) ) {
	ConvertTimescale( arg, compilerDirectiveUnit, 
				compilerDirectivePrecision);
	compilerDirectiveTimescaleValid   = TRUE;
	    
   // ??? mdh - not implemented 
    } else if( !strncmp( "`default_nettype", line, directiveLen ) ) {
	if( !strncmp( "wire", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiWire;
	} else if( !strncmp( "tri", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiTri;
	} else if( !strncmp( "tri0", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiTri0;
	} else if( !strncmp( "wand", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiWand;
	} else if( !strncmp( "triand", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiTriAnd;
	} else if( !strncmp( "tri1", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiTri1;
	} else if( !strncmp( "wor", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiWor;
	} else if( !strncmp( "trior", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiTriOr;
	} else if( !strncmp( "trireg", arg, argLen ) ) {
	    compilerDirectiveDefaultNetType = vpiTriReg;
	}
    } else {
	/*
 	 * treat all others as comments
	 */
	lex_appendComment( line );
    }
}

/*****************************************************
	yywrap
	- handle end of file processing
 *****************************************************/

int yywrap( void )
{
	fstack_t* oldStack;
	int	  result;
	int	  complex;	
	int	  wasFile = !fstack;
	const char*  macroName = NULL;
	int 	  timestamp;
	int	  minDependentZone;
	int	  insertionZone;

	result = 1;
	complex = FALSE;

	/*
	 * if a file or macro was pushed, pop it.
	 */
	if( fstack ) {
		/*
		 * pop context
		 */
		if( fstack->argList ) {
			if( fstack->argList[0] ) {
				tempStack.Free( fstack->argList[0] );
			}
			temp1Stack.Free( fstack->argList );
		}		
			
		complex = fstack->complex;
		macroName = fstack->macroName;
		timestamp = fstack->timestamp;
		minDependentZone = fstack->minDependentZone;
		insertionZone = fstack->insertionZone;
		oldStack = fstack;
		fstack = oldStack->next;
		yy_delete_buffer( oldStack->state );
		if( fstack && fstack->minDependentZone > minDependentZone ) {
		    fstack->minDependentZone = minDependentZone;
		}
//		YY_CURRENT_BUFFER = 0;
		wasFile = FALSE;
		if( oldStack->f ) {
			wasFile = TRUE;
			fclose( oldStack->f );
			/*
			 * must eat rest of include line 
			 */
			BEGIN EAT;
		}
		shell_xfree( oldStack, sizeof(fstack_t) );	

		/*
		 * restore previous context
		 */
		if( fstack ) {
			yy_switch_to_buffer( fstack->state );
			loc.lineno = fstack->lineno;
			loc.filename = fstack->filename;
			if( !quietFlag && wasFile ) {
				logprintf( "Continuing compilation of source file : %s\n", loc.filename );
			}
		} else {
			if( initialState == NULL ) {
				initialState = yy_create_buffer( NULL, YY_BUF_SIZE );
			}
			yy_switch_to_buffer( initialState );
			if(!interactive){
				return 1;
			}
		}
		result = 0;
	}

	if( complex ) {
		char* 		body;
		fstack_t*	stack;
		/*
		 * Add whitespace to end of macro to ensure
		 * it doesn't cross token boundries. Add a null
		 * to terminate string.
		 */
		tempStack.Grow( " ", 2 );
		body = (char*) tempStack.Finish();
		/*
		 * save top level context if needed 
		 */
		if( !initialStateSaved ) {
			initialState = YY_CURRENT_BUFFER;
			initialStateSaved = TRUE;
		}
		if( fstack ) {
			fstack->lineno = loc.lineno;
		}
		/*
		 * push a context pointing to macro text
		 */
		stack = (fstack_t*) shell_xmalloc( sizeof(fstack_t) );
		stack->argList = macroArgList;
		stack->complex = FALSE;
	        stack->eof = FALSE;
		stack->eof = FALSE;
		stack->s = body;
		stack->c = body;
		stack->e = stack->c + strlen( stack->c  );
		stack->state = yy_create_buffer( NULL, YY_BUF_SIZE );
		stack->f = NULL;
		stack->next = fstack;
		stack->filename = loc.filename;
		stack->timestamp = timestamp;
		stack->macroName = macroName;
		stack->serialNumber = fstackSerialNumber++;
		stack->minDependentZone  = minDependentZone;
	 	stack->insertionZone = insertionZone;
		stack->macroBytesConsumed = 0;
		fstack = stack;
		/* 
		 * switch in context
		 */
		yy_switch_to_buffer( fstack->state );
		BEGIN INITIAL;
	}
	if( wasFile ) {
		BEGIN INITIAL;
	}
	return result;
}

/*****************************************************
	processString
	- in place conversion of special characters
	  \n,\t,\\,\",\ddd
 *****************************************************/

void processString( char* sPtr )
{
	char*	dPtr;
	
	dPtr = sPtr;
	while( *sPtr ) {
		if( !strncmp( sPtr, "\\n", 2 ) ) { 
			*dPtr++ = '\n';
			sPtr += 2;
		} else if( !strncmp( sPtr, "\\t", 2 ) ) { 
			*dPtr++ = '\t';
			sPtr += 2;
		} else if( !strncmp( sPtr, "\\\\", 2 ) ) { 
			*dPtr++ = '\\';
			sPtr += 2;
		} else if( !strncmp( sPtr, "\\\"", 2 ) ) { 
			*dPtr++ = '\"';
			sPtr += 2;
		} else if( *sPtr == '\\' && 
			*(sPtr+1) >= '0' && *(sPtr+1) <= '3' && 
			*(sPtr+2) >= '0' && *(sPtr+2) <= '7' && 
			*(sPtr+3) >= '0' && *(sPtr+3) <= '7' ) {
			*dPtr++ = (*(sPtr+1) - '0') * 64 +  
				  (*(sPtr+2) - '0') * 8 +
				  *(sPtr+3) - '0' ;
			sPtr += 4;
		} else if( *sPtr == '\\' && 
			*(sPtr+1) >= '0' && *(sPtr+1) <= '7' && 
			*(sPtr+2) >= '0' && *(sPtr+2) <= '7' ) {
			*dPtr++ = (*(sPtr+1) - '0') * 8 +  
				  *(sPtr+2) - '0' ;
			sPtr += 3;
		} else if( *sPtr == '\\' && 
			*(sPtr+1) >= '0' && *(sPtr+1) <= '7' ) {
			*dPtr++ = *(sPtr+1) - '0' ;
			sPtr += 2;
		} else if( !strncmp( sPtr, "\\", 1 ) ) { 
			*dPtr++ = *(sPtr+1);
			sPtr += 2;
		} else {
			*dPtr++ = *sPtr++;
		}
	}
	*dPtr = 0;
}

/*******************************************************
	doIfdef
	- `ifdef processing
 *******************************************************/

void doIfdef( const char* macroName )
{
	int	defined;
	CMacro*	macro;
	int pragma = pragmaTickDefines.find(macroName) != 
						pragmaTickDefines.end();

	/*
	 * ignore passthru pragmas in disabled conditionals
	 */
	pragma &= (disableDepth==0);
	    
	ifDepth++;
	if( pragma && !ifPragma ) {
	    ifPragma = ifDepth;
	    ifPragmaSense = 1;
	}
	endifCountStack.push(endifCount);
	endifCount = 1;
	macro = CMacro::LookupMacro( macroName );
	/*
         * enter new preprocessor zone if this macro is pass thru or
	 * dependent up on a macro from a different zone
 	 */
	if( pragma || !macro || macro->GetZone() != preprocessorZone ) {
	    preprocessorZone++;
	}
	defined = !ifPragma && macro != NULL && macro->IsDefined();
	if( !defined || disableDepth > 0 ) {
		disableDepth++;
	}
	if( disableDepth == 0 ) {
		BEGIN INITIAL;
	} else if( ifPragma ) {
		yymore();
		BEGIN IFPRAGMA;
	} else {
		BEGIN IFSCAN;
	}
	TRACE( "`ifdef -> %s, ifDepth=%d,disableDepth=%d\n", 
		macroName, ifDepth, disableDepth ); 
}

/*******************************************************
	doIfndef
	- `ifndef processing
 *******************************************************/

void doIfndef( const char* macroName )
{
	int	defined;
	CMacro*	macro;
	int pragma = pragmaTickDefines.find(macroName) != 
						pragmaTickDefines.end();
	/*
	 * ignore passthru pragmas in disabled conditionals
	 */
	pragma &= (disableDepth==0);

	ifDepth++;
	if( pragma && !ifPragma ) {
	    ifPragma = ifDepth;
	    ifPragmaSense = 0;
	}
	endifCountStack.push(endifCount);
	endifCount = 1;
	macro = CMacro::LookupMacro( macroName );
	/*
         * enter new preprocessor zone if this macro is pass thru or
	 * dependent up on a macro from a different zone
 	 */
	if( pragma || !macro || macro->GetZone() != preprocessorZone ) {
	    preprocessorZone++;
	}
	defined = ifPragma || macro != NULL && macro->IsDefined();
	if( defined || disableDepth > 0 ) {
		disableDepth++;
	}
	if( disableDepth == 0 ) {
		BEGIN INITIAL;
	} else if( ifPragma ) {
		yymore();
		BEGIN IFPRAGMA;
	} else {
		BEGIN IFSCAN;
	}
	TRACE( "`ifndef -> %s, ifDepth=%d,disableDepth=%d\n", 
		macroName, ifDepth, disableDepth ); 
}

/*******************************************************
	doElsif
	- `elseif processing
 *******************************************************/

void doElsif( const char* macroName )
{
	int	defined;
	CMacro*	macro;
	int pragma = pragmaTickDefines.find(macroName) != 
						pragmaTickDefines.end();
	/*
	 * ignore passthru pragmas in disabled conditionals
	 */
	pragma &= (disableDepth==0);

	if( pragma && !ifPragma ) {
		yyerror( "`elseif cannot start pass through block" );
	}
	if( ifDepth == 0 ) {
		yyerror( "`elseif not paired with `ifdef" );
		return;
	}
	if( ifPragma ) {
		yymore();
	} else if( disableDepth == 0 ) {
		disableDepth++;
	} else if( disableDepth == 1 ) {
		disableDepth--;
	}

	ifDepth++;
	endifCount++;
	macro = CMacro::LookupMacro( macroName );
	defined = !ifPragma && macro != NULL && macro->IsDefined();
	if( !defined || disableDepth > 0 ) {
		disableDepth++;
	}
	if( disableDepth == 0 ) {
		BEGIN INITIAL;
	} else if( ifPragma ) {
		BEGIN IFPRAGMA;
	} else {
		BEGIN IFSCAN;
	}
	TRACE( "`elseif -> %s, ifDepth=%d,disableDepth=%d\n", 
		macroName, ifDepth, disableDepth ); 
}

/*******************************************************
	doElse
	- `else processing
 *******************************************************/

void doElse( )
{ 
	if( ifDepth == 0 ) {
		yyerror( "`else not paired with `ifdef" );
		return;
	}
	if( ifPragma ) {
	    yymore();
	} else if( disableDepth == 0 ) {
		disableDepth++;
	} else if( disableDepth == 1 ) {
		disableDepth--;
	}
	if( disableDepth == 0 ) {
		BEGIN INITIAL;
	} else if( ifPragma ) {
		BEGIN IFPRAGMA;
	} else { 
		BEGIN IFSCAN;
	}
	TRACE( "`else ifDepth=%d,disableDepth=%d\n", 
		ifDepth, disableDepth ); 
}	

/*******************************************************
	doEndif
	- `endif processing
	   returns 1 if pragma is ending
 *******************************************************/

int	doEndif( )
{
	int result = 0;
	while( endifCount ) {
		if( ifPragma == ifDepth ) {
	     		ifPragma = 0;
			result = 1;
		}
		if( ifDepth == 0 ) {
			yyerror( "`end not paired with `ifdef" );
			return result;
		}
		ifDepth--; 
		if( disableDepth > 0 ) {
			disableDepth--;
		}
		if( disableDepth == 0 ) {
			BEGIN INITIAL;
		} else {
			BEGIN IFSCAN;
		}
		endifCount--;
	}
	if( endifCountStack.size() == 0 ) {
	    /*
             * attempt to gracefully handle mismatched `if/`endifs
	     */
	    endifCount = 0;
	} else {
	    endifCount = endifCountStack.top();
	    endifCountStack.pop();
	}
	if( ifPragma ) {
	        yymore();
		BEGIN IFPRAGMA;
	}
	    
	TRACE( "`endif ifDepth=%d,disableDepth=%d\n", 
		ifDepth, disableDepth ); 

	return result;
}	

/*******************************************************
	doUndef
	- undefine macro 
 *******************************************************/

void 	doUndef( const char* macroName )
{
	CMacro*	m;

	m = CMacro::LookupMacro( macroName );
	if( m == NULL ) {
		return;
	}
	m->SetDefined( FALSE );	
}

/*******************************************************
	IsMacroDefined
	- if macro is undefined
 *******************************************************/

int 	IsMacroDefined( char* macroName )
{

	CMacro* macro = CMacro::LookupMacro( macroName );
	return macro != NULL && macro->IsDefined();
}

/*******************************************************
	doSimpleMacro
	- simple macro processing 
	  returns TRUE if macro is simple or undefined
 *******************************************************/

int 	doSimpleMacro( char* macroName )
{
	char*		body;
	CMacro*		macro;

	macro = CMacro::LookupMacro( macroName );
	if( macro == NULL || !macro->IsDefined() ) {
	    if( macrocheck ) {
		error( &loc, "macro '%s' is undefined", macroName );
	    }
	    return TRUE;
	}
	if( !macro->IsComplex() ) {
		body = macro->GetBody();
		TRACE( "simple macro -> %s = %s\n", macroName, body );
	} else {
		currentMacro = macro;
		TRACE( "complex macro -> %s\n", macroName );
		return FALSE;
	}

	/*
	 * save top level context if needed 
	 */
	if( !initialStateSaved ) {
		initialState = YY_CURRENT_BUFFER;
		initialStateSaved = TRUE;
	}
	if( fstack ) {
		fstack->lineno = loc.lineno;
	}
	/*
	 * push a context pointing to macro text
	 */
	fstack_t* stack = (fstack_t*) shell_xmalloc( sizeof(fstack_t) );
	stack->argList = NULL;
	stack->complex = FALSE;
	stack->eof = FALSE;
	stack->s = body;
	stack->c = body;
	stack->e = stack->c + strlen( stack->c );
	stack->state = yy_create_buffer( NULL, YY_BUF_SIZE );
	stack->f = NULL;
	stack->next = fstack;
	stack->filename = loc.filename;
	stack->timestamp = tokenTimestamp;
	unitStack->Grow( "`", 1 );
	unitStack->Grow( macro->GetName(), 
			 strlen(macro->GetName() ) );
	unitStack->Grow( "", 1 );
	stack->macroName = (char*)unitStack->Finish();
	stack->serialNumber = fstackSerialNumber++;
	stack->minDependentZone  = macro->GetZone();
	stack->insertionZone = preprocessorZone;
	stack->macroBytesConsumed = 0;
	fstack = stack;
	/* 
	 * switch in context
	 */
	yy_switch_to_buffer( fstack->state );


	return TRUE;
}
/*******************************************************
	doComplexMacro
	- complex macro processing 
 *******************************************************/

void	doComplexMacro( char** argList )
{
	char*		body;
	fstack_t*	stack;
	char**		p1;
	char**		p2;

	body = currentMacro->GetBody();
	realArgList = currentMacro->GetArgList();
	macroArgList = argList;

	for( p1 = realArgList, p2 = macroArgList; 
		*p1 != NULL && *p2 != NULL; p1++, p2++ );
	if( *p1 || *p2 ) {
		yyerror( "incorrect number of arguments for macro" );
		if( argList[0] ) {
			tempStack.Free( argList[0] );
		}
		temp1Stack.Free( argList );
		return;
	}


	BEGIN SUB;

	/*
	 * save top level context if needed 
	 */
	if( !initialStateSaved ) {
		initialState = YY_CURRENT_BUFFER;
		initialStateSaved = TRUE;
	}
	if( fstack ) {
		fstack->lineno = loc.lineno;
	}
	/*
	 * push a context pointing to macro text
	 */
	stack = (fstack_t*) shell_xmalloc( sizeof(fstack_t) );
	stack->argList = NULL;
	stack->complex = TRUE;
	stack->eof = FALSE;
	stack->s = body;
	stack->c = body;
	stack->e = stack->c + strlen( stack->c );
	stack->state = yy_create_buffer( NULL, YY_BUF_SIZE );
	stack->f = NULL;
	stack->next = fstack;
	stack->filename = loc.filename;
	stack->timestamp = tokenTimestamp;
	unitStack->Grow( "`", 1 );
	unitStack->Grow( currentMacro->GetName(), 
			 strlen(currentMacro->GetName() ) );
	unitStack->Grow( "", 1 );
	stack->macroName = (char*)unitStack->Finish();
	stack->serialNumber = fstackSerialNumber++;
	stack->minDependentZone  = currentMacro->GetZone();
	stack->insertionZone = preprocessorZone;
	stack->macroBytesConsumed = 0;
	fstack = stack;
	/* 
	 * switch in context
	 */
	yy_switch_to_buffer( fstack->state );


}

/*******************************************************
	startMacro
	- start macro definition 
 *******************************************************/
void startMacro()
{
	BEGIN DMSTEXT; 
	inMacroDef = TRUE;
}

/*******************************************************
	terminateMacro
	- end macro definition 
 *******************************************************/
void terminateMacro()
{
	char* ptr;
	inMacroDef = FALSE;
	permStack->Grow( "", 1 ); 
	ptr = (char*)permStack->Finish();
	currentMacro->SetBody( ptr );
	currentMacro->SetZone( preprocessorZone );
	TRACE( "macro body -> %s\n", ptr );
	BEGIN INITIAL; 
}

/*******************************************************
	doInclude
	- include file processing 
 *******************************************************/

void doInclude( const char* aFilename )
{
	FILE*	fin;
	fstack_t*	stack;
        string fullPath;
	
	TRACE( "include file -> %s\n", aFilename );
	/*
	 * open file with search
	 */
	fin = fopenSearch( aFilename, "r", incSearchList, 1, &fullPath );
	if( fin == NULL ) {
		message( &loc, mCOIDF, aFilename );
		return;
	}
	/*
	 * save top level context if needed
	 */
	if( !initialStateSaved ) {
		initialState = YY_CURRENT_BUFFER;
		initialStateSaved = TRUE;
	}

	if( fstack ) {
		fstack->lineno = loc.lineno;
	}
	/*
	 * push a context point to file
	 */
	stack = (fstack_t*) shell_xmalloc( sizeof(fstack_t) );
	stack->argList = NULL;
	stack->complex = FALSE;
	stack->eof = FALSE;
	stack->s = NULL;
	stack->c = NULL;
	stack->e = NULL;
	stack->filename = aFilename;
	stack->lineno = 0;
	stack->state = yy_create_buffer( fin, YY_BUF_SIZE );
	stack->f = fin;
	stack->next = fstack;
	stack->timestamp = tokenTimestamp;
	unitStack->Grow( "`include \"", 10 );
	unitStack->Grow( aFilename, strlen(aFilename) );
	unitStack->Grow( "\"", 2 );
	stack->macroName = (char*)unitStack->Finish();
	stack->serialNumber = fstackSerialNumber++;
	stack->minDependentZone  =  preprocessorZone;
	stack->insertionZone = preprocessorZone;
	stack->macroBytesConsumed = 0;
	fstack = stack;
	/* 
	 * switch in context
	 */
	yy_switch_to_buffer( fstack->state );
	loc.lineno = 1;
	loc.filename = aFilename;

	if( !quietFlag ) {
	    logprintf( "Compiling included source file : %s\n", 
		verboseFlag ? fullPath.c_str() : aFilename );
	}
}

/*******************************************************
	fopenSearch
	- version of fopen that searchs standard places 
 *******************************************************/

FILE*	fopenSearch( const char* name, const char* options, 
		     CSymbol** searchList, int isInclude,
		     string* resolvedPath )
{
	INT32	i;
	FILE*	fin;
	string	path = name;
		
	fin = fopen( name, options );

	if( fin == NULL && searchList ) {
		for( i = 0; searchList[i]; i++ ) {
			path = searchList[i]->GetName();
			path += PATH_SEPARATOR;
			path += name;					
			fin = fopen( path.c_str(), options );
			if( fin != NULL ) {
				if( isInclude ) {
				    includeFiles[strdup(path.c_str())] = 1;
				}
				break;
			}
		}
	} else if( fin && isInclude ) {
	    includeFiles[strdup(name)] = 1;
	}

	if( fin && resolvedPath ) {
	    *resolvedPath = path;
	}
	    
	return fin;
}

/*******************************************************
	yyerror
	- lexer error handler 
 *******************************************************/

void	yyerror( const char* message )
{
	error( &loc, message );
}
	
/*******************************************************
	yy_input
	- lexer input routine 
 *******************************************************/


int	yy_input( char* buf, int ms )
{
	if( fstack && fstack->eof ) {
	    return 0;
	}

	int	result;

	/*
	 * force ms to be less than 1024
	 * to allow room for pushing back 
	 * macros
	 */
	if( ms > 1024 ) {
		ms = 1024;
	}	

	if( fstack && !fstack->f ) {
		ms = 1;
	}

	/*
	 * if eof has happen
	 * return 0
	 */
	if( eofCount != 0 ) {
		return 0;
	}

	/*
	 * if context has been pushed by 
	 * macro insertion then supply
	 * data from buffer 
	 */
	if( fstack && fstack->c ) {
		result = fstack->e - fstack->c;
		while( !fstack->f && !fstack->complex ) {
		    if( result > 1 && 
			fstack->c[0] == '`' &&
			fstack->c[1] == '"' ) {
			result -= 1;
			fstack->c += 1;
			break;
		    }
		    if( result > 4 && 
			fstack->c[0] == '`' &&
			fstack->c[1] == '\\' &&
			fstack->c[2] == '`' &&
			fstack->c[3] == '"' ) {
			result -= 2;
			fstack->c += 2;
			fstack->c[0] = '\\';
			break;
		    } 
		    if( result > 1 && 
			fstack->c[0] == '`' &&
			fstack->c[1] == '`' ) {
			result -= 2;
			fstack->c += 2;
		    } else {
			break; 
		    } 

		}
		if( result == 0 ) {
			buf[0] = '\n';
			fstack->eof = TRUE;
			return 1;
		}
		if( result > ms ) {
			result = ms;
		}
		memcpy( buf, fstack->c, result );
		fstack->c += result;
		return result; 
	} 

	
	/*	
	 * else use flex code
	 */
	if( YY_CURRENT_BUFFER->yy_is_interactive ) {
		int c = '*';
		int n;
		for( n = 0; n < ms &&
			(c = getc( yyin )) != EOF && c != '\n'; ++n ){
			buf[n] = (char) c;
		}
		if( c == '\n' ) {
			buf[n++] = (char) c;
		}
		if( c == EOF && ferror( yyin ) ) {
			YY_FATAL_ERROR( "input in flex scanner failed" );
		}
		result = n;
	} else {
		result = fread( buf, 1, ms, yyin );
		if( result == 0 ) {
		    if( ferror( yyin ) ) {
			YY_FATAL_ERROR( "input in scanner failed" );
		    } else {
	                /*
 	                 * if in the middle of a macro definition 
	                 * terminate macro. Happens if no \n at end of file
 	                 */
	                if( inMacroDef ) {
	                    result = 1;
	                    buf[0] = '\n';
	                }
			buf[result] = '\001';
			result++;
			fstack->eof = TRUE;
		    }
		}	
	}
	return result;
 }
/*******************************************************
    lex_appendComment
    - Append string to list of comments. returns FALSE
      if comment was a pragma and not added to comment list
 *******************************************************/
int lex_appendComment( char* ptr )
{
    int result = 1;

    /* 
     * if no pragmas are defined, add sysnopsys pragma
     */
    if( !regex_init ) {
	regex_init = 1;
        if( !pragmas.size() ) {
            pragmas.push_back( "[/* \t]*synopsys[\t ]" );
        }
    	list<string>::iterator pptr;
        for( pptr = pragmas.begin(); pptr != pragmas.end(); ++pptr ) {
	    regex_t preg;
            MASSERT( !regcomp( &preg, pptr->c_str(), 
    		     REG_EXTENDED|REG_NOSUB|REG_NEWLINE ) ); 
	    regex.push_back( preg );
        }
    }
    /*
     * check comment to see if it is
     * a pragma.
     */
    regex_t preg;
    list<regex_t>::iterator rptr;
    for( rptr = regex.begin(); rptr != regex.end(); ++rptr ) {
        if( !regexec( &(*rptr), ptr, 0, NULL, 0 ) ) {
    	    result = 0;
	    break;
        }
    }
    if( result ) { 
    	int len = strlen( ptr );
        commentStack.Grow( ptr, len );
    }
    return result;
}
/*******************************************************
    lex_copyStringWithSubstitution
    - return copy of string with simple `defines evaluated
 *******************************************************/
char* lex_copyStringWithSubstitution( char *ptr, int size )
{
    int phase = 0;
    int tickStart = 0;
    unitStack->Grow( ptr, size );
    unitStack->Grow( "\000", 1 );
    size++;
    for( int i = 0; i < size; i++ ) {
	int bt = ptr[i] == '`';
	int first = ('a' <= ptr[i] && ptr[i] <= 'z') ||
		    ('A' <= ptr[i] && ptr[i] <= 'Z') ||
		    ptr[i] == '_';
	int second = ('0' <= ptr[i] && ptr[i] <= '9') ||
		     ptr[i] == '$' ||
		     first;
	int done = (i+1 == size);
		    
	if( phase == 0 && bt ) {
	    tickStart = i;
	    phase++;
	} else if( phase == 1 ) {
	    phase++;
	    if( !first || done ) {
		phase++;
	    } 
	} else if( phase == 2 ) {
	    if( !second || done ) {
		phase++;
	    }
	}
	if( phase == 3 ) {
	    char*   body;
	    CMacro* macro;
	    char*   key;

	    key = strndup( &ptr[tickStart+1], i-tickStart-1 );
	    macro = CMacro::LookupMacro( key );
	    if( macro != NULL && macro->IsDefined() && 
		(!keepMacros || macro->GetZone() != preprocessorZone) &&
	        !macro->IsComplex() ) {
/*
printf( "String: **%s**\n", &ptr[tickStart] );
printf( "key: **%s**\n", key );
printf( "macro: **%s**\n", macro->GetBody() );
*/
		int ksize = strlen(key)+1;
		char* mptr = macro->GetBody();
		int msize = strlen(mptr);
		ptr = (char*)unitStack->GetBase();
		if( ksize != msize ) {
		    if( ksize < msize ) {
			unitStack->Grow( msize-ksize ); 
		        ptr = (char*)unitStack->GetBase();
		    }
		    memmove( &ptr[tickStart+msize], 
			     &ptr[tickStart+ksize], size-tickStart-ksize );
		    // Note this may result in unused byte being allocated
		    size += msize - ksize;
		}
		int len = strlen(mptr);
		memcpy( &ptr[tickStart], mptr, len );
		/*
		 * replace new lines with spaces
		 * to not break tick def replacement
		 * in cpp style comments
		 */
		for( int j = 0; j < len; ++j ) {
		    char c = ptr[tickStart+j];
		    if( c == '\n' || c == '\r' ) {
			ptr[tickStart+j] = ' ';
		    }
		}
		i = tickStart-1;
	    }
	    free( key );
	    phase = 0;
// printf( "Result<%d>: **%s**\n\n", loc.lineno, &ptr[tickStart] );
	}
    }
    char* pr = (char*) unitStack->Finish();
    return pr;
}

/*******************************************************
	lex_getComments
	- return ptr to accumulated comments
 *******************************************************/

char*	lex_getComments()
{
	INT32	size;
	char*	comment;

	size = (char*)commentStack.NextFree() - (char*)commentStack.GetBase();
	comment = lex_copyStringWithSubstitution( 
			(char*)commentStack.GetBase(), size );

	commentStack.Free( NULL );
	return comment;
}
/*******************************************************
	lex_ifdefPragma
	- return ptr to accumulated pragma
 *******************************************************/

char*	lex_ifdefPragma( char* s, int sense )
{
	string  buf;
	char*   ptr;

	if( sense ) {
		buf = "`ifdef ";
	} else {
		buf = "`ifndef ";
	}
	buf += s;
	ptr = (char*)buf.c_str();
	char* result =  lex_copyStringWithSubstitution( ptr, buf.length() );
	return result;
}
/*******************************************************
	lex_init
	- lexer initialization routine 
 *******************************************************/

void  lex_init()
{
	eofCount = 0;
	interactive = 1;
	loc.filename = "stdin";
	loc.lineno = 1;
}
/*******************************************************
	lex_openFile
	- open file to be lexed
 *******************************************************/

int	lex_openFile( char* aFilename ) 
{
	fstack_t*	stack;
	FILE*		fin;
	

	preprocessorZone++;
	eofCount = 0;
	interactive = 0;
	fin = fopenSearch( aFilename, "r", NULL, 0, NULL );
	if( fin == NULL ) {
		error( (Coord_t*)NULL, "cannot open file '%s'\n", aFilename );
		return FALSE;
	}
	/*
	 * save orignal state if first open 
	 */
	if( !initialStateSaved ) {
		initialState = YY_CURRENT_BUFFER;
		initialStateSaved = TRUE;
	}

	/*
	 * create an initial context pointing to file
	 */
	stack = (fstack_t*) shell_xmalloc( sizeof(fstack_t) );
	stack->argList = NULL;
	stack->complex = FALSE;
	stack->eof = FALSE;
	stack->s = NULL;
	stack->c = NULL;
	stack->e = NULL;
	stack->filename = aFilename;
	stack->lineno = 0;
	stack->state = yy_create_buffer( fin, YY_BUF_SIZE );
	stack->f = fin;
	stack->next = fstack;
	stack->timestamp = tokenTimestamp;
	stack->macroName = NULL;
	stack->minDependentZone  =  preprocessorZone;
	stack->insertionZone = preprocessorZone;
	stack->macroBytesConsumed = 0;
	fstack = stack;

	yy_switch_to_buffer( fstack->state );
	loc.lineno = 1;
	loc.filename = aFilename;

	if( !quietFlag ) {
	    logprintf( "Compiling source file : %s\n", aFilename );
	}
	
	return TRUE;
}

/*******************************************************
	clean_identifier
	- remove escape sequences from identifiers 
 *******************************************************/

void clean_identifier( char *s )
{
	char* sptr;
	char* dptr;
	int   inEscape;

	dptr     = s;
	sptr     = s;
	inEscape = FALSE;

	while( *sptr != 0 ) {
		if( !inEscape && *sptr == '\\' ) {
			sptr++;
			inEscape = TRUE;	
			continue;
		}
		if( inEscape && 
		    ( *sptr == ' ' ||
		      *sptr == '\t' ||
		      *sptr == '\n' ||
		      *sptr == '\f' ) ) {
			sptr++;
			inEscape = FALSE;
			continue;
		}
		*dptr++ = *sptr++;
	}
	*dptr = 0;
}

/***********************************************************
	lex_startUDP
	- put parser in UDP mode
 ***********************************************************/

void lex_startUDP()
{
    BEGIN UDP;
}

/***********************************************************
	InvalidMacroToken
	- Initialize null macro info structure
	  that won't match. Used as place holder in parser.
 ***********************************************************/
MacroToken InvalidMacroToken()
{
    /*
     * determine history depth
     */
    int depth = 0;
    
    MacroToken info;
    info.null = FALSE;
    info.initialTimestamp = 0;
    info.finalTimestamp = 0;
    info.historyDepth = 0;
    info.text = NULL;
    info.nodePtr = NULL;
    info.attribute = NULL;
    info.previousNodePtr = NULL;
    info.history = NULL;
    return info;
}

/***********************************************************
	NullMacroToken
	- Initialize null macro info structure
	  for use as a default argument to functions
 ***********************************************************/
MacroToken NullMacroToken()
{
    /*
     * determine history depth
     */
    int depth = 0;
    
    MacroToken info;
    info.null = TRUE;
    info.initialTimestamp = 0;
    info.finalTimestamp = 0;
    info.historyDepth = 0;
    info.text = NULL;
    info.nodePtr = NULL;
    info.attribute = NULL;
    info.previousNodePtr = NULL;
    info.history = NULL;
    return info;
}

/***********************************************************
	AllocateMacroToken
	- Allocate and initialize macro info structure
 ***********************************************************/
MacroToken AllocateMacroToken( const char* text, int leng, int consumed )
{
    /*
     * deal with default sentinal
     */
    if( consumed < 0 ) {
	consumed = leng;
    }
    /*
     * determine history depth
     */
    int depth = 0;
    fstack_t* ptr = fstack;
    if( keepMacros ) {
        while( ptr ) {
            if( ptr->macroName == NULL || ptr->complex ) { 
                break;
            }
            if( ptr->minDependentZone < ptr->insertionZone ) {
                break;
            }
            ptr = ptr->next;
            depth++;
        }
    }
 
    MacroToken info;
    if( depth > 0 ) {
        info.history = (MacroHistory*)tokenStack->Alloc(
					depth*sizeof(MacroHistory));
    } else {
	info.history = NULL;
    }
    info.null = FALSE;
    info.initialTimestamp = tokenTimestamp;
    info.finalTimestamp = tokenTimestamp;
    info.historyDepth = depth;
    info.text = (char*)unitStack->Copy0( text, leng );
    info.nodePtr = NULL;
    info.attribute = NULL;
    info.previousNodePtr = NULL;
    ptr = fstack;
    fstack_t* prevPtr = NULL;
    int prevFinal = FALSE;
    for( int i = 0; i < depth; ++i ) {
	if( prevPtr ) {
	    if( prevFinal ) {
	        ptr->macroBytesConsumed += strlen(prevPtr->macroName);
	    }
	} else {
	    ptr->macroBytesConsumed += consumed;
	}
	info.history[i].macroName = ptr->macroName;
	info.history[i].initialTimestamp = ptr->timestamp;
	info.history[i].finalTimestamp = ptr->timestamp;
	if( ptr->f ) {
	    info.history[i].final = FALSE;
	} else {
//	    info.history[i].final = ptr->c == ptr->e;
	    info.history[i].final = (ptr->e - ptr->s) <= ptr->macroBytesConsumed;
	}
	info.history[i].serialNumber = ptr->serialNumber;
	prevPtr = ptr;
	prevFinal = info.history[i].final;
	ptr = ptr->next;
    }
//printf( "MDH creating:\n" );
//DumpMacroInfo( info );
    return info;
}

/***********************************************************
	MergeToken
	- merge macro info from two tokens. return NULL 
	  if merge is not possible.
 ***********************************************************/
MacroToken MergeToken( const MacroToken& a, const MacroToken& b )
{
//printf( "MDH a:\n" );
//DumpMacroInfo( a );
//printf( "MDH b:\n" );
//DumpMacroInfo( b );
    if( !keepMacros ) {
	return NullMacroToken();
    }
    /*
     * merge requires:
     * - sucessive timestamps
     * - a merge point
     */
    if( a.finalTimestamp+1 != b.initialTimestamp ) {
        return NullMacroToken();
    }
    int aMergeDepth = -1;
    int bMergeDepth = -1;
    for( int i = 0; i < a.historyDepth; ++i ) {
        for( int j = 0; j < b.historyDepth; ++j ) {
	    if( a.history[i].serialNumber == b.history[j].serialNumber ) {
		aMergeDepth = i;
		bMergeDepth = j;
	        goto FOUND;
	    }
        }
    }
FOUND:
    if( aMergeDepth < 0 ) {
        return NullMacroToken();
    }
    /*
     * merge entries
     */
    int aDepth = aMergeDepth;
    int bDepth = bMergeDepth;
    int depth = 0;
    while( 1 ) {
	if( aDepth >= a.historyDepth ) {
	    break;
	}
	if( bDepth >= b.historyDepth ) {
	    break;
	}
	if( a.history[aDepth].serialNumber != b.history[bDepth].serialNumber ) {
	    break;
	}
	aDepth++;
	bDepth++;
	depth++;
    }
    MacroToken info;
    info.history = (MacroHistory*)tokenStack->Alloc(depth*sizeof(MacroHistory));
    info.null = FALSE;
    info.initialTimestamp = a.initialTimestamp;
    info.finalTimestamp = b.finalTimestamp;
    info.historyDepth = depth;
    info.text = NULL;
    info.nodePtr = NULL;
    info.attribute = NULL;
    info.previousNodePtr = NULL;
    for( int i = 0; i < depth; ++i ) {
	info.history[i].macroName = a.history[aMergeDepth+i].macroName;
	info.history[i].initialTimestamp = 
				a.history[aMergeDepth+i].initialTimestamp;
	info.history[i].finalTimestamp = b.finalTimestamp;
	info.history[i].final = b.history[bMergeDepth+i].final;
	info.history[i].serialNumber = a.history[aMergeDepth+i].serialNumber;
    }
//printf( "MDH info:\n" );
//DumpMacroInfo( info );
    return info;
}

/***********************************************************
	CompleteToken
	- see if token can be represented by a single macro.
	  returns history structure if possible, NULL otherwise.
 ***********************************************************/
const MacroHistory* CompleteToken( const MacroToken& macroInfo )
{
    int result = -1;
    for( int i = 0; i < macroInfo.historyDepth; ++i ) {
	if( !macroInfo.history[i].final ) {
	    break;
	}
	if( macroInfo.finalTimestamp != macroInfo.history[i].finalTimestamp || 
	    macroInfo.initialTimestamp != macroInfo.history[i].initialTimestamp ) {
	    break;
	}
	result = i;
    }
    if( result >= 0 ) {
	return &macroInfo.history[result];
    } 
    return NULL;
}

/***********************************************************
	DumpMacroInfo
	- print macro info for debug
 ***********************************************************/
void DumpMacroInfo( const MacroToken& a )
{
    printf( "  null=%d\n", a.null );
    printf( "  timestamp=(%d,%d)\n", a.initialTimestamp, a.finalTimestamp );
    printf( "  text=%s\n", a.text );
    if( const_cast<MacroToken&>(a).nodePtr != NULL ) {
        printf( "  node=" ); 
	const_cast<MacroToken&>(a).nodePtr->Dump(stdout); 
	printf( "\n" );
    }
    if( const_cast<MacroToken&>(a).previousNodePtr != NULL ) {
        printf( "  previous node=" ); 
	const_cast<MacroToken&>(a).previousNodePtr->Dump(stdout); 
	printf( "\n" );
    }
    for( int i = 0; i < a.historyDepth; ++i ) {
	printf( "    name=%s\n",  a.history[i].macroName );
	printf( "    timestamp=(%d,%d)\n",  a.history[i].initialTimestamp, 							a.history[i].finalTimestamp );
	printf( "    final=%d\n",  a.history[i].final );
	printf( "    sn=%d\n",  a.history[i].serialNumber );
    }
}

/***********************************************************
	Macro2String
	- convert macro to text 
 ***********************************************************/
char* Macro2String( CMacro* macro )
{
    const char* prefix = "`define ";
    const char* name = macro->GetName();
    char* body = const_cast<char*>(macro->GetBody());
    unitStack->Grow( prefix, strlen(prefix) );
    unitStack->Grow( name, strlen(name) );
    if( macro->IsComplex() ) {
	char** arglist = macro->GetArgList();
        unitStack->Grow( "(", 1 );
	int first = 1;
	while( *arglist ) {
	    if( !first ) {
                unitStack->Grow( ",", 1 );
	    } else {
	        first = 0;
	    }
    	    unitStack->Grow( *arglist, strlen(*arglist) );
	    arglist++;
	}
        unitStack->Grow( ")", 1 );
    }
    unitStack->Grow( " ", 1 );
    int blen = strlen(body);
    for( int i = 0; i < blen; ++i ) {
	switch( body[i] ) {
	case '\n':
            unitStack->Grow( "\\", 1 );
            unitStack->Grow( &body[i], 1 );
	    break;
	default:
            unitStack->Grow( &body[i], 1 );
	    break;
	}
    }
    unitStack->Grow( " ", 1 );
    unitStack->Grow( "", 1 ); 
    return (char*)unitStack->Finish();
}

/*************************************************
     Macro2Node
     - return node associated with macro
 ************************************************/
CNode* Macro2Node( const MacroToken& m )
{ 
    return cLINK( const_cast<MacroToken&>(m).previousNodePtr,
		  const_cast<MacroToken&>(m).nodePtr ); 
}
